#ifndef _COLOR_SPACES_H
#define _COLOR_SPACES_H  
////////////////////////////////////////////////////////////////////////////////////
#include "loki_TypeManip.h" 
#include "loki_TypeTraits.h" 
#include "loki_static_check.h" 
//#include "alexandresku.h"
////////////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
namespace ImAlg
{
/** @addtogroup ElementaryFiltersDoxGroup  Elementary filters
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @{
 */
	/*!
     * \brief   ,  .   
	 *        .    PixByPix3to1 
	 *   PixByPix3to3 :      . !  
	 *   YCbCr.
     *******************************************************************************************/
	namespace C
	{
		enum NormalizationType
		{
			nNoScale=0, //> 
			nStretch=1,  //>  ,    .      .
			nSaveRatio=2,//>     ,     ,        .
			nSaveChrom=3, //> :      nSaveRatio,        (.  YCbCr     Cb/Cr ).
			nDefault=nStretch
		};
		enum SystemType
		{
			sys_other=0,
			sys_chrom=1,
			sys_hue=2
		};

		template<typename T,class ColorSystem,NormalizationType Norm=nDefault> class Normalizer;

		template<typename T,class ColorSystem>
		class Normalizer<T,ColorSystem,nNoScale>
		{
		public:
			typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
			const T maxcolor;
			const T mincolor;
			const T_temp mult01;
			Normalizer(T MaxColor=T(255),T MinColor=0,T ZeroColor=T(128))
				: maxcolor(MaxColor),mincolor(MinColor),mult01(T_temp(MaxColor-MinColor))
			{
			}
			template<typename TT>
			void RGBto01(T r,T g,T b,TT& r01,TT& g01,TT& b01) const
			{
				r01=TT((r-mincolor)/mult01);
				g01=TT((g-mincolor)/mult01);
				b01=TT((b-mincolor)/mult01);
			}
			template<typename TT>
			void from01toRGB(TT r01,TT g01,TT b01,T& r,T& g,T& b) const
			{
				r=T(r01*mult01+mincolor);
				g=T(g01*mult01+mincolor);
				b=T(b01*mult01+mincolor);
			}
			///returns sample_normalized from sample_unnormalized, dual to  toRGB()
			template<int n>
			T fromRGB(T sample_unnorm) const { return sample_unnorm;}
			///returns rgb_normalized from rgb_unnormalized, dual to  fromRGB()
			template<int n>
			T toRGB(T rgb_unnorm) const { return rgb_unnorm;}
			template<int n,typename MultiplierType>
			MultiplierType Offset() const {return 0;}
			template<int n,typename MultiplierType>
			MultiplierType Mult() const {return 1;}
		};
		template<typename T,class ColorSystem>
		class Normalizer<T,ColorSystem,nStretch>
		{
			typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
			T maxcolor;
			T mincolor;
			T_temp mult01;
			T_temp a[ColorSystem::n_components];
			T_temp b[ColorSystem::n_components];
		private:
			template<int n,typename TT>
			struct Cof
			{
				template<typename T1>
				static void build(TT* a,TT* b,T1 MaxColor,T1 MinColor)
				{
					a[n-1]=TT(MaxColor-MinColor)/(ColorSystem::sample_max<n-1,TT>(MaxColor)-ColorSystem::sample_min<n-1,TT>(MaxColor)); 
					b[n-1]=TT(MinColor)-a[n-1]*ColorSystem::sample_min<n-1,TT>(MaxColor);
					Cof<n-1,TT>::build(a,b,MaxColor,MinColor);
				}
			};
			template<typename TT>
			struct Cof<0,TT>
			{
				template<typename T1>
				static void build(TT* a,TT* b,T1 MaxColor,T1 MinColor){}
			};
		public:
/*
Normalizer(const Normalizer& nn)
				: maxcolor(nn.maxcolor),mincolor(nn.mincolor),mult01(nn.mult01)
			{
				Cof<ColorSystem::n_components,T_temp>::build(a,b,maxcolor,mincolor);
			}
			Normalizer& operator=(const Normalizer& nn)
				: maxcolor(nn.maxcolor),mincolor(nn.mincolor),mult01(nn.mult01)
			{
				Cof<ColorSystem::n_components,T_temp>::build(a,b,maxcolor,mincolor);
			}
*/
			Normalizer(T MaxColor_=T(255),T MinColor_=0,T ZeroColor_=T(128))
				: maxcolor(MaxColor_),mincolor(MinColor_),mult01(T_temp(MaxColor_-MinColor_))
			{
				Cof<ColorSystem::n_components,T_temp>::build(a,b,maxcolor,mincolor);

/*				std::cout << "Normalizer:" << std::endl 
					<< "maxcolor= " << maxcolor << std::endl 
					<< "mincolor= " << mincolor << std::endl 
					<< "mult01= " << mult01 << std::endl ;
				for(int i=0 ; i<ColorSystem::n_components; ++i)
				{
					std::cout << "a["<<i<<"]=" << a[i] << "     b["<<i<<"]=" << b[i] <<std::endl; 
				}
*/
			}
			template<typename TT>
			void RGBto01(T r,T g,T b,TT& r01,TT& g01,TT& b01) const
			{
				r01=TT((r-mincolor)/mult01);
				g01=TT((g-mincolor)/mult01);
				b01=TT((b-mincolor)/mult01);
			}
			template<typename TT>
			void from01toRGB(TT r01,TT g01,TT b01,T& r,T& g,T& b) const
			{
				r=T(r01*mult01+mincolor);
				g=T(g01*mult01+mincolor);
				b=T(b01*mult01+mincolor);
			}
			T MaxColor() const {return maxcolor;}
			///returns sample_normalized from sample_unnormalized, dual to  toRGB()
			template<int n>
			T fromRGB(T sample_unnorm) const { return Mult<n,T_temp>()*sample_unnorm+Offset<n,T_temp>();}
			///returns rgb_normalized from rgb_unnormalized, dual to  fromRGB()
			template<int n>
			T toRGB(T rgb_unnorm) const { return (rgb_unnorm-Offset<n,T_temp>())/Mult<n,T_temp>();}

			template<int n,typename MultiplierType>
			MultiplierType Offset() const 
			{
				LOKI_STATIC_CHECK(((0<=n)&&(n<ColorSystem::n_components)),Sample_number_is_illegal_in_template_parameter_n);
				return b[n];
			}
			template<int n,typename MultiplierType>
			MultiplierType Mult() const 
			{
				LOKI_STATIC_CHECK(((0<=n)&&(n<ColorSystem::n_components)),Sample_number_is_illegal_in_template_parameter_n);
				return a[n];
			}
		};
		template<typename T,class ColorSystem>
		class Normalizer<T,ColorSystem,nSaveRatio>
		{
			typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
			const T maxcolor;
			const T mincolor;
			const T_temp mult01;
			T_temp a;
			T_temp b[ColorSystem::n_components];
		private:
			template<int n,typename TT>
			struct Cof
			{
				template<typename T1>
				static void build(TT& a,T1 MaxColor)
				{
					Cof<n-1,TT>::build(a,b);
					a[n-1]=TT(MaxColor-MinColor)/(ColorSystem::sample_max<n-1,TT>(MaxColor)-ColorSystem::sample_min<n-1,TT>(MaxColor)); 
				}
			};
			template<typename TT>
			struct Cof<0,TT>
			{
				static void build(TT& a){ a=0;}
			};
			template<int n,typename TT>
			struct Ofs
			{
				static void build(TT a,TT* b)
				{
					b[n-1]=TT(MinColor)-a*ColorSystem::sample_min<n-1,TT>();
					Ofs<n-1,TT>::build(a,b);
				}
			};
			template<typename TT>
			struct Ofs<0,TT>
			{
				static void build(TT a,TT* b){}
			};
		public:
			Normalizer(T MaxColor=T(255),T MinColor=0,T ZeroColor=T(128))
				: maxcolor(MaxColor),mincolor(MinColor),mult01(T_temp(MaxColor-MinColor))
			{
				Cof<ColorSystem::n_components,T_temp>::build();
				Ofs<ColorSystem::n_components,T_temp>::build();
			}
			template<typename TT>
			void RGBto01(T r,T g,T b,TT& r01,TT& g01,TT& b01)
			{
				r01=TT((r-mincolor)/mult01);
				g01=TT((g-mincolor)/mult01);
				b01=TT((b-mincolor)/mult01);
			}
			template<typename TT>
			void from01toRGB(TT r01,TT g01,TT b01,T& r,T& g,T& b)
			{
				r=T(r01*mult01+mincolor);
				g=T(g01*mult01+mincolor);
				b=T(b01*mult01+mincolor);
			}
			///returns sample_normalized from sample_unnormalized, dual to  toRGB()
			template<int n>
			T fromRGB(T sample_unnorm) const { return Mult<n,T_temp>()*sample_unnorm+Offset<n,T_temp>();}
			///returns rgb_normalized from rgb_unnormalized, dual to  fromRGB()
			template<int n>
			T toRGB(T rgb_unnorm) const { return (rgb_unnorm-Offset<n,T_temp>())/Mult<n,T_temp>();}

			template<int n,typename MultiplierType>
			MultiplierType Offset() const 
			{
				LOKI_STATIC_CHECK(((0<=n)&&(n<ColorSystem::n_components)),Sample_number_is_illegal_in_template_parameter_n);
				return b;
			}
			template<int n,typename MultiplierType>
			MultiplierType Mult() const 
			{
				LOKI_STATIC_CHECK(((0<=n)&&(n<ColorSystem::n_components)),Sample_number_is_illegal_in_template_parameter_n);
				return a;
			}
		};
		template<typename T,class ColorSystem>
		class Normalizer<T,ColorSystem,nSaveChrom>
		{
			typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
			const T maxcolor;
			const T mincolor;
			const T_temp mult01;
			T_temp a;
			T_temp b[ColorSystem::n_components];
		private:
		public:
			Normalizer(T MaxColor=T(255),T MinColor=0,T ZeroColor=T(128))
				: maxcolor(MaxColor),mincolor(MinColor),mult01(T_temp(MaxColor-MinColor))
			{
			}
			template<typename TT>
			void RGBto01(T r,T g,T b,TT& r01,TT& g01,TT& b01)
			{
				r01=TT((r-mincolor)/mult01);
				g01=TT((g-mincolor)/mult01);
				b01=TT((b-mincolor)/mult01);
			}
			template<typename TT>
			void from01toRGB(TT r01,TT g01,TT b01,T& r,T& g,T& b)
			{
				r=T(r01*mult01+mincolor);
				g=T(g01*mult01+mincolor);
				b=T(b01*mult01+mincolor);
			}
			///returns sample_normalized from sample_unnormalized, dual to  toRGB()
			template<int n>
			T fromRGB(T sample_unnorm) { return a*sample_unnorm+b;}
			///returns rgb_normalized from rgb_unnormalized, dual to  fromRGB()
			template<int n>
			T toRGB(T rgb_unnorm) { return (rgb_unnorm-b)/a;}

			template<int n>
			double Offset(T MaxColor,T MinColor)
			{
				typedef double Real_t;
				if(is_sample_sign_alternating<n>())
				{
				}
				else
				{
				}
			}
		};
/*
			Real_t min=sample_min<n,Real_t>(); 
			Real_t max=sample_max<n,Real_t>(); 
			Real_t maxabs=sample_maxabs<n,Real_t>();
			Real_t maxsat=sample_max_saturation<n,Real_t>(); 
			is_sample_sign_alternating<n>();
			is_sample_cyclic_var<n>();
*/
		/// HSV and HSL systems use the same definition of Hue, here delta=max(r,g,b)-min(r,g,b)
		/// Returns value in bounds 0..6.
		inline double HSV_HSL_hue(double max,double r,double g,double b,double delta)
		{
			double hh;
			if( r == max )
			{
				hh = (g-b)/delta;		// between yellow & magenta
				if(hh<0)
					hh+=6;
			}
			else if( g == max )
				hh = 2 + (b-r)/delta;	// between cyan & yellow
			else //( b == max )
				hh = 4 + (r-g)/delta;	// between magenta & cyan
			return hh;
		}

		/*!
		* \brief converts RGB709 to CIE XYZ color coordinates.
		*
		*  The CIE XYZ color system describes a color in terms of its components
		*  of X, Y and Z primaries.  In ordinary circumstances, all three of
		*  these components must be nonnegative.
		*  
		*  RGB values in a particular set of primaries can be transformed to and 
		*  from CIE XYZ via a 3x3 matrix transform. These transforms involve tristimulus 
		*  values, that is a set of three linear-light components that conform to the 
		*  CIE color-matching functions. CIE XYZ is a special set of tristimulus values. 
		*  In XYZ, any color is represented as a set of positive values.
		*  
		*  To transform from XYZ to RGB (with D65 white point), the matrix transform 
		*  used is:
		*
		*  \code
		*  	[ R ]   [  3.240479 -1.537150 -0.498535 ]   [ X ]
		*  	[ G ] = [ -0.969256  1.875992  0.041556 ] * [ Y ]
		*  	[ B ]   [  0.055648 -0.204043  1.057311 ]   [ Z ]
		*  \endcode
		*
		*  The range for valid R, G, B values is [0,1]. Note, this matrix has negative 
		*  coefficients. Some XYZ color may be transformed to RGB values that are 
		*  negative or greater than one. This means that not all visible colors can 
		*  be produced using the RGB system.
		*  
		*  The inverse transformation matrix is as follows:
		*  
		*  \code
		*  	[ X ]   [  0.412453  0.357580  0.180423 ]   [ R ] 
		*  	[ Y ] = [  0.212671  0.715160  0.072169 ] * [ G ]
		*  	[ Z ]   [  0.019334  0.119193  0.950227 ]   [ B ]
		*  \endcode
 		*  	
		*  	 , 2- . (.68):
		*  	[X]= [ 0.49018626 0.30987954 0.19993420 ]   [Rc]
		*  	[Y]= [ 0.17701522 0.81232418 0.01066060 ] * [Gc]
		*  	[Z]= [ 0.00000000 0.01007720 0.98992280 ]   [Bc]
		*  	----------------------------------------------
		*  	[Rc]   [ 2.36353918 -0.89582361 0.46771557 ]   [X]
		*  	[Gc] = [-0.51511248  1.42643694  0.08867553 ] * [Y]
		*  	[Bc]   [ 0.00524373 -0.01452082  1.00927709 ]   [Z]
		*  
		*  
		***********************************************************************************************/
		struct XYZ
		{
			enum { n_components=3};
			enum { is_metric=1 };
			enum { is_system=sys_other };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("X","Y","Z");}
			static std::string name() { return "XYZ";}

			//Unified name functions
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<0> ) { return X_f(r,g,b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<1> ) { return Y_f(r,g,b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<2> ) { return Z_f(r,g,b);}
			template<typename T>
			static T rgb(T x,T y,T z,Loki::Int2Type<0> ) { return R_f(x,y,z);}
			template<typename T>
			static T rgb(T x,T y,T z,Loki::Int2Type<1> ) { return G_f(x,y,z);}
			template<typename T>
			static T rgb(T x,T y,T z,Loki::Int2Type<2> ) { return B_f(x,y,z);}
			template<int n,typename T>
			static T sample(T r,T g,T b)  { return sample(r,g,b,Loki::Int2Type<n>());}
			template<int n,typename T>
			static T rgb(T x,T y,T z) { return rgb(x,y,z,Loki::Int2Type<n>());}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.0,0.0,0.0));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.950456,1.0,1.088754));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.950456,1.0,1.088754));
			}
			template<int n>
			static bool is_sample_sign_alternating() { return false;}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}


			template<typename T>
			static T X_f(T r,T g,T b) { return T( 0.412453 * r + 0.357580 * g + 0.180423 * b);}
			template<typename T>
			static T Y_f(T r,T g,T b) { return T( 0.212671 * r + 0.715160 * g + 0.072169 * b);}
			template<typename T>
			static T Z_f(T r,T g,T b) { return T( 0.019334 * r + 0.119193 * g + 0.950227 * b);}
			template<typename T>
			static T R_f(T x,T y,T z) { return T( 3.240479 * x - 1.537150 * y - 0.498535 * z);}
			template<typename T>
			static T G_f(T x,T y,T z) { return T(-0.969256 * x + 1.875992 * y + 0.041556 * z);}
			template<typename T>
			static T B_f(T x,T y,T z) { return T( 0.055648 * x - 0.204043 * y + 1.057311 * z);}
			/*!
			* \brief  XYZ: RGB -> X   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct X		
			{
				typedef XYZ ColorSystem;
				T operator()(T r,T g,T b) const { return X_f(r,g,b);}
			};
			/*!
			* \brief  XYZ: RGB -> Y   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct Y		
			{
				typedef XYZ ColorSystem;
				T operator()(T r,T g,T b) const { return Y_f(r,g,b);}
			};
			/*!
			* \brief  XYZ: RGB -> Z   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct Z		
			{
				typedef XYZ ColorSystem;
				T operator()(T r,T g,T b) const { return X_f(r,g,b);}
			};
			/*!
			* \brief  XYZ: XYZ -> R   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct R		
			{
				typedef XYZ ColorSystem;
				T operator()(T x,T y,T z) const { return R_f(x,y,z);}
			};
			/*!
			* \brief  XYZ: XYZ -> G   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct G		
			{
				typedef XYZ ColorSystem;
				T operator()(T x,T y,T z) const { return G_f(x,y,z);}
			};
			/*!
			* \brief  XYZ: XYZ -> B   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct B		
			{
				typedef XYZ ColorSystem;
				T operator()(T x,T y,T z) const { return B_f(x,y,z);}
			};

			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,XYZ,nDefault>,typename Iter=T&>
			struct fromRGB
			{
				typedef XYZ ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,XYZ,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T r,T g,T b,Iter x,Iter y,Iter z) const 
				{
					x=norm.fromRGB<0>(X_f(r,g,b));
					y=norm.fromRGB<1>(Y_f(r,g,b));
					z=norm.fromRGB<2>(Z_f(r,g,b));
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,XYZ,nDefault>,typename Iter=T&>
			struct toRGB
			{
				typedef XYZ ColorSystem;
				NormType norm;

				toRGB(NormType normalizer=Normalizer<T,XYZ,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T x,T y,T z,Iter r,Iter g,Iter b) const 
				{
					r = norm.toRGB<0>(R_f(x,y,z));
					g = norm.toRGB<0>(G_f(x,y,z));
					b = norm.toRGB<0>(B_f(x,y,z));
				}
			};

		};

		/*!
		* \brief   YCbCr
		* 
		*  \code
		*    Y  =  0.299   * R + 0.587   * G + 0.114   * B
		*    Cb = -0.16874 * R - 0.33126 * G + 0.50000 * B
		*    Cr =  0.50000 * R - 0.41869 * G - 0.08131 * B
		*    ----------------------------------------------
		*    R = Y + 0.00000 * Cb + 1.40200 * Cr
		*    G = Y - 0.34414 * Cb - 0.71414 * Cr
		*    B = Y + 1.77200 * Cb + 0.00000 * Cr
		*  \endcode
		*******************************************************************************************/
		struct YCbCr
		{
			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_chrom };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("Y","Cb","Cr");}
			static std::string name() { return "YCbCr";}

			//Unified name functions
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<0> ) { return T( 0.29900 * r + 0.58700 * g + 0.11400 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<1> ) { return T(-0.16874 * r - 0.33126 * g + 0.50000 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<2> ) { return T( 0.50000 * r - 0.41869 * g - 0.08131 * b);}
			template<typename T>
			static T rgb(T y,T cb,T cr,Loki::Int2Type<0> ) { return T( y + 0.00000 * cb + 1.40200 * cr );}
			template<typename T>
			static T rgb(T y,T cb,T cr,Loki::Int2Type<1> ) { return T( y - 0.34414 * cb - 0.71414 * cr );}
			template<typename T>
			static T rgb(T y,T cb,T cr,Loki::Int2Type<2> ) { return T( y + 1.77200 * cb + 0.00000 * cr );}
			template<int n,typename T>
			static T sample(T r,T g,T b)  { return sample(r,g,b,Loki::Int2Type<n>());}
			template<int n,typename T>
			static T rgb(T y,T cb,T cr) { return rgb(y,cb,cr,Loki::Int2Type<n>());}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.0,-0.5,-0.5));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,0.5,0.5));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,0.5,0.5));
			}
			/// specific for cromatic-like color systems, =sqrt(c1^2+c2^2)
			template<typename T>
			static T sample_max_saturation() 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(0.7071067811865);
			}
			template<int n>
			static bool is_sample_sign_alternating() { return Vtl::Alg::Return1stArg<n==0>(false,true);}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}


			template<typename T>
			static T Y_f(T r,T g,T b)  { return sample<0>(r,g,b);	}
			template<typename T>
			static T Cb_f(T r,T g,T b) { return sample<1>(r,g,b);	}
			template<typename T>
			static T Cr_f(T r,T g,T b) { return sample<2>(r,g,b);	}
			template<typename T>
			static T R_f(T y,T cb,T cr){ return rgb<0>(y,cb,cr);}
			template<typename T>
			static T G_f(T y,T cb,T cr){ return rgb<1>(y,cb,cr);}
			template<typename T>
			static T B_f(T y,T cb,T cr){ return rgb<2>(y,cb,cr);}

			/*!
			* \brief  YCbCr: RGB -> Y   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct Y		
			{
				typedef YCbCr ColorSystem;
				T operator()(T r,T g,T b) const { return Y_f(r,g,b);}
			};
			/*!
			* \brief  YCbCr: RGB -> Cb   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct Cb		
			{
				typedef YCbCr ColorSystem;
				T operator()(T r,T g,T b) const {return Cb_f(r,g,b);}
			};
			/*!
			* \brief  YCbCr: RGB -> Cr   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct Cr		
			{
				typedef YCbCr ColorSystem;
				T operator()(T r,T g,T b) const {return Cr_f(r,g,b);}
			};
			/*!
			* \brief  YCbCr -> R   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct R		
			{
				typedef YCbCr ColorSystem;
				T operator()(T y,T cb,T cr) const {return R_f(y,cb,cr);}
			};
			/*!
			* \brief  YCbCr -> G   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct G		
			{
				typedef YCbCr ColorSystem;
				T operator()(T y,T cb,T cr) const {return G_f(y,cb,cr);}
			};
			/*!
			* \brief  YCbCr -> B   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct B
			{
				typedef YCbCr ColorSystem;
				T operator()(T y,T cb,T cr) const {return B_f(y,cb,cr);}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,YCbCr,nDefault>,typename Iter=T&>
			struct fromRGB
			{
				typedef YCbCr ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,YCbCr,nDefault>(255,0,128)) : norm(normalizer) {}
				//fromRGB(const fromRGB& f) : norm(f.norm) {}
				//fromRGB& operator=(const fromRGB& f) { norm=f.norm; return *this;}

				void operator()(T r,T g,T b,Iter y,Iter cb,Iter cr) const 
				{
					y =norm.fromRGB<0>(YCbCr::Y_f(r,g,b));
					cb=norm.fromRGB<1>(YCbCr::Cb_f(r,g,b));
					cr=norm.fromRGB<2>(YCbCr::Cr_f(r,g,b));
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,YCbCr,nDefault>,typename Iter=T&>
			struct toRGB
			{
				typedef YCbCr ColorSystem;
				const NormType norm;

				toRGB(NormType normalizer=Normalizer<T,YCbCr,nDefault>()) : norm(normalizer) {}

				void operator()(T y,T cb,T cr,Iter r,Iter g,Iter b) const 
				{
					r = norm.toRGB<0>(YCbCr::R_f(y,cb,cr));
					g = norm.toRGB<1>(YCbCr::G_f(y,cb,cr));
					b = norm.toRGB<2>(YCbCr::B_f(y,cb,cr));
				}
			};
		};

///////////////////////////////////////////////////////
		/*!
		* \brief  ,    \ref Lit_ColorInvariance .  
		*    ,  .. (shading, highlighting)
		* 
		*  \code
		*	 E                   | 0.06 0.63 0.27 | |R|
		*	 E_lambda         =  | 0.3 0.04 -0.35 |*|G|
		*	 E_lambda_lambda     | 0.34 -0.6 0.17 | |B|
        * 
		*	 |R|   |1.1844   1.5685   1.3482  | |E               |     
		*	 |G| = |0.99088  0.47562 -0.59453 |*|E_lambda        |
		*	 |B|   |1.1284  -1.4583   1.0876  | |E_lambda_lambda |
		*  \endcode
		*******************************************************************************************/
		struct EElEll
		{
			enum { n_components=3};
			enum { is_metric=1 };
			enum { is_system=sys_chrom };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("E","El","Ell");}
			static std::string name() { return "EElEll";}
			//Unified name functions
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<0> ) { return T( 0.06 * r  + 0.63 * g   + 0.27 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<1> ) { return T( 0.3  * r  + 0.04 * g   - 0.35 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<2> ) { return T( 0.34 * r  - 0.6  * g   + 0.17 * b);}
			template<typename T>
			static T rgb(T e,T el,T ell,Loki::Int2Type<0> ) { return T( 1.1844 *e + 1.5685 *el + 1.3482 *ell);}
			template<typename T>
			static T rgb(T e,T el,T ell,Loki::Int2Type<1> ) { return T( 0.99088*e + 0.47562*el - 0.59453*ell);}
			template<typename T>
			static T rgb(T e,T el,T ell,Loki::Int2Type<2> ) { return T( 1.1284 *e - 1.4583 *el + 1.0876 *ell);}
			template<int n,typename T>
			static T sample(T r,T g,T b)  { return sample(r,g,b,Loki::Int2Type<n>());}
			template<int n,typename T>
			static T rgb(T e,T el,T ell) { return rgb(e,el,ell,Loki::Int2Type<n>());}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.0,-0.35,-0.6));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.96,0.34,0.51));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.96,0.35,0.6));
			}
			/// specific for cromatic-like color systems, =sqrt(c1^2+c2^2)
			template<typename T>
			static T sample_max_saturation(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*0.6946221994725);
			}
			template<int n>
			static bool is_sample_sign_alternating() { return Vtl::Alg::Return1stArg<n==0>(false,true);}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}

/*
			template<typename T>
			static T E_f(T R,T G,T B)    { return T( 0.06 * R  + 0.63 * G   + 0.27 * B);}
			template<typename T>
			static T El_f(T R,T G,T B)   { return T( 0.3  * R  + 0.04 * G   - 0.35 * B);}
			template<typename T>
			static T Ell_f(T R,T G,T B)  { return T( 0.34 * R  - 0.6  * G   + 0.17 * B);}
			template<typename T>
			static T R_f(T E,T El,T Ell) { return T( 1.1844 *E + 1.5685 *El + 1.3482 *Ell);}
			template<typename T>
			static T G_f(T E,T El,T Ell) { return T( 0.99088*E + 0.47562*El - 0.59453*Ell);}
			template<typename T>
			static T B_f(T E,T El,T Ell) { return T( 1.1284 *E - 1.4583 *El + 1.0876 *Ell);}
			template<typename T>
			static T El_f_toPositive(T el,unsigned bits=8)   { return T( el*(0.5/0.35)+(1<<(bits-1)));}
			template<typename T>
			static T Ell_f_toPositive(T ell,unsigned bits=8)   { return T( ell*(0.5/0.6)+(1<<(bits-1)));}
*/
			//Specific name functions
			template<typename T>
			static T E_f(T r,T g,T b)    { return sample<0>(r,g,b);}
			template<typename T>
			static T El_f(T r,T g,T b)   { return sample<1>(r,g,b);}
			template<typename T>
			static T Ell_f(T r,T g,T b)  { return sample<2>(r,g,b);}
			template<typename T>
			static T R_f(T e,T el,T ell) { return rgb<0>(e,el,ell);}
			template<typename T>
			static T G_f(T e,T el,T ell) { return rgb<1>(e,el,ell);}
			template<typename T>
			static T B_f(T e,T el,T ell) { return rgb<2>(e,el,ell);}


			template<typename T>
			static T El_f_toPositive(T el,unsigned bits=8)   { return T( el*(0.5/0.35)+(1<<(bits-1)));}
			template<typename T>
			static T Ell_f_toPositive(T ell,unsigned bits=8)   { return T( ell*(0.5/0.6)+(1<<(bits-1)));}
			/*!
			* \brief  EElEll: RGB -> E   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct E		
			{
				typedef EElEll ColorSystem;
				T operator()(T r,T g,T b) const { return E_f(r,g,b);}
			};
			/*!
			* \brief  EElEll: RGB -> E_lambda   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct El		
			{
				typedef EElEll ColorSystem;
				T operator()(T r,T g,T b) const { return El_f(r,g,b);}
			};
			/*!
			* \brief  EElEll: RGB -> E_lambda_lambda   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct Ell		
			{
				typedef EElEll ColorSystem;
				T operator()(T r,T g,T b) const { return Ell_f(r,g,b);}
			};
			/*!
			* \brief  EElEll -> R   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct R		
			{
				typedef EElEll ColorSystem;
				T operator()(T e,T el,T ell) const {return R_f(e,el,ell);}
			};
			/*!
			* \brief  EElEll -> G   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct G		
			{
				typedef EElEll ColorSystem;
				T operator()(T e,T el,T ell) const {return G_f(e,el,ell);}
			};
			/*!
			* \brief  EElEll -> B   PixByPix3to1.
			*******************************************************************************************/
			template<typename T>
			struct B
			{
				typedef EElEll ColorSystem;
				T operator()(T e,T el,T ell) const {return B_f(e,el,ell);}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,EElEll,nDefault>,typename Iter=T&>
			class fromRGB
			{
			public:
				typedef EElEll ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,EElEll,nDefault>(255,0,128)) : norm(normalizer) {}

				void operator()(T r,T g,T b,Iter e,Iter el,Iter ell) const 
				{
					e  =norm.fromRGB<0>(sample<0>(r,g,b));
					el =norm.fromRGB<1>(sample<1>(r,g,b));
					ell=norm.fromRGB<2>(sample<2>(r,g,b));
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,EElEll,nDefault>,typename Iter=T&>
			class toRGB
			{
			public:
				typedef EElEll ColorSystem;
				NormType norm;

				toRGB(NormType normalizer=Normalizer<T,EElEll,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T e,T el,T ell,Iter r,Iter g,Iter b) const 
				{
					r = norm.toRGB<0>(rgb<0>(e,el,ell));
					g = norm.toRGB<1>(rgb<1>(e,el,ell));
					b = norm.toRGB<2>(rgb<2>(e,el,ell));
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<unsigned bits,typename T,typename Iter=T&>
			class fromRGB_positive
			{
			public:
				typedef EElEll ColorSystem;
				fromRGB_positive() {}
				void operator()(T r,T g,T b,Iter e,Iter el,Iter ell) const 
				{
					fromRGB<T,Iter> f;
					f(r,g,b,e,el,ell);
					el=El_f_toPositive(el,bits);
					ell=Ell_f_toPositive(ell,bits);
				}
			};
		};

///////////////////////////////////////////////////////
		/*!
		* \brief   YIQ.
		* 
		* The YIQ system is the color primary system adopted by National Television 
		* System Committee (NTSC) for color TV broadcasting. The YIQ color solid is 
		* made by a linear transformation of the RGB cube. Its purpose is to exploit 
		* certain characteristics of the human eye to maximize the utilization of a 
		* fixed bandwidth. The human visual system is more sensitive to changes in 
		* luminance than to changes in hue or saturation, and thus a wider bandwidth 
		* should be dedicated to luminance than to color information. Y is similar to 
		* perceived luminance, I and Q carry color information and some luminance 
		* information. The Y signal usually has 4.2 MHz bandwidth in a 525 line 
		* system. Originally, the I and Q had different bandwidths (1.5 and 0.6 MHz), 
		* but now they commonly have the same bandwidth of 1 MHz. 
		* 
		*  \code
		*    Y = 0.299 * R + 0.587 * G + 0.114 * B 
		*    I = 0.596 * R - 0.275 * G - 0.321 * B 
		*    Q = 0.212 * R - 0.523 * G + 0.311 * B 
		*    ----------------------------------------------
		*    R = Y + 0.956 * I + 0.621 * Q 
		*    G = Y - 0.272 * I - 0.647 * Q 
		*    B = Y - 1.105 * I + 1.702 * Q 
		*  \endcode
		*******************************************************************************************/
		struct YIQ
		{
			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_chrom };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("Y","I","Q");}
			static std::string name() { return "YIQ";}

			//Unified name functions
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<0> ) { return T( 0.299 * r + 0.587 * g + 0.114 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<1> ) { return T( 0.596 * r - 0.275 * g - 0.321 * b );}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<2> ) { return T( 0.212 * r - 0.523 * g + 0.311 * b);}
			template<typename T>
			static T rgb(T y,T i,T q,Loki::Int2Type<0> ) { return T( y + 0.956 * i + 0.621 * q );}
			template<typename T>
			static T rgb(T y,T i,T q,Loki::Int2Type<1> ) { return T( y - 0.272 * i - 0.647 * q );}
			template<typename T>
			static T rgb(T y,T i,T q,Loki::Int2Type<2> ) { return T( y - 1.105 * i + 1.702 * q );}
			template<int n,typename T>
			static T sample(T r,T g,T b)  { return sample(r,g,b,Loki::Int2Type<n>());}
			template<int n,typename T>
			static T rgb(T y,T i,T q) { return rgb(y,i,q,Loki::Int2Type<n>());}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.0,-0.596,-0.523));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,0.596,0.523));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,0.596,0.523));
			}
			/// specific for cromatic-like color systems, =sqrt(c1^2+c2^2)
			template<typename T>
			static T sample_max_saturation(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*0.7929344235181);
			}
			template<int n>
			static bool is_sample_sign_alternating() { return Vtl::Alg::Return1stArg<n==0>(false,true);}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}

			template<typename T>
			static T Y_f(T r,T g,T b) {return sample<0>(r,g,b);}
			template<typename T>
			static T I_f(T r,T g,T b) {return sample<1>(r,g,b);}
			template<typename T>
			static T Q_f(T r,T g,T b) {return sample<2>(r,g,b);}
			template<typename T>
			static T R_f(T y,T i,T q) {return rgb<0>(y,i,q);}
			template<typename T>
			static T G_f(T y,T i,T q) {return rgb<1>(y,i,q);}
			template<typename T>
			static T B_f(T y,T i,T q) {return rgb<2>(y,i,q);}
			/*!
			* \brief  YIQ: RGB -> Y  .
			*******************************************************************************************/
			template<typename T>
			struct Y		
			{
				typedef YIQ ColorSystem;
				T operator()(T r,T g,T b) const {return Y_f(r,g,b);}
			};
			/*!
			* \brief  YIQ: RGB -> I  .
			*******************************************************************************************/
			template<typename T>
			struct I		
			{
				typedef YIQ ColorSystem;
				T operator()(T r,T g,T b) const {return I_f(r,g,b);}
			};
			/*!
			* \brief  YIQ: RGB -> Q  .
			*******************************************************************************************/
			template<typename T>
			struct Q		
			{
				typedef YIQ ColorSystem;
				T operator()(T r,T g,T b) const {return Q_f(r,g,b);}
			};
			/*!
			* \brief  YIQ -> R  .
			*******************************************************************************************/
			template<typename T>
			struct R		
			{
				typedef YIQ ColorSystem;
				T operator()(T y,T i,T q) const {return R_f(y,i,q);}
			};
			/*!
			* \brief  YIQ -> G  .
			*******************************************************************************************/
			template<typename T>
			struct G		
			{
				typedef YIQ ColorSystem;
				T operator()(T y,T i,T q) const {return G_f(y,i,q);}
			};
			/*!
			* \brief  YIQ -> B  .
			*******************************************************************************************/
			template<typename T>
			struct B
			{
				typedef YIQ ColorSystem;
				T operator()(T y,T i,T q) const {return B_f(y,i,q);}
			};
			/*!
			* \brief   PixByPix3to3,  RGB -> YIQ.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,YIQ,nDefault>,typename Iter=T&>
			struct fromRGB
			{
				typedef YIQ ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,YIQ,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T r,T g,T b,Iter y,Iter i,Iter q) const 
				{
					y = norm.fromRGB<0>(YIQ::Y_f(r,g,b));
					i = norm.fromRGB<1>(YIQ::I_f(r,g,b));
					q = norm.fromRGB<2>(YIQ::Q_f(r,g,b));
				}
			};
			/*!
			* \brief   PixByPix3to3,  YIQ -> RGB.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,YIQ,nDefault>,typename Iter=T&>
			struct toRGB
			{
				typedef YIQ ColorSystem;
				NormType norm;

				toRGB(NormType normalizer=Normalizer<T,YIQ,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T y,T i,T q,Iter r,Iter g,Iter b) const 
				{
					r = norm.toRGB<0>(YIQ::R_f(y,i,q));
					g = norm.toRGB<1>(YIQ::G_f(y,i,q));
					b = norm.toRGB<2>(YIQ::B_f(y,i,q));
				}
			};
		};

///////////////////////////////////////////////////////
		/*!
		* \brief   YUV.
		* 
		* 
		*  \code
		*    Y =   0.299 * R + 0.587 * G + 0.114 * B
		*    U = - 0.147 * R - 0.289 * G + 0.437 * B
		*    V =   0.615 * R - 0.515 * G - 0.100 * B
		*    ----------------------------------------------
		*    R = Y + 0.000 * U + 1.140 * V
		*    G = Y - 0.396 * U - 0.581 * V
		*    B = Y + 2.029 * U + 0.000 * V
		*  \endcode
		*******************************************************************************************/
		struct YUV
		{
			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_chrom };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("Y","U","V");}
			static std::string name() { return "YUV";}

			//Unified name functions
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<0> ) { return T( 0.299 * r + 0.587 * g + 0.114 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<1> ) { return T(-0.147 * r - 0.289 * g + 0.437 * b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<2> ) { return T( 0.615 * r - 0.515 * g - 0.100 * b);}
			template<typename T>
			static T rgb(T y,T u,T v,Loki::Int2Type<0> ) { return T( y + 1.140 * v );}
			template<typename T>
			static T rgb(T y,T u,T v,Loki::Int2Type<1> ) { return T( y - 0.396 * u - 0.581 * v );}
			template<typename T>
			static T rgb(T y,T u,T v,Loki::Int2Type<2> ) { return T( y + 2.029 * u );}
			template<int n,typename T>
			static T sample(T r,T g,T b)  { return sample(r,g,b,Loki::Int2Type<n>());}
			template<int n,typename T>
			static T rgb(T y,T u,T v) { return rgb(y,i,q,Loki::Int2Type<n>());}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.0,-0.436,-0.615));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,0.437,0.615));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,0.437,0.615));
			}
			/// specific for cromatic-like color systems
			template<typename T>
			static T sample_max_saturation(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*0.7544494681554);
			}
			template<int n>
			static bool is_sample_sign_alternating() { return Vtl::Alg::Return1stArg<n==0>(false,true);}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}



			template<typename T>
			static T Y_f(T r,T g,T b) {return sample<0>(r,g,b);}
			template<typename T>
			static T U_f(T r,T g,T b) {return sample<1>(r,g,b);}
			template<typename T>
			static T V_f(T r,T g,T b) {return sample<2>(r,g,b);}
			template<typename T>
			static T R_f(T y,T u,T v) {return rgb<0>(y,u,v);;}
			template<typename T>
			static T G_f(T y,T u,T v) {return rgb<1>(y,u,v);;}
			template<typename T>
			static T B_f(T y,T u,T v) {return rgb<2>(y,u,v);}
			/*!
			* \brief  YUV: RGB -> Y  .
			*******************************************************************************************/
			template<typename T>
			struct Y		
			{
				typedef YUV ColorSystem;
				T operator()(T r,T g,T b) const {return Y_f(r,g,b);}
			};
			/*!
			* \brief  YUV: RGB -> U  .
			*******************************************************************************************/
			template<typename T>
			struct U		
			{
				typedef YUV ColorSystem;
				T operator()(T r,T g,T b) const {return U_f(r,g,b);}
			};
			/*!
			* \brief  YUV: RGB -> V  .
			*******************************************************************************************/
			template<typename T>
			struct V		
			{
				typedef YUV ColorSystem;
				T operator()(T r,T g,T b) const {return V_f(r,g,b);}
			};
			/*!
			* \brief  YUV -> R  .
			*******************************************************************************************/
			template<typename T>
			struct R		
			{
				typedef YUV ColorSystem;
				T operator()(T y,T u,T v) const {return R_f(y,u,v);}
			};
			/*!
			* \brief  YUV -> G  .
			*******************************************************************************************/
			template<typename T>
			struct G		
			{
				typedef YUV ColorSystem;
				T operator()(T y,T u,T v) const {return G_f(y,u,v);}
			};
			/*!
			* \brief  YUV -> B  .
			*******************************************************************************************/
			template<typename T>
			struct B
			{
				typedef YUV ColorSystem;
				T operator()(T y,T u,T v) const {return B_f(y,u,v);}
			};
			/*!
			* \brief   PixByPix3to3,  RGB -> YUV.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,YUV,nDefault>,typename Iter=T&>
			struct fromRGB
			{
				typedef YUV ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,YUV,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T r,T g,T b,Iter y,Iter u,Iter v) const 
				{
					y = norm.fromRGB<0>(Y_f(r,g,b));
					u = norm.fromRGB<1>(U_f(r,g,b));
					v = norm.fromRGB<2>(V_f(r,g,b));
				}
			};
			/*!
			* \brief   PixByPix3to3,  YUV -> RGB.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,YUV,nDefault>,typename Iter=T&>
			struct toRGB
			{
				typedef YUV ColorSystem;
				NormType norm;

				toRGB(NormType normalizer=Normalizer<T,YUV,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T y,T u,T v,Iter r,Iter g,Iter b) const 
				{
					r = norm.toRGB<0>(R_f(y,u,v));
					g = norm.toRGB<1>(G_f(y,u,v));
					b = norm.toRGB<2>(B_f(y,u,v));
				}
			};
		};
	   /*!
 		* \brief   ,  struct HSI
		*
		*  \code  2.4494897427831780981972840747059
		*    a = 1/3         =0.33333333333333333333333333333333
		*    b = 1/sqrt(6.)  =0.40824829046386301636621401245113 //098
		*    c = sqrt(6.)/3  =0.81649658092772603273242802490167 //0196
		*    d = sqrt(6.)/2  =1.2247448713915890490986420373525  //529
		*
 		*    I  =   a * R + a * G +  a * B
		*    V1 = - b * R - b * G +  c * B
		*    V2 =   b * R - b * G +  0  * B
		*    ----------------------------------------------
		*    R = I -  b * V1 + d * V2
		*    G = I +  b * V1 - d * V2
		*    B = I +  c * V1 + 0 * V2
		*  \endcode
	    */
		struct IV1V2
		{
#define COLOR_SPACE_CONST_IV1V2_A 0.33333333333333333333333333333333
#define COLOR_SPACE_CONST_IV1V2_B 0.40824829046386301636621401245098
#define COLOR_SPACE_CONST_IV1V2_C 0.81649658092772603273242802490196
#define COLOR_SPACE_CONST_IV1V2_D 1.2247448713915890490986420373529
			
#define COLOR_SPACE_CONST_IV1V2_S 0.91287092917527685576161630466804

			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_chrom };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("I","V1","V2");}
			static std::string name() { return "IV1V2";}
			//Unified name functions
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<0> ) { return I_f(r,g,b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<1> ) { return V1_f(r,g,b);}
			template<typename T>
			static T sample(T r,T g,T b,Loki::Int2Type<2> ) { return V2_f(r,g,b);}
			template<typename T>
			static T rgb(T i,T v1,T v2,Loki::Int2Type<0> ) { return R_f(i,v1,v2);}
			template<typename T>
			static T rgb(T i,T v1,T v2,Loki::Int2Type<1> ) { return G_f(i,v1,v2);}
			template<typename T>
			static T rgb(T i,T v1,T v2,Loki::Int2Type<2> ) { return B_f(i,v1,v2);}
			template<int n,typename T>
			static T sample(T r,T g,T b)  { return sample(r,g,b,Loki::Int2Type<n>());}
			template<int n,typename T>
			static T rgb(T i,T v1,T v2) { return rgb(y,v1,v2,Loki::Int2Type<n>());}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(0.0,-COLOR_SPACE_CONST_IV1V2_C,-COLOR_SPACE_CONST_IV1V2_B));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,COLOR_SPACE_CONST_IV1V2_C,COLOR_SPACE_CONST_IV1V2_B));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*Vtl::Alg::ReturnArg<n>(1.0,COLOR_SPACE_CONST_IV1V2_C,COLOR_SPACE_CONST_IV1V2_B));
			}
			/// specific for cromatic-like color systems
			template<typename T>
			static T sample_max_saturation(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(MaxColor*COLOR_SPACE_CONST_IV1V2_S);
			}
			template<int n>
			static bool is_sample_sign_alternating() { return Vtl::Alg::Return1stArg<n==0>(false,true);}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}

			template<typename T>
			static T I_f(T r,T g,T b) {return T(COLOR_SPACE_CONST_IV1V2_A * (r + g + b));}
			template<typename T>
			static T V1_f(T r,T g,T b) {return T(- COLOR_SPACE_CONST_IV1V2_B * r - COLOR_SPACE_CONST_IV1V2_B * g + COLOR_SPACE_CONST_IV1V2_C * b);}
			template<typename T>
			static T V2_f(T r,T g,T b) {return T(  COLOR_SPACE_CONST_IV1V2_B * r - COLOR_SPACE_CONST_IV1V2_B * g );}
			template<typename T>
			static T R_f(T i,T v1,T v2) {return T( i - COLOR_SPACE_CONST_IV1V2_B * v1 + COLOR_SPACE_CONST_IV1V2_D * v2 );}
			template<typename T>
			static T G_f(T i,T v1,T v2) {return T( i + COLOR_SPACE_CONST_IV1V2_B * v1 - COLOR_SPACE_CONST_IV1V2_D * v2 );}
			template<typename T>
			static T B_f(T i,T v1,T v2) {return T( i + COLOR_SPACE_CONST_IV1V2_C * v1  );}

			/*!
			* \brief  IV1V2: RGB -> I.
			*******************************************************************************************/
			template<typename T>
			struct I		
			{
				typedef IV1V2 ColorSystem;
				T operator()(T r,T g,T b) const {return I_f(r,g,b);}
			};
			/*!
			* \brief  IV1V2: RGB -> V1  .
			*******************************************************************************************/
			template<typename T>
			struct V1		
			{
				typedef IV1V2 ColorSystem;
				T operator()(T r,T g,T b) const {return V1_f(r,g,b);}
			};
			/*!
			* \brief  IV1V2: RGB -> V2 .
			*******************************************************************************************/
			template<typename T>
			struct V2		
			{
				typedef IV1V2 ColorSystem;
				T operator()(T r,T g,T b) const {return V2_f(r,g,b);}
			};
			/*!
			* \brief  IV1V2 -> R  .
			*******************************************************************************************/
			template<typename T>
			struct R		
			{
				typedef IV1V2 ColorSystem;
				T operator()(T i,T v1,T v2) const {return R_f(i,v1,v2);}
			};
			/*!
			* \brief  IV1V2 -> G  .
			*******************************************************************************************/
			template<typename T>
			struct G		
			{
				typedef IV1V2 ColorSystem;
				T operator()(T i,T v1,T v2) const {return G_f(i,v1,v2);}
			};
			/*!
			* \brief  IV1V2 -> B  .
			*******************************************************************************************/
			template<typename T>
			struct B
			{
				typedef IV1V2 ColorSystem;
				T operator()(T i,T v1,T v2) const {return B_f(i,v1,v2);}
			};
			/*!
			* \brief   PixByPix3to3,  RGB -> IV1V2.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,IV1V2,nDefault>,typename Iter=T&>
			struct fromRGB
			{
				typedef IV1V2 ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,IV1V2,nDefault>(255,0,128)) : norm(normalizer) {}

				void operator()(T r,T g,T b,Iter i,Iter v1,Iter v2) const 
				{
					i  = norm.fromRGB<0>( I_f(r,g,b));
					v1 = norm.fromRGB<1>(V1_f(r,g,b));
					v2 = norm.fromRGB<2>(V2_f(r,g,b));
				}
			};
			/*!
			* \brief   PixByPix3to3,  IV1V2 -> RGB.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,IV1V2,nDefault>,typename Iter=T&>
			struct toRGB
			{
				typedef IV1V2 ColorSystem;
				NormType norm;

				toRGB(NormType normalizer=Normalizer<T,IV1V2,nDefault>(255,0,128)) : norm(normalizer) {}
				void operator()(T i,T v1,T v2,Iter r,Iter g,Iter b) const 
				{
					r = norm.toRGB<0>(R_f(i,v1,v2));
					g = norm.toRGB<1>(G_f(i,v1,v2));
					b = norm.toRGB<2>(B_f(i,v1,v2));
				}
			};
		};

		/*!
 		 * \brief   HSV .
		 * The Hue/Saturation/Value  
		 * model was created by A. R. Smith in 1978. It is 
		 * based on such intuitive color characteristics as tint, shade and tone 
		 * (or family, purety and intensity). The coordinate system is cylindrical, 
		 * and the colors are defined inside a hexcone. 
		 * The hue value H runs from 0 to 360?. 
		 * The saturation S is the degree of strength or purity and is from 0 to 1. 
		 * Purity is how much white is added to the color, so S=1 makes the purest 
		 * color (no white). 
		 * 		Brightness V also ranges from 0 to 1, where 0 is the black.
		 * 
		 * There is no transformation matrix for RGB/HSV conversion, but the algorithm 
		 * follows:
		 * 
		 * r,g,b values are from 0 to 1
		 * h = [0,360], s = [0,1], v = [0,1]
		 *		if s == 0, then h = -1 (undefined)
		 */
		struct HSV
		{
			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_hue };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("H","S","V");}
			static std::string name() { return "HSV";}
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,typename Iter=T&>
			struct fromRGB
			{
				typedef HSV ColorSystem;
				void operator()(T r,T g,T b,Iter h,Iter s,Iter v) const 
				{
					double min = Vtl::min( r, g, b );
					v = Vtl::max( r, g, b );
					double max = v;
					double delta = max - min;
					s= T(( max > 0 ) ? ( (255.*delta)/max ): 0);
					if(s <=0)
					{
						// r = g = b = 0		// s = 0, v is undefined
					//	h = -1;
						h=0;//yurin
					}
					else
					{
					//h *= 60;				// degrees
						h =T(HSV_HSL_hue(max,r,g,b,delta)*255.0f/6);//yurin
					//	s *=255;//yurin
					}
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,typename Iter=T&>
			struct toRGB
			{
				typedef HSV ColorSystem;
				void operator()(T h,T s,T v,Iter r,Iter g,Iter b) const 
				{
					if( s == 0 ) 
					{
						// achromatic (grey)
						r = g = b = v;
						return;
					}
					else
					{
						double ss=double(s)/255.;//yurin
						double hh=double(h)/(255./6);//yurin
						// sector 0 to 5
						int i = int(hh);
						double f = hh - i;			// factorial part of h
						T p = T( double(v) * ( 1.0 - ss ) );
						//float q = v * ( 1 - s * f );
						//float t = v * ( 1 - s * ( 1 - f ) );
						T tt= T( double(v) * ( 1.0 - ss * ((i&0x01)?f:( 1 - f )) ) );
						switch( i ) 
						{
							case 0:
								r = v;
								g = tt;//t;
								b = p;
								break;
							case 1:
								r = tt;//q;
								g = v;
								b = p;
								break;
							case 2:
								r = p;
								g = v;
								b = tt;//t;
								break;
							case 3:
								r = p;
								g = tt;//q;
								b = v;
								break;
							case 4:
								r = tt;//t;
								g = p;
								b = v;
								break;
							default:		// case 5:
								r = v;
								g = p;
								b = tt;//q;
								break;
						};
					}
				}
			};
		};
		/*!
 		 * \brief   HSL .
		 * The Hue/Saturation/Value  
		 */
		struct HSL
		{
			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_hue };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("H","S","L");}
			static std::string name() { return "HSL";}


			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(0.0,0.0,0.0));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(6.0,1.0,1.0));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(6.0,1.0,1.0));
			}
			template<int n>
			static bool is_sample_sign_alternating() { return false;}
			template<int n>
			static bool is_sample_cyclic_var() { return Vtl::Alg::Return1stArg<n==0>(true,false);}
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,HSL,nDefault>,typename Iter=T&>
			struct fromRGB
			{
				typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
				typedef HSL ColorSystem;
				NormType norm;

				fromRGB(NormType normalizer=Normalizer<T,HSL,nDefault>()) : norm(normalizer) {}
				void operator()(T r,T g,T b,Iter h,Iter s,Iter l) const 
				{
					T_temp rr,gg,bb;
					norm.RGBto01(r,g,b,rr,gg,bb);
					T_temp min = Vtl::min( rr, gg, bb );
					T_temp max = Vtl::max( rr, gg, bb );
					T_temp l2 = (max+min);
					T_temp l_ = T(l2/2);
					T_temp delta = max - min;
					T_temp s_;
					T_temp h_;
					if(delta<=0)  
					{
					//	h = -1;
						s_=0;
						h_=0;//yurin
					}
					else
					{
						if(l2<=1)
							s_=delta/l2;
						else
							s_=delta/(2-l2);
					//h *= 60;				// degrees
						h_ =T_temp(HSV_HSL_hue(max,rr,gg,bb,delta));//yurin
					}
					h = norm.fromRGB<0>(h_);
					s = norm.fromRGB<1>(s_);
					l = norm.fromRGB<2>(l_);
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,HSL,nDefault>,typename Iter=T&>
			struct toRGB
			{
				const NormType& norm;
				typedef HSL ColorSystem;

				toRGB(const NormType& normalizer=Normalizer<T,HSL,nDefault>()) : norm(normalizer) {}

				void operator()(T h,T s,T l,Iter r,Iter g,Iter b) const 
				{
					double q=(l<0.5)?(l*(1+s)):(l+s-l*s);
					double p=2*l-q;
					double hh=h/6;
					r=T(ColorC(p,q,hh+1/3.));
					g=T(ColorC(p,q,hh));
					b=T(ColorC(p,q,hh-1/3.));
				}
			private:
				double ColorC(double p,double q,double t)
				{
					if(t<0)
						t+=1;
					if(t>1)
						t-=1;

					if(t<(1/6.))
						return p+(q-p)*6*t;
					else if(t<0.5)
						return q;
					else if(t<(2/3.))
						return p+(q-p)*6*(2/3.-t);
					else
						return p;
				}
			};
		};
		/*!
 		 * \brief   HSI (or may be refered as IHS).
		 * The Hue/Saturation/Intensity  
		 */
		template<class LinSys=Vtl::ImAlg::C::IV1V2> 
		struct HSI
		{
		public:
			enum { n_components=3};
			enum { is_metric=0 };
			enum { is_system=sys_hue };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("H","S","I");}
			static std::string name() 
			{ 
				if(Loki::IsSameType<LinSys,Vtl::ImAlg::C::IV1V2>::value )
					return "HSI";
				else
				{
					std::ostringstream os;
					os << "HSI("<<LinSys::name() << ")";
					return os.str();
				}
				//return Loki::IsSameType<LinSys,Vtl::ImAlg::C::IV1V2>::value ? "HSI" : "HSI_ex";
			}

			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(0);
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(255.0,1.0,200.0));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				return sample_max<n,T>(); 
			}
			template<int n>
			static bool is_sample_sign_alternating() { return false;}
			template<int n>
			static bool is_sample_cyclic_var() { return  Vtl::Alg::Return1stArg<n==0>(true,false);}
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,typename Iter=T&>
			struct fromRGB
			{
				typedef HSI ColorSystem;
				void operator()(T r,T g,T b,Iter h,Iter s,Iter i) const 
				{
					//*i = PrimColorSys::I_f(r,g,b);
					//T v1 = PrimColorSys::V1_f(r,g,b);
					//T v2 = PrimColorSys::V2_f(r,g,b);
					typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;

					const float MaxColor=256.0f;
					const float Offset=128.0f;

					T_temp v1;
					T_temp v2;
					//PrimColorSys::fromRGB<T>(r,g,b,i,&v1,&v2); 
					typedef Normalizer<float,LinSys,nNoScale> NoScaleNorm;
					NoScaleNorm no_norm;
					LinSys::fromRGB<T,NoScaleNorm> f(no_norm);
					//IV1V2::fromRGB<T> f;
					f(r,g,b,i,v1,v2); 
					//s = sqrt(Vtl::sqr(v1)+Vtl::sqr(v2));
					//h = atan2(v2,v1);
					s = T(MaxColor* sqrt(Vtl::sqr(v1)+Vtl::sqr(v2)) / 	LinSys::sample_max_saturation<float>(MaxColor)); 
					h = T(MaxColor*atan2(v2,v1)/M_PI+Offset);
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,typename Iter=T&>
			struct toRGB
			{
				typedef HSI ColorSystem;


				void operator()(T h,T s,T i,Iter r,Iter g,Iter b) const 
				{
					typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
					
					const float MaxColor=256.0f;
					const float Offset=128.0f;

					T_temp hh = T_temp(double(h-Offset)/MaxColor*M_PI);
					//s /=1.2;
					s*=LinSys::sample_max_saturation<float>(MaxColor);
					s/=256.;
					T_temp v1 = s*cos(hh);
					T_temp v2 = s*sin(hh);
					//PrimColorSys::toRGB<T>(i,v1,v2,r,g,b); 
					typedef Normalizer<float,LinSys,nNoScale> NoScaleNorm;
					NoScaleNorm no_norm;
					LinSys::toRGB<T,NoScaleNorm> f(no_norm);
					f(i,(T)v1,(T)v2,r,g,b); 
				}
			};
		};
		/*!
 		 * \brief   L*a*b*  
		 */
		struct Lab
		{
			enum { n_components=3};
			enum { is_metric=1 };
			enum { is_system=sys_other };

			template<int n>
			static const char *sample_name() { return Vtl::Alg::ReturnArg<n>("L","a","b");}
			static std::string name() { return "Lab";}
			template<int n,typename T>
			static T sample_min(T MaxColor) 
			{
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(0.0,-500.0,-200.0));
			}
			template<int n,typename T>
			static T sample_max(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(100.0,500.0,200.0));
			}
			template<int n,typename T>
			static T sample_maxabs(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(Vtl::Alg::ReturnArg<n>(100.0,500.0,200.0));
			}
			/// specific for cromatic-like color systems, =sqrt(c1^2+c2^2)
			template<typename T>
			static T sample_max_saturation(T MaxColor) 
			{ 
				LOKI_STATIC_CHECK(Loki::TypeTraits<T>::isFloat,Type_must_be_one_of_floating_point_types);
				return T(538.5164807135);
			}
			template<int n>
			static bool is_sample_sign_alternating() { return Vtl::Alg::Return1stArg<n==0>(false,true);}
			template<int n>
			static bool is_sample_cyclic_var() { return false;}

			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,Lab,nDefault>,typename Iter=T&>
			class fromRGB
			{
			public:
				typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
				typedef Lab ColorSystem;
				NormType norm;

			private:
				T_temp x0;
				T_temp y0;
				T_temp z0;
			private:
				/*!
				 * 
				 *
				 **************************************************************************/
				template<typename F>
				static F f(F w) 
				{
					if(w >  (F)(0.0088564516790356308171716757554635) )
						return pow(w,(F)(0.333333333333333333333));
					else
						return (F)(7.787037037037037037037037037037*w+0.13793103448275862068965517241379);
				}
			public:

				fromRGB(NormType normalizer=Normalizer<T,Lab,nDefault>(255,0,128)) 
					: norm(normalizer) 
				{
					x0=XYZ::X_f<T_temp>(normalizer.MaxColor(),normalizer.MaxColor(),normalizer.MaxColor());
					y0=XYZ::Y_f<T_temp>(normalizer.MaxColor(),normalizer.MaxColor(),normalizer.MaxColor());
					z0=XYZ::Z_f<T_temp>(normalizer.MaxColor(),normalizer.MaxColor(),normalizer.MaxColor());
				}

				void operator()(T r,T g,T b,Iter lc,Iter ac,Iter bc) const 
				{
					T_temp x=XYZ::X_f<T_temp>(r,g,b)/x0;
					T_temp y=XYZ::Y_f<T_temp>(r,g,b)/y0;
					T_temp z=XYZ::Z_f<T_temp>(r,g,b)/z0;
					lc=norm.fromRGB<0>(116*f(y)-16);
					ac=norm.fromRGB<1>(500*(f(x)-f(y)));
					bc=norm.fromRGB<2>(200*(f(y)-f(z)));
					//T_temp n=2.0;
					//T_temp ofs=128;
					//lc=lc*2+16;
					//ac=ac*n+ofs;
					//bc=bc*1.25f+ofs;
				}
			};
			/*!
			* \brief   PixByPix3to3.
			*******************************************************************************************/
			template<typename T,class NormType=Normalizer<T,Lab,nDefault>,typename Iter=T&>
			class toRGB
			{
			public:
				typedef typename Loki::Select<Loki::TypeTraits<T>::isFloat,T,double>::Result T_temp;
				typedef Lab ColorSystem;
				NormType norm;
			public:
				T_temp x0;
				T_temp y0;
				T_temp z0;
			private:
				/*!
				 * 
				 *
				 **************************************************************************/
				template<typename F>
				static F f_rev(F w) 
				{
					//delta=0.20689655172413793103448275862069
					//3*delta^2=0.12841854934601664684898929845422
					if(w>(F)(0.20689655172413793103448275862069))
						return pow(w,(F)(3));
					else
						return (F)((w-0.13793103448275862068965517241379)*0.12841854934601664684898929845422);
				}
			public:

				toRGB(NormType normalizer=Normalizer<T,Lab,nDefault>(255,0,128)) 
					: norm(normalizer) 
				{
					x0=XYZ::X_f<T_temp>(normalizer.MaxColor(),normalizer.MaxColor(),normalizer.MaxColor());
					y0=XYZ::Y_f<T_temp>(normalizer.MaxColor(),normalizer.MaxColor(),normalizer.MaxColor());
					z0=XYZ::Z_f<T_temp>(normalizer.MaxColor(),normalizer.MaxColor(),normalizer.MaxColor());
				}

				void operator()(T lc,T ac,T bc,Iter r,Iter g,Iter b) const 
				{
					T_temp y=(lc+16.0f)/116;
					T_temp x=y+ac/500.0f;
					T_temp z=y-bc/200.0f;
					y=f_rev(y)*y0;
					x=f_rev(y)*x0;
					z=f_rev(y)*z0;
					r=norm.toRGB<0>(XYZ::R_f<T_temp>(x,y,z));
					g=norm.toRGB<1>(XYZ::G_f<T_temp>(x,y,z));
					b=norm.toRGB<2>(XYZ::B_f<T_temp>(x,y,z));
				}
			};
		};
	}
/** @} */ // end of ElementaryFiltersDoxGroup  
}//namespace ImAlg
}//namespace Vtl
////////////////////////////////////////////////////////////////////////////////////
#endif //_COLOR_SPACES_H
