#ifndef _IMALG_TOOLS_H 
#define _IMALG_TOOLS_H 
////////////////////////////////////////////////////////////////////////////////////////////////
#define _USE_MATH_DEFINES
#include <math.h>
#include <alexandresku.h>
#include "plf_strip_buffer_support.h"
////////////////////////////////////////////////////////////////////////////////////////////////
namespace Vtl 
{
namespace ImAlg 
{

/*!
 * \brief C   ,    , 
 *  . ,    , - 
 *      ...
 * 
 *
 *****************************************************************************************************/
namespace Tools 
{
////////////////////////////////////////////////////////////////////////////////////////////////

	template<typename T> 
	void xchg(T& v1,T& v2) 
	{
		T v =v1;
		v1=v2;
		v2=v;
	}


////////////////////////////////////////////////////////////////////////////////////////////////
	template <typename RandomIter, class F> 
	void copy_1_1(RandomIter src_dst,size_t width,size_t height,int nls,F functor )
	{
		nls-=width;
		for( ; height-- ; src_dst += nls )
		{
			for(int x=width; x-- ; ++src_dst)
				*src_dst = functor(*src_dst);
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_1_1(ConstRandomIter src,size_t width,size_t height,int nls_s,RandomIter dst,int nls_d,F functor )
	{
		nls_s-=width;
		nls_d-=width;
		for( ; height-- ; src+=nls_s, dst+=nls_d )
		{
			for(int x=width; x-- ;)
				*dst++ = functor(*src++);
		}
	}
	template <typename ConstRandomIter1, typename ConstRandomIter2, typename RandomIter, class F> 
	void copy_2_1(ConstRandomIter1 src_1,size_t width,size_t height,int nls_s1,ConstRandomIter2 src_2,int nls_s2,RandomIter dst,int nls_d,F functor )
	{
		nls_s1-=width;
		nls_s2-=width;
		nls_d-=width;
		for( ; height-- ; src_1+=nls_s1, src_2+=nls_s2, dst+=nls_d )
		{
			for(int x=width; x-- ;)
				*dst++ = functor(*src_1++,*src_2++);
		}
	}
	template <typename ConstRandomIter1, typename ConstRandomIter2, typename RandomIter, class F> 
	void copy_2_2(ConstRandomIter1 src_1,size_t width,size_t height,int nls_s1,ConstRandomIter2 src_2,int nls_s2,RandomIter dst1,int nls_d1,RandomIter dst2,int nls_d2,F functor )
	{
		nls_s1-=width;
		nls_s2-=width;
		nls_d1-=width;
		nls_d2-=width;
		for( ; height-- ; src_1+=nls_s1, src_2+=nls_s2, dst1+=nls_d1, dst2+=nls_d2 )
		{
			for(int x=width; x-- ;)
				functor(*src_1++,*src_2++,*dst1++,*dst2++);
		}
	}
	template <typename ConstRandomIter1, typename ConstRandomIter2, typename RandomIter, class F> 
	void copy_2_3(ConstRandomIter1 src_1,size_t width,size_t height,int nls_s1,ConstRandomIter2 src_2,int nls_s2,RandomIter dst1,int nls_d1,RandomIter dst2,int nls_d2,RandomIter dst3,int nls_d3,F functor )
	{
		nls_s1-=width;
		nls_s2-=width;
		nls_d1-=width;
		nls_d2-=width;
		nls_d3-=width;
		for( ; height-- ; src_1+=nls_s1, src_2+=nls_s2, dst1+=nls_d1, dst2+=nls_d2 , dst3+=nls_d3 )
		{
			for(int x=width; x-- ;)
				functor(*src_1++,*src_2++,*dst1++,*dst2++,*dst3++);
		}
	}
	template <typename ConstRandomIter1, typename ConstRandomIter2, typename RandomIter, class F> 
	void copy_2_4(ConstRandomIter1 src_1,size_t width,size_t height,int nls_s1,ConstRandomIter2 src_2,int nls_s2,RandomIter dst1,int nls_d1,RandomIter dst2,int nls_d2,RandomIter dst3,int nls_d3,RandomIter dst4,int nls_d4,F functor )
	{
		nls_s1-=width;
		nls_s2-=width;
		nls_d1-=width;
		nls_d2-=width;
		nls_d3-=width;
		nls_d4-=width;
		for( ; height-- ; src_1+=nls_s1, src_2+=nls_s2, dst1+=nls_d1, dst2+=nls_d2 , dst3+=nls_d3 , dst4+=nls_d4 )
		{
			for(int x=width; x-- ;)
				functor(*src_1++,*src_2++,*dst1++,*dst2++,*dst3++,*dst4++);
		}
	}
	template <typename ConstRandomIter1, typename ConstRandomIter2, typename ConstRandomIter3, typename RandomIter, class F> 
	void copy_3_1(ConstRandomIter1 src_1,size_t width,size_t height,int nls_s1,ConstRandomIter2 src_2,int nls_s2,ConstRandomIter3 src_3,int nls_s3,RandomIter dst,int nls_d,F functor )
	{
		nls_s1-=width;
		nls_s2-=width;
		nls_s3-=width;
		nls_d-=width;
		for( ; height-- ; src_1+=nls_s1, src_2+=nls_s2, src_3+=nls_s3, dst+=nls_d )
		{
			for(int x=width; x-- ;)
				*dst++ = functor(*src_1++,*src_2++,*src_3++);
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_3_2(ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, RandomIter out1, int nls1o, RandomIter out2, int nls2o, F functor)
	{
		for(int y=0; y<height; y++)
		{
			for(int x=0; x<width; x++)
				functor(in1[x],in2[x],in3[x],out1[x],out2[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			out1+=nls1o;
			out2+=nls2o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_3_3(ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, RandomIter out1, int nls1o, RandomIter out2, int nls2o, RandomIter  out3, int nls3o, F functor)
	{
		for(int y=0; y<height; y++)
		{
			for(int x=0; x<width; x++)
				functor(in1[x],in2[x],in3[x],out1[x],out2[x],out3[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			out1+=nls1o;
			out2+=nls2o;
			out3+=nls3o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_3_4(ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, RandomIter out1, int nls1o, RandomIter out2, int nls2o, RandomIter  out3, int nls3o, RandomIter  out4, int nls4o, F functor)
	{
		for(int y=0; y<height; y++)
		{
			for(int x=0; x<width; x++)
				functor(in1[x],in2[x],in3[x],out1[x],out2[x],out3[x],out4[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			out1+=nls1o;
			out2+=nls2o;
			out3+=nls3o;
			out4+=nls4o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_3_5(ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, RandomIter out1, int nls1o, RandomIter out2, int nls2o, RandomIter  out3, int nls3o, RandomIter  out4, int nls4o,  RandomIter  out5, int nls5o, F functor)
	{
		for(int y=0; y<height; y++)
		{
			for(int x=0; x<width; x++)
				functor(in1[x],in2[x],in3[x],out1[x],out2[x],out3[x],out4[x],out5[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			out1+=nls1o;
			out2+=nls2o;
			out3+=nls3o;
			out4+=nls4o;
			out5+=nls5o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_5_1(ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, ConstRandomIter in4, int nls4, ConstRandomIter in5, int nls5, RandomIter  out1, int nls1o, F functor)
	{
		for(int y=0; y<height; y++)
		{
			for(int x=0; x<width; x++)
				out1[x]=functor(in1[x],in2[x],in3[x],in4[x],in5[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			in4+=nls4;
			in5+=nls5;
			out1+=nls1o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void copy_6_1(ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, ConstRandomIter in4, int nls4, ConstRandomIter in5, int nls5,  ConstRandomIter in6, int nls6, RandomIter  out1, int nls1o, F functor)
	{
		for(int y=0; y<height; y++)
		{
			for(int x=0; x<width; x++)
				out1[x]=functor(in1[x],in2[x],in3[x],in4[x],in5[x],in6[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			in4+=nls4;
			in5+=nls5;
			in6+=nls6;
			out1+=nls1o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void envTransf_3e1_1(Vtl::Plf::Extent ext,ConstRandomIter in1, int width, int height, int nls1, ConstRandomIter in2, int nls2, ConstRandomIter in3, int nls3, RandomIter out1, int nls1o, F functor)
	{
		width-=ext.right;
		height-=ext.forward;
		for(int y=0; y<ext.before; ++y)
		{
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			out1+=nls1o;
		}
		functor.set_IterNextLineShift(nls1);

		for(int y=ext.before; y<height; y++)
		{
			for(int x=ext.left; x<width; x++)
				out1[x]=functor(in1+x,in2[x],in3[x]);
			in1+=nls1;
			in2+=nls2;
			in3+=nls3;
			out1+=nls1o;
		}
	}
	template <typename ConstRandomIter, typename RandomIter, class F> 
	void envTransf_1_1(Vtl::Plf::Extent ext,ConstRandomIter in1, int width, int height, int nls1, RandomIter out1, int nls1o, F functor)
	{
		width-=ext.right;
		height-=ext.forward;
		for(int y=0; y<ext.before; ++y)
		{
			in1+=nls1;
			out1+=nls1o;
		}
		functor.set_IterNextLineShift(nls1);

		for(int y=ext.before; y<height; y++)
		{
			for(int x=ext.left; x<width; x++)
				out1[x]=functor(in1+x);
			in1+=nls1;
			out1+=nls1o;
		}
	}

////////////////////////////////////////////////////////////////////////////////////////////////
	namespace findMax
	{
		template<int num_equal_neihbours>
		struct is_center_max_in3x3
		{
			template<typename T1,class F,typename T2>
			void operator()(const T1* src,int left,int right,int up,int down,F functor,T2 x,T2 y)
			{
				T1 v=src[0];
				if(v < src[left] )  return;
				if(v < src[right] ) return;
				if(v < src[up] )	return;
				if(v < src[down] )	return;
				if(v < src[up+left] )   return;
				if(v < src[up+right] )   return;
				if(v < src[down+left] ) return;
				if(v < src[down+right] ) return;
				//now central point is not less than 8 neighbours,
				// check for strict maximum
				int cnt=0;			
				if(v == src[left] )  ++cnt;
				if(v == src[right] ) ++cnt;
				if(v == src[up] )	 ++cnt;
				if(v == src[down] )	 ++cnt;
				if(cnt > num_equal_neihbours)
					return;
				if(v == src[up+left] )    ++cnt;
				if(v == src[up+right] )   ++cnt;
				if(v == src[down+left] )  ++cnt;
				if(v == src[down+right] ) ++cnt;
				if(cnt <= num_equal_neihbours)
					functor(x,y,v);
			}
		};
		namespace Impl
		{
			template<int stride,class ICM,typename T1,class F>
			const T1* InRow_BoundsNo(const T1* src,int y,int width,int up,int down,ICM& is_center_max_in3x3_,F functor)
			{
				src+=stride;
				--width;
				for(int x=1; x<width; ++x ,src+=stride )
					is_center_max_in3x3_(src,-stride,stride,up,down,functor,x,y);
				return src;
			}
			template<int stride,class ICM,typename T1,class F>
			void InRow_BoundsCyclic(const T1* src,int y,int width,int up,int down,ICM& is_center_max_in3x3_,F functor)
			{
				is_center_max_in3x3_(src,(width-1)*stride,stride,up,down,functor,0,y);
				src=InRow_BoundsNo<stride>(src,y,width,up,down,is_center_max_in3x3_,functor);
				is_center_max_in3x3_(src,-stride,-(width-1)*stride,up,down,functor,(width-1),y);
			}
		}
		template<typename T,int stride=1>
		class SetPixel
		{
			T* src;
			int nls;
		public:
			SetPixel(T* data_ptr,int nls_) : src(data_ptr),nls(nls_) {}
			template<typename C,typename V>
			void operator()(C x,C y,V v)
			{
				src[y*nls+x*stride]=v;
			}
		};
		template<int stride,class ICM,typename T1,class F>
		void BoundsNo(const T1* src,size_t width,size_t height,int nls,ICM& is_center_max_in3x3_, F functor )
		{
			src+=nls;
			for( int y=1 ; y < int(height-1) ; ++y , src+=nls )
				Impl::InRow_BoundsNo<stride>(src,y,width,-nls,nls,is_center_max_in3x3_,functor );
		}
		template<int stride,class ICM,typename T1,class F>
		void BoundsCyclic(const T1* src,size_t width,size_t height,int nls,ICM& is_center_max_in3x3_, F functor )
		{
			Impl::InRow_BoundsCyclic<stride>(src,0,width,nls*(height-1),nls,is_center_max_in3x3_,functor );
			src+=nls;
			for( int y=1 ; y < int(height-1) ; ++y , src+=nls )
				Impl::InRow_BoundsCyclic<stride>(src,y,width,-nls,nls,is_center_max_in3x3_,functor );
			Impl::InRow_BoundsCyclic<stride>(src,int(height-1),width,-nls,-nls*(height-1),is_center_max_in3x3_,functor );
		}
	}

////////////////////////////////////////////////////////////////////////////////////////////////
}//namespace Tools 
}//namespace ImAlg 
}//namespace Vtl
////////////////////////////////////////////////////////////////////////////////////////////////
#endif //_IMALG_TOOLS_H 
