
#include "AutoCalibration.h"

//#define USE_KRUPPA
#define USE_DIRECT_SEARCH

namespace Basics3D
{
	// After debugging essential matrix out_E won't be needed to be returned
	void C_AutoCalibration::GetRightP
		                (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4>> &out_P,
										 Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_K,
										 Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_K1,
										 Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_E,
										 const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F)
	{
		std::cout << std::endl << std::endl << "GetRightP";

		// obtain focal length
#ifdef USE_KRUPPA
		double f[2];
		double f1[2];
		FocalLength(f, f1, in_F);
		std::cout << std::endl << std::endl << "f[0] = " << f[0] << "  f[1] = " << f[1];
		std::cout << std::endl << std::endl << "f1[0] = " << f1[0] << "  f1[1] = " << f1[1];

		// compose calibration matrices
		// (both variants of essential matrix will be printed
		//  but only the last (corresponding to f[0] and f1[0]
		//  will be used futher)

		// use f[1] and f1[1]
		out_K(0,0) = sqrt(fabs(f[1])); out_K(0,1) = 0.0;              out_K(0,2) = 0.0;
		out_K(1,0) = 0.0;              out_K(1,1) = sqrt(fabs(f[1])); out_K(1,2) = 0.0;
		out_K(2,0) = 0.0;              out_K(2,1) = 0.0;              out_K(2,2) = 1.0;

		out_K1(0,0) = sqrt(fabs(f1[1])); out_K1(0,1) = 0.0;               out_K1(0,2) = 0.0;
		out_K1(1,0) = 0.0;               out_K1(1,1) = sqrt(fabs(f1[1])); out_K1(1,2) = 0.0;
		out_K1(2,0) = 0.0;               out_K1(2,1) = 0.0;               out_K1(2,2) = 1.0;

		//std::cout << std::endl << std::endl << "K" << std::endl;
		//T_MatrixOperations::Print(out_K);
		//std::cout << std::endl << std::endl << "K1" << std::endl;
		//T_MatrixOperations::Print(out_K1);

		EssentialMatrix(out_E, in_F, out_K, out_K1);
		std::cout << std::endl << std::endl << "E" << std::endl;
		T_MatrixOperations::Print(E);

		// use f[0] and f1[0]
		out_K(0,0) = sqrt(fabs(f[0])); out_K(0,1) = 0.0;              out_K(0,2) = 0.0;
		out_K(1,0) = 0.0;              out_K(1,1) = sqrt(fabs(f[0])); out_K(1,2) = 0.0;
		out_K(2,0) = 0.0;              out_K(2,1) = 0.0;              out_K(2,2) = 1.0;

		out_K1(0,0) = sqrt(fabs(f1[0])); out_K1(0,1) = 0.0;               out_K1(0,2) = 0.0;
		out_K1(1,0) = 0.0;               out_K1(1,1) = sqrt(fabs(f1[0])); out_K1(1,2) = 0.0;
		out_K1(2,0) = 0.0;               out_K1(2,1) = 0.0;               out_K1(2,2) = 1.0;

		//std::cout << std::endl << std::endl << "K" << std::endl;
		//T_MatrixOperations::Print(out_K);
		//std::cout << std::endl << std::endl << "K1" << std::endl;
		//T_MatrixOperations::Print(out_K1);

		EssentialMatrix(out_E, in_F, out_K, out_K1);
		std::cout << std::endl << std::endl << "E" << std::endl;
		T_MatrixOperations::Print(out_E);

#elif defined(USE_DIRECT_SEARCH)
		double fl = DirectSearchOfFocalLenth(in_F);
		
		out_K(0,0) = fl;  out_K(0,1) = 0.0; out_K(0,2) = 0.0;
		out_K(1,0) = 0.0; out_K(1,1) = fl;  out_K(1,2) = 0.0;
		out_K(2,0) = 0.0; out_K(2,1) = 0.0; out_K(2,2) = 1.0;

		out_K1 = out_K;

		EssentialMatrix(out_E, in_F, out_K, out_K);
		std::cout << std::endl << std::endl << "E" << std::endl; T_MatrixOperations::Print(out_E);
#else
#error "unrealized, use USE_KRUPPA or USE_DIRECT_SEARCH"
#endif
		// obtain camera matrices
		C_ProjectiveMatrix::Right(out_P, out_E);

		// print obtained camera matrices
		// and calibration and rotation matrices corresponding to each of them
		for (int i = 0; i < 4; i++)
		{
			std::cout << std::endl << std::endl << "P'" << i << std::endl; T_MatrixOperations::Print(out_P[i]);

			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K;
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> R;
			Vtl::NumAlg::Lin::ublas::c_vector<double, 3> t;
			DecomposeProjMatr(K, R, t, out_P[i]);
		}		
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_AutoCalibration::DirectSearchOfFocalLenth
			                     (const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F)
	{
		// check focal lenthes from -10 to 10 to be the most proper
		// and output estimation of closeness of singular values
		// of ess. m. to file
		std::ofstream focus;
		focus.open("focus.txt", std::ios::out);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> E;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K;

		double min_d = 1;
		double min_f = 10.0;
		double min_sv[2];
		
		// coarse estimation of focal lenth (with the step 0.1)
		for (double fl = 0.0; fl < 10.0; fl += 0.1)
		{
			// compose calibration matrix
			K(0,0) = fl;  K(0,1) = 0.0; K(0,2) = 0.0;
			K(1,0) = 0.0; K(1,1) = fl;  K(1,2) = 0.0;
			K(2,0) = 0.0; K(2,1) = 0.0; K(2,2) = 1.0;

			// obtain essential matrix
			EssentialMatrix(E, in_F, K, K);
			
			double sv[2];
			SingularValues(sv, E);
			double d = fabs(sv[0] - sv[1]) / (sv[0] + sv[1]);
			focus << fl << " " << d << "\n";

			if (d < min_d)
			{
				min_d = d;
				min_f = fl;
				min_sv[0] = sv[0];
				min_sv[1] = sv[1];
			}
		}
		focus.close();

		std::cout << std::endl << std::endl << "Coarse estimation of focal lenth" << std::endl << 
			"d = " << min_d << " f = " << min_f << "  sv: " << min_sv[0] << " " << min_sv[1];

		// fine estimation of focal lenth (with the step 0.001)
		for (double fl = min_f - 0.5; fl < min_f + 0.5; fl += 0.001)
		{
			// compose calibration matrix
			K(0,0) = fl;  K(0,1) = 0.0; K(0,2) = 0.0;
			K(1,0) = 0.0; K(1,1) = fl;  K(1,2) = 0.0;
			K(2,0) = 0.0; K(2,1) = 0.0; K(2,2) = 1.0;

			// obtain essential matrix
			EssentialMatrix(E, in_F, K, K);
			
			double sv[2];
			SingularValues(sv, E);
			double d = fabs(sv[0] - sv[1]) / (sv[0] + sv[1]);
			
			if (d < min_d)
			{
				min_d = d;
				min_f = fl;
				min_sv[0] = sv[0];
				min_sv[1] = sv[1];
			}
		}

		std::cout << std::endl << std::endl << "Fine estimation of focal lenth" << std::endl << 
			"d = " << min_d << " f = " << min_f << "  sv: " << min_sv[0] << " " << min_sv[1];

		return min_f;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_AutoCalibration::FocalLength
		                        (double out_f[2],
														 double out_f1[2],
		                         const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_F)
	{
		// SVD decomposition of the fundamental matrix
		//const int N=3;
		enum {N=3};
		Vtl::NumAlg::Lin::ublas::c_matrix<double, N, N> U;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, N, N> V;
		Vtl::NumAlg::Lin::ublas::c_vector<double, N> W;
		Vtl::NumAlg::Lin::ublas::c_vector<double, N> RV1;

		U = in_F;

		std::cout << std::endl << std::endl << "FM" << std::endl;
		T_MatrixOperations::Print(U);

		Vtl::NumAlg::Lin::svd::decomp(U, W, V, RV1);

		std::cout << std::endl << std::endl << "U" << std::endl;
		T_MatrixOperations::Print(U);
		std::cout << std::endl << std::endl << "W" << std::endl;
		std::cout << W(0) << "  " << W(1) << "  " << W(2);
		std::cout << std::endl << std::endl << "V" << std::endl;
		T_MatrixOperations::Print(V);

		// form equation
		SortS<double> s[3];
		s[0]=SortS<double>(W(0),0);
		s[1]=SortS<double>(W(1),1);
		s[2]=SortS<double>(W(2),2);
		std::sort(s,s+3);

		const int k1 = s[2].index;
		const int k2 = s[1].index;
		std::cout << std::endl << std::endl << "k1 = " << k1 << "  k2 = " << k2;
		const double s1 = W(k1);
		const double s2 = W(k2);

		/*double p = U(k1,0) * U(0,k2) + U(k1,1) * U(1,k2);
		double q = U(k2,0) * U(0,k2) + U(k2,1) * U(1,k2);
		double f = V(k1,0) * V(0,k1) + V(k1,1) * V(1,k1);
		double h = V(k1,0) * V(0,k2) + V(k1,1) * V(1,k2);

		double a = s1 * s2 * h * q + s1 * s1 * f * p;
		double b = s1 * s1 * (p * V(k1,2) * V(2,k1) + f * U(k1,2) * U(2,k2))
			       + s1 * s2 * (h * U(k2,2) * U(2,k2) + q * V(k1,2) * V(2,k2));
		double c = s1 * U(2,k2) * V(k1,2) * (s2 * U(k2,2) * V(2,k2) + s1 * U(k1,2) * V(2,k1));*/

		//double p = U(k1,0) * U(0,k2) + U(k1,1) * U(1,k2);
		//double q = U(k2,0) * U(0,k2) + U(k2,1) * U(1,k2);
		//double r = U(k1,0) * U(0,k1) + U(k1,1) * U(1,k1);
		//double f = V(k1,0) * V(0,k1) + V(k1,1) * V(1,k1);
		//double h = V(k1,0) * V(0,k2) + V(k1,1) * V(1,k2);
		//double g = V(k2,0) * V(0,k2) + V(k2,1) * V(1,k2);

		//// A1 * x * y + B1 * y + C1 * x + D1 = 0;
		//double A1 = s1 * s2 * q * h + s1 * s1 * f * p;
		//double B1 = s1 * V(k1,2) * (s2 * q * V(2,k2) + s1 * p * V(2,k1));
		//double C1 = s1 * U(2,k2) * (s2 * h * U(k2,2) + s1 * f * U(k1,2));
		//double D1 = s1 * V(k1,2) * U(2,k2) * (s1 * V(2,k1) * U(k1,2) + s2 * U(k2,2) * V(2,k2));

		//// A2 * x * y + B2 * y + C2 * x + D2 = 0;
		//double A2 = s2 * s2 * p * g + s1 * s2 * h * r;
		//double B2 = s2 * V(2,k2) * (s2 * p * V(k2,2) + s1 * r * V(k1,2));
		//double C2 = s2 * U(k1,2) * (s2 * g * U(2,k2) + s1 * h * U(2,k1));
		//double D2 = s2 * U(k1,2) * V(2,k2)*(s2 * U(2,k2) * V(k2,2) + s1 * V(k1,2) * U(2,k1));

		double u12 = U(0,k1) * U(0,k2) + U(1,k1) * U(1,k2);
		double u_2  = U(0,k2) * U(0,k2) + U(1,k2) * U(1,k2);
		double u_1  = U(0,k1) * U(0,k1) + U(1,k1) * U(1,k1);
		double v_1  = V(0,k1) * V(0,k1) + V(1,k1) * V(1,k1);
		double v12 = V(0,k1) * V(0,k2) + V(1,k1) * V(1,k2);
		double v_2  = V(0,k2) * V(0,k2) + V(1,k2) * V(1,k2);

		// A1*x*y + B1*y + C1*x + D1 = 0;
		double A1 = s1 * s2 * u_2 * v12 + s1 * s1 * v_1 * u12;
		double B1 = s1 * s2 * u_2 * V(2,k1) * V(2,k2) + s1 * s1 * u12 * V(2,k1) * V(2,k1);
		double C1 = s1 * s2 * v12 * U(2,k2) * U(2,k2) + s1 * s1 * v_1 * U(2,k1) * U(2,k2);
		double D1 = s1 * s1 * U(2,k1) * U(2,k2) * V(2,k1) * V(2,k1)
			      + s1 * s2  * U(2,k2) * U(2,k2)* V(2,k1) * V(2,k2);

		// A2 * x * y + B2 * y + C2 * x + D2 = 0;
		double A2 = s2 * s2 * u12 * v_2 + s1 * s2 * v12 * u_1;
		double B2 = s2 * s2 * u12 * V(2,k2) * V(2,k2) + s1 * s2 * u_1 * V(2,k1) * V(2,k2);
		double C2 = s2 * s2 * v_2 * U(2,k1) * U(2,k2) + s1 * s2 * v12 * U(2,k1) * U(2,k1);
		double D2 = s2 * s2 * U(2,k1) * U(2,k2) * V(2,k2) * V(2,k2)
			      + s1 * s2 * U(2,k1) * U(2,k1) * V(2,k2) * V(2,k1);

		// x^2 * (A2*C1 - A1*C2) + x * (-A1*D2 - B1*C2 + A2*D1 + B2*C1) + (B2*D1 - B1*D2) = 0;
		double a = A2 * C1 - A1 * C2;
		double b = -A1 * D2 - B1 * C2 + A2 * D1 + B2 * C1;
		double c = B2 * D1 - B1 * D2;
		std::cout << std::endl << std::endl << "A1 = " << A1 << "  B1 = " << B1 << "  C1 = "
			<< C1 << "  D1 = " << D1 << std::endl << "A2 = " << A2 << "  B2 = " << B2
			<< "  C2 = " << C2 << "  D2 = " << D2;
		std::cout << std::endl << std::endl << "a = " << a << "  b = " << b << "  c = " << c;

		// solve equation
		double eps = 0.00000001;
		if (fabs(a) < eps)
		{
			if (fabs(b) > eps)
			{
				/*temp*/std::cout << std::endl << "The only root";
				out_f[0] = c / b;
				//out_a[0] = sqrt(out_a[0]);
			}
			else
			{
				/*temp*/std::cout << std::endl << "No roots";
				out_f[0] = 1.0;
			}

			out_f[1] = 1.0;
		}
		else
		{
			double D = b * b - 4.0 * a * c;
			/*temp*/std::cout << std::endl << "  D = " << D;
			D = sqrt(D);
			/*temp*/std::cout << "  sqrt(D) = " << D;
			out_f[0] = (-b + D) / (2.0 * a);
			out_f[1] = (-b - D) / (2.0 * a);
			std::cout << std::endl << std::endl << "sqr(f1) = " << out_f[0]
				<< "  sqr(f2) = " << out_f[1];
	
			/*for (int i = 0; i < 2; i++)
			{
				if (out_a[i] > 0.0)
				{
					out_a[i] = sqrt(out_a[i]);
				}
				else
				{
					out_a[i] = 1.0;
				}
			}*/
		}

		std::cout << std::endl << std::endl << "quad eq: ";
		for (int i = 0; i < 2; i ++)
		{
			std::cout << a * out_f[i] * out_f[i] + b * out_f[i] + c << "  ";
		}
		// y = - (C2 * x + D2) / (A2 * x + B2)
		
		for (int i = 0; i < 2; i++)
		{
			out_f1[i] = - (C2 * out_f[i] + D2) / (A2 * out_f[i] + B2);
			std::cout << std::endl << std::endl << "eq1: "
				<< A1 * out_f[i] * out_f1[i] + B1 * out_f1[i] + C1 * out_f[i] + D1 << "  "
				<< A2 * out_f[i] * out_f1[i] + B2 * out_f1[i] + C2 * out_f[i] + D2;
		}
		std::cout << std::endl << std::endl << "sqr(y) = " << out_f1[0]
			<< "  sqr(y) = " << out_f1[1];
		
			using namespace  Vtl::NumAlg::Lin;
		// check 
		ublas::c_matrix<double, 3, 1> u1;
		u1(0, 0) = U(0, k1); u1(1, 0) = U(1, k1); u1(2, 0) = U(2, k1);

		ublas::c_matrix<double, 1, 3> u1T;
		u1T = boost::numeric::ublas::trans(u1); //u1T(0, 0) = U(k1, 0); u1T(0, 1) = U(k1, 1); u1T(0, 2) = U(k1, 2);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> u2;
		u2(0, 0) = U(0, k2); u2(1, 0) = U(1, k2); u2(2, 0) = U(2, k2);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 1, 3> u2T;
		u2T = boost::numeric::ublas::trans(u2); //u2T(0, 0) = U(k2, 0); u2T(0, 1) = U(k2, 1); u2T(0, 2) = U(k2, 2);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> v1;
		v1(0, 0) = V(0, k1); v1(1, 0) = V(1, k1); v1(2, 0) = V(2, k1);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 1, 3> v1T;
		v1T = boost::numeric::ublas::trans(v1); //v1T(0, 0) = V(k1, 0); v1T(0, 1) = V(k1, 1); v1T(0, 2) = V(k1, 2);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> v2;
		v2(0, 0) = V(0, k2); v2(1, 0) = V(1, k2); v2(2, 0) = V(2, k2);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 1, 3> v2T;
		v2T = boost::numeric::ublas::trans(v2); //v2T(0, 0) = V(k2, 0); v2T(0, 1) = V(k2, 1); v2T(0, 2) = V(k2, 2);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> w;
		w(0,0) = out_f[1]; w(0,1) = 0.0;      w(0,2) = 0.0;
		w(1,0) = 0.0;      w(1,1) = out_f[1]; w(1,2) = 0.0;
		w(2,0) = 0.0;      w(2,1) = 0.0;      w(2,2) = 1.0;

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> w1;
		w1(0,0) = out_f1[1]; w1(0,1) = 0.0;       w1(0,2) = 0.0;
		w1(1,0) = 0.0;       w1(1,1) = out_f1[1]; w1(1,2) = 0.0;
		w1(2,0) = 0.0;       w1(2,1) = 0.0;       w1(2,2) = 1.0;

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 1, 3> tmp13;
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 1, 1> tmp11;

		tmp13 = boost::numeric::ublas::prod(u2T, w1);
		tmp11 = boost::numeric::ublas::prod(tmp13, u2);
		double A = tmp11(0, 0);

		tmp13 = boost::numeric::ublas::prod(v1T, w);
		tmp11 = boost::numeric::ublas::prod(tmp13, v2);
		tmp11 = tmp11 * s1;
		tmp11 = tmp11 * s2;
		double B = tmp11(0, 0);

		tmp13 = boost::numeric::ublas::prod(v1T, w);
		tmp11 = boost::numeric::ublas::prod(tmp13, v1);
		tmp11 = tmp11 * s1;
		tmp11 = tmp11 * s1;
		double C = tmp11(0, 0);

		tmp13 = boost::numeric::ublas::prod(u1T, w1);
		tmp11 = boost::numeric::ublas::prod(tmp13, u2);
		double D = tmp11(0, 0);

		tmp13 = boost::numeric::ublas::prod(u1T, w1);
		tmp11 = boost::numeric::ublas::prod(tmp13, u1);
		double E = tmp11(0, 0);

		tmp13 = boost::numeric::ublas::prod(v2T, w);
		tmp11 = boost::numeric::ublas::prod(tmp13, v2);
		tmp11 = tmp11 * s2;
		tmp11 = tmp11 * s2;
		double F = tmp11(0, 0);


		std::cout << std::endl << "A / C = " << A / C << "  D / B = " << D / B << "  E / F = " << E / F;



		///*temp*/std::cout << std::endl << std::endl << "Focal length:" << std::endl;
		///*temp*/for (int i = 0; i < 2; i++)
		///*temp*/{
		///*temp*/	std::cout << out_a[i] << "  " 
		///*temp*/		<< a * out_a[i] * out_a[i] * out_a[i] * out_a[i] + 
		///*temp*/		  b * out_a[i] * out_a[i] + c << " | ";
		///*temp*/}
	}


	///////////////////////////////////////////////////////////////////////////////////////
	int C_AutoCalibration::GetH
		                (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4>> &out_H,
			               const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P,
			               const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P1)
	{
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 16> Q;
		AbsoluteDualQuadric quadric;
		int nQ = quadric.Get(Q, in_P, in_P1);

		for (int i = 0; i < nQ; i++)
		{
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> Qi;
			
			for (int j = 0; j < 4; j++)
			{
				for (int k = 0; k < 4; k++)
				{
					Qi(j, k) = Q(i, j * 4 + k); 
				}
			}
			std::cout << std::endl << std::endl << "input Q" << i << std::endl;
			T_MatrixOperations::Print(Qi);

			// eigenvalue decomposition for Jacobi's algorithm
			std::vector<double> d(4);
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> v;
			int nrot;
			jacobi(Qi, d, v, &nrot);

			std::cout << std::endl << std::endl << "nrot = " << nrot;

			std::cout << std::endl << std::endl << "v" << std::endl;
			T_MatrixOperations::Print(v);

			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> diag;
			for (int k = 0; k < 4; k++)
			{
				for (int l = 0; l < 4; l++)
				{
					diag(k, l) = 0.0;
				}
				diag(k, k) = d[k];
			}
			
			std::cout << std::endl << std::endl << "d" << std::endl;
			T_MatrixOperations::Print(diag);

			// Check D == V^T * A * V
			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> vT;
			vT = boost::numeric::ublas::trans(v);

			Qi = boost::numeric::ublas::prod(vT, Qi);
			Qi = boost::numeric::ublas::prod(Qi, v);

			std::cout << std::endl << std::endl << "v^T * A * v" << std::endl;
			T_MatrixOperations::Print(Qi);

			// Check inversion
			//std::cout << std::endl << std::endl << "Check inversion" << std::endl;

			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> v_1;
			v_1 = Vtl::NumAlg::Lin::svd::inverse_save(v, 0.000001);
			//std::cout << std::endl << std::endl << "v * v^-1" << std::endl;
			//T_MatrixOperations::Print(boost::numeric::ublas::prod(v, v_1));

			Vtl::NumAlg::Lin::ublas::c_matrix<double, 4, 4> vT_1;
			vT_1 = Vtl::NumAlg::Lin::svd::inverse_save(vT, 0.000001);
			//std::cout << std::endl << std::endl << "v^T * (v^T)^-1" << std::endl;
			//T_MatrixOperations::Print(boost::numeric::ublas::prod(vT, vT_1));

			// Check (v^T)^-1 * D * v^-1 == A
			Qi = boost::numeric::ublas::prod(vT_1, diag);
			Qi = boost::numeric::ublas::prod(Qi, v_1);

			std::cout << std::endl << std::endl << "(v^T)^-1 * d * v^-1" << std::endl;
			T_MatrixOperations::Print(Qi);
			std::cout << "";

			// d -> (1, 1, 1, 0)
			/*double s0 = sqrt(fabs(d[0]));
			double s1 = sqrt(fabs(d[1]));
			double s2 = sqrt(fabs(d[2]));
			std::cout << std::endl << std::endl << "sqrt(d(1,1)) = " << s0 <<
				"   sqrt(d(2,2)) = " << s1 << "   sqrt(d(3,3)) = " << s2;

			std::cout << std::endl << std::endl << "(v^T)^-1" << std::endl;
			T_MatrixOperations::Print(vT_1);

			for (int k = 0; k < 4; k++)
			{
				vT_1(k, 0) *= s0;
				vT_1(k, 1) *= s1;
				vT_1(k, 2) *= s2;
			}
			for (int k = 0; k < 3; k++)
			{
				diag(k, k) /= fabs(d[k]);
			}
			std::cout << std::endl << std::endl << "After scaling"
				<< std::endl << "v" << std::endl;
			T_MatrixOperations::Print(vT_1);
			std::cout << std::endl << std::endl << "d" << std::endl;
			T_MatrixOperations::Print(diag);

			v_1 = boost::numeric::ublas::trans(vT_1);

			Qi = boost::numeric::ublas::prod(vT_1, diag);
			Qi = boost::numeric::ublas::prod(Qi, v_1);

			std::cout << std::endl << std::endl << "(v^T)^-1 * d * v^-1" << std::endl;
			T_MatrixOperations::Print(Qi);*/
			// Move the least d[i] to the last position
			/*double d_min = d[0];
			int k_min = 0;
			for (int k = 1; k < 4; k++)
			{
				if (d[k] < d_min)
				{
					d_min = d[k];
					k_min = k;
				}
			}
			std::cout << std::endl << std::endl << "k_min = " << k_min << std::endl;

			if (k_min != 3)
			{
				for (int k = 0; k < 4; k++)
				{
					double tmp = vT_1(k, 3);
					vT_1(k, 3) = vT_1(k, k_min);
					vT_1(k, k_min) = tmp;
				}
			}
			v_1 = boost::numeric::ublas::trans(vT_1);

			for (int k = 0; k < 4; k++)
			{
				for (int l = 0; l < 4; l++)
				{
					diag(k, l) = 0.0;
				}
				diag(k, k) = d[k];
			}

			std::cout << std::endl << std::endl << "After changing"
				<< std::endl << "d" << std::endl;
			T_MatrixOperations::Print(diag);

			Qi = boost::numeric::ublas::prod(vT_1, diag);
			Qi = boost::numeric::ublas::prod(Qi, v_1);

			std::cout << std::endl << std::endl << "(v^T)^-1 * d * v^-1" << std::endl;
			T_MatrixOperations::Print(Qi);*/


			out_H[i] = vT_1;
			std::cout << std::endl << std::endl << "H" << std::endl;
			T_MatrixOperations::Print(out_H[i]);
		}	
		return nQ;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_AutoCalibration::DecomposeProjMatr
		                        (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_K,
			                       Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_R,
	                           Vtl::NumAlg::Lin::ublas::c_vector<double, 3> &out_t,
			                       const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 4> &in_P)
	{
		// copy the part of the projective matrix
		double dm3 = sqrt(in_P(2, 0) * in_P(2, 0) + 
			                in_P(2, 1) * in_P(2, 1) + in_P(2, 2) * in_P(2, 2));

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> M;
		for (int i = 0; i < 3; i++)
		{
			for (int j = 0; j < 3; j++)
			{
				M(i, j) = in_P(i, j) / dm3;
			}
		}

		// decompose, get calibration K and rotation R matrices
		RQDecomposition(out_K, out_R, M);

		/*temp*/std::cout << std::endl << std::endl << "K" << std::endl;
		/*temp*/T_MatrixOperations::Print(out_K);
		/*temp*/std::cout << std::endl << std::endl << "R" << std::endl;
		/*temp*/T_MatrixOperations::Print(out_R);

		// get translation vector t
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> K_1;
		K_1 = Vtl::NumAlg::Lin::svd::inverse_save(out_K, 0.000000001);

		Vtl::NumAlg::Lin::ublas::c_vector<double, 3> p4;
		/*temp*/std::cout << std::endl << std::endl << "p4 = K*t  ";
		for (int i = 0; i < 3; i++)
		{
			p4(i) = in_P(i, 3);
			/*temp*/std::cout << p4(i) << "  ";
		}
		out_t = boost::numeric::ublas::prod(K_1, p4);

		/*temp*/std::cout << std::endl << std::endl << "Translation t  "
		/*temp*/	<< out_t(0) << "  " << out_t(1) << "  " << out_t(2);
	}
}
