
#include "Combination.h"

namespace Basics3D
{
	/////////////////////////////////////////////////////////////////////////////////////////
	C_Combination::C_Combination(const T_ImagesSet<AImageMem8> &in_img)
	: m_img(in_img)
	{
		int n = m_img.Size();
		
		for (int i = 0; i < n; i++)
		{
			if (m_img[i] == 0)
			{
				throw "\nOne of images was not recieved\n";
				return;
			}
		}

		// create image to combine
		int w = 0;
		int h = m_img[0]->Height();

		for (int i = 0; i < n; i++)
		{
			w += (m_img[i])->Width();
		}

		m_comb = dynamic_cast<AImageMem8 *>
			(::AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, m_img[0]->Bits()));

		if (m_comb == 0)
		{
			throw "\nCan't create combined image\n";
			return;
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	// Makes combination of two member images
	void C_Combination::Make()
	{
		AImageMem8::Pixel *p_c = (*m_comb)[0];
		
		int n = m_img.Size();
		
		std::vector<AImageMem8::Pixel *> p(n);

		for (int i = 0; i < n; i++)
		{
			p[i] = (*(m_img[i]))[0];
		}

		/*Vtl::Plf::F11::DoNothing<AImageMem8::Pixel > func;
		Vtl::ImgAlg::Tools::copy(p_1, m_im1->Width(), m_im1->Height(),
				 m_im1->NextLineShift(), p_c,  m_comb->NextLineShift(), func);
		Vtl::ImgAlg::Tools::copy(p_2, m_im2->Width(), m_im2->Height(),
				 m_im2->NextLineShift(), p_c +m_im1->Width(), m_comb->NextLineShift(), func);*/


		for (int y = 0; y < m_comb->Height(); y++)
		{
			int prevX = 0;
			
			for (int i = 0; i < n; i++)
			{
				for (int x = prevX; x < prevX + m_img[i]->Width(); x++)
				{
					*p_c = *(p[i]);
					p_c++;
					(p[i])++;
				}

				prevX += m_img[i]->Width();
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	// Marks corresponding points on parts of combined image
	// by connection them by white line
	// fileName - *.dist.key - output file of Corr containing
	//                         information about correspondences
	// N - number of correspondences to mark
	//     use -1 to mark all correspondences
	void C_Combination::MarkCorrespondences(const std::vector<T_Points2DSet<C_Point2D>>
		                                                                       &in_PointsSet,
																		      int in_N)
	{
		AImageMem8::Pixel color_pen_wh = AImageMem8::Pixel((1 << m_comb->Bits()) - 1);
		typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor_wh;
		DrawFunctor_wh draw_functor_wh(color_pen_wh);
		typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor_wh> 
																																				SetPixelFunctor_wh;
		SetPixelFunctor_wh set_pix_wh
															((*m_comb)[0], m_comb->NextLineShift(), draw_functor_wh);
		Vtl::ImAlg::Bres::Line<SetPixelFunctor_wh> line_wh(set_pix_wh);


		AImageMem8::Pixel color_pen_bl = AImageMem8::Pixel(0);
		typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor_bl;
		DrawFunctor_bl draw_functor_bl(color_pen_bl);
		typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor_bl> 
																																				SetPixelFunctor_bl;
		SetPixelFunctor_bl set_pix_bl
															((*m_comb)[0], m_comb->NextLineShift(), draw_functor_bl);
		Vtl::ImAlg::Bres::Line<SetPixelFunctor_bl> line_bl(set_pix_bl);

		int n = in_PointsSet.size();
		
		int prevX = 0;

		for (int j = 0; j < n; j++)
		{
			int max = in_PointsSet[j].Size();

			if (in_N == -1 || in_N > max)
			{
				in_N = max;
			}

			for (int i = 0; i < max; i += int(max / in_N))
			{
				if ((in_PointsSet[j])[2 * i].Numb() < 0)
				{
				  continue;
				}
				double x1 = (in_PointsSet[j])[2 * i].X();
				double y1 = /**/m_img[j]->Height() - (in_PointsSet[j])[2 * i].Y();
				double x2 = (in_PointsSet[j])[2 * i + 1].X();
				double y2 = /**/m_img[j]->Height() - (in_PointsSet[j])[2 * i + 1].Y();

				if (j - (j / 2) * 2)
				{
					set_pix_bl.Set0(int(x1) + prevX, int(y1));
					line_bl.Draw(int(x2 - x1 + (m_img[j])->Width()), int(y2 - y1));
				}
				else
				{
					set_pix_wh.Set0(int(x1) + prevX, int(y1));
					line_wh.Draw(int(x2 - x1 + (m_img[j])->Width()), int(y2 - y1));
				}
			}

			prevX += (m_img[j])->Width();
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	void C_Combination::MarkCorrespondences(const T_Points2DSet<C_Point2D> &in_PointsSet,
																					int in_n,//number of points in group
																					int in_N)// number of corr. to draw
	{
		AImageMem8::Pixel color_pen_wh = AImageMem8::Pixel((1 << m_comb->Bits()) - 1);
		typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor_wh;
		DrawFunctor_wh draw_functor_wh(color_pen_wh);
		typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor_wh> 
																																		  SetPixelFunctor_wh;
		SetPixelFunctor_wh set_pix_wh((*m_comb)[0], m_comb->NextLineShift(), draw_functor_wh);
		Vtl::ImAlg::Bres::Line<SetPixelFunctor_wh> line_wh(set_pix_wh);


		AImageMem8::Pixel color_pen_bl = AImageMem8::Pixel(0);
		typedef Vtl::ImAlg::Bres::Draw<AImageMem8::Pixel> DrawFunctor_bl;
		DrawFunctor_bl draw_functor_bl(color_pen_bl);
		typedef Vtl::ImAlg::Bres::SetPixelGray<AImageMem8::Pixel, DrawFunctor_bl> 
																																			SetPixelFunctor_bl;
		SetPixelFunctor_bl set_pix_bl((*m_comb)[0], m_comb->NextLineShift(), draw_functor_bl);
		Vtl::ImAlg::Bres::Line<SetPixelFunctor_bl> line_bl(set_pix_bl);

		int n = in_PointsSet.Size(); // number of correspondences

		if (in_N == -1 || in_N > n)
		{
			in_N = n;
		}
		
		for (int i = 0; i < n; i += int(n / in_N))	// iteration for correspondences
		{
			int prevX = 0;

			for (int j = 0;  j < in_n - 1; j++)					// iteration for images
			{
				double x1 = in_PointsSet[i * in_n + j].X();
				double y1 = /**/m_img[j]->Height() - in_PointsSet[i * in_n + j].Y();
				double x2 = in_PointsSet[i * in_n + j + 1].X();
				double y2 = /**/m_img[j]->Height() - in_PointsSet[i * in_n + j + 1].Y();

				if (j - (j / 2) * 2)
				{
					set_pix_bl.Set0(int(x1) + prevX, int(y1));
					line_bl.Draw(int(x2 - x1 + (m_img[j])->Width()), int(y2 - y1));
				}
				else
				{
					set_pix_wh.Set0(int(x1) + prevX, int(y1));
					line_wh.Draw(int(x2 - x1 + (m_img[j])->Width()), int(y2 - y1));
				}

				prevX += (m_img[j])->Width();
			}
		}
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	// Saves image m_comb
	// outputImageName - filename to save combined image
	void C_Combination::Save (const char* in_outputImageName)
	{
		::AImageSave(m_comb, in_outputImageName);
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	C_Combination::~C_Combination()
	{
		delete m_comb;
	}
}
