#include "Basics3D/HomographyMatrix.h"

namespace Basics3D
{
	/////////////////////////////////////////////////////////////////////////////////////////
	// Returns the minimal number of points pairs requared for computation of homogr. matrix
	int C_HomographyMatrix::SampleSize()
	{
		return 4;
	}

	/////////////////////////////////////////////////////////////////////////////////////////
	// Returns the distance threshold
	// in_sigma is inaccuracy of feature points detector
	double C_HomographyMatrix::DistThresh(double in_sigma)
	{
		return sqrt(5.99) * in_sigma;
	}
	/////////////////////////////////////////////////////////////////////////////////////////
	void C_HomographyMatrix::CreateSystem(Vtl::NumAlg::Lin::ublas::matrix<double> &out_M,
																				const T_Points2DSet<C_Point2D> &in_Points)
	{
		int pairsNumb = in_Points.Size();

		out_M.resize(2 * pairsNumb, 9, false);
		
		for (int i = 0; i < pairsNumb; i++)
		{
			int str1 = 2 * i;
			int str2 = 2 * i + 1;

			out_M(str1, 0) = out_M(str1, 1) = out_M(str1, 2) = 0;
			out_M(str1, 3) = -in_Points[2 * i].X();
			out_M(str1, 4) = -in_Points[2 * i].Y();
			out_M(str1, 5) = -1;
			out_M(str1, 6) = in_Points[2 * i].X() * in_Points[2 * i + 1].Y();
			out_M(str1, 7) = in_Points[2 * i].Y() * in_Points[2 * i + 1].Y();
			out_M(str1, 8) = in_Points[2 * i + 1].Y();

			out_M(str2, 0) = in_Points[2 * i].X();
			out_M(str2, 1) = in_Points[2 * i].Y();
			out_M(str2, 2) = 1;
			out_M(str2, 3) = out_M(str2, 4) = out_M(str2, 5) = 0;
			out_M(str2, 6) = -in_Points[2 * i].X() * in_Points[2 * i + 1].X();
			out_M(str2, 7) = -in_Points[2 * i].Y() * in_Points[2 * i + 1].X();
			out_M(str2, 8) = -in_Points[2 * i + 1].X();
		}
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_HomographyMatrix::Find
											(std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out,
											 const Vtl::NumAlg::Lin::ublas::matrix<double> &in_System)
	{
		Vtl::NumAlg::Lin::ublas::c_vector<double, 9> vect;

		SolveSystem(vect, in_System);

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> H;
		
		for (int i = 0; i < 3; i++)
		{
			for (int j = 0; j < 3; j++)
			{
			  H(i, j) = vect(i * 3 + j);
			}
		}

		out.push_back(H);
	}

	
	///////////////////////////////////////////////////////////////////////////////////////
	// in_p1 is a calculated poind on 2d image
	// in_p2 is a point on 2d image
	double C_HomographyMatrix::Dist(const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> 
		                                                                              &in_p1,
	                                const C_Point2D &in_p2) 
	{
		return sqrt((in_p2.X() - in_p1(0,0)) * (in_p2.X() - in_p1(0,0)) +
								(in_p2.Y() - in_p1(1,0)) * (in_p2.Y() - in_p1(1,0)));
	}

	///////////////////////////////////////////////////////////////////////////////////////
	double C_HomographyMatrix::MistFunc(const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                                const C_Point2D &in_p1,
			                                const C_Point2D &in_p2)
	{
	  Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> p2;

		p2(0, 0) = in_p1.X() * in_matrix[0] + in_p1.Y() * in_matrix[1] + in_matrix[2];
    p2(1, 0) = in_p1.X() * in_matrix[3] + in_p1.Y() * in_matrix[4] + in_matrix[5];
		p2(2, 0) = in_p1.X() * in_matrix[6] + in_p1.Y() * in_matrix[7] + in_matrix[8];

		p2(0, 0) /= p2(2, 0);
		p2(1, 0) /= p2(2, 0);
		p2(2, 0) = 1;

		double f = (p2(0, 0) - in_p2.X()) * (p2(0, 0) - in_p2.X()) +
								(p2(1, 0) - in_p2.Y()) * (p2(1, 0) - in_p2.Y());

		return f;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_HomographyMatrix::DerivativeOfMistFunc
		                           (Vtl::NumAlg::Lin::ublas::vector<double> &out,
			                          const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix,
			                          const C_Point2D &in_p1,
			                          const C_Point2D &in_p2)
	{
	  out.resize(9);

		double m00, m01, m02, m10, m11, m12, m20, m21, m22;
		m00 = in_matrix[0]; m01 = in_matrix[1]; m02 = in_matrix[2];
		m10 = in_matrix[3]; m11 = in_matrix[4]; m12 = in_matrix[5];
		m20 = in_matrix[6]; m21 = in_matrix[7]; m22 = in_matrix[8];

		double x2 = in_p1.X() * m00 + in_p1.Y() * m01 + m02;
		double y2 = in_p1.X() * m10 + in_p1.Y() * m11 + m12;
		double z2 = in_p1.X() * m20 + in_p1.Y() * m21 + m22;

		x2 /= z2;
		y2 /= z2;
		
		double x1 = in_p2.X();
		double y1 = in_p2.Y();

		double X = x2 - x1;
		double Y = y2 - y1;
		double A = X * x2 + Y * y2;

		out[0] = - 2 * X * x1 / z2;
		out[1] = - 2 * X * y1 / z2;
		out[2] = - 2 * X / z2;
		out[3] = - 2 * Y * x1 / z2;
		out[4] = - 2 * Y * y1 / z2;
		out[5] = - 2 * Y / z2;
		out[6] = 2 * A * x1 / z2;
		out[7] = 2 * A * y1 / z2;
		out[8] = 2 * A / z2;
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_HomographyMatrix::NormMatrix
		                             (Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out_M,
													        Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &in_M,
														      const double in_err)
	{
		out_M = Vtl::NumAlg::Lin::svd::inverse_save(in_M, in_err); 
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_HomographyMatrix::ApplyRansac
		           (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<T_Points2DSet<C_Point2D>> &in_pointsSet,
		            const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
		            const int in_j,
		            const int in_size,
		            const double in_sigma)
	{
		std::cout << "RANSAC FOR HOMOGRAPHY" << std::endl;

		T_Model<C_HomographyMatrix> Model(in_pointsSet[in_i]);
		Ransac<C_HomographyMatrix>(out_matrix[in_j], out_inliers[in_i], Model,
			                         in_pointsSet[in_i], in_size, in_sigma);

		in_pointsSet[in_i].Disable(out_inliers[in_i]);

		MakeDifferenceImage(in_img, out_matrix, in_i, in_j, 0);
		MarkCorrespondences("Corr_", in_j, in_img, in_pointsSet, -1);
	}

	///////////////////////////////////////////////////////////////////////////////////////
	void C_HomographyMatrix::RecomputeWithAllPoints
		           (std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> &out_matrix,
		            std::vector<T_Points2DSet<C_Point2D>> &out_inliers,
								std::vector<C_Points2DParamSet> &in_pointsParamSet,
								const T_ImagesSet<AImageMem8> &in_img,
		            const int in_i,
								const int in_j,
								const double in_sigma,
								const double in_t)
	{
		std::cout << std::endl << "----------------------------------------"
			<< std::endl << "RECOMPUTE HOMOGRAPHY WITH WHOLE SETS OF POINTS" << std::endl;

		out_inliers[in_i].Clear();
		T_Model<C_HomographyMatrix> Model;
		RecompMatrix<C_HomographyMatrix>(out_matrix[in_j], out_inliers[in_i], Model,
									 in_pointsParamSet[in_i], in_pointsParamSet[in_i + 1], in_sigma, in_t);
		
		in_pointsParamSet[in_i].Disable(out_inliers[in_i], 0);
		in_pointsParamSet[in_i + 1].Disable(out_inliers[in_i], 1);
		
		PrintAndSave("Homography", out_matrix, in_i, in_j);
		MakeDifferenceImage(in_img, out_matrix, in_i, in_j, 1);
	}
}