#ifndef MISTAKE_FUNC_H_
#define MISTAKE_FUNC_H_

#include <iostream>
#include <vector>
#include <math.h>

// library includings
#include "num_linalg.h"

// own includings
#include "Point2D.h"
#include "Points2DSet.h"

namespace Basics3D
{
	template <class Operations>
	class T_MistakeFunc
	{
	private:
		T_Points2DSet<C_Point2D> m_Points;
		int m_M;

	public:
		/////////////////////////////////////////////////////////////////////////////////////
		T_MistakeFunc(const T_Points2DSet<C_Point2D> &in_Points,
									int in_M)
		{
		  m_Points = in_Points;
			m_M = in_M;
		}

		/////////////////////////////////////////////////////////////////////////////////////
		double Func(const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix)
		{
			double f = 0;

			int size = m_Points.Size();

			for (int i = 0; i < m_Points.Size(); i ++)
			{
				f += Operations::MistFunc(in_matrix, m_Points[2 * i], m_Points[2 * i + 1]);
			}

			return f;
		}

		/////////////////////////////////////////////////////////////////////////////////////
		void Derivative(Vtl::NumAlg::Lin::ublas::vector<double> &out_dy,
			              const Vtl::NumAlg::Lin::ublas::vector<double> &in_matrix)
		{
			for (int i = 0; i < m_M; i++)
			{
				out_dy[i] = 0;
			}

			int size = m_Points.Size();

			for (int i = 0; i < size; i++)
			{
				Vtl::NumAlg::Lin::ublas::vector<double> tmp;
				Operations::DerivativeOfMistFunc(tmp, in_matrix, m_Points[2 * i],
					                               m_Points[2 * i + 1]);

				for (int j = 0; j < m_M; j++)
				{
				  out_dy[j] += tmp[j];
				}
			}
		}
	};
}
#endif // MISTAKE_FUNC_H_
