
#include "RecompHomography.h"

/////////////////////////////////////////////////////////////////////////////////////////
C_RecompHomography::C_RecompHomography
                         (const Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> *in_Homography,
		                      char *in_filename1,
		                      char *in_filename2,
										      float in_t1,
										      float in_t2)
{
	m_Homography = *in_Homography;

	// Load 2 sets of points
	m_PointsParamSet1 = Basics3D::C_Points2DParamSet(1);
	m_PointsParamSet2 = Basics3D::C_Points2DParamSet(1);

  m_PointsParamSet1.Load(in_filename1);
  m_PointsParamSet2.Load(in_filename2);

	m_t1 = in_t1;
	m_t2 = in_t2;
}

/////////////////////////////////////////////////////////////////////////////////////////
void C_RecompHomography::Apply(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> *out,
		                           int in_width2,
		                           int in_height2)
{
	using namespace Basics3D;
	int size1 = m_PointsParamSet1.Size();
	int size2 = m_PointsParamSet2.Size();

	int PairsNumb = 1;
  int PrevPairsNumb = 0;

	std::vector<C_Point2D> Points;				// found corresponding pairs of points

	while (PairsNumb > PrevPairsNumb)
	{
		PrevPairsNumb = PairsNumb;

		// for each point from the 1st set
		for (int i = 0; i < size1; i++)
		{
			const C_Point2DParam* p_Point1 = m_PointsParamSet1[i];

			// if the correspondence was already set for this point 
			if (p_Point1->Numb() < 0)
			{
				continue;
			}
			
			// transform point with the homography matrix
			C_Point2D PointTr(p_Point1->X(), p_Point1->Y(), i);
      Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> PointTrM(3, 1);
			PointTr.Matrix(&PointTrM);
			PointTrM = boost::numeric::ublas::prod(m_Homography, PointTrM);
      PointTrM(0, 0) = PointTrM(0, 0) / PointTrM(2, 0);
			PointTrM(1, 0) = PointTrM(1, 0) / PointTrM(2, 0);
			PointTrM(2, 0) = 1;
		
			// if transformed point is out of the 2nd image
			if (PointTrM(0, 0) < 0 || PointTrM(1, 0) < 0 ||
					PointTrM(0, 0) >= in_width2 || PointTrM(1, 0) >= in_height2)
			{
				continue;
			}

			C_Point2D prevPoint(-1, -1, -1);
			C_Point2D bestPoint(-1, -1, -1);
			double prevDist = 1;
			double bestDist = 1000000;

			// for each point of the 2nd set
			for (int j = 0; j < size2; j++)
			{
				const C_Point2DParam* p_Point2 = m_PointsParamSet2[j];

				// if the correspondence was already set for this point 
				if (p_Point2->Numb() < 0)
				{
					continue;
				}

				// if PointTrM is close enought to p_Point2
				if ( (p_Point2->X() - PointTrM(0, 0)) * (p_Point2->X() - PointTrM(0, 0)) +
						 (p_Point2->Y() - PointTrM(1, 0)) * (p_Point2->Y() - PointTrM(1, 0)) < m_t1)
				{
					// distance between parameter vectors of p_Point1 and p_Point2
					double dist = p_Point1->SqrDistance(p_Point2);

					if (dist < bestDist)
					{
						prevDist = bestDist;
						prevPoint = bestPoint;
						bestDist = dist;
						bestPoint = *p_Point2;
					}
				}
			}

			// if difference between the best corresponding point and 
			// the next cor. point is big
			if (prevDist / bestDist > m_t2)
			{
				Points.push_back(C_Point2D(p_Point1->X(), p_Point1->Y(), p_Point1->Numb()));
				Points.push_back(C_Point2D(bestPoint.X(), bestPoint.Y(), bestPoint.Numb()));

				(m_PointsParamSet1[i])->SetNumb(-1);
				(m_PointsParamSet2[bestPoint.Numb()])->SetNumb(-1);
			}
		}

		PairsNumb = Points.size() / 2;
		std::cout << "\n\nPairs number = " << Points.size()/2 << "\n\n";

		// Recalculate homography matrix for the new set of correspondences
		C_Points2DSet PointsSet_N(&Points);
		//PointsSet_N.Normalize();
		std::vector<C_Point2D> Points_N = PointsSet_N.Vector();

		Vtl::NumAlg::Lin::ublas::matrix<double> System(0, 0);
		C_HomographyMatrix::CreateSystem(&System, Points_N);
		std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> Matrices;
		C_HomographyMatrix::Find(&Matrices, &System);
		*out = Matrices[0];
		//Points_N.~vector<C_Point2D>();

		/*Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T(3, 3);
		C_NormalizationMatrix::Create
			(&T, PointsSet_N.MeanX1(), PointsSet_N.MeanY1(), PointsSet_N.MeanDev1());
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T1(3, 3);
		C_NormalizationMatrix::Create
			(&T1, PointsSet_N.MeanX2(), PointsSet_N.MeanY2(), PointsSet_N.MeanDev2());
		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> T1_1(3, 3);
		C_MatrixOperations<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>>::Invert
			                                                                      (&T1_1, &T1);*/
		
		C_MatrixOperations<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>>::Denormalize
			                                         (&m_Homography, &m_Homography, &T1_1, &T);
		//m_Homography = m_Homography.DeNormalize(T1_1, T);

		C_MatrixOperations<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>>::Print(&m_Homography); 
		std::cout << "\n_________________________________________\n";
	}
}