#ifndef NEWTON_RAPHSON_H_
#define NEWTON_RAPHSON_H_

#include <math.h>
#include <iostream>

// own includings
#include "Func.h"

#define JMAX 20			// Set to maximum number of iterations.
#define MAXIT 100		// Maximum allowed number of iterations.

namespace Basics3D
{
	// Using the Newton-Raphson method, find the root of a function known
	// to lie in the interval [x1; x2]. The root rtnewt will be refined
	// until its accuracy is known within xacc. funcd is a user-supplied
	// routine that returns both the function value and the first
	// derivative of the function at the point x.
	double rtnewt(const IFuncWithDer *g,
								double x1,
								double x2,
								double xacc);
	
	/////////////////////////////////////////////////////////////////////////////////////////
	// Using a combination of Newton-Raphson and bisection, find the root 
	// of a function bracketed between x1 and x2. The root, returned as
	// the function value rtsafe, will be refined until its accuracy is
	// known within xacc. funcd is a user-supplied routine that returns
	// both the function value and the first derivative of the function.
	double rtsafe(const IFuncWithDer *g,
								double x1,
								double x2,
								double xacc);
	
}
#endif // NEWTON_RAPHSON_H_
