#ifndef RECOMP_MATR_H_
#define RECOMP_MATR_H_

#include <vector>

//library includings
#include "num_linalg.h"

// own includings
#include "Points2DParamSet.h"
#include "Points2DSet.h"
#include "MatrixOperations.h"
#include "FundamentalMatrix.h"
#include "HomographyMatrix.h"
#include "TransformMatrices.h"
#include "Point2D.h"
#include "Point2DParam.h"
#include "Model.h"

namespace Basics3D
{
	template <class Operations>
	void RecompMatrix(Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> &out,
										T_Points2DSet<C_Point2D> &outInliers,
										T_Model<Operations> &in_Model,
										C_Points2DParamSet &in_PointsParamSet1,
										C_Points2DParamSet &in_PointsParamSet2,
										double in_sigma,
										double in_t2)
	{
		int size1 = in_PointsParamSet1.Size();
		int size2 = in_PointsParamSet2.Size();

		int pairsNumb = 1;
		int prevPairsNumb = 0;

		Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> temp = out;
		
		while (pairsNumb > prevPairsNumb)
		{
			T_Points2DSet<C_Point2D> inliers(2);				// found corresponding pairs of points

			prevPairsNumb = pairsNumb;

			// for each point from the 1st set
			for (int i = 0; i < size1; i++)
			{
				const C_Point2DParam *p_Point1 = in_PointsParamSet1[i];

				// points, which are not to be operated on, have number -1
				if (p_Point1->Numb() < 0)
				{
				  continue;
				}

				Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 1> PointTrM(3, 1);
				in_Model.Transform
					(PointTrM, temp, C_Point2D(p_Point1->X(), p_Point1->Y(), p_Point1->Numb()));
				
				C_Point2D prevPoint(-1, -1, -1);
				C_Point2D bestPoint(-1, -1, -1);
				double prevDist = 1;
				double bestDist = 1000;

				// for each point of the 2nd set
				for (int j = 0; j < size2; j++)
				{
					const C_Point2DParam *p_Point2 = in_PointsParamSet2[j];

					// points, which are not to be operated on, have number -1
					if (p_Point2->Numb() < 0)
					{
						continue;
					}

					// if PointTrM is close enought to p_Point2
					if (Operations::Dist(PointTrM, *p_Point2) < Operations::DistThresh(in_sigma))
					{
						// distance between parameter vectors of p_Point1 and p_Point2
						double dist = p_Point1->Distance(*p_Point2);

						if (dist < bestDist)
						{
							prevDist = bestDist;
							prevPoint = bestPoint;
							bestDist = dist;
							bestPoint = *p_Point2;
						}
					}
				}

				// if difference between the best corresponding point and 
				// the next cor. point is big enought
				if (prevDist / bestDist > in_t2)
				{
					std::vector<C_Point2D> P;
					P.push_back(C_Point2D(p_Point1->X(), p_Point1->Y(), p_Point1->Numb()));
					P.push_back(C_Point2D(bestPoint.X(), bestPoint.Y(), bestPoint.Numb()));
					inliers.Add(P);
				}
			}

			pairsNumb = inliers.Size();
			std::cout << std::endl << std::endl << "Pairs number = " << pairsNumb
								<< std::endl << std::endl;

			std::cout << "\n_________________________________________\n";

			// Recalculate matrix for the new set of correspondences
			if (pairsNumb > prevPairsNumb)
			{
				out = temp;
				outInliers = inliers;

				std::vector<Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3>> Matrices;
				in_Model.ComputeMatrices(Matrices, inliers, in_sigma);
				temp = Matrices[0];

				/**/Vtl::NumAlg::Lin::ublas::c_matrix<double, 3, 3> forPrint;
				/**/T_MatrixOperations::Scale(forPrint, temp);
				/**/T_MatrixOperations::Print(forPrint); 
			}
		}
	}
}
#endif // RECOMP_MATR_H_
