#ifndef SOLVE_SYSTEM_H_
#define SOLVE_SYSTEM_H_

//library includings
#include "num_linalg.h"

namespace Basics3D
{
	template<class SystemType, class SolutionType>
  void SolveSystem(SolutionType &out_Vect,
	                 const SystemType &in_System)
	{
		const int m = in_System.size1();
		const int n = in_System.size2();

		if (out_Vect.size() != n)
		{
		  throw "Not appropriate size of solution";
			return;
		}

		typedef typename SystemType::value_type Real_t;
		Vtl::NumAlg::Lin::ublas::matrix<Real_t> U(m, n);
		Vtl::NumAlg::Lin::ublas::matrix<Real_t> V(n, n);
		Vtl::NumAlg::Lin::ublas::vector<Real_t> W(n);
		Vtl::NumAlg::Lin::ublas::vector<Real_t> RV1(n);

		U = in_System;

		Vtl::NumAlg::Lin::svd::decomp(U, W, V, RV1);

		double min = fabs(W(0));
		int k = 0;
		
		for (int j = 1; j < n; j++)
		{
			if  (std::abs(W(j)) < min)
			{
				min = std::abs(W(j));
				k = j;
			}
		}

		for (int i = 0; i < n; i++)
		{
			out_Vect(i) = V(i, k);
		}
	}
}

#endif // SOLVE_SYSTEM_H_
