#define REAL double

#include <stdio.h>
#include <stdlib.h>
#include <cmath>
#include <iostream>
#include <string.h>

#include "sfm_mesh_TriResult.h"

/*****************************************************************************/
/*                                                                           */
/*  report()   Print the input or output.                                    */
/*                                                                           */
/*****************************************************************************/

void TriResult::report(struct triangulateio *io)
{
	for (int i = 0; i < io->numberofpoints; i++) 
	{
		printf("Point %4d:", i);
		for (int j = 0; j < 2; j++) 
			printf("  %.6g", io->pointlist[i * 2 + j]);
	    printf("\n");
	}
    
	for (int i = 0; i < io->numberoftriangles; i++) 
	{
        printf("Triangle %4d points:", i);
        for (int j = 0; j < io->numberofcorners; j++) 
	          printf("  %4d", io->trianglelist[i * io->numberofcorners + j]);
        if (io->numberoftriangleattributes > 0) 
	        printf("   attributes");
        for (int j = 0; j < io->numberoftriangleattributes; j++) 
			printf("  %.6g", io->triangleattributelist[i * io->numberoftriangleattributes + j]);
        printf("\n");
	}
    printf("\n");
}

/*****************************************************************************/
/*                                                                           */
/*  RunTri()   Create and refine a mesh.                                       */
/*                                                                           */
/*****************************************************************************/

void TriResult::RunTri(Matrix& W)
{
	P=W.size2();//W.Col();		
	F=W.size1();//W.Row()/2;
	uv.resize(1,2*P);//	uv.Create(1,2*P);
	for(int i=0;i<P;i++)
	{
		uv(0,2*i)   = W(0,i);
		uv(0,2*i+1) = W(1,i);
	}
	ExeTri();
}

void TriResult::ExeTri()
{
	 /* Define input points. */

	in.numberofpoints = P;
	in.numberofpointattributes = 1;
	in.pointlist = (REAL *) malloc(in.numberofpoints * 2 * sizeof(REAL));
	
	for(int i=0;i<2*in.numberofpoints;i++)
	{
		in.pointlist[i] = uv(0,i);
	}
	
	in.pointattributelist = (REAL *) malloc(in.numberofpoints *
                                          in.numberofpointattributes *
                                          sizeof(REAL));
	in.pointmarkerlist = (int *) malloc(in.numberofpoints * sizeof(int));
  
	for(int i=0;i<in.numberofpoints;i++)
	{
		in.pointattributelist[i] = 0.0;
		in.pointmarkerlist[i] = 0;
	}
  
	in.numberofsegments = 0;
	in.numberofholes = 0;
	in.numberofregions = 1;
	in.regionlist = (REAL *) malloc(in.numberofregions * 4 * sizeof(REAL));
	in.regionlist[0] = 0.5;
	in.regionlist[1] = 5.0;
	in.regionlist[2] = 7.0;            /* Regional attribute (for whole mesh). */
	in.regionlist[3] = 0.1;          /* Area constraint that will not be used. */

	/* Make necessary initializations so that Triangle can return a */
	/*   triangulation in `mid' and a voronoi diagram in `vorout'.  */

	mid.pointlist = (REAL *) NULL;            /* Not needed if -N switch used. */
	/* Not needed if -N switch used or number of point attributes is zero: */
	mid.pointattributelist = (REAL *) NULL;
	mid.pointmarkerlist = (int *) NULL; /* Not needed if -N or -B switch used. */
	mid.trianglelist = (int *) NULL;          /* Not needed if -E switch used. */
	/* Not needed if -E switch used or number of triangle attributes is zero: */
	mid.triangleattributelist = (REAL *) NULL;
	mid.neighborlist = (int *) NULL;         /* Needed only if -n switch used. */
	/* Needed only if segments are output (-p or -c) and -P not used: */
	mid.segmentlist = (int *) NULL;
	/* Needed only if segments are output (-p or -c) and -P and -B not used: */
	mid.segmentmarkerlist = (int *) NULL;
	mid.edgelist = (int *) NULL;             /* Needed only if -e switch used. */
	mid.edgemarkerlist = (int *) NULL;   /* Needed if -e used and -B not used. */

	vorout.pointlist = (REAL *) NULL;        /* Needed only if -v switch used. */
	/* Needed only if -v switch used and number of attributes is not zero: */
	vorout.pointattributelist = (REAL *) NULL;
	vorout.edgelist = (int *) NULL;          /* Needed only if -v switch used. */
	vorout.normlist = (REAL *) NULL;         /* Needed only if -v switch used. */

	/* Triangulate the points.  Switches are chosen to read and write a  */
	/*   PSLG (p), preserve the convex hull (c), number everything from  */
	/*   zero (z), assign a regional attribute to each element (A), and  */
	/*   produce an edge list (e), a Voronoi diagram (v), and a triangle */
	/*   neighbor list (n).                                              */

	//triangulate("pczevn", &in, &mid, &vorout);
	triangulate("pczAevnQ", &in, &mid, &vorout);
	 
	/* Make necessary initializations so that Triangle can return a */
	/*   triangulation in `out'.                                    */

	out.pointlist = (REAL *) NULL;            /* Not needed if -N switch used. */
	out.pointattributelist = (REAL *) NULL;
	out.trianglelist = (int *) NULL;          /* Not needed if -E switch used. */
	out.triangleattributelist = (REAL *) NULL;

	/* Refine the triangulation according to the attached */
	/*   triangle area constraints.                       */

	//triangulate("prazBP", &mid, &out, (struct triangulateio *) NULL);
	triangulate("pzBPQ", &mid, &out, (struct triangulateio *) NULL);
	
	//printf("Refined triangulation:\n\n");
	//report(&out);

	nTri=out.numberoftriangles;		
	nCorn=out.numberofcorners;
	
	tri.resize(nTri, nCorn); //tri.Create(nTri, nCorn);
	
	for(int i=0;i<nTri;i++)
		for(int j=0;j<nCorn;j++)
			tri(i,j)=out.trianglelist[i * nCorn + j];

	//tri.Save(".\\mtr\\Triangulation.mtr");


	/* Free all allocated arrays, including those allocated by Triangle. */
	free(in.pointlist);
	free(in.pointattributelist);
	free(in.pointmarkerlist);
	free(in.regionlist);
	free(mid.pointlist);
	free(mid.pointattributelist);
	free(mid.pointmarkerlist);
	free(mid.trianglelist);
	free(mid.triangleattributelist);
	//free(mid.trianglearealist);
	free(mid.neighborlist);
	free(mid.segmentlist);
	free(mid.segmentmarkerlist);
	free(mid.edgelist);
	free(mid.edgemarkerlist);
	free(vorout.pointlist);
	free(vorout.pointattributelist);
	free(vorout.edgelist);
	free(vorout.normlist);
}

void TriResult::Build_texture(const Matrix& W,int numpic, Matrix& W_tex)							//    
{
	if(numpic>0)
	{
		P=W.size2();		//P=W.Col();		
		F=W.size1()/2;     //F=W.Row()/2;

		W_tex.resize(2,P);//W_tex.Create(2,P);
		for(int i=0;i<P;i++)
		{
			W_tex(0,i) = 0.5 + W(2*numpic-2,i);
			W_tex(1,i) = 0.5 - W(2*numpic-1,i);

		}
		//W_tex.CopyRow(0,2*numpic-2,W);
		//W_tex.CopyRow(1,2*numpic-1,W);
	}
}


void TriResult::BuildVRML(const char *fname, const Matrix& X, const Matrix& R,
						  const Matrix& t,const char *texfile, const Matrix& W_tex,const Matrix& g)
{
	FILE * fp=::fopen(fname,"wt");
	
	if(!fp) throw "TriResult::BuildVRML() : Can't open output file";
	
	::fprintf(fp,   "#VRML V2.0 utf8 \n\n");
	::fprintf(fp,	"NavigationInfo \n");
	::fprintf(fp,	"{ \n");    
	::fprintf(fp,	"\t  type \"EXAMINE\" \n");    
	::fprintf(fp,	"} \n");

	for(int i=0;i<F;i++)
	{
		::fprintf(fp,	"Viewpoint \n");
		::fprintf(fp,	"{ \n");    
		::fprintf(fp,	"\t position %+1e %+1e %+1e \n",t(0,i),t(1,i),t(2,i));
		::fprintf(fp,	"\t orientation %+1e %+1e %+1e 3.1415926 \n",R(3*i+1,0),R(3*i+1,1),R(3*i+1,2));//3.1415926
		::fprintf(fp,	"\t description \"%d\" \n",i);
		::fprintf(fp,	"} \n");
	}
	
	::fprintf(fp,	"Shape \n");
	::fprintf(fp,	"{\n");
	::fprintf(fp,	"\t appearance Appearance \n");
	::fprintf(fp,	"\t { \n");
	::fprintf(fp,   "\t\t material Material \n");
	::fprintf(fp,   "\t\t { \n");
	::fprintf(fp,   "\t\t\t diffuseColor 0.8 0.55 0.4 \n");
	::fprintf(fp,   "\t\t\t shininess         0.3 \n");
	::fprintf(fp,   "\t\t\t specularColor     0 0 0.2 \n");
	::fprintf(fp,   "\t\t\t emissiveColor     0.1 0.1 0.1 \n");
	::fprintf(fp,   "\t\t\t ambientIntensity  0.5 \n"); 
	::fprintf(fp,   "\t\t\t transparency      0.0 \n");  
	::fprintf(fp,   "\t\t } \n");
	
	if(strcmp("-1",texfile))
	{
		::fprintf(fp,   "\t\t texture ImageTexture \n"); 
		::fprintf(fp,   "\t\t { \n");
		::fprintf(fp,   "\t\t\t url  \"%s\" \n",texfile);
		::fprintf(fp,   "\t\t } \n");
	}
	
	::fprintf(fp,	"\t } \n");
	::fprintf(fp,	"\t geometry IndexedFaceSet \n");
	::fprintf(fp,	"\t { \n");
	::fprintf(fp,	"\t\t coordIndex \n" );
	::fprintf(fp,	"\t\t [ \n"); 

	for(int i=0; i < nTri ; i++)
	{
		::fprintf(fp,"\t\t\t ");		
		for(int j=0; j < nCorn ; j++)
			::fprintf(fp,"%d ",out.trianglelist[i * nCorn + j]);
		::fprintf(fp,"-1 \n");
		::fprintf(fp,"\t\t\t ");		
		for(int j=nCorn-1; j >=0 ; j--)
			::fprintf(fp,"%d ",out.trianglelist[i * nCorn + j]);
		::fprintf(fp,"-1 \n");
	}
	
	::fprintf(fp,	"\t\t ] \n");
	::fprintf(fp,	"\t\t coord Coordinate \n");
	::fprintf(fp,	"\t\t { \n");
	::fprintf(fp,   "\t\t\t point \n");
	::fprintf(fp,	"\t\t\t [ \n");  

	double x;
	for(int i=0; i < P ; i++)
	{
		::fprintf(fp,"\t\t\t\t ");		
		for(int j=0; j < 3 ; j++)
		{
			x=X(j,i)*g(0,0);
			::fprintf(fp,"%+le ",x);
		}
		::fprintf(fp,"\n");
	}

	::fprintf(fp,	"\t\t\t ] \n"); 
	::fprintf(fp,	"\t\t } \n");
	
	if(strcmp("-1",texfile))
	{
		::fprintf(fp,	"\t\t texCoord TextureCoordinate \n");
		::fprintf(fp,	"\t\t { \n"); 
		::fprintf(fp,	"\t\t\t	point \n");
		::fprintf(fp,	"\t\t\t [ \n");

		for(int i=0; i < P ; i++)
		{
			::fprintf(fp,"\t\t\t\t ");		
			::fprintf(fp,"%+le %+le ",W_tex(0,i),W_tex(1,i));
			::fprintf(fp,"\n");
		}

		::fprintf(fp,	"\t\t\t ] \n"); 
		::fprintf(fp,	"\t\t } \n");
	}

	::fprintf(fp,	"\t } \n");
	::fprintf(fp,	" } \n");

	::fclose(fp);

	free(out.pointlist);
	free(out.pointattributelist);
	free(out.trianglelist);
	free(out.triangleattributelist);
}
