#ifndef _SS_RIDGE_FUNCTORS_H 
#define _SS_RIDGE_FUNCTORS_H
///////////////////////////////////////////////////////////////////////////////////////
//#include "num_linalg.h"
//#include "AImage.h"
//#include "plf_functors.h"
//#include "plf_kernels_1d.h"
///////////////////////////////////////////////////////////////////////////////////////
namespace Hanina
{
	template<typename T1,typename T2>
	class Comb_ridge
	{
		T2 val;
	public:
		Comb_ridge(T2 value)
			: val(value)
		{
		}
		T2 operator()(T1 v1,T2 v2)
		{
			return (v1>4) ? val : v2;
		}
	};
	template<typename T>
	class Check_4_grad_sign_change
	{
		const int nls;
		T sprod(int ofs,T hx,T hy,const T* grad_x,const T* grad_y)
		{
			return hx*grad_x[ofs]+hy*grad_y[ofs];
		}
	public:
		Check_4_grad_sign_change(int next_line_shift)
			:nls(next_line_shift)
		{
		}
		T operator()(T image,T hx,T hy,const T* grad_x,const T* grad_y)
		{
			if(image!=T(0))
			{
				T v1;
				T v2;
				if(ABS(hx) < ABS(hy))
				{//     .   -    
					T alpha=hx/hy;
					if(alpha<0)
					{// (hx< hy) && ((hx<0 hy >0 )|| (hx>0 hy < 0 ) ) 
						//    , '+' -  , 
						// alpha=0 -  (|) , alpha=1 -  (/)
						//
						//    +-+=+ 
						//    | | |
						//    +-+-+ 
						//    | | |
						//    +=+-+ 
						//
						v1=(1+alpha)*sprod(-nls,hx,hy,grad_x,grad_y)-alpha*sprod(-nls+1,hx,hy,grad_x,grad_y);
						v2=(1+alpha)*sprod(+nls,hx,hy,grad_x,grad_y)-alpha*sprod(+nls-1,hx,hy,grad_x,grad_y);
					}
					else
					{// (gx< gy) && ( (gx>=0 gy >0) || ( gx<=0 gy < 0 ) ) 
						//    , '+' -  , 
						// alpha=0 -  (|) , alpha=1 -  (\)
						//
						//    +=+-+ 
						//    | | |
						//    +-+-+ 
						//    | | |
						//    +-+=+ 
						//
						v1=(1-alpha)*sprod(-nls,hx,hy,grad_x,grad_y)+alpha*sprod(-nls-1,hx,hy,grad_x,grad_y);
						v2=(1-alpha)*sprod(+nls,hx,hy,grad_x,grad_y)+alpha*sprod(+nls+1,hx,hy,grad_x,grad_y);
					}
				}
				else
				{//     .  -    
					T alpha=hy/hx;
					if(alpha<0)
					{// (gx >= gy) && ( (gx>0 , gy<=0) || ( gx <0 gy >= 0 ) ) 
						//   (#)  , '+' -  , 
						// alpha=0 -  (-) , alpha=1 -  (/)
						//
						//
						//    +-+-+ 
						//    | | #
						//    +-+-+ 
						//    # | |
						//    +-+-+ 
						//
						v1=(1+alpha)*sprod(-1,hx,hy,grad_x,grad_y)-alpha*sprod(-1+nls,hx,hy,grad_x,grad_y);
						v2=(1+alpha)*sprod(+1,hx,hy,grad_x,grad_y)-alpha*sprod(+1-nls,hx,hy,grad_x,grad_y);
					}
					else
					{// (gx >= gy) && ( (gx>0 , gy>=0) || ( gx <0 gy <= 0 ) ) 
						//   (#)  , '+' -  , 
						// alpha=0 -  (-) , alpha=1 -  (\)
						//
						//    +-+-+ 
						//    # | |
						//    +-+-+ 
						//    | | #
						//    +-+-+ 
						//
						v1=(1-alpha)*sprod(-1,hx,hy,grad_x,grad_y)+alpha*sprod(-1-nls,hx,hy,grad_x,grad_y);
						v2=(1-alpha)*sprod(+1,hx,hy,grad_x,grad_y)+alpha*sprod(+1+nls,hx,hy,grad_x,grad_y);
					}
				}
				if( v1*v2 < 0 )
					return image;
			}
			return 0;
		}
	};
	
	//   33     x, y  t
	//  - . ..  3   ..,   
	class FillingHessianMatrix3x3
	{
		private:
			float mult;			
		public:
			Vtl::NumAlg::Lin::ublas::c_matrix<float,3,3> m33;
			Vtl::NumAlg::Lin::ublas::c_matrix<float,3,3> t33;
			Vtl::NumAlg::Lin::ublas::c_matrix<float,3,3> tmp33;
			Vtl::NumAlg::Lin::ublas::c_vector<float,3> work;
			Vtl::NumAlg::Lin::ublas::c_vector<float,3> eigen;
		/*
					   |Ixx Ixy Ixt|
			matr_hes = |Ixy Iyy Iyt|
					   |Ixt Iyt Itt|
		*/

		FillingHessianMatrix3x3(float sigma) : mult(sigma*sigma)
		{
		}

		template<typename T1> 
		void operator()(T1 Ixx, T1 Iyy, T1 Ixy, T1 Ixxx, T1 Iyyy, T1 Ixyy, T1 Iyxx, 
						T1 Ixxxx, T1 Iyyyy, T1 Ixxyy, T1 &max_eig, 
						T1 &vx1, T1 &vy1, T1 &vt1, T1 &vx2, T1 &vy2, T1 &vt2) 
		{
			
			build_3D_hess( Ixx,Iyy,Ixy, Ixxx,Iyyy,Ixyy,Iyxx, Ixxxx,Iyyyy,Ixxyy);

			//eigen_val_vec_3D(max_eig, vx1,vy1,vt1, vx2,vy2,vt2);
			eigen_val_vec_2D_proj_3D(max_eig, vx1,vy1,vt1, vx2,vy2,vt2);
		}
	private:
		template<typename T1> 
		void eigen_val_vec_3D(T1 &max_eig, T1 &vx1, T1 &vy1, T1 &vt1, T1 &vx2, T1 &vy2, T1 &vt2)
		{
			Vtl::NumAlg::Lin::eigen::decomp(m33,eigen,work);
			//   - eigen(2)
			max_eig = mult* (eigen(2));//-eigen(1)
			// vx1 -     e1
			vx1 = m33(0,2);
			vy1 = m33(1,2);
			vt1 = m33(2,2);
			// vx2 -     e2
			vx2 = m33(0,1);
			vy2 = m33(1,1);
			vt2 = m33(2,1);
			//// vx3 -     e3
			//vx3 = m33(0,0);
			//vy3 = m33(1,0);
			//vt3 = m33(2,0);
		}
		template<typename T1> 
		void eigen_val_vec_2D_proj_3D(T1 &max_eig, T1 &vx1, T1 &vy1, T1 &vt1, T1 &vx2, T1 &vy2, T1 &vt2)
		{
			T1 vx;
			T1 vy;

			calc_2D_eigmax(m33(0,0), m33(0,1), m33(1,1), vx1, vx, vy);
			vx1=vx;
			vt1=vy;
			build_3D_hess_projection(vx1, vt1);
			//  ..  .
			vx2=vt1;
			vt2=-vx1;
			if( std::abs(vt1) > std::abs(vx1) )
			{
				vx2=vx1;
				vt2=vt1;
			}
			else
			{
				vx2=vt1;
				vt2=-vx1;
			}
			//  ,       vx,vy;
			vy1=vy*vx1;
			vx1=vx*vx1;
			vy2=vy*vx2;
			vx2=vx*vx2;
			
			//, ,     2D 
			vx1=vx;
			vy1=vy;
			vt1=0;
			//vx2=vx;
			//vy2=vy;
			//vt2=0;

			vx2=0;
			vy2=0;
			vt2=1;


			Vtl::NumAlg::Lin::eigen::decomp(m33,eigen,work);
			//   - eigen(2)
			max_eig = mult* (sqrt(eigen(2)));//-eigen(1)
		}
		template<typename T1> 
		void calc_2D_eigmax(T1 Ixx, T1 Ixy, T1 Iyy, T1 &maxeig, T1 &vx, T1 &vy)
		{
			using namespace Vtl;
			//  det(H-lambda*E)=0;, .
			T1 b = (Ixx+Iyy)/2;
			T1 D = b-Iyy; //(Ixx-Iyy)/2;
			D = sqrt(sqr(D)+sqr(Ixy)); //sqrt  
			maxeig = (b>=0) ? b+D : b-D ; //.   .(       . .
			//  .. 
			b=Ixx-maxeig;
			D=sqrt(sqr(Ixy)+sqr(b));
			//   1
			vx=Ixy/D;
			vy=-b/D;
		}
		template<typename T1> 
		void build_3D_hess(T1 Ixx, T1 Iyy, T1 Ixy, T1 Ixxx, T1 Iyyy, T1 Ixyy, T1 Iyxx,T1 Ixxxx, T1 Iyyyy, T1 Ixxyy)
		{
			T1 Ixt = 1/2*(Ixxx + Ixyy); 
			T1 Iyt = 1/2*(Iyxx + Iyyy); 
			m33(0,0)= Ixx;  m33(1,0)= Ixy;  m33(2,0)= Ixt;
			m33(0,1)= Ixy;  m33(1,1)= Iyy;  m33(2,1)= Iyt; 
			m33(0,2)= Ixt;  m33(1,2)= Iyt; 
			m33(2,2)= 1/4*(Ixxxx + 2*Ixxyy + Iyyyy); //Itt 
			/// ,   ,  ..  , ..  , ..    .
			t33=m33;
			tmp33=m33;
			m33=boost::numeric::ublas::prod(t33,tmp33);
		}
		template<typename T1> 
		void build_3D_hess_projection(T1& vx, T1& vy)
		{
			T1 H11 = m33(0,0)*vx*vx+m33(0,1)*vx*vy+m33(1,1)*vy*vy;
			T1 H12 = m33(0,2)*vx+m33(1,2)*vy;
			T1 H22 = m33(2,2);
			T1 maxeig;
			calc_2D_eigmax(H11, H12, H22, maxeig, vx,vy);
		}
	};
	//////////////////////////////////////////////////////////////////////////////////////////
	///
	///
	///
	//////////////////////////////////////////////////////////////////////////////////////////
			template<typename T,int i_alpha,int i_beta>
			struct p33;
			template<typename T>
			class p33<T,0,1> // X- alpha, Y- beta, Z- main
			{
				typedef const T* Planes_t;
				const Planes_t* plane;
				size_t ofs_to_cur;
				int nls;
			public:
				p33(const Planes_t* plane_,size_t ofs_to_cur_,int nls_)
					:plane(plane_),ofs_to_cur(ofs_to_cur_),nls(nls_)
				{
				}
				template<int d_a,int d_b>
				T f()
				{
					return plane[2][ofs_to_cur+d_a+d_b*nls];
				}
				template<int d_a,int d_b>
				T b()
				{
					return plane[0][ofs_to_cur+d_a+d_b*nls];
				}
			};
			template<typename T>
			class p33<T,1,2> // Y- alpha, Z- beta, X- main
			{
				typedef const T* Planes_t;
				const Planes_t* plane;
				size_t ofs_to_cur;
				int nls;
			public:
				p33(const Planes_t* plane_,size_t ofs_to_cur_,int nls_)
					:plane(plane_),ofs_to_cur(ofs_to_cur_),nls(nls_)
				{
				}
				template<int d_a,int d_b>
				T f()
				{
					return plane[d_b+1][ofs_to_cur+1+d_a*nls];
				}
				template<int d_a,int d_b>
				T b()
				{
					return plane[d_b+1][ofs_to_cur-1+d_a*nls];
				}
			};
			template<typename T>
			class p33<T,0,2> // X- alpha, Z- beta, Y- main
			{
				typedef const T* Planes_t;
				const Planes_t* plane;
				size_t ofs_to_cur;
				int nls;
			public:
				p33(const Planes_t* plane_,size_t ofs_to_cur_,int nls_)
					:plane(plane_),ofs_to_cur(ofs_to_cur_),nls(nls_)
				{
				}
				template<int d_a,int d_b>
				T f()
				{
					return plane[d_b+1][ofs_to_cur+d_a+nls];
				}
				template<int d_a,int d_b>
				T b()
				{
					return plane[d_b+1][ofs_to_cur+d_a-nls];
				}
			};

			template<int d_a,int d_b,typename T,class F>
			void v12_bilin(T alpha,T beta,T& v1, T& v2,F p)
			{
				alpha = ABS(alpha);
				beta  = ABS(beta);

				T a1 = (1-alpha) * p.f<0,0>()     + alpha* p.f<d_a,0>();
				T a2 = (1-alpha) * p.f<0,d_b>()   + alpha* p.f<d_a,d_b>();
				v1 =(1-beta)*a1+beta*a2;

				  a1 = (1-alpha) * p.b<0,0>()     + alpha* p.b<-d_a,0>();
				  a2 = (1-alpha) * p.b<0,-d_b>()   + alpha* p.b<-d_a,-d_b>();
				v2 =(1-beta)*a1+beta*a2;
			}
			template<typename T,class F>
			void v12_bilinear(T alpha,T beta,T& v1, T& v2,F p)
			{
				if(alpha >= 0)
				{
					if( beta >= 0 )
					{ // alpha > 0 && beta > 0
						v12_bilin<1,1>(alpha,beta,v1,v2,p);
					}
					else
					{ // alpha > 0 && beta < 0
						v12_bilin<1,-1>(alpha,beta,v1,v2,p);
					}
				}
				else
				{
					if( beta >= 0 )
					{ // alpha < 0 && beta > 0
						v12_bilin<-1,1>(alpha,beta,v1,v2,p);
					}
					else
					{ // alpha < 0 && beta < 0
						v12_bilin<-1,-1>(alpha,beta,v1,v2,p);
					}
				}
			}

	//////////////////////////////////////////////////////////////////////////////////////////
	///
	///       333
	///
	//////////////////////////////////////////////////////////////////////////////////////////
	template<typename T> 
	class ScaleScan3x3
	{
			T thr;
			int nls;
		public:
			int Nls() const {return nls;}
			typedef const T* Planes_t;
			ScaleScan3x3(int nls_set) 
			{			
				thr = 0;
				nls = nls_set;
			}
			T operator()(const Planes_t* modul, size_t ofs_to_cur,T v1x,T v1y,T v1z,T v2x,T v2y,T v2z)
			{
				//return modul[1][ofs_to_cur];

				if( dir_check(modul,ofs_to_cur,v1x,v1y,v1z)  &&  dir_check(modul,ofs_to_cur,v2x,v2y,v2z) )
				{ 
					//     true,   
					return  modul[1][ofs_to_cur];
				}
				else
					return 0;

			}
			void dir_check_max_X(const Planes_t* plane,size_t ofs_to_cur,T vx,T vy,T vz,T& v1,T& v2)
			{
				T alpha=vy/vx;
				T beta=vz/vx;
				p33<T,1,2> p(plane,ofs_to_cur,nls);// Y- alpha, Z- beta, X- main

				v12_bilinear(alpha,beta,v1,v2,p);
			}
			void dir_check_max_Y(const Planes_t* plane,size_t ofs_to_cur,T vx,T vy,T vz,T& v1,T& v2)
			{
				T alpha=vx/vy;
				T beta=vz/vy;
				p33<T,0,2> p(plane,ofs_to_cur,nls);// X- alpha, Z- beta, Y- main

				v12_bilinear(alpha,beta,v1,v2,p);
			}
			void dir_check_max_Z(const Planes_t* plane,size_t ofs_to_cur,T vx,T vy,T vz,T& v1,T& v2)
			{
				//Vtl::NumAlg::sign(1,alpha); -             
				using namespace Vtl::NumAlg;
				T alpha=vx/vz;
				T beta=vy/vz;
				p33<T,0,1> p(plane,ofs_to_cur,nls);// X- alpha, Y- beta, Z- main

				v12_bilinear(alpha,beta,v1,v2,p);
			}

			bool dir_check(const Planes_t* plane,size_t ofs_to_cur,T vx,T vy,T vz)
			{
				T v0=plane[1][ofs_to_cur];
				if(v0 >= thr)
				{
					T v1;
					T v2;
					if(ABS(vx) < ABS(vz))
					{
						if(ABS(vy) < ABS(vz))
						{//   X & Y  < Z
							dir_check_max_Z(plane,ofs_to_cur,vx,vy,vz,v1,v2);
						}
						else 
						{//   X  < Z <= Y
							dir_check_max_Y(plane,ofs_to_cur,vx,vy,vz,v1,v2);
						}
					}
					else  // Z <= X
					{
						if(ABS(vy) < ABS(vx))
						{//   Z & Y  < X
							dir_check_max_X(plane,ofs_to_cur,vx,vy,vz,v1,v2);
						}
						else 
						{//   Z  <=  X <= Y
							dir_check_max_Y(plane,ofs_to_cur,vx,vy,vz,v1,v2);
						}
					}

					if( (v0 > v1) && (v0 > v2) )
					{
						return true;
					}
				}
				return false;
			}
	};
}//namespace Hanina


///////////////////////////////////////////////////////////////////////////////////////
#endif //_SS_RIDGE_FUNCTORS_H