#ifndef CENTER_CLUSTERIZATION_H
#define CENTER_CLUSTERIZATION_H

//system includings
#include <vector>
//#include <iostream>
//#include <math.h>

// library includings


// own includings
#include "KMeans.h"

namespace Vtl
{
namespace CurveDetection
{
  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  struct CircleCenter
  {
    int x;
    int y;

    std::vector< DatumR<T> > points;

    //-------------------------------------------------------------------------
  public:
    CircleCenter(){};

    //-------------------------------------------------------------------------
    CircleCenter(int in_x,
                 int in_y)
    {
       x = in_x;
       y = in_y;
    }

    //-------------------------------------------------------------------------
    double Dist(const CircleCenter<T> &c) const 
    {
      double dx = x - c.x;
		  double dy = y - c.y;
      return sqrt(dx * dx + dy * dy);
    }

    //-------------------------------------------------------------------------
    /*CircleCenter<T> &operator = (const CircleCenter<T> &c)
    {
      x = c.x;
      y = c.y;
      points = c.points;

      return (*this);
    }*/

    //-------------------------------------------------------------------------
    void AddPoints(const CircleCenter<T> &c)
    {
      int size1 = points.size();
      int size2 = c.points.size();

      points.resize(size1 + size2);

      for (int i = 0; i < size2; i++)
      {
        points[size1 + i] = c.points[i];
      }
    }
  };

  //////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  void Normalize(IterType img,
                 int width,
                 int height,
                 T maxVal)
  {
    float coef = 255.0f /*/ 2.0f*/ / maxVal;

    for (int y = 0; y < height; y++)
    {
      for (int x = 0; x < width; x++)
      {
        (*img) *= coef;
        img++;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  int IsMax(IterType img,
            int width,
            int height,
            int nls,
            int x,
            int y,
            int neigh)
  {
    T val = *(img + y * nls + x);

    int maxX = std::min(x + neigh, width - 1);
    int maxY = std::min(y + neigh, height - 1);

    for (int i = std::max(x - neigh, 0); i <= maxX; i++)
    {
      for (int j = std::max(y - neigh, 0); j <= maxY; j++)
      {
        if (i == x && j == y)
        {
          continue;
        }

        if ( *(img + j * nls + i) >= val)
        {
          return 0;
        }
      }
    }
    return 1;
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  void FindMax(IterType img,
               std::vector< CircleCenter<T> > &centers,
               std::vector<int>  &max,
               int width,
               int height,
               int nls,
               T thresh)
  {
    max.clear();

    for (int y = 0; y < height; y++)
    {
      for (int x = 0; x < width; x++)
      {        
        if (IsMax<IterType, T>(img, width, height, nls, x, y, 2))
        {
          for (int i = std::min(y*nls + x, (int)centers.size() - 1); i >= 0; i--)
          {
            if (centers[i].x == x && centers[i].y == y)
            {
              if (centers[i].points.size() > thresh)
              {
                max.push_back(i);
              }
              break;
            }
          }
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename T> 
  void CalcMaxDev(const std::vector<int> &centMax,
                  const std::vector< CircleCenter<T> > &centers,
                  std::vector<T> &maxDev)
  {
    maxDev.resize((int)centMax.size());

    std::cout << std::endl << std::endl << "Calculate maximum allowed deviation";

    for (int i = 0; i < (int)centMax.size(); i++)
    {
      T max = -1;
      T prevMax = -1;

      int j = centMax[i];
      int size = centers[j].points.size();

      std::cout << std::endl << centers[j].x << " " << centers[j].y << "   ";

      for (int k = 0; k < size; k++)
      {
        T val = centers[j].points[k].val;
        if (val > max)
        {
          prevMax = max;
          max = val;
        }
        else if (val > prevMax)
        {
          prevMax = val;
        }
      }

      std::cout << "max = " << Vtl::Curvature::CurvatureScale<T>::value_to_radius(max)
        << "   prevMax = " << Vtl::Curvature::CurvatureScale<T>::value_to_radius(prevMax);
      maxDev[i] = T(Vtl::Curvature::CurvatureScale<T>::value_to_radius(prevMax));
      maxDev[i] *= T(0.25) / 3;
      std::cout << "   maxDev = " << maxDev[i] << "  points size = "
        << centers[j].points.size();

     }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  void MarkCenters(IterType img,
                   int nls,
                   const std::vector<int> &centMax,
                   const std::vector< CircleCenter<T> > &centers)
  {
    std::cout << std::endl << "Mark centers: ";
    for (int i = 0; i < centMax.size(); i++)
    {
      std::cout << i << "  *";
      int j = centMax[i];
      std::cout << " **";
      int x = centers[j].x;
      std::cout << " ***";
      int y = centers[j].y;
      std::cout << " ****";

      *(img + y * nls + x) = T(255);
      std::cout << " *****";
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T> 
  void CalcClustRad(IterType img,
                    int nls,
                    const std::vector<int> &centMax,
                    const std::vector< CircleCenter<T> > &centers,
                    std::vector<T> &clRad)
  {
    clRad.resize((int)centMax.size());

    std::cout << std::endl << std::endl << "Calculate clusters' radii";

    int size = centers.size();

    for (int i = 0; i < (int)centMax.size(); i++)
    {
      int j = centMax[i];
      int x1 = centers[j].x;
      int y1 = centers[j].y;

      std::cout << std::endl << centers[j].x << " " << centers[j].y << "   ";

      double v1 = *(img + y1 * nls + x1);

      double v = 0;
      v += *(img + (y1 - 1) * nls + x1);
      v += *(img + (y1 + 1) * nls + x1);
      v += *(img + y1 * nls + (x1 - 1));
      v += *(img + y1 * nls + (x1 + 1));

      if (v > 0)
      {
        v /= 4.0;
        T sigma = T(sqrt(- 0.5 / log(v / v1)));
        clRad[i] = 5 * sigma;
      }
      else
      {
        clRad[i] = T(1);
      }

      std::cout << "v1 = " << v1 << "   v = " << v << "   sigma = " << clRad[i];
     }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  void BuildCenterClusters(IterType img,
                           int width,
                           int height,
                           int nls,
                           std::vector< CircleCenter<T> > &centers,
                           const std::vector<int> &centMax,
                           const std::vector<T> &maxDev)
  {
    //std::cout << std::endl << "BuildCenterClusters:  ";
    int clustNumb = centMax.size();

    std::vector< CircleCenter<T> > clusters(clustNumb);

    for (int i = 0; i < clustNumb; i++)
    {
      clusters[i] = CircleCenter<T>(centers[centMax[i]].x, centers[centMax[i]].y);
    }

    int size = centers.size();

    for (int i = 0; i < size; i++)
    {
      //std::cout << std::endl << i << "  ";
      int numb = 0;
      double min = centers[i].Dist(clusters[0]);
      //std::cout << std::endl << "0  " << min << "  !!!";

      for (int j = 1; j < clustNumb; j++)
      {
        double dist = centers[i].Dist(clusters[j]);
        //std::cout << std::endl << j << "  " << dist;

        if (dist < min)
        {
          //std::cout << "  !!!";
          min = dist;
          numb = j;
        }
      }

      if (min <= maxDev[numb])
      {
        //std::cout << std::endl << "Add:  maxDev = " << maxDev[numb];
        clusters[numb].AddPoints(centers[i]);

        int x = centers[i].x;
        int y = centers[i].y;

        if (x >= 0 && x < width && y >= 0 && y < height)
        {
          *(img + y * nls + x) = T(128);
        }
        //std::cout << std::endl << "added";
      }
      //else
      //{
      //  std::cout << std::endl << "Too far";
      //}
    }

    centers.clear();
    centers = clusters;
  }

  ///////////////////////////////////////////////////////////////////////////////
  // Draw edge points of the same cluster.
  // Also draw mean center of cluster with point 3*3
  template <typename IterType, typename T>
  void DrawCentCluster(CircleCenter<T> &cluster,
                       IterType img,
                       int width,
                       int height,
									     int nls,
                       T color)
  {
    // draw all cluster data
    int size = cluster.points.size();

    for (int i = 0; i < size; i++)
	  {
      int x = cluster.points[i].x;
		  int y = cluster.points[i].y;
		  *(img + y*nls + x) = color;
	  }

    // draw cluster center
    int midX = (cluster.x + width / 3) * 3;
    int midY = (cluster.y + height / 3) * 3;
    int maxX = std::min(midX + 1, width - 1);
    int maxY = std::min(midY + 1, height - 1);

    for (int y = std::max(midY - 1, 0); y <= maxY; y++)
    {
      for (int x = std::max(midX - 1, 0); x <= maxX; x++)
      {
        *(img + y*nls + x) = color;
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType, typename T>
  void ThreshDraw(IterType img,
                  int width,
                  int height,
                  T thresh,
                  T color)
  {
    for (int i = 0; i < width * height; i++)
    {
      if (fabs(*img) > thresh)
      {
        (*img) = color;
      }
      img++;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename IterType>
  void TwoStepsClusterization(IterType comb,
                              const IterType GradX,
                              const IterType GradY,
                              AImageMem8* img,
                              AImageMem8* img_pal,
                              int w,
                              int h,
                              const char *fname,
                              Vtl::Plf::F11::Convert2ByteBounded &c_to_Byte,
															Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert2ByteBounded > &adj_3mod)
  {
    // clusterization

    // circle centers
    std::vector<float> cent(w * h, 0);
    float *centImg = &(cent[0]);

    std::vector< Vtl::CurveDetection::CircleCenter<float> > centers;
    float max = Vtl::CurveDetection::MarkCenters(centImg, centers, comb, GradX,
                                                 GradY, w, h, w, 1.0f);
    std::cout << std::endl << std::endl << "centers.size = "
      << (int)centers.size() << std::endl << "max = " << max;

    if (max > 0)
    {
      Vtl::CurveDetection::Normalize(centImg, w, h, 2 * max);
    }

    Tests::SaveImage(img, centImg, w, h, w, adj_3mod, 0.0f, fname, "1_CenterAccum");

    // blur center accumulator
    float sigmaCent = 1.0f;
    float nsigmas_env = 3.5f;
    typedef Vtl::ImAlg::WndFun::Gauss<float, 0, false> FF_; 
		FF_ f_(sigmaCent, nsigmas_env);
		typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K_;
		K_ ker_(f_);
		Vtl::ImAlg::Conv::ConvInplace<float, K_> g_(ker_);

    Vtl::ImAlg::Conv::Conv2D::conv2DInplace(centImg, w, h, w, g_, g_);
    Tests::SaveImage(img, centImg, w, h, w, adj_3mod, sigmaCent, fname, "2_CenterAccum_blured");

    // search maximum in accumulator
    float tCentNumb = 0.15f * 2.0f * float(M_PI) * 10.0f;

    std::vector<int> centMax;
    Vtl::CurveDetection::FindMax(centImg, centers, centMax, w, h, w, tCentNumb);
    Tests::SaveImage(img, centImg, w, h, w, adj_3mod, 0.0f, fname, "3_Center_max");

    // calculate maximum deviation allowed in every maxima    
    std::vector<float> maxDev;
    //Vtl::CurveDetection::CalcMaxDev(centMax, centers, maxDev);
    Vtl::CurveDetection::CalcClustRad(centImg, w, centMax, centers, maxDev);
    
    Vtl::CurveDetection::BuildCenterClusters(centImg, w, h, w, centers, centMax, maxDev);
    //Vtl::CurveDetection::MarkCenters(centImg, w, centMax, centers);
    Tests::SaveImage(img, centImg, w, h, w, adj_3mod, 0.0f, fname, "4_Center_clust");

    Vtl::CurveDetection::ThreshDraw(comb, w, h, 0.0f, 255.0f);

    int cClustNumb = centMax.size();

    const int colorsNumb = 10;
    std::vector<float> color;

    if (colorsNumb >= cClustNumb)
    {
      color.resize(colorsNumb);
      color[0] = 1;
      color[1] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(12);
      color[2] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(32);
      color[3] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(40);
      color[4] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(47);
      color[5] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(105);
      color[6] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(150);
      color[7] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(200);
      color[8] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(350);
      color[9] = Vtl::Curvature::CurvatureScale<float>::radius_to_value(1500);
    }
    else
    {
      color.resize(cClustNumb);
      float min = 25.0f;
      float step = 125.0f / cClustNumb;

      for (int i = 0; i < cClustNumb; i++)
      {
        color[i] = min + i * step;
      }
    }
    
    std::cout << std::endl << std::endl << "cClustNumb = " << cClustNumb;

    for (int i = 0; i < cClustNumb; i++)
    {
      Vtl::CurveDetection::DrawCentCluster(centers[i], comb, w, h, w, color[i]);
    }
    //std::cout << std::endl << "save center clusters";
    Tests::SaveImage(img_pal, comb, w, h, w, c_to_Byte, 0.0f, fname, "CentClusters");
    //std::cout << " . . . saved";
    //std::cout << std::endl << "draw edges with black";
    Vtl::CurveDetection::ThreshDraw(comb, w, h, 0.0f, 255.0f);
    //std::cout << " . . . drawn";

        
    // clusterization over curvature radii
    
    // select number of clusters over radii
    std::cout << std::endl << "CLUSTERIZATION OVER CURVATURE RADII";
    for (int i = 0; i < cClustNumb; i++)
    {
      //std::cout << std::endl << "cent cluster:  " << i;
      std::vector< Cluster1D< float, DatumR<float> > > rClusters;
      float dev1;
      int rClustNumb;

      for (rClustNumb = 1; rClustNumb <= 15; rClustNumb++)
      {
        //std::cout << std::endl << "number of radius clusters:  " << rClustNumb;
        std::vector<float> dev;
		    KMeans(centers[i].points, rClusters, dev, rClustNumb);

        if (rClusters.size() == 0)
        {
          rClustNumb = 15;
          continue;
        }
        
        float maxDev = 0.0f;

        for (int j = 0; j < rClustNumb; j++)
        {
          std::cout << " | " << dev[j] << "  " << rClusters[j].val << "  " << Vtl::Curvature::CurvatureScale<float>::value_to_radius(rClusters[j].val);
          if (maxDev < dev[j] / rClusters[j].val)
          {
            maxDev = dev[j] / rClusters[j].val;  std::cout << " ! ";
          }
        }
        std::cout << std::endl << std::endl << "maxDev = " << maxDev;
        
        if (maxDev < 0.035)
        {
          break;
        }

        if (rClustNumb == 1)
        {
          dev1 = maxDev;
        }
        else
        {
          if (maxDev < 0.5 * dev1)
          {
            break;
          }
        }
      }

      if (rClustNumb > 15)
      {
        continue;
      }

      std::cout << std::endl << std::endl << centers[i].x << "  "
        << centers[i].y << "  " << rClustNumb;

      // draw and save points clusterized over their curvature radii
	    for (int j = 0; j < rClustNumb; j++)
	    {
        float r = rClusters[j].val;
        r = Vtl::Curvature::CurvatureScale<float>::value_to_radius(r);
        std::cout << std::endl << r << "  " << rClusters[j].data.size()
          << "  " << rClusters[j].data.size() / (2.0 * M_PI * r)
          << std::endl;
        if (rClusters[j].data.size() > 0.1 * 2 * M_PI * r)
        {
          std::cout << "  draw";
	        DrawRadCluster<float*, float>(&(rClusters[j]), comb, w);
          //std::cout << "  drawn";
        }
        else
        {
          std::cout << "  don't draw";
        }
	    }

      //std::cout << std::endl << "save";
      std::ostringstream os;
	    os << "RadClust" << "_" << i << "_" << rClustNumb;
      Tests::SaveImage(img_pal, comb, w, h, w, c_to_Byte, 0.0f, fname, os.str().c_str());
      //std::cout << std::endl << "saved";
    }

    std::cout << "";
  }

}
}


#endif // CENTER_CLUSTERIZATION_H