#ifndef CLUST_VIA_ACCUM_H
#define CLUST_VIA_ACCUM_H

// system includings
#include <vector>

// own includings
#include "ImageContainer.h"

namespace Vtl
{
namespace CurveDetection
{
  template <typename IterType, typename T>
  bool IsMax(std::vector<IterType> iters,
            int c_x, 
            int c_y,
            int nls)
  {
    T val = *(iters[1] + c_y * nls + c_x);

    for (int x = c_x - 1; x <= c_x + 1; x++)
    {
      for (int y = c_y - 1; y <= c_y + 1; y++)
      {
        int shift = y * nls + x;

        for (int r = 0; r < 3; r++)
        {
          if (r == 1 && x == c_x && y == c_y)
          {
            continue;
          }

          if (*(iters[r] + shift) >= val)
          {
            return false;
          }
        }
      }
    }
    return true;
  }

  /////////////////////////////////////////////////////////////////////////////
  struct Circle
  {
    int x;
    int y;
    
    std::vector<int> pointsNumbs;

    //-------------------------------------------------------------------------
  public:
    //CircleCenter(){};

    //-------------------------------------------------------------------------
    Circle(int in_x,
           int in_y)
    {
       x = in_x;
       y = in_y;
    }

    //-------------------------------------------------------------------------
    Circle(int in_x,
           int in_y,
           int in_p)
    {
       x = in_x;
       y = in_y;
       pointsNumbs.push_back(in_p);
    }

    //-------------------------------------------------------------------------
    double Dist(const Circle &c) const 
    {
      double dx = x - c.x;
		  double dy = y - c.y;
      return sqrt(dx * dx + dy * dy);
    }

    //-------------------------------------------------------------------------
    /*CircleCenter<T> &operator = (const CircleCenter<T> &c)
    {
      x = c.x;
      y = c.y;
      points = c.points;

      return (*this);
    }*/

    //-------------------------------------------------------------------------
    void AddPoints(const Circle &c)
    {
      int size1 = pointsNumbs.size();
      int size2 = c.pointsNumbs.size();

      pointsNumbs.resize(size1 + size2);

      for (int i = 0; i < size2; i++)
      {
        pointsNumbs[size1 + i] = c.pointsNumbs[i];
      }
    }
  };

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  void DetectMax(std::vector< std::vector<int> > &maxima,
                 ImageContainer< std::vector<T> > &accum,
                 std::vector< std::vector<Circle> > &centers,
                 int w,
                 int h,
                 int nls,
                 int thresh)
  {
    std::cout << std::endl << std::endl << std::endl << "Maxima:";

    int rNumb = accum.size();

    // for each radius layer
    for (int r = 1; r < rNumb - 1; r++)
    {
      std::vector<T> *_radLayer0 = accum[r - 1];
      std::vector<T> *_radLayer1 = accum[r];
      std::vector<T> *_radLayer2 = accum[r + 1];

      std::vector<float *> radL(3);
      radL[0] = &((*_radLayer0)[0]);
      radL[1] = &((*_radLayer1)[0]);
      radL[2] = &((*_radLayer2)[0]);

      // for each point in layer
      for (int c_x = 1; c_x < w - 1; c_x++)
      {
        for (int c_y = 1; c_y < h - 1; c_y++)
        {
          if (*(radL[1] + c_y * nls + c_x) > 0)
          {
            if (IsMax<float *, float>(radL, c_x, c_y, nls))
            {
              int n = FindCell(centers[r], c_x, c_y);

              if (n >= 0)
              {
                if (int(centers[r][n].pointsNumbs.size()) >= thresh)
                {
                  std::cout << std::endl << "max: x = " << c_x << "  y = " << c_y
                  << "  r = " << (5.0 * pow(1.1, r)) << "  pNumb = "
                  << centers[r][n].pointsNumbs.size();

                  maxima[r].push_back(n);

                  *(radL[1] + c_y * nls + c_x) = T(255);
                }
              }
            }
          }
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  void ClassifyViaAccum(std::vector< std::vector< Circle > > &clust,
                        std::vector< std::vector<T> > &d,
                        std::vector< std::vector<int> > &maxima,
                        std::vector< std::vector< Circle > > &centers)
  {
    int rNumb = centers.size(); 

    // Circle.pointsNumbs containt numbers of centers here
    clust.resize(rNumb);
    d.resize(rNumb);

    // points from each radius layer classify between maxima 
    for (int r = 0; r < rNumb; r++)
    {
      int maxNumb = maxima[r].size();  // number of maxima on the current layer

      //std::cout << std::endl << "r = " << (5.0 * pow(1.15, r))
      // << " maxNumb = " << maxNumb;

      if (maxNumb > 0)
      {
        for (int m = 0; m < maxNumb; m++)
        {
          int i = maxima[r][m];
          int x = centers[r][i].x;
          int y = centers[r][i].y;

          clust[r].push_back(Circle(x, y));

          //std::cout << std::endl << "      " << x << " " << y;
        }

        d[r].resize(maxNumb, 0);

        int centNumb = centers[r].size(); // number of centers on the layer
        
        // for each center on the layer define the closest maxima
        for (int c = 0; c < centNumb; c++)
        {
          //std::cout << std::endl << "c = " << c << "  x = " << centers[r][c].x
          //  << "  y = " << centers[r][c].y;

          T minDist = T(clust[r][0].Dist(centers[r][c]));
          int minI = 0;
          //std::cout << "  " << minDist;

          for (int j = 1; j < maxNumb; j++)
          {
            T dist = T(clust[r][j].Dist(centers[r][c]));

            //std::cout << "  " << dist;

            if (dist < minDist)
            {
              minDist = dist;
              minI = j;
            }
          }

          //std::cout << "  minI = " << minI;

          clust[r][minI].pointsNumbs.push_back(c);

          if (minDist > d[r][minI])
          {
            d[r][minI] = minDist;
          }
        }

        for (int m = 0; m < maxNumb; m++)
        {
          std::cout << std::endl << int(clust[r][m].pointsNumbs.size());
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  void DrawCentClusters(ImageContainer< std::vector<T> > &accum,
                        std::vector< std::vector<Circle> > &clust,
                        std::vector< std::vector<Circle> > &centers,
                        int w,
                        int h,
                        int nls)
  {
    int rNumb = accum.size();

    for (int r = 0; r < rNumb; r++)
    {
      std::vector<float> *_radLayer = accum[r];
      float *radL = &((*_radLayer)[0]);

      ThreshDraw(radL, w, h, T(0), T(0));

      int clustNumb = clust[r].size();

      for (int cl = 0; cl < clustNumb; cl++)
      {
        T color = T(255.0 / clustNumb * cl + 1);

        int pNumb = clust[r][cl].pointsNumbs.size();

        for (int p = 0; p < pNumb; p++)
        {
          int i = clust[r][cl].pointsNumbs[p];
          int x = centers[r][i].x;
          int y = centers[r][i].y;

          //std::cout << std::endl << x << "  " << y;

          *(radL + y * nls + x) = color;
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename T, typename IterType>
  void DrawPointsClust(Circle &clust,
                       std::vector< DatumR<T> > &points,
                       IterType img,
                       int w,
                       int h,
                       int nls,
                       T color)
  {
    int pNumb = clust.pointsNumbs.size();

    for (int p = 0; p < pNumb; p++)
    {
      int n = clust.pointsNumbs[p];
      int x = points[n].x;
      int y = points[n].y;

      *(img + y * nls + x) = color;
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  template <typename T>
  void ClustCenters(std::vector< std::vector< Circle > > &clust,
                    std::vector< std::vector<T> > &d,
                    std::vector< std::vector< Circle > > &centers)
  {
    std::cout << std::endl << std::endl << std::endl << "ClustCenters";
    int rNumb = centers.size();

    for (int r = 0; r < rNumb; r++)
    {
      for (int cl = 0; cl < int(clust[r].size()); cl++)
      {
        std::cout << std::endl << std::endl << "   clust " << cl << " "
          << clust[r][cl].x << " " << clust[r][cl].y;

        while (1)
        {
          int pNumb = clust[r][cl].pointsNumbs.size();

          std::cout << std::endl << "      pNumb = " << pNumb;

          double dev = 0.0;
          int n = 0;

          for (int p = 0; p < pNumb; p++)
          {
            int i = clust[r][cl].pointsNumbs[p];
            int k = centers[r][i].pointsNumbs.size();
            double dist = clust[r][cl].Dist(centers[r][i]);

            if (dist < 0.25 * (5.0 * pow(1.1, r)))
            {
              dev += dist * k;
              n += k;
            }
            //std::cout << std::endl << "         " << clust[r][cl].Dist(centers[r][i])
            //  << "  " << k;
          }

          dev /= n;

          std::cout << " => n = " << n << "  dev = " << dev
            << "  r = " << d[r][cl] << std::endl;

          if (d[r][cl] > 3 * dev)
          {
            std::cout << "  need to erase some points";

            double maxD = 3 * dev;

            d[r][cl] = T(0);
            
            for (int p = pNumb - 1; p >= 0; p--)
            {
              int i = clust[r][cl].pointsNumbs[p];
              double dist = clust[r][cl].Dist(centers[r][i]);

              //std::cout << std::endl << "         " << dist;

              if (dist > maxD)
              {
                //std::cout << " => erase";
                clust[r][cl].pointsNumbs.erase(clust[r][cl].pointsNumbs.begin()
                                               + p);
              }
              else
              {
                if (dist > d[r][cl])
                {
                  d[r][cl] = dist;
                }
              }
            }
          }
          else
          {
            break;
          }
        }
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  void ClustEdgePoints(std::vector< Circle >  &pointsClust,
                       std::vector< std::vector< Circle > > &centClust,
                       std::vector< std::vector< Circle > > &centers,
                       int thresh)
  {
    int rNumb = centers.size();

    for (int r = 0; r < rNumb; r++)
    {
      int clustNumb = int(centClust[r].size());

      for (int cl = 0; cl < clustNumb; cl++)
      {
        pointsClust.push_back(Circle(centClust[r][cl].x, centClust[r][cl].y));
        int n = pointsClust.size() - 1;

        int centNumb = centClust[r][cl].pointsNumbs.size();

        for (int c = 0; c < centNumb; c++)
        {
          int i = centClust[r][cl].pointsNumbs[c];
          pointsClust[n].AddPoints(centers[r][i]);
        }

        if (int(pointsClust[n].pointsNumbs.size()) < thresh)
        {
          pointsClust.erase(pointsClust/*.end()*/.begin() + n);
        }
        else
        {
          std::cout << std::endl << "r = " << (5.0 * pow(1.1, r)) << "  x = "
            << pointsClust[n].x << "  y = " << pointsClust[n].y << "  pNumb = "
            << pointsClust[n].pointsNumbs.size();
        }
      }
    }

  }

}
}


#endif // CLUST_VIA_ACCUM_H