#ifndef SINOGRAM_H
#define SINOGRAM_H

// system includings
#define _USE_MATH_DEFINES
#include <cmath>

// library includings
#include "integral_transforms.h"
#include "imalg_resampling.h"


namespace Vtl
{
namespace CurveDetection
{
  template <typename Real_t>
  void BuildSinogram(Real_t* img,Real_t* sinogr,size_t w,size_t h,const char *base_fname,const char *postfix)
	{
			//  :

		//////	std::ostringstream os;
		//////os <<	base_fname << "_Original.tif";
		//////std::string outfname = os.str();
		SaveSinogr(img, w, h, base_fname,postfix,"Original");

//    SubtrAver(img, w, h);

		size_t size[] = {w, h};
		Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2, false>(img, size);
		Vtl::NumAlg::IntTransf::Hartley::fht2D(img, w, h);
		Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<2, false>(img, size);

		//////std::ostringstream os1;
		//////os1 <<	base_fname << "_Swaped.tif";
		//////outfname = os1.str();
		SaveSinogr(img, w, h, base_fname,postfix,"Swaped");
		
		Real_t c = 2 * Real_t(1.4);
		Vtl::ImAlg::Resamp::LinGrid<Real_t> lng(-Real_t(h / c), Real_t(h / c), h);
		Vtl::ImAlg::Resamp::PhiGrid<Real_t> phig(Real_t(0), Real_t(M_PI), w);
		Vtl::ImAlg::Resamp::InterpBilinear<Real_t> interp(img, w, h, w, Real_t(0));
		Vtl::ImAlg::Resamp::CartToPolar(interp, Real_t(w / 2.0), Real_t(h / 2.0),
			                              sinogr, w, h, w, lng, phig);

		//////std::ostringstream os2;
		//////  os2 <<	base_fname << "_Polar.tif";
		//////outfname = os2.str();
		SaveSinogr(sinogr, w, h, base_fname,postfix,"Polar");

		typedef Vtl::ImAlg::WndFun::DoG<Real_t, 0> Filt;
		Filt f(4, w, 1);
		Vtl::Plf::Kernel::FKern<Real_t, Vtl::Plf::Kernel::KT_Win, false> ker(f, w,
			                                               Real_t(w / 2), Real_t(0));
		for(size_t y = 0; y < h; ++y)
		{
			Real_t *py = sinogr + h * y;
			Vtl::Plf::Kernel::WindowData(py, ker.begin(), ker.end(), 1, 0);

			for(size_t x = 0; x < w; ++x)
			{
				c = 1;
				c -= exp(-Vtl::sqr((x - w/2.0f) / (4.0f)) / 2.0f);
				py[x] *= c;
			}
		}

		//////std::ostringstream os3;
		//////os3 <<	base_fname << "_Polar_f.tif";
		//////outfname = os3.str();
		SaveSinogr(sinogr, w, h, base_fname,postfix,"Polar_f");
		
		Real_t *py = sinogr;
		size_t nn[] = {w};

		for(size_t y = 0; y < h; ++y, py += w)
		{
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1, false>(py, nn);
			Vtl::NumAlg::IntTransf::Hartley::fht1D(py, w);
			Vtl::NumAlg::IntTransf::Utils::SwapQuadrants<1,false>(py, nn);
		}

		//////std::ostringstream os4;
		//////os4 <<	base_fname << "_Sinogram.tif";
		//////outfname = os4.str();
		SaveSinogr(sinogr, w, h, base_fname,postfix,"Sinogram");
	}

	//---------------------------------------------------------------------------
  template<typename T>
	std::pair<T,T> MinMaxImage(const T* p, int w, int h, int nls)
	{
		T min_v = p[0];
		T max_v = p[0];
		//std::cout << std::endl << "p[0] = " << p[0];

		for(int y = 0; y < h; ++y, p += nls)
			for(int x = 0; x < w; ++x)
			{
				if(p[x] < min_v)
				{
					min_v = p[x];
					//std::cout << std:: endl << "min: " << x << "  " << y << "  " << p[x];
				}
				else if(p[x] > max_v)
				{
					max_v = p[x];
					//std::cout << std:: endl << "max: " << x << "  " << y << "  " << p[x];
				}
			}
		
		return std::pair<T,T>(min_v, max_v);
	}

	//---------------------------------------------------------------------------
	template <typename T>
	void SaveSinogr(const T* img,
		              int w,
									int h,
									const char* base_name,const char* suffix,const char* postfix)
	{
			std::ostringstream os;
			os <<	YFileSys::getNameWithoutExt(base_name)
				<< "_#_FHT_" << suffix << "_" << postfix
				<< YFileSys::getExt(base_name);
			std::string outfname=os.str();


		std::pair<T, T> min_max = MinMaxImage(img, w, h, w);
		std::cout << std::endl << outfname << ": " << min_max.first << " to "
			        << min_max.second << std::endl;

		T min = min_max.first;
		T coef = min_max.second - min_max.first;
		if (coef > 0)
		{
		  coef =  T(255) / coef;
		}
		else
		{
		  coef = T(1);
		}

	  std::auto_ptr<AImage> _dst(::AImageCreate(0, AImage::IMGC_GRAYSCALE, w, h, 8));
		AImageMem8* dst = dynamic_cast<AImageMem8*>(_dst.get());

		AImageMem8::Pixel *p_d = (*dst)[0];

		for (int y = 0; y < h; y++)
		{
			for (int x = 0; x < w; x++)
			{
				(*p_d) = AImageMem8::Pixel(((*img) - min) * coef);
				//(*p_d) = (*img);
				
				p_d++;
				img++;
			}
		}

		::AImageSave(dst, outfname.c_str());

		min_max = MinMaxImage((*dst)[0], w, h, w);
		std::cout << outfname << ": " << min_max.first << " to "
			        << min_max.second << std::endl;
	}

	//---------------------------------------------------------------------------
	template <typename T>
	void SubtrAver(T* src,
		             int w,
								 int h)
	{
	  T aver = T(0);
		T c = T(w * h);
    T* begin = src;

		for (int y = 0; y < h; y++)
		  for (int x = 0; x < w; x++)
		{
			aver += (*src) / c;
			src++;
		}
		
		src = begin;
    for (int y = 0; y < h; y++)
		  for (int x = 0; x < w; x++)
		{
		  (*src) -= aver;
			src++;
		}

		std::cout << std::endl << "average = " << aver;
	}
}
}

#endif // SINOGRAM_H