#ifndef HYSTERESIS_H
#define HYSTERESIS_H

// std includings
#include <vector>

template< typename T>
class Hysteresis
{
private:
  T* src;
  T* dst;
  int w;
  int h;
  T thStart;
  T thEnd;
  std::vector<T> mask;

  //============================================================================
public:
  Hysteresis(T* in_src,
             int in_w,
             int in_h)
  {
    src = in_src;
    w = in_w;
    h = in_h;
    
    mask.resize(w * h, T(0));
  }

  //---------------------------------------------------------------------------
  void Apply(T in_thStart,
             T in_thEnd,
             T* in_dst,
             T resVal = T(0))
  {
    dst = in_dst;

    thStart = in_thStart;
    thEnd = in_thEnd;

    BuildMask();

    int sz = w * h;

    // if resVal == 0 draw extended areas with colour from src
    // otherwise - with the passed value
    if (resVal == T(0))
    {
      for (int i = 0; i < sz; i++)
      {
        *(dst + i) = mask[i] * *(src + i);
      }
    }
    else
    {
      for (int i = 0; i < sz; i++)
      {
        *(dst + i) = mask[i] * resVal;
      }
    }
  }

  //---------------------------------------------------------------------------
private:

  void AssignMask()
  {
    int size = w * h;

    for (int i = 0; i < size; i++)
    {
      mask[i] = T(0);
    }
  }

  //---------------------------------------------------------------------------
  void BuildMask()
  {
    for (int y = 0; y < h; y++)
    {
      for (int x = 0; x < w; x++)
      {
        if (*(src + y * w + x) > thStart)
        {
          FollowLine(x, y);
        }
      }
    }
  }

  //---------------------------------------------------------------------------
  void FollowLine(int x,
                  int y)
  {
    mask[y * w + x] = T(1);

    for (int i = -1; i <= 1; i++)
    {
      for (int j = -1; j <= 1; j++)
      {
        if (i == 0 && j == 0)
        {
          continue;
        }

        CheckNeighbour(x + i, y + j);
      }
    }
  }

  //---------------------------------------------------------------------------
  void CheckNeighbour(int x,
                      int y)
  {
    if (x >= 0)
      if (x < w)
        if (y >= 0)
          if (y <= h)
            if (*(src + y * w + x) > thEnd)
              if (mask[y * w + x] == T(0))
                FollowLine(x, y);
  }
};

#endif // HYSTERESIS_H