
#ifndef STAT_DIFF_H
#define STAT_DIFF_H

// system includings
#define _USE_MATH_DEFINES
#include <cmath>

// library includings
#include "convolve.h"
#include "plf_kernels_1d.h"

template< typename T >
void GetStatCharact(T* mean,
                    T* sq,
                    const T* img,
                    const int w,
                    const int h,
                    float sigma,
                    float nsigmas_env)
{
  //    
	memcpy(mean, img, w * h * sizeof(T));
	Vtl::ImAlg::Tools::copy_1_1(img, w, h, w, sq, w, Vtl::Plf::F11::sqr<T>);
	
	// ,    
	// 
	typedef Vtl::ImAlg::WndFun::Gauss<float, 0, false> FF0; 
	FF0 f0(sigma, nsigmas_env);
	typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K0;
	K0 ker0(f0);
	Vtl::ImAlg::Conv::ConvInplace<float, K0> g0(ker0);

	//  
	Vtl::ImAlg::Conv::Conv2D::conv2DInplace(mean, w, h, w, g0, g0);
	Vtl::ImAlg::Conv::Conv2D::conv2DInplace(sq, w, h, w, g0, g0);

  //   .
	Vtl::Plf::F21::Disp<T> dsp;
	Vtl::ImAlg::Tools::copy_2_1(mean, w, h, w, sq, w, sq, w, dsp);
}

//===================================================================
// Func: Vtl::Plf::F31::StatDiff<T> or Vtl::Plf::F31::StatDiffThresh<T>
template< typename T, typename Func >
void StatDiff(T* result,
              const T* img,
              const int w,
              const int h,
              const float sigma,
              const float nsigmas_env,
              Func stDiffFunc)
{
  const int sz = w * h;

  //      (). 
  //    1-    ,  
  //  , ,  .
	std::vector<T> _data0(sz);
	std::vector<T> _data1(sz);
	std::vector<T> _data2(sz);

  //      .
	T *mean = &(_data0[0]);
	T *sq = &(_data1[0]);
	
  GetStatCharact(mean, sq, img, w, h, sigma, nsigmas_env);

  //   
  Vtl::ImAlg::Tools::copy_3_1(img, w, h, w, mean, w, sq, w, result, w, stDiffFunc);
}

//=============================================================================
namespace Vtl
{
namespace Plf
{
namespace F31
{
  template<typename T>
	class StatDiffThresh
	{
		float A;
		float sigma_d;
		float m_d;
		float alpha;
		float thresh;

	public:
		StatDiffThresh(float amplification = 6.0f,
			             float sigma_desired = 85.0f,
									 float mean_desired = 128.0f,
									 float phone_to_edges = 0.8f,
									 float in_thresh = 5.0f) 
		{
      A = amplification;
			sigma_d = sigma_desired;
			m_d = mean_desired;
			alpha = phone_to_edges;
			thresh = in_thresh;
		}

		//-------------------------------------------------------------------------
		T operator()(T f,
			           T f_mean,
								 T f_sigma) const 
		{
			T coef = /*T(1);*/T(atan((f_mean - thresh) / 2.0) / M_PI_2 + 0.5);

			return /*(f - f_mean) * (A * sigma_d / (A * f_sigma + sigma_d)) * coef + 
				+ alpha * m_d + (1 - alpha) * f_mean*/
        f * (A * sigma_d / (A * f_sigma + sigma_d)) * coef;
		}
	};
}
}
}

//=============================================================================
template< typename T >
void GradStatDiff(T *gradMod,
                  T *statDiff,
                  T *Lx,
                  T *Ly,
                  int w,
                  int h)
{
  for (int y = 0; y < h; y++)
  {
    for (int x = 0; x < w; x++)
    {
      T coef;
			
			if ((*gradMod) == 0)
			{
			  coef = 1.0f;
			}
			else
			{
			  coef = (*statDiff) / (*gradMod);
			}

			std::cout << std::endl << coef;
      *Lx *= coef;
      *Ly *= coef;

      statDiff++;
      gradMod++;
      Lx++;
      Ly++;
    }
  }
}

#endif // STAT_DIFF_H