#ifndef CURVATURE_FUNC_H
#define CURVATURE_FUNC_H

// std includings
#include <string>
#include <sstream>
#include <memory> 
#include <list>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <time.h>
#include <assert.h>

// library includings
#include "yfilesys.h"
#include "aimage.h"
#include "color_spaces.h"
#include "convolve.h"
#include "imalg_tools.h"
#include "plf_functors.h"
#include "plf_kernels_1d.h"
#include "tests.h"
#include "consolap.h"
#include "edt_meijster_2000.h"

// own includings
#include "curvature/CurvatureScale.h"
#include "ImageContainer.h"
#include "curvature/CombinedKappaLines.h"
#include "curvature/Histogramm.h"
#include "curvature/Sigma.h"
#include "SaveImageArray.h"
#include "EdgeDetection/CannyEdgeDetector.h"
#include "curvature/EdtEdgeBlur.h"
#include "EdtEdgeCoord.h"
#include "curvature/Preprocessing.h"
#include "curvature/EdgeCurvature.h"
#include "curvature/IsolineCurvature.h"

namespace Vtl
{
namespace Curvature
{
  template<typename T>
	T if_not_zero(T value,T mask)
	{
		return (mask!=T(0)) ? value : T(0);
	}

	/////////////////////////////////////////////////////////////////////////////	
  /////////////////////////////////////////////////////////////////////////////	

	template<typename T>
	void calc_means_and_stddev(const T& hist,double& mean,double& stddev)
	{
		double sum=0;
		double sumsqr=0;
		double N=0;
		for(size_t i=0; i<hist.size(); ++i)
		{
			double v = fabs(CurvatureScale<double>::value_to_radius(i)); 
			double n = hist[i];
			N += n;
			sum += v * n;
			sumsqr += v*v * n;
			/*std::cout << std::endl << i << "  v = " << v << "   n = " << n
				<< "   sum = " << sum << "  sumSqr = " << sumsqr;*/
		}
		mean = sum/N;
		stddev = sqrt( sumsqr/N - mean*mean );
	}

	/////////////////////////////////////////////////////////////////////////////
  /////////////////////////////////////////////////////////////////////////////
	
  void MultiscaleCurvatureEst(const char *fname,
		                          const int w,
															const int h,
															float *data,
															float *Canny,
															float *comb,
															float *GradX,
															float *GradY,
															std::vector<double> &hist,
															AImageMem8 *img_pal,
															AImageMem8 *hist_img,
															Vtl::Plf::F11::Convert2ByteBounded &c_to_Byte,
															Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert2ByteBounded > &adj_3mod)
	{
// to estimate multiscale curvature from isolines
#define CDI_ESTIMATE_CURVATURE_USING_ISOLINES_FORMULA

// to estimate multiscale curvature from edges
#ifndef CDI_ESTIMATE_CURVATURE_USING_ISOLINES_FORMULA
#define CDI_ESTIMATE_CURVATURE_USING_EDGE_FORMULA
#endif

// to save auxiliary images
//#define CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS

    //-----------------------------------------------------------------------------
    // PREPROCESSING 
	  //! Select only one proprocessing!

    int size = w * h;
    std::vector<float> _data15(size);
    float *preproc = &(_data15[0]);

#define PREPROC_ALG_TO_ISO 1

#if PREPROC_ALG_TO_ISO == 1
		preproc_image_DoNothing<float>(data, w, h, preproc);
#ifdef CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS
		//Tests::SaveImage(img     ,preproc ,w,h,w,  c_to_Byte, 0,fname,"PreProc_DoNothing");
		Tests::SaveImageArray(preproc, w, h, fname, "_PreProc_DoNothing");
#endif

#elif PREPROC_ALG_TO_ISO == 2
		preproc_image_StatDiff(data, w, h, preproc, 32.0f, 2.0f);
#ifdef CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS
		//Tests::SaveImage(img     ,preproc ,w,h,w,  c_to_Byte, 32,fname,"PreProc_StatDiff");
		Tests::SaveImageArray(preproc, w, h, fname, "_PreProc_StatDiff");
#endif

#elif PREPROC_ALG_TO_ISO == 3
		memcpy(data, preproc, sz * sizeof(float));// Double preprocessing StatDiff->ZeroCross
		preproc_image_ZeroCross(data, w, h, preproc, 4.0f,4 .0f);
#ifdef CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS
		//Tests::SaveImage(img     ,preproc ,w,h,w,  adj_3mod , 4,fname,"PreProc_ZeroCross");
    Tests::SaveImageArray(preproc, w, h, fname, "_PreProc_ZeroCross");
#endif
#else
	#error 'unknown preprocessing'
#endif
	//-------------------------------------------------

		memcpy(data, preproc, size * sizeof(float));

		ImageContainer< std::vector<float> > kappaSet;

		float minSigma = 2.5f;
		float maxSigma = 25.0f;
		Vtl::Curvature::Sigma<float> sigma(minSigma, float(pow(2.0f, 1/3.0f)),
                                                                     maxSigma);
		const float nsigmas_env = 6.0f;

		indicator->start("Generate step edge curvature map", 
                           int(100 * log(minSigma)), int(100 * log(maxSigma)));
		// Loop in sigma (Scale space levels)
		for (float s = sigma.Get(); s > 0; s = sigma.Next())
		{
			indicator->current(int(100 * log(s)));

#ifdef CDI_ESTIMATE_CURVATURE_USING_EDGE_FORMULA
      EdgeCurvature<float> curvEstimator(data, w, h);
#endif
#ifdef CDI_ESTIMATE_CURVATURE_USING_ISOLINES_FORMULA
			IsolineCurvature<float> curvEstimator(data, w, h);      
#endif

      curvEstimator.CalculateDerivatives_direct(s, nsigmas_env);			

      std::vector<float> _data11(size);
      std::vector<float> _data18(size);
      float *Kappa = &(_data11[0]);
      float *Kappa_lines = &(_data18[0]);
			
			//  edge
			for(int i = 0; i < size; ++i)
			{
        Kappa[i] = curvEstimator(i);
      }
				
      Vtl::ImAlg::Tools::copy_2_1(Kappa, w, h, w, Canny, w, Kappa_lines, w,
				                          Vtl::Curvature::if_not_zero<float>);
			kappaSet.Add(_data18);
#ifdef CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS
			Tests::SaveImage(img_pal, Kappa,       w, h, w, c_to_Byte, s, fname, "Kappa");
			Tests::SaveImage(img_pal, Kappa_lines, w, h, w, c_to_Byte, s, fname, "Kappa_lines");
#endif

			//Histogramm::Build(hist,Kappa_lines,w,h,w,Histogramm::TtoInt);

#ifdef CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS
			{//Build and save curvatures histogramm as image at each scale, print data for matlab accuracy analysis for multiscale algorithm.
				Vtl::Histogramm::TtoAbsIntBounded<float> t2uint(hist.size() - 1);
				double mean;
				double stddev;
				Vtl::Histogramm::Build(hist, Kappa_lines, w, h, w, t2uint);
				hist[0] = 0;

				Vtl::Histogramm::Draw(&(hist[0]), hist.size(), (*hist_img)[0],
				hist_img->Width(), hist_img->Height(), hist_img->NextLineShift());
				Tests::SaveImage(hist_img, s, fname, "Hist");
				
				Vtl::Curvature::calc_means_and_stddev(hist, mean, stddev);

				// This output was used to build graphs of curvature estimation accuracy 
				// versus on sigma and actual curvature.
				// The image is supposed synthetic, 
				// containing a (set of) (fragment(s)) circle(s) of single known radius.
				std::cout << std::endl << "  " << s << "    " << mean << " " 
							<< stddev << " " 	<< stddev / mean;
			}
#endif

	//{
			//	std::cout << "hist = " << std::endl;
			//	for(size_t i=0; i<hist.size() ; ++i)
			//		std::cout << "[" << i <<"]= " << hist[i] << std::endl;
			//	std::cout <<" --- " << std::endl;
			//}
/*
      Vtl::Histogramm::Draw(&(hist[0]),hist.size(),(*Hist_img)[0],Hist_img->Width(),Hist_img->Height(),Hist_img->NextLineShift());
      SaveImage(Hist_img,sigma,fname,"Hist");
			std::vector<double>::iterator hist_max=std::max_element(hist.begin(),hist.end());
			int i_hist_max=hist_max-hist.begin();
//			std::cout << "i_hist_max= " << i_hist_max << "   hist[" << i_hist_max <<"]= " << hist[i_hist_max]<< std::endl;
			double alpha;
			int l_min_ent;
			int l_max_ent;
			double *hst=&(hist[0]);
			alpha=Vtl::Histogramm::FindSecondMax(hst,i_hist_max,hist.size()-1,l_min_ent,l_max_ent);
//			std::cout << "l_max_ent= " << l_max_ent << "   l_min_ent=" << l_min_ent <<"   alpha= " << alpha << std::endl;
			alpha=Vtl::Histogramm::FindSecondMax(hst,i_hist_max,0,l_min_ent,l_max_ent);
//			std::cout << "l_max_ent= " << l_max_ent << "   l_min_ent=" << l_min_ent <<"   alpha= " << alpha << std::endl;
//			std::cout << "  ======== " << alpha << std::endl;
*/	}
		indicator->finish();

    Vtl::Curvature::CombinedKappaLines(comb, Canny, GradX, GradY, w, h, 5.0f,
			                                 kappaSet, sigma);
		Tests::SaveImage(img_pal, comb, w, h, w, c_to_Byte, 0.0f, fname, "Multiscale");

		const int radius_min = 0;
		const int radius_max = 10000000;
		Vtl::Curvature::SelectCurvatureRange<float>(comb,w,h,radius_min,radius_max);
		Tests::SaveImage(img_pal, comb, w, h, w, c_to_Byte, 0.0f, fname, "Multiscale_no_high");

		Vtl::Histogramm::TtoAbsIntBounded<float> t2uint(hist.size() -1);
		Vtl::Histogramm::Build(hist, comb, w, h, w, t2uint);  
    hist[0] = 0;

#ifdef CDI_SAVE_1SCALE_CURVATURES_AND_HISTOGRAMS
		Vtl::Histogramm::Draw(&(hist[0]), hist.size(), (*hist_img)[0],
			       hist_img->Width(), hist_img->Height(), hist_img->NextLineShift());
		Tests::SaveImage(hist_img, 0.0f, fname, "Hist");

		double mean;
		double stddev;
		Vtl::Curvature::calc_means_and_stddev(hist, mean, stddev);
		std::cout << std::endl << "result" << std::endl << mean << " "
			        << stddev << " " << stddev / mean;
#endif
	}

  ///////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////
}
}



#endif //CURVATURE_FUNC_H


///////////////////////////////////////////////////////////////////////////////
/*    const float hisThStart = 10.0f;
      const float hisThEnd = 0.0f;
      
      CannyEdgeDetector<float> canny(data, w, h);
      canny.Apply(s, nsigmas_env, hisThStart, hisThEnd);

			const float edgeLevel = 128.0f;
      const int sz = w * h;

			//Levelling(&(canny.edges[0]), sz, edLevel);
      std::vector<float> edt(sz);
      std::vector<int> x(sz);
		  std::vector<int> y(sz);

      Vtl::ImAlg::Edt::EdtEdgeCoord::edt(&(canny.edges[0]), w, h, w, &(edt[0]),
                        &(x[0]), &(y[0]), w, Vtl::ImAlg::Edt::isEdge<float>());	

      EdtEdgeBlur(&(canny.edges[0]), Lx, Ly, &(edt[0]), &(x[0]), &(y[0]), w, h,               
                                                                 s, edgeLevel);
      //SaveNormImageArray(&(edt[0]), w, h, "test", "_EDT");
      //SaveNormImageArray(&(canny.edges[0]), w, h, "test", "_canny");
      //SaveNormImageArray(Lx, w, h, "test", "_Lx");
      //SaveNormImageArray(Ly, w, h, "test", "_Ly");
								       
			//typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0_bg; 
			//FF0_bg f0_bg(2 * s, nsigmas_env);
			//typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0_bg;
			//K0_bg ker0_bg(f0_bg);
			//Vtl::ImAlg::Conv::ConvInplace<float,K0_bg> g0_bg(ker0_bg);

			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(canny.Lx[0]),w,h,w,g0_bg,g0_bg);
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(canny.Ly[0]),w,h,w,g0_bg,g0_bg);

      memcpy(Lxx, &(canny.Lx[0]), sz * sizeof(float));
			memcpy(Lxy, &(canny.Lx[0]), sz * sizeof(float));
			memcpy(Lyy, &(canny.Ly[0]), sz * sizeof(float));
				
			//2 
			//typedef Vtl::ImAlg::WndFun::Gauss<float,2,false> FF2; 
			//FF2 f2(s,nsigmas_env);
			//typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K2;
			//K2 ker2(f2);
			//Vtl::ImAlg::Conv::ConvInplace<float,K2> g2(ker2);

			// 2 
			//  Lxx:
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxx_,w,h,w,g2,g0);
			//  Lxy:
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxy_,w,h,w,g1,g1);
			//  Lyy:
			//Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lyy_,w,h,w,g0,g2);

			
			typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0_sm; 
			FF0_sm f0_sm(2.0f,nsigmas_env);
			typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0_sm;
			K0_sm ker0_sm(f0_sm);
			Vtl::ImAlg::Conv::ConvInplace<float,K0_sm> g0_sm(ker0_sm);
			
			typedef Vtl::ImAlg::WndFun::Gauss<float,1,false> FF1_sm; 
			FF1_sm f1_sm(2.0f,nsigmas_env);
			typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K1_sm;
			K1_sm ker1_sm(f1_sm);
			Vtl::ImAlg::Conv::ConvInplace<float,K1_sm> g1_sm(ker1_sm);
			
			// 2   1
			//  Lxx:
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxx,w,h,w,g1_sm,g0_sm);
			//  Lxy:
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lxy,w,h,w,g0_sm,g1_sm);
			//  Lyy:
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lyy,w,h,w,g0_sm,g1_sm);

      //  1 
      Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(canny.Lx[0]),w,h,w,g0_sm,g0_sm);
			Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(canny.Ly[0]),w,h,w,g0_sm,g0_sm);*/
