#ifndef CURVATURE_SCALE_H
#define CURVATURE_SCALE_H

namespace Vtl
{
namespace Curvature
{
	// The class provides functions to make convertions between curvature 
	// value (kappa), curvature radius and "savable" presentation of radius
  template<typename T>
	class CurvatureScale
	{
		enum { Rmax = 100000 };
    enum { Vmax = 255 };

	public:
		static T value_to_radius(T v)
		{
			T absV = fabs(v);
			T sgnV = ( v != 0 ) ? v / absV : 1 ;
			return T( sgnV * exp( absV/Vmax * log(float(Rmax)) ) ); //=R;
		}

		static T radius_to_value(T R)
		{
			T absR = fabs(R);
			T sgnR = ( R != 0 ) ? R / absR : 1 ;
			return sgnR * log(std::max(T(1),absR)) / log(float(Rmax)) * Vmax;
		}

		static T kappa_to_radius(T kappa_nominator,T kappa_denominator)
		{
			T nom=abs(kappa_nominator);
			T den=abs(kappa_denominator);
			return T( ( nom*Rmax > den ) ?
				       /*den/nom*/kappa_denominator/kappa_nominator: Rmax ); //    ,    !!!
		}
	};
} // namespace Curvature
} // namespace Vtl


#endif CURVATURE_SCALE_H