#ifndef ISOLINE_CURVATURE_H
#define ISOLINE_CURVATURE_H

// own includes
#include "CurvatureScale.h"

namespace Vtl
{
namespace Curvature
{
  template< typename T >
  class IsolineCurvature
  {
    //  Arrays for keeping auxiliary data (images)
    std::vector<T> Lx;
	  std::vector<T> Ly;
	  std::vector<T> Lxx;
	  std::vector<T> Lxy;
	  std::vector<T> Lyy;
  	
    T* data;

    int w;
    int h;
    int size;

  	
	  //-------------------------------------------------------------------------
  public:
	  IsolineCurvature(T* in_data,
                     int in_w,
		                 int in_h)
	  {
      data = in_data;

      w = in_w;
      h = in_h;
      size = w * h;

		  Lx.resize(size);
		  Ly.resize(size);
      Lxx.resize(size);
      Lxy.resize(size);
      Lyy.resize(size);
	  }

    //-------------------------------------------------------------------------
    void CalculateDerivatives_direct(T sigma,
                                     T nSigmasEnv)
    {
      //    
	    memcpy(&(Lx[0]), data, size * sizeof(T));
	    memcpy(&(Ly[0]), data, size * sizeof(T));
  	
	    memcpy(&(Lxx[0]), data, size * sizeof(T));
	    memcpy(&(Lxy[0]), data, size * sizeof(T));
	    memcpy(&(Lyy[0]), data, size * sizeof(T));
  	  
      //  ,    
      //     1,2 
      //  
	    typedef Vtl::ImAlg::WndFun::Gauss<float, 0, false> FF0; 
	    FF0 f0(sigma, nSigmasEnv);
	    typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K0;
	    K0 ker0(f0);
	    Vtl::ImAlg::Conv::ConvInplace<float, K0> g0(ker0);

	    // 1 
	    typedef Vtl::ImAlg::WndFun::Gauss<float, 1, false> FF1; 
	    FF1 f1(sigma, nSigmasEnv);
	    typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K1;
	    K1 ker1(f1);
	    Vtl::ImAlg::Conv::ConvInplace<float, K1> g1(ker1);

      //2 
		  typedef Vtl::ImAlg::WndFun::Gauss<float, 2, false> FF2; 
		  FF2 f2(sigma, nSigmasEnv);
		  typedef Vtl::Plf::Kernel::FKern<float, Vtl::Plf::Kernel::KT_Conv, true> K2;
		  K2 ker2(f2);
		  Vtl::ImAlg::Conv::ConvInplace<float,K2> g2(ker2);

	    //  
	    // 1 
	    //  x:
	    Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lx[0]), w, h, w, g1, g0);
	    //  y:
	    Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Ly[0]), w, h, w, g0, g1);

      // 2 
		  //  xx:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxx[0]), w, h, w, g2, g0);
		  //  xy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lxy[0]), w, h, w, g1, g1);
		  //  yy:
		  Vtl::ImAlg::Conv::Conv2D::conv2DInplace(&(Lyy[0]), w, h, w, g0, g2);
    }

    //===========================================================================
    T operator()(int i)
	  {
      T Lx2 = Lx[i] * Lx[i];
		  T Ly2 = Ly[i] * Ly[i];
		  T M = sqrt(Lx2 + Ly2);
		  //if (M > 0.0001)
		  //{
		    T nom = 2 * Lxy[i] * Lx[i] * Ly[i] - Lx2 * Lyy[i] - Ly2 * Lxx[i];
		    T den = M * M * M;
        T R = CurvatureScale<T>::kappa_to_radius(nom, den);
		    return CurvatureScale<T>::radius_to_value(R);
		  //}
		  //return CurvatureScale<T>::radius_to_value(0);
	  }
  };
} // namespace Curvature
} // namespace Vtl

#endif // ISOLINE_CURVATURE_H