#ifndef SIGMA_H
#define SIGMA_H

//system includings
#include <math.h>

namespace Vtl
{
namespace Curvature
{
	template <typename T>
	class Sigma
	{
	private:
		const T m_start;
		const T m_mult;
		T m_stop;
		T m_current;

		//---------------------------------------------------------------------------
	public:
		Sigma(T start, T mult, T stop)
			:m_start(start), m_mult(mult), m_stop(stop), m_current(start) {};
		//---------------------------------------------------------------------------
		T Next()
		{
			m_current *= m_mult;
			
			if (m_current > m_stop)
			{
				m_current /= m_mult;
				m_stop = m_current;
				return T(-1);
			}
			return m_current;
		}
		//---------------------------------------------------------------------------
		T Next(T sigma) const
		{
		  return sigma * m_mult;
		}
		//---------------------------------------------------------------------------
		/*T Previous(T sigma) const
		{
		  return sigma / m_mult;
		}*/
		//---------------------------------------------------------------------------
		T Get() const
		{
			return m_current;
		}
		//---------------------------------------------------------------------------
		T Get(int i) const
		{
		  if (i >= 0)
			{
				T res = m_start * T(pow(double(m_mult), double(i)));
			  if (res <= m_stop)
				{
				  return res;
				}
				return m_stop;
			}
			return 0;
		}
		//---------------------------------------------------------------------------
		/*int GetIndex(double sigma) const
		{
			if (sigma >= m_start)
			{
				//if (sigma > m_stop)
				//{
				//  sigma = m_stop;
				//}

				int i = int(log(sigma/m_start) / log(m_mult));
				i = std::max(0, i);
				
				//std::cout << std::endl << "i = " << i << "   sigma = " << sigma <<
				//		"   s * m^i = " << m_start * pow(double(m_mult), double(i)) <<
				//		"  stop = " << m_stop;
				return i;
			}
			else
			{
			  return 0;
			}
		}*/
	};
}
}
#endif //SIGMA_H
