#ifndef _PLF_FILTER_IMAGES_H
#define _PLF_FILTER_IMAGES_H
/////////////////////////////////////////////////////////////////////////////
#include <algorithm>

#include "plf_kernels_1d.h"

#include "plf_filter_elem.h"
#include "plf_io_filter.h"
#include "plf_iterators.h"
#include "plf_gistogram.h"

namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
	/** 
	 *  @brief   .
	 *
	 *   -  .
	 *
	 *   : .................
	 **********************************************************************************************/
	template <typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class medianFilter
	: public FilterElementary
	{
	public:
		Vtl::Plf::OLSI::Gistogramm<T> gistogr;
		InPort<T> in;
		OutPort<T,P> out;
		int r;
		int _bits;
	public:
		medianFilter()
			: FilterElementary(1,1)
		{
		}
		medianFilter(int rr,int bits, std::string name,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(rr,bits,name,post_proc);
		}
		void Init(int rr,int bits,std::string name,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			r=rr;
			_bits=bits;
			AddPort(in,0,Extent(r,r,r+1,r));
			AddPort(out,0);
		}

		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();

			ImgSize img_size=in.getImageSize(); 
			gistogr.Init(_bits,img_size.width,r);
		}
		
		virtual ~medianFilter()
		{
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			const int w=ims.width;
			
			//int i= in.get 
			out.beginUpdate();
			ImageRef<T> img=in.getImg(); 
			const T* input=img[0];
			
			OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
	
			if(Y_pos()==0)
			{
				for(int y=-r; y<=r; y++)
				{	
					input=img[y];
					//printf(" %d\n",int(input[0]));
					gistogr.initGis_of_one_line(input-r);
				}
				
				gistogr.initMediansForGistogramms();
				
				for(int x=0; x<w; x++)    
				{
					
					*output++ =gistogr.getMediana_value(x);
				}
			}
			else
			{
				 const T* add = img[r];
				 const T* del = img[-r-1];
				gistogr.add_del_lines_to_gist(add-r, del-r);
				for(int x=0; x<w; x++)    
				{
					*output++ = gistogr.getMediana_value(x);			
				}
			}
			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};

	/** 
	 *  @brief    ,     .
	 *
	 *   -  .
	 *
	 *   : .................
	 **********************************************************************************************/
	template <typename T,class P=Vtl::Plf::F11::DoNothing<T> >
	class medianFilter2
	: public FilterElementary
	{
	public:
		Vtl::Plf::OLSI::Gistogramm2<T> gistogr;
		InPort<T> in;
		OutPort<T,P> out;
		int r;
		int _bits;
	public:
		medianFilter2()
			: FilterElementary(1,1)
		{
		}
		medianFilter2(int rr,int bits, std::string name,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(1,1)
		{
			Init(rr,bits,name,post_proc);
		}
		void Init(int rr, int bits,std::string name,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			out.setPostProc(post_proc);
			r=rr;
			_bits=bits;
			AddPort(in,0,Extent(r,r,r+1,r));
			AddPort(out,0);
		}

		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			FilterElementary::set_OutPortsImagesSizeAndInPortsExtents();

			ImgSize img_size=in.getImageSize(); 
			gistogr.Init(_bits,img_size.width,r);
		}
		
		virtual ~medianFilter2()
		{
		}
		virtual void processLine() 
		{
			ImgSize ims=in.getImageSize();
			int w = ims.width; 
			out.beginUpdate();
			ImageRef<T> img=in.getImg(); 
			const T* input=img[0];
			
			OutPortNoPostProc<T>::BufType::Iter_Line output=out.begin();
			if(Y_pos()==0)
			{
				ImageRef<T> img=in.getImg(); 
				for(int y=-r; y<=r; y++)
				{	
					const T* input=img[y];
					gistogr.initGis_of_one_line(input-r);
				}
				
				gistogr.initMediansForGistogramms();
				
				for(int x=0; x<w; x++)    
				{
					*output++ = gistogr.getMediana_value(x);
				}
			}
			else
			{
				 const T* add = img[r];
				 const T* del = img[-r-1];
				gistogr.add_del_lines_to_gist(add-r, del-r);
				for(int x=0; x<w; x++)    
				{
					*output++ = gistogr.getMediana_value(x);			
				}
			}
			in.ToNextLine(Y_pos()+1);
			out.endUpdate();
		}
	};
} //namespace Plf
} //namespace Vtl
#endif //_PLF_FILTER_IMAGES_H
