// system includings
//#include <memory> 
#include <vector>
#include <iostream>

// library includings
#include "aimage.h"
#include "consolap.h"
//#include "imalg_tools.h"
//#include "plf_functors.h"

// own includings
#include "SaveImageArray.h"
//#include "ImgFromGrad.h"
//#include "StatDiff.h"
#include "EdgeDetection/DiZenzoEdgeDetector.h"
#include "EdgeDetection/RayleighEdgeDetector.h"
#include "EdgeDetection/CannyEdgeDetector.h"


namespace Tests
{
	/*enum EdgeDetectorType
	{
	  DiZenzo,
		Rayleigh,
	};*/

	/////////////////////////////////////////////////////////////////////////////
  int TestEdgeDetectors(int argc, char *argv[])
  {
		// !!! DEFINE TYPE OF DETECTOR TO TEST
		//int TYPE = EdgeDetectorType::DiZenzo;

    // Load input image
    if(argc < 2) 
    {
		  std::cout << "Argument - input image" << std::endl;
      return 0;
    }

    const char *fname = argv[1];

		const float sigma = 5.0f;
		const float nSigmasEnv = 3.0;
		const float histThStart = 10.0f;
		const float histThEnd = 0.0f;

		if (1/*TYPE == EdgeDetectorType::DiZenzo*/)
		{
			std::auto_ptr<AImage> _img(::AImageLoad(0, true, fname, indicator));
			AImageMemRgb8* img = dynamic_cast<AImageMemRgb8*>(_img.get());
			if(!img)
			{
				return 0;
			}

			int w = img->Width();
			int h = img->Height();
			int size = w * h;

			std::vector<float> edgeStrength(size);
			std::vector<float> xDir(size);
			std::vector<float> yDir(size);

			DiZenzoEdgeDetector<float> diZenzo;
			diZenzo.Apply(img, &(edgeStrength[0]), &(xDir[0]), &(yDir[0]), sigma,
																						 nSigmasEnv, histThStart, histThEnd);

			Tests::SaveNormImageArray(&(edgeStrength[0]), w, h, fname, "_DiZenzo");
		}
		else if (0/*TYPE == EdgeDetectorType::Rayleigh*/)
		{
		  std::auto_ptr<AImage> _img(::AImageLoad(0, true, fname, indicator));
			AImageMem8* img = dynamic_cast<AImageMem8*>(_img.get());
			if(!img)
			{
				return 0;
			}

			int w = img->Width();
			int h = img->Height();
			int size = w * h;

			std::vector<float> edgeStrength(size);
			std::vector<float> xDir(size);
			std::vector<float> yDir(size);

			//DiZenzoEdgeDetector<float> diZenzo;
			//diZenzo.Apply(img, &(edgeStrength[0]), &(xDir[0]), &(yDir[0]), sigma,
			//																			 nSigmasEnv, histThStart, histThEnd);

			Tests::SaveNormImageArray(&(edgeStrength[0]), w, h, fname, "_Rayleigh");
		}
		else if (0/*Canny*/)
		{
			std::auto_ptr<AImage> _imgRGB(::AImageLoad(0, true, fname, indicator));
			AImageMemRgb8* imgRGB = dynamic_cast<AImageMemRgb8*>(_imgRGB.get());
			if(!imgRGB)
			{
				return 0;
			}
		  AImageMem8* img = imgRGB->GetPlane8(0);

			// image characteristics
			const int w = img->Width();
			const int h = img->Height();
			const int sz = w * h;
			const int nls = img->NextLineShift();

			//         float
			std::vector<float> _src(sz);
			float *src  = &(_src[0]);

			Vtl::Plf::F11::Convert<float> c_to_float;
			Vtl::ImAlg::Tools::copy_1_1((*img)[0], w, h, nls, src, w, c_to_float);

			CannyEdgeDetector<float> canny(src, w, h);
			canny.Apply(sigma, nSigmasEnv, histThStart, histThEnd);

			Tests::SaveNormImageArray(&(canny.grMod[0]), w, h, fname, "_GrayGrad");
			Tests::SaveImageArray(&(canny.edges[0]), w, h, fname, "_Canny");
		}

		return 1;
  }
}