
// system includings
#include <memory> 
#include <vector>
#include <iostream>

// library includings
#include "aimage.h"
#include "consolap.h"
#include "imalg_tools.h"
#include "plf_functors.h"

// own includings
#include "EdgeDetection/CannyEdgeDetector.h"
#include "SaveImageArray.h"
#include "ImgFromGrad/ImgFromGrad.h "
#include "StatDiff.h"


namespace Tests
{
  int TestImgFromGrad(int argc, char *argv[])
  {
    // Load input image
    if(argc < 2) 
    {
		  std::cout << "Argument - input image" << std::endl;
      return 0;
    }

    const char *fname = argv[1];
	  std::auto_ptr<AImage> _img(::AImageLoad(0, true, fname, indicator));
    AImageMem8* img = dynamic_cast<AImageMem8*>(_img.get());
	  if(!img)
    {
		  return 0;
    }

    // image characteristics
    const int w = img->Width();
	  const int h = img->Height();
	  const int sz = w * h;
    const int nls = img->NextLineShift();

    //         float
    std::vector<float> _src(sz);
    float *src  = &(_src[0]);

    Vtl::Plf::F11::Convert<float> c_to_float;
	  Vtl::ImAlg::Tools::copy_1_1((*img)[0], w, h, nls, src, w, c_to_float);

    // Find edges and gradient components
    const float sigma = 1.0f;
	  const float nSigmasEnv = 3.5;
    const float hisThStart = 7.5f;
    const float hisThEnd = 2.0f;

    CannyEdgeDetector<float> canny(src, w, h);
    canny.Apply(sigma, nSigmasEnv, hisThStart, hisThEnd);

    float* grMod = &(canny.grMod[0]);
    float* edges = &(canny.edges[0]);
    float* Lx = &(canny.Lx[0]);
    float* Ly = &(canny.Ly[0]);

	  SaveImageArray(grMod, w, h, fname, "_GradMod");
    SaveImageArray(edges, w, h, fname, "_Edges");
		SaveNormImageArray(Lx, w, h, fname, "_Lx");
		SaveNormImageArray(Ly, w, h, fname, "_Ly");

		// 
    Levelling(edges, sz, 100.0f);
    Tests::SaveImageArray(edges, w, h, fname, "_Edges_levelled");


    EdgeComponents(edges, grMod, Lx, Ly, sz);
    Tests::SaveNormImageArray(Lx, w, h, fname, "_Lx_levelled");
		Tests::SaveNormImageArray(Ly, w, h, fname, "_Ly_levelled");

		/*typedef Vtl::ImAlg::WndFun::Gauss<float,0, false> FF0_bg; 
		FF0_bg f0_bg(2 * sigma, nSigmasEnv);
		typedef Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv,true> K0_bg;
		K0_bg ker0_bg(f0_bg);
		Vtl::ImAlg::Conv::ConvInplace<float,K0_bg> g0_bg(ker0_bg);

		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Lx, w, h, w, g0_bg, g0_bg);
		Vtl::ImAlg::Conv::Conv2D::conv2DInplace(Ly, w, h, w, g0_bg, g0_bg);*/

    ImgFromGrad(canny.Lx, canny.Ly, h, w, fname);
    
    return 1;
  }
} // namespace Tests