
#include <string>
#include <sstream>
#include <memory> 
#include <list>
#include <iostream>
#include <fstream>
#include <iomanip>

#include <time.h>
#include <assert.h>

#include "yfilesys.h"
#include "aimage.h"
#include "color_spaces.h"

#include "tests.h"
#include "consolap.h"

#include "CurveAnalysis/edge_det.h"

namespace Students
{
	template<typename T>
	void step_edge_detect(AImage* img,T *edge,T * gx,T * gy,T* fri,float sigma)
	{
		using namespace Curve;

		const float n_sigmas_in_eviron=4.0f;

		const char *BAD_String="BAD lib: starage type defined incorrectly";
		if(img==0)
			throw "Null image in step_edge_detect";
		const int w=img->Width();
		const int h=img->Height();
		const int sz=w*h;
		switch(img->GetImageStore())
		{
		case AImage::IMGS_AImageMem8:
			{
				AImageMem8* im=dynamic_cast<AImageMem8*>(img);
				if(im==0) throw BAD_String; //Impossible due to library, but check it for correctness.
				Vtl::Plf::F11::Convert<T> cvt;
				//copy_1_1 -  -,       
				//        (*im)[0].
				Vtl::ImAlg::Tools::copy_1_1((*im)[0],w,h,img->NextLineShift(),edge,w,cvt);
				EdgeDetect<float> ed(sigma,n_sigmas_in_eviron);
				ed.Canny(w,h,w,edge,gx,gy,fri);
			}
			break;
		case AImage::IMGS_AImageMem16:
			{
				AImageMem16* im=dynamic_cast<AImageMem16*>(img);
				if(im==0) throw BAD_String; //Impossible due to library, but check it for correctness.
				Vtl::Plf::F11::Convert<T> cvt;
				//copy_1_1 -  -,       
				//        (*im)[0].
				Vtl::ImAlg::Tools::copy_1_1((*im)[0],w,h,img->NextLineShift(),edge,w,cvt);
				EdgeDetect<float> ed(sigma,n_sigmas_in_eviron);
				ed.Canny(w,h,w,edge,gx,gy,fri);
			}
			break;
		case AImage::IMGS_AImageMemRgb8:
			{
				AImageMemRgb8* im=dynamic_cast<AImageMemRgb8*>(img);
				if(im==0) throw BAD_String; //Impossible due to library, but check it for correctness.
				AImageMem8 * r=im->GetPlane8(2);
				AImageMem8 * g=im->GetPlane8(1);
				AImageMem8 * b=im->GetPlane8(0);
				Vtl::Plf::F11::Convert<T> cvt;
				//copy_1_1 -  -,       
				//        (*im)[0].
				Vtl::ImAlg::Tools::copy_1_1((*r)[0],w,h,r->NextLineShift(),edge,w,cvt);
				Vtl::ImAlg::Tools::copy_1_1((*g)[0],w,h,g->NextLineShift(),gx,w,cvt);
				Vtl::ImAlg::Tools::copy_1_1((*b)[0],w,h,b->NextLineShift(),gy,w,cvt);

				std::vector<float> tmp(sz*2);
				float *tmp1=&(tmp[0]);
				float *tmp2=tmp1;
				EdgeDetect<float> ed(sigma,n_sigmas_in_eviron);
				ed.DiZenzo(w,h,w,edge,gx,gy,fri,tmp1,tmp2);
			}
			break;
		case AImage::IMGS_AImageMemRgb16:
			{
				AImageMemRgb16* im=dynamic_cast<AImageMemRgb16*>(img);
				if(im==0) throw BAD_String; //Impossible due to library, but check it for correctness.
				AImageMem16 * r=im->GetPlane16(2);
				AImageMem16 * g=im->GetPlane16(1);
				AImageMem16 * b=im->GetPlane16(0);
				Vtl::Plf::F11::Convert<T> cvt;
				//copy_1_1 -  -,       
				//        (*im)[0].
				Vtl::ImAlg::Tools::copy_1_1((*r)[0],w,h,r->NextLineShift(),edge,w,cvt);
				Vtl::ImAlg::Tools::copy_1_1((*g)[0],w,h,g->NextLineShift(),gx,w,cvt);
				Vtl::ImAlg::Tools::copy_1_1((*b)[0],w,h,b->NextLineShift(),gy,w,cvt);

				std::vector<float> tmp(sz*2);
				float *tmp1=&(tmp[0]);
				float *tmp2=tmp1;
				EdgeDetect<float> ed(sigma,n_sigmas_in_eviron);
				ed.DiZenzo(w,h,w,edge,gx,gy,fri,tmp1,tmp2);
			}
			break;
		default:
			throw "unsupported image storage";
		};
	}
}//namespace Students


namespace Tests
{

	int test_curve_analysis(int argc,char *argv[])
	{
		using namespace Students;

		if(argc < 2) 
		{
			std::cout << "Command string:\n test.exe input_img" <<std::endl;
			return 1;
		}
		
		const float sigma=2.5f;


		// ,  auto_ptr        {}
		std::auto_ptr<AImage> _img(::AImageLoad(0,true,argv[1],indicator));
		AImage* img=_img.get();

		//   
		const int w=img->Width();
		const int h=img->Height();
		const int sz=w*h;


		//     ,  
		// 1)    - 0,    -   ,  .   1  
		//    ( ,  100%  ).
		// 2)   ,   x  y     . 
		//           .   ,     
		//    .   2 .
		//       :
		std::vector<float> data(sz*4);
		float *edges = &(data[0]); //   
		float *gx = edges +   sz;  // x- 
		float *gy = edges + 2*sz;  // y- 
		float *fri = edges + 3*sz; // Feature Response Immage 


		step_edge_detect(img,edges,gx,gy,fri,sigma);


		//   
		{
			//   .
			std::auto_ptr<AImage> _out(::AImageCreate(0,AImage::IMGC_GRAYSCALE,w,h,8));
			AImageMem8* out=dynamic_cast<AImageMem8*>(_out.get());

			//    8   ,        
			Vtl::Plf::F11::Convert<uint8_t> c_to_Byte;
//			Vtl::Plf::F11::Lin<Real_t> lin(16.0f,128.0f);
			Vtl::Plf::F11::GammaCorMult<float> gcm(255.0f,2.2f,1.0f);
			//  ,      
			Vtl::Plf::F11::Comb<uint8_t,Vtl::Plf::F11::GammaCorMult<float>,Vtl::Plf::F11::Convert<uint8_t> > adj(gcm,c_to_Byte);
			//          ()
			Vtl::ImAlg::Tools::copy_1_1(edges,w,h,w,(*out)[0],out->NextLineShift(),adj);
			::AImageSave(out,"crv_anlz-edge.tif",indicator);
			//          ()
			Vtl::ImAlg::Tools::copy_1_1(fri,w,h,w,(*out)[0],out->NextLineShift(),adj);
			::AImageSave(out,"crv_anlz-feature_response_image.tif",indicator);
		}


		//>       .
		// edges -   0,  edge
		// gx,gy -   .


		return 0;
	}
}

	
