#include <stdio.h>
#include <stdlib.h> //rand(),srand()
#include <time.h> //time()

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <memory> //auto_ptr()

#include <list>

#include "loki_typeManip.h"
#include "plf_kernels_1d.h"
#include "plf_functors.h"
#include "imalg_tools.h"
#include "aimage.h"
#include "num_linalg.h"
#include "consolap.h"
#include "tests.h"


#include "ytime.h"
/*
	class Conv3D
	{
		public:
		template<typename T, class 1, class 2> 
		static void conv3D(const T* data_in,int width,int height,int dept, int nls_in, int nls_z, T* data_out,int nls_out,1& cx,2& cy)
		{
			/////////????????
			const T* di=data_in;
			T* dout=data_out;
			for(int y=0; y<height; ++y, di+=nls_in, dout+=nls_out)
				cx(di,di+width,data_out);
			for(int x=0; x<width; ++x)
			{
				SliceIter<T,0> si(data_in++,nls);
				cy(si,si+height,data_out++);
			}
			for(int z=0; z<dept; ++z );
		}
	};
*/

namespace Img
{
	/// Type of Width, Height etc. Signed value, because position may be negative (eg. out of image fragment) so, there will be position comparisons with eg. Width.
	typedef int size_type;
	/// Type of pointer difference. Negative values shold be processed too!
	typedef int DistType;

	/// Image layout in memory/file (type of ordering)
	enum Interleave
	{
		BIP=0, //!< band interleaved by pixel 
		BIL=1, //!< band interleaved by line 
		BSQ=2, //!< band sequential
		BII=3, //!< band interleaved by 2D image, analog of BIL for N-Dimensional images (for 2D images -is the same as BSQ!) 
		SEP=-1  //!< bands at separated location, NOT COMPATIBLE with ENVI
	};
	///Pixel sample data types (fully ENVI-compatible, as in IMG/HDR file)
	///Do NOT change numerical values for format support work correctly!
	enum SampleType
	{
		ST_Unknown=255, //!< any non standart value, ERROR, Not supported by ENVI   
		ST_u8   =1,  //!<     u   1 =    8 bit byte; 
		ST_s8   =254,//!<     s   1 =    8 bit signed byte; NOT ENVI COMPATIBLE! Extension due to TIFF! 
		ST_s16  =2,  //!<     s   2 =   16-bit signed integer; 
		ST_s32  =3,  //!<     s   3 =   32-bit signed long integer; 
		ST_f32  =4,  //!<     f   4 =   32-bit floating point; 
		ST_f64  =5,  //!<     f   5 =   64-bit double precision floating point; 
		ST_c32f =6,  //!<     f   6 = 2x32-bit complex, real-imaginary pair of double precision; ??SINGLE PRECISION?? 
		ST_c64f =9 , //!<     f   9 = 2x64-bit double precision complex, real-imaginary pair of double precision; 
		ST_u16  =12, //!<     u  12 =   16-bit unsigned integer; 
		ST_u32  =13, //!<     u  13 =   32-bit unsigned integer; 
		ST_s64  =14, //!<     s  14 =   64-bit integer; 
		ST_u64  =15  //!<     u  15 =   64-bit unsigned integer.
	};

	/// Operatione with containers when SIZE KNOWN AT COMPILE TIME.
	namespace KnownSize
	{
		template<size_t n,class Iter,class ConstIter> 
		void copy(Iter dst,ConstIter src,Loki::Int2Type<n>)
		{
			*dst=*src;
			copy(++dst,++src,Loki::Int2Type<n-1>());
		}
		template<class Iter,class ConstIter> 
		void copy(Iter dst,ConstIter src,Loki::Int2Type<0>)
		{
		}
		//Copy from one iterator to another, size is known at compile time.
		template<size_t n,class Iter,class ConstIter> 
		void copy(Iter dst,ConstIter src)
		{
			copy(dst,src,Loki::Int2Type<n>());
		}
	}

	// Strided iterator throw array. Stride size is known at COMPILER TIME (faster!).
	template<typename T,int stride_> 
	class Iter1D
	{
	public:
		typedef T value_type;
		typedef T& reference_type;
		typedef const T& const_reference_type;
		typedef T* pointer_type;
		typedef const T* const_pointer_type;
	private:
		T* ptr;
		enum {stride=stride_};
	public:
		Iter1D(const Iter1D &rhs) :ptr(rhs.ptr) {}
		Iter1D(T* data,int strd=stride_): ptr(data) {}
		Iter1D& operator++() { ptr+=stride; return *this; }
		Iter1D operator++(int) { Iter1D prev(ptr); ptr+=stride; return prev; }
		Iter1D& operator--() { ptr-=stride; return *this; }
		Iter1D operator--(int) { Iter1D prev(ptr); ptr-=stride; return prev; }
		Iter1D& operator+=(DistType d) { ptr+=stride*d; return *this; }
		Iter1D& operator-=(DistType d) { ptr-=stride*d; return *this; }
		Iter1D& operator+(DistType d) { Iter1D prev(ptr); prev+=d; return prev; }
		Iter1D& operator-(DistType d) { Iter1D prev(ptr); prev-=d; return prev; }
		T& operator*() {return *ptr;}
		const T& operator*() const {return *ptr;}
		bool operator!=(const Iter1D &rhs) const { return ptr!=rhs.ptr; }
		Iter1D& operator=(const Iter1D &rhs) { ptr=rhs.ptr; return *this; }
	};
	// Strided iterator throw array (specialization). Stride size is known at RUN TIME (slower!).
	template<typename T> 
	class Iter1D<typename T,0> 
	{
	public:
		typedef T value_type;
		typedef T& reference_type;
		typedef const T& const_reference_type;
		typedef T* pointer_type;
		typedef const T* const_pointer_type;
	private:
		T* ptr;
		int stride;
	public:
		Iter1D(const Iter1D &rhs) :ptr(rhs.ptr), stride(rhs.stride) {}
		Iter1D(T* data,int stride_): ptr(data),stride(stride_) {}
		Iter1D& operator++() { ptr+=stride; return *this; }
		Iter1D operator++(int) { Iter1D prev(ptr); ptr+=stride; return prev; }
		Iter1D& operator--() { ptr-=stride; return *this; }
		Iter1D operator--(int) { Iter1D prev(ptr); ptr-=stride; return prev; }
		Iter1D& operator+=(DistType d) { ptr+=stride*d; return *this; }
		Iter1D& operator-=(DistType d) { ptr-=stride*d; return *this; }
		Iter1D& operator+(DistType d) { Iter1D prev(ptr); prev+=d; return prev; }
		Iter1D& operator-(DistType d) { Iter1D prev(ptr); prev-=d; return prev; }
		T& operator*() {return *ptr;}
		const T& operator*() const {return *ptr;}
		bool operator!=(const Iter1D &rhs) const { return ptr!=rhs.ptr; }
		Iter1D& operator=(const Iter1D &rhs) { ptr=rhs.ptr; stride=rhs.stride; return *this; }
	};
	template<typename T_sample,int T_n=2,int T_Nbands=1,Interleave T_Layout=BSQ>
	class ImageLayout
	{
	private:
		template<int n,Interleave i>
		void calcStrides(Loki::Int2Type<n>,Loki::Int2Type<i> )
		{
		}
		template<int n,Interleave i>
		void calcStrides(Loki::Int2Type<n>,Loki::Int2Type<BIP> )
		{
		}
	};
	template<typename T_sample,int T_n=2,int T_Nbands=1,Interleave T_Layout=BSQ>
	class ImageNDref
	{
		typedef T_sample value_type;
		enum {N=T_n};
		enum {Nbands=T_Nbands};
		enum {Layout=T_Layout};
		size_type size[N];
		size_type samples_per_size[N];
		size_type stride[N];
		value_type* data;
	public:
		virtual ~ImageNDref() {}
		ImageNDref(value_type* data_,size_type* sz,size_type* samples_per_dim_line)
			:data(data_)
		{
			KnownSize::copy<N>(size,sz);
			KnownSize::copy<N>(samples_per_size,samples_per_dim_line);

		}
	};
}//namespace Img

namespace Tests
{
/*
	class YFileRead
	{
//		char handle[16];
		typedef std::basic_ifstream<uint8_t> InStream;
		InStream ifs;
	private:
		YFileRead(const YFileRead&);
		void constructor(const char * filename)
		{
			//InStream ifs(fname,std::ios_base::in|std::ios_base::binary);
			//Mode: app, ate, binary, in, out, trunc
			ifs.open(filename,std::ios_base::in|std::ios_base::binary);
			if(!ifs)
			{
				std::ostringstream os;
				os << "Can't open for read file " << filename;
				throw os.str();
			}
		}
	public:
		YFileRead(std::string filename) {constructor(filename.c_str());}
		YFileRead(const char *filename) {constructor(filename);}
		virtual ~YFileRead()
		{
			_close();
		}
		void _close()
		{
			ifs.close();
		}
		void SetPosition(int64_t offset)
		{
			ifs.seekg(offset);
		}
		int read(void* data,int sz)
		{
			ifs.read((uint8_t*)data,sz);
			size_t cnt=ifs.gcount();
			return cnt;
		}
		int read(int64_t offset,void* data,int sz)
		{
			SetPosition(offset);
			read(data,sz);
		}
		int64_t GetFileSize() //const
		{
			uint64_t pos_0=ifs.tellg();
			ifs.seekg(0,std::ios_base::end);
			uint64_t pos=ifs.tellg();
			ifs.seekg(pos_0,std::ios_base::beg);
		}
		YDataTime GetTime() const
		{
			
		}
	};
*/

const int X_size=256;
const int Y_size=256;
const int Z_size=256;

	template<typename T>
	class Image3D
	{
	public:
		const int X_size;
		const int Y_size;
		const int Z_size;
		typedef T value_type;
		typedef Img::Iter1D<value_type,1> Iter_0; 
		typedef Img::Iter1D<value_type,0> Iter_1; 
		typedef Img::Iter1D<value_type,0> Iter_2; 
		template<int i>
		struct Iter1D_Type
		{
			typedef typename Loki::Select<(i==0),Iter_0,typename Loki::Select<(i==1),Iter_1,Iter_2>::Result >::Result Result; 
		};
	private:
		typedef uint16_t FilePixelType;
		typedef std::basic_ifstream<uint8_t> InStream;
		std::vector<T> data;
	public:
		Image3D(int X_size_,int Y_size_,int Z_size_)
			: X_size(X_size_),Y_size(Y_size_),Z_size(Z_size_)
		{
			data.resize(X_size*Y_size*Z_size);
		}
		T* ptr(int x,int y,int z) {return &(data[0])+x+(y+z*Y_size)*X_size; }
		//int stepX() const {return 1;}
		//int stepY() const {return X_size;}
		//int stepZ() const {return X_size*Y_size;}
		template<int i>
		int step() const {return step(Loki::Int2Type<i>() );}
		template<int i>
		int length() const {return length(Loki::Int2Type<i>() );}
	private:
		int step(Loki::Int2Type<0> )  const {return 1;}
		int step(Loki::Int2Type<1> ) const {return X_size;}
		int step(Loki::Int2Type<2> ) const {return X_size*Y_size;}
		int length(Loki::Int2Type<0> )  const {return X_size;}
		int length(Loki::Int2Type<1> ) const {return Y_size;}
		int length(Loki::Int2Type<2> ) const {return Z_size;}
	public:
		template<int i>
		typename Iter1D_Type<i>::Result begin_1D(int x,int y,int z)
		{
			return begin_1D(x,y,z,Loki::Int2Type<i>());
		}
		Iter_0  begin_1D(int x,int y,int z,Loki::Int2Type<0>)
		{
			return Iter_0(ptr(0,y,z),step<0>());
		}
		Iter_1  begin_1D(int x,int y,int z,Loki::Int2Type<1>)
		{
			return Iter_1(ptr(x,0,z),step<1>());
		}
		Iter_2  begin_1D(int x,int y,int z,Loki::Int2Type<2>)
		{
			return Iter_2(ptr(x,y,0),step<2>());
		}
		template<int i>
		typename Iter1D_Type<i>::Result end_1D(int x,int y,int z)
		{
			return end_1D(x,y,z,Loki::Int2Type<i>());
		}
		Iter_0  end_1D(int x,int y,int z,Loki::Int2Type<0>)
		{
			return Iter_0(ptr(length<0>(),y,z),step<0>());
		}
		Iter_1  end_1D(int x,int y,int z,Loki::Int2Type<1>)
		{
			return Iter_1(ptr(x,length<1>(),z),step<1>());
		}
		Iter_2  end_1D(int x,int y,int z,Loki::Int2Type<2>)
		{
			return Iter_2(ptr(x,y,length<2>()),step<2>());
		}
		int size() const {return X_size*Y_size*Z_size;}
		void Load(const char *fname)
		{
			//Mode: app, ate, binary, in, out, trunc
			InStream ifs(fname,std::ios_base::in|std::ios_base::binary);
			if(!ifs)
			{
				std::ostringstream os;
				os << "Can't open for read file " << fname;
				throw os.str();
			}
			load_type(ifs,Loki::Int2Type<(Loki::IsSameType<T,FilePixelType>::value==true)?1:0>() );
		}
	private:
		/// same type
		void load_type(InStream& ifs,Loki::Int2Type<1> )
		{
// , getline()  n-1        .
			//ifs.getline(im.ptr(0,0,0),sizeof(T)*im.size());
			const size_t sz=sizeof(T)*im.size(); 
			ifs.read(im.ptr(0,0,0),sz);
			size_t cnt=ifs.gcount();
			if(sz!=cnt)
			{
				std::ostringstream os;
				os << sz << " Error read " << cnt;
				throw os.str();
			}
		}
		/// different type, need conversion on fly
		void load_type(InStream& ifs,Loki::Int2Type<0> )
		{
			enum {isGreater= (sizeof(T)>=sizeof(FilePixelType)) ? 1 :0 };
			load_to_different_pixel_size(ifs,Loki::Int2Type<isGreater>() );

		}
		/// different type, need conversion on fly, 
		/// Destination type is SMALLER then pixel size in file
		void load_to_different_pixel_size(InStream& ifs,Loki::Int2Type<0> )
		{
			throw "Pixel size in container Image3D is smaller than in file";

		}
		/// different type, need conversion on fly
		/// Destination type is LARGER OR SAME SIZE then pixel size in file
		void load_to_different_pixel_size(InStream& ifs,Loki::Int2Type<1> )
		{
			T* ptr1=ptr(0,0,0);
			const size_t sz=size()*sizeof(FilePixelType); 

			const uint8_t* ptr_end=(uint8_t*)(ptr1+size());
			FilePixelType * ptr2=(FilePixelType*)(ptr_end-sz);
			
// , getline()  n-1        .
//			ifs.getline(ptr2,size()*sizeof(FilePixelType));
			ifs.read((uint8_t*)ptr2,sz);
			size_t cnt=ifs.gcount();
			if(sz!=cnt)
			{
				std::ostringstream os;
				os << sz << " Error read " << cnt;
				throw os.str();
			}

			for(int i=0; i< size(); ++i)
				ptr1[i]=T(ptr2[i]);
		}
	};
/*
	template<typename T>
	class Iter1D
	{
		T* ptr;
		int stride;
	public:
		Iter1D(T* p,int step) 
			:ptr(p),stride(step)
		{}
		Iter1D& operator++() { ptr+=stride; return *this;}
		Iter1D& operator--() { ptr-=stride; return *this;}
		T operator*() const { return *ptr;}
		T& operator*() { return *ptr;}
		bool operator!=(Iter1D& it2) { return ptr!=it2.ptr;}
	};
*/
	class Conv3D
	{
	public:
		template<typename T, class C1, class C2, class C3>
		static void conv3DInplace(T& im, C1& gx,C2& gy, C3& gz)
		{
		   for(int y=0 ; y < im.Y_size ; ++y )
			for(int z=0 ; z < im.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(im.ptr(0,y,z), im.stepX());
				Iter1D<uint16_t> i_end(im.ptr(im.X_size,y,z), im.stepX())
				gx(i,i_end);
			}

			for(int x=0 ; x < im.X_size ; ++x )
			for(int y=0 ; y < im.Y_size ; ++y )
			{
				Iter1D<uint16_t> i(im.ptr(x,y,0), im.stepZ());
				Iter1D<uint16_t> i_end(im.ptr(x,y,im.stepZ() ), im.stepZ())
				gy(i,i_end);
			}

			for(int x=0 ; x < im.X_size ; ++x )
			for(int z=0 ; y < im.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(im.ptr(x,0,z), im.stepY());
				Iter1D<uint16_t> i_end(im.ptr(x,im.stepY() ,z), im.stepY())
				gz(i,i_end);
			}
		}

		template<class Image_T,class F1, class F2, class F3>
		static void conv3D(Image_T& img, F1 f1, F2 f2, F3 f3)
		{

			for(int y=0 ; y < img.Y_size ; ++y )
			for(int z=0 ; z < img.Z_size ; ++z )
			{
				typename Image_T::Iter1D_Type<0>::Result i1=img.begin_1D<0>(0,y,z);
				typename Image_T::Iter1D_Type<0>::Result i2=img.end_1D<0>(0,y,z);
				f1(i1,i2);
			}

			for(int x=0 ; x < img.X_size ; ++x )
			for(int y=0 ; y < img.Y_size ; ++y )
			{
				typename Image_T::Iter1D_Type<2>::Result i1=img.begin_1D<2>(x,y,0);
				typename Image_T::Iter1D_Type<2>::Result i2=img.end_1D<2>(x,y,0);
				f1(i1,i2);
			}

			for(int x=0 ; x < img.X_size ; ++x )
			for(int z=0 ; z < img.Z_size ; ++z )
			{
				typename Image_T::Iter1D_Type<1>::Result i1=img.begin_1D<1>(x,0,z);
				typename Image_T::Iter1D_Type<1>::Result i2=img.end_1D<1>(x,0,z);
				f1(i1,i2);
			}

		}
	};

	int test_medical_3d_image(int argc, char* argv[])
	{
		typedef Image3D<float> Img3D;	
		typedef float Pixel;
		
		std::cout << "=======  Start  ===  test_medical_3d_image   =============" << std::endl;
		Img3D	im(X_size,Y_size,Z_size);
		im.Load("..\\..\\Cube_1.img");
		int sigma = 10;
		Vtl::ImAlg::Conv::FastGaussConvolve<Img3D::value_type,0> g0(sigma);
        Vtl::ImAlg::Conv::FastGaussConvolve<Img3D::value_type,1> g1(sigma);
        Vtl::ImAlg::Conv::FastGaussConvolve<Img3D::value_type,2> g2(sigma);
		{
			Conv3D::conv3D(im, g0,g0,g0);

		}
#if 0		
		Img3D imxx(X_size,Y_size,Z_size);
		Img3D imyy(X_size, Y_size, Z_size);
		Img3D imzz(X_size, Y_size, Z_size);
	
        Img3D imx(X_size,Y_size,Z_size);
		Img3D imy(X_size, Y_size, Z_size);
		Img3D imz(X_size, Y_size, Z_size);


		Img3D imxy(X_size,Y_size,Z_size);
		Img3D imyz(X_size, Y_size, Z_size);
		Img3D imxz(X_size, Y_size, Z_size);

		Img3D im2x(X_size,Y_size,Z_size);
		Img3D im2y(X_size, Y_size, Z_size);
		Img3D im2z(X_size, Y_size, Z_size);
	

		
		
        
		memcpy(imxx.ptr(0,0,0),im.ptr(0,0,0),im.size()*sizeof(uint16_t));
		memcpy(imyy.ptr(0,0,0),im.ptr(0,0,0),im.size()*sizeof(uint16_t));
		memcpy(imzz.ptr(0,0,0),im.ptr(0,0,0),im.size()*sizeof(uint16_t));
		
		memcpy(imxy.ptr(0,0,0),im.ptr(0,0,0),im.size()*sizeof(uint16_t));
		memcpy(imyz.ptr(0,0,0),im.ptr(0,0,0),im.size()*sizeof(uint16_t));
		memcpy(imxz.ptr(0,0,0),im.ptr(0,0,0),im.size()*sizeof(uint16_t));
		
	
        

		for(int y=0 ; y< im.Y_size ; ++y)
		for(int z=0 ; z< im.Z_size ; ++z)
		{
			float *p=im.ptr(0,y,z);
			g0(p,p+im.X_size);
		}
		
		for(int x=0 ; x< im.X_size ; ++x)
		for(int z=0 ; z< im.Z_size ; ++z)
		{
			Iter1D<uint16_t>  i2(im.ptr(x,0,z),im.stepY());
			Iter1D<uint16_t> i2_end(im.ptr(x,im.Y_size,z),im.stepY());
			g0(i2,i2_end);
		}

		for(int x=0 ; x< im.X_size ; ++x)
		for(int y=0 ; y< im.Z_size ; ++y)
		{
			Iter1D<uint16_t>  i2(im.ptr(x,y,0),im.stepZ());
			Iter1D<uint16_t> i2_end(im.ptr(x,y,im.Z_size),im.stepY());
			g0(i2,i2_end);
		}

		//Calcuation Iyy
		
		{
			for(int y=0 ; y < imyy.Y_size ; ++y )
			for(int z=0 ; z < imyy.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imyy.ptr(0,y,z), imyy.stepX());
				Iter1D<uint16_t> i_end(im.ptr(imyy.X_size,y,z), imyy.stepX());
				g0(i,i_end);
			}

			for(int x=0 ; x < imyy.X_size ; ++x )
			for(int y=0 ; y < imyy.Y_size ; ++y )
			{
				Iter1D<uint16_t> i(imyy.ptr(x,y,0), imyy.stepZ());
				Iter1D<uint16_t> i_end(imyy.ptr(x,y,imyy.stepZ()), imyy.stepZ());
				g0(i,i_end);
			}

			for(int x=0 ; x < imyy.X_size ; ++x )
			for(int z=0 ; z < imyy.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imyy.ptr(x,0,z), imyy.stepY());
				Iter1D<uint16_t> i_end(imyy.ptr(x,imyy.stepY(),z), imyy.stepY());
				g2(i,i_end);
			}
		}

			//Calcuation Izz
		
		{
			for(int y=0 ; y < imzz.Y_size ; ++y )
			for(int z=0 ; z < imzz.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imzz.ptr(0,y,z), imzz.stepX());
				Iter1D<uint16_t> i_end(imzz.ptr(imzz.X_size,y,z), imzz.stepX());
				g0(i,i_end);
			}

			for(int x=0 ; x < imzz.X_size ; ++x )
			for(int y=0 ; y < imzz.Y_size ; ++y )
			{
				Iter1D<uint16_t> i(imzz.ptr(x,y,0), imzz.stepZ());
				Iter1D<uint16_t> i_end(imzz.ptr(x,y,imzz.stepZ()), imzz.stepZ());
				g2(i,i_end);
			}

			for(int x=0 ; x < imzz.X_size ; ++x )
			for(int z=0 ; z < imzz.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imzz.ptr(x,0,z), imzz.stepY());
				Iter1D<uint16_t> i_end(im.ptr(x,imzz.stepY(),z), imzz.stepY());
				g0(i,i_end);
			}
		}
//Calculation Laplassian
		float norm_const_for_laplass =1;//??
		for (int x=0;x<imxx.X_size;++x)
			for(int y=0;y<imxx.Y_size;++y)
			   for(int z=0; z<imxx.Z_size;++z)
			   {
				   *imxx.ptr(x,y,z) = norm_const_for_laplass * (*imxx.ptr(x,y,z)+ *imyy.ptr(x,y,z)+ *imzz.ptr(x,y,z));
			   }
		//Calcuation Ix
		
		{
			for(int y=0 ; y < imx.Y_size ; ++y )
			for(int z=0 ; z < imx.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imx.ptr(0,y,z), imx.stepX());
				Iter1D<uint16_t> i_end(imx.ptr(imx.X_size,y,z), imx.stepX());
				g1(i,i_end);
			}

			for(int x=0 ; x < imx.X_size ; ++x )
			for(int y=0 ; y < imx.Y_size ; ++y )
			{
				Iter1D<uint16_t> i(imx.ptr(x,y,0), imx.stepZ());
				Iter1D<uint16_t> i_end(imx.ptr(x,y,imx.stepZ()), imx.stepZ());
				g0(i,i_end);
			}

			for(int x=0 ; x < imx.X_size ; ++x )
			for(int z=0 ; z < imx.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imx.ptr(x,0,z), imx.stepY());
				Iter1D<uint16_t> i_end(imx.ptr(x,imx.stepY() ,z), imx.stepY());
				g0(i,i_end);
			}
		}

		//Calcuation Iy
		
		{
			for(int y=0 ; y < imy.Y_size ; ++y )
			for(int z=0 ; z < imy.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imy.ptr(0,y,z), imy.stepX());
				Iter1D<uint16_t> i_end(imy.ptr(im.X_size,y,z), imy.stepX());
				g0(i,i_end);
			}

			for(int x=0 ; x < imyy.X_size ; ++x )
			for(int y=0 ; y < imyy.Y_size ; ++y )
			{
				Iter1D<uint16_t> i(imy.ptr(x,y,0), imy.stepZ());
				Iter1D<uint16_t> i_end(imy.ptr(x,y,imy.stepZ() ), imy.stepZ());
				g0(i,i_end);
			}

			for(int x=0 ; x < imy.X_size ; ++x )
			for(int z=0 ; z < imy.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imy.ptr(x,0,z), imy.stepY());
				Iter1D<uint16_t> i_end(imy.ptr(x,imy.stepY() ,z), imy.stepY());
				g1(i,i_end);
			}
		}

			//Calcuation Iz
		
		{
			for(int y=0 ; y < imz.Y_size ; ++y )
			for(int z=0 ; z < imz.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imz.ptr(0,y,z), imz.stepX());
				Iter1D<uint16_t> i_end(imz.ptr(imz.X_size,y,z), imz.stepX());
				g0(i,i_end);
			}

			for(int x=0 ; x < imz.X_size ; ++x )
			for(int y=0 ; y < imz.Y_size ; ++y )
			{
				Iter1D<uint16_t> i(imz.ptr(x,y,0), imz.stepZ());
				Iter1D<uint16_t> i_end(imz.ptr(x,y,imz.stepZ() ), imz.stepZ());
				g1(i,i_end);
			}

			for(int x=0 ; x < imz.X_size ; ++x )
			for(int z=0 ; z < imz.Z_size ; ++z )
			{
				Iter1D<uint16_t> i(imz.ptr(x,0,z), imz.stepY());
				Iter1D<uint16_t> i_end(im.ptr(x,imz.stepY() ,z), imz.stepY());
				g0(i,i_end);
			}
		}
		for (int x=0 ; x<im.X_size ; ++x)
			for(int y=0 ; y<im.Y_size ; ++y)
			   for(int z=0 ; z<im.Z_size ;++z)
			   {
					*im2x.ptr(x,y,z)=sqrt(float(*imx.ptr(x,y,z)));
					*im2y.ptr(x,y,z)=sqrt(float(*imy.ptr(x,y,z)));
					*im2z.ptr(x,y,z)=sqrt(float(*imz.ptr(x,y,z)));
			   }


		for(int x=0 ; x<imxy.X_size ; ++x)
			for(int y=0 ; y<imxy.Y_size ; ++y)
			   for(int z=0 ; z<imxy.Z_size ; ++z)
			   {
				   *imxy.ptr(x,y,z)=*imx.ptr(x,y,z) * *imy.ptr(x,y,z);
				   *imxz.ptr(x,y,z)=*imx.ptr(x,y,z) * *imz.ptr(x,y,z);
				   *imyz.ptr(x,y,z)=*imy.ptr(x,y,z) * *imz.ptr(x,y,z);
			   }

//conv g(x), g(y), g(z)
       float s=1.7f; 
	   Vtl::ImAlg::Conv::FastGaussConvolve<uint16_t,0> gu0(s*sigma);
		{
			for(int y=0 ; y < im.Y_size ; ++y )
			for(int z=0 ; z < im.Z_size ; ++z )
			{
				Iter1D<uint16_t> i2x(im2x.ptr(0,y,z), im2x.stepX());
				Iter1D<uint16_t> i2x_end(im2x.ptr(im2x.X_size,y,z), im2x.stepX());
				gu0(i2x,i2x_end);
				Iter1D<uint16_t> i2y(im2y.ptr(0,y,z), im2x.stepX());
				Iter1D<uint16_t> i2y_end(im2y.ptr(im2y.X_size,y,z), im2y.stepX());
				gu0(i2y,i2y_end);
				Iter1D<uint16_t> i2z(im2z.ptr(0,y,z), im2z.stepX());
				Iter1D<uint16_t> i2z_end(im2z.ptr(im2z.X_size,y,z), im2z.stepX());
				gu0(i2z,i2z_end);
				Iter1D<uint16_t> ixy(imxy.ptr(0,y,z), imxy.stepX());
				Iter1D<uint16_t> ixy_end(imxy.ptr(im2y.X_size,y,z), imxy.stepX());
				gu0(ixy,ixy_end);
				Iter1D<uint16_t> ixz(imxz.ptr(0,y,z), imxz.stepX());
				Iter1D<uint16_t> ixz_end(imxz.ptr(imxz.X_size,y,z), imxz.stepX());
				gu0(ixz,ixz_end);
				Iter1D<uint16_t> iyz(imyz.ptr(0,y,z), imyz.stepX());
				Iter1D<uint16_t> iyz_end(imyz.ptr(imyz.X_size,y,z), imyz.stepX());
				gu0(iyz,iyz_end);
			}

			for(int x=0 ; x < im.X_size ; ++x )
			for(int y=0 ; y < im.Y_size ; ++y )
			{
				Iter1D<uint16_t> i2x(im2x.ptr(x,y,0), im2x.stepZ());
				Iter1D<uint16_t> i2x_end(im2x.ptr(x,y,im2x.Z_size), im2x.stepZ());
				gu0(i2x,i2x_end);
				Iter1D<uint16_t> i2y(im2y.ptr(x,y,0), im2x.stepZ());
				Iter1D<uint16_t> i2y_end(im2y.ptr(x,y,im2x.Z_size), im2y.stepZ());
				gu0(i2y,i2y_end);
				Iter1D<uint16_t> i2z(im2z.ptr(x,y,0), im2z.stepZ());
				Iter1D<uint16_t> i2z_end(im2z.ptr(x,y,im2x.Z_size), im2z.stepZ());
				gu0(i2z,i2z_end);
				Iter1D<uint16_t> ixy(imxy.ptr(x,y,0), imxy.stepZ());
				Iter1D<uint16_t> ixy_end(imxy.ptr(x,y,im2x.Z_size), imxy.stepZ());
				gu0(ixy,ixy_end);
				Iter1D<uint16_t> ixz(imxz.ptr(x,y,0), imxz.stepZ());
				Iter1D<uint16_t> ixz_end(imxz.ptr(x,y,im2x.Z_size), imxz.stepZ());
				gu0(ixz,ixz_end);
				Iter1D<uint16_t> iyz(imyz.ptr(x,y,0), imyz.stepZ());
				Iter1D<uint16_t> iyz_end(imyz.ptr(x,y,im2x.Z_size), imyz.stepZ());
				gu0(iyz,iyz_end);
			}

			for(int x=0 ; x < im.X_size ; ++x )
			for(int z=0 ; z < im.Z_size ; ++z )
			{
				Iter1D<uint16_t> i2x(im2x.ptr(x,0,z), im2x.stepZ());
				Iter1D<uint16_t> i2x_end(im2x.ptr(x,im2x.Y_size,z), im2x.stepY());
				gu0(i2x,i2x_end);
				Iter1D<uint16_t> i2y(im2y.ptr(x,0,z), im2x.stepZ());
				Iter1D<uint16_t> i2y_end(im2y.ptr(x,im2x.Y_size,z), im2y.stepY());
				gu0(i2y,i2y_end);
				Iter1D<uint16_t> i2z(im2z.ptr(x,0,z), im2z.stepZ());
				Iter1D<uint16_t> i2z_end(im2z.ptr(x,im2x.Y_size,z), im2z.stepY());
				gu0(i2z,i2z_end);
				Iter1D<uint16_t> ixy(imxy.ptr(x,0,z), imxy.stepZ());
				Iter1D<uint16_t> ixy_end(imxy.ptr(x,im2x.Y_size,z), imxy.stepY());
				gu0(ixy,ixy_end);
				Iter1D<uint16_t> ixz(imxz.ptr(x,0,z), imxz.stepZ());
				Iter1D<uint16_t> ixz_end(imxz.ptr(x,im2x.Y_size,z), imxz.stepY());
				gu0(ixz,ixz_end);
				Iter1D<uint16_t> iyz(imyz.ptr(x,0,z), imyz.stepZ());
				Iter1D<uint16_t> iyz_end(imyz.ptr(x,im2x.Y_size,z), imyz.stepY());
				gu0(iyz,iyz_end);
			}
		}

		Vtl::NumAlg::Lin::ublas::c_matrix<float,3,3> A;
		Vtl::NumAlg::Lin::ublas::c_vector<float,3> e;
		Vtl::NumAlg::Lin::ublas::c_vector<float,3> work;

		Vtl::NumAlg::Lin::eigen::decomp(A,e,work); 
		{
// e - min sobstvennie zna4enia
		}
        //  2-    
#endif
		return 0;
	}
}//namespace Tests

