#include <stdio.h>
#include <stdlib.h> //rand(),srand()
#include <time.h> //time()

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <memory> //auto_ptr()

#include <list>

#include "plf_kernels_1d.h"
#include "plf_functors.h"
#include "imalg_tools.h"
#include "aimage.h"
#include "consolap.h"
#include "tests.h"



namespace Tests
{
		template<typename T,unsigned D> class Gauss_;
		///  ( )
		template<typename T> class Gauss_<T,0>
		{
			const T s2;
		public:
			Gauss_(T sigma) : s2(2*sigma*sigma) {}
			template<class TT> T operator()(TT x) const
			{
				return exp(-(x*x)/s2);
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,1>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma){}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T(-x*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,2>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma){}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T((t-1)*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,3>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T(x*(3-t)*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,4>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T((3+(t-6)*t)*exp(-t/2.));
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,5>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T(-x*((t-10)*t+15)*exp(-t/2.));
				
			}
		};
		///    ( )
		template<typename T> class Gauss_<T,6>
		{
			const T s;
		public:
			Gauss_(T sigma) : s(sigma) {}
			template<class TT> T operator()(TT x) const
			{
				double t=Vtl::sqr(x/s);
				return T((((t-15)*t+45)*t-15)*exp(-t/2.));
			}
		};
		/*!
		 * \brief   (Hann).   
		 * \f$ w(x)= { 1 \over 2 } [ 1 - cos( { 2 \pi x \over size - 1 ) }] \f$
		 *  0 <= x <= size-1
		 *
		 *   \f$ c = { 2 \pi  \over size - 1 }  \f$ ,    
		 *   : \f[ w(x)= { 1 \over 2 }  [1-cos( c x )] \f]
		 *
		 *   : \n
		 * \f$ w'(x) = { d w(x) \over d x } = + { 1 \over 2 }  c  sin( c x ) \f$ \n
		 * \f$ w''(x)     = + { 1 \over 2 }  c^2 cos( c x )  \f$ \n
		 * \f$ w^{(3)}(x) = - { 1 \over 2 }  c^3 sin( c x )  \f$ \n
		 * \f$ w^{(4)}(x) = - { 1 \over 2 }  c^4 cos( c x )  \f$ \n
		 * \f$ w^{(5)}(x) = + { 1 \over 2 }  c^5 sin( c x )  \f$ \n
		 * :\n
		 *	\f[ w^{(n)}(x) = { 1 \over 2 } c^n  (-1)^{ [ n/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( c x ) , \qquad if \quad n \quad is \quad odd \\ 
		 *		cos( c x ) , \qquad if \quad n \quad is \quad even
		 *	\end{array} \right. \f] 
		 *
		 *   \f[ dsin(x,n)= { d^n sin \over d x^n}(x) = (-1)^{ [ (n+1)/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( x ) , \qquad if \quad n \quad is \quad even \\ 
		 *		cos( x ) , \qquad if \quad n \quad is \quad odd
		 *	\end{array} \right. \f] 
		 *  \f$ { [ ... ] } \f$         . 
		 * \see Vtl::Alg::dsin
		 * 
		 *   :
		 *	\f[ w^{(n)}(x) = { 1 \over 2 } c^n  dsin<n-1>( c x )  \f] 
		 *
		 *
		 *****************************************************************************/
		template<typename T,unsigned D> class Hann_
			: public Vtl::ImAlg::WndFun::Traits<T,true,true,D,D>
		{
			const T c;
		public:
			static const char * Name() { return "Hann"; }
			template<class TT> 
			Hann_(TT size_) : Traits(T(size_),T(size_/2.0f)), c(T(2*M_PI/(size_-1))) {}
			template<class TT> 
			Hann_(TT size_, const Hann_& ) : Traits(T(size_),T(size_/2.0f)), c(T(2*M_PI/(size_-1))) {}
			template<class TT> T operator()(TT x) const
			{
				return result( T(x), Loki::Int2Type<(D==0)>() );
			}
		private:
			T result(T x,Loki::Int2Type<true>) const
			{
				return T(0.5*(1-cos(c*x))); 
			}
			T result(T x,Loki::Int2Type<false>) const
			{
				return T(0.5*Vtl::Alg::x_to_N<D>::value(c)*Vtl::Alg::dsin<D-1>(c*x));;
			}
		};
		namespace Kernel
		{
			/**
			* class Gauss
			* 
			* \brief       \f$ \exp(-x^2/(2*\sigma^2))\f$ 
			*    6  .    
			*   (float, double),  -   (0- ), 
			*   (. Vtl::Plf::Kernel::Type).  wnd_half_size -    
			*    sigma       
			* wnd_half_size=sigma*sigmas_number. .   void Init()  .
			* 
			* \htmlinclude kernels_1d_conv_gauss.htm
			*	
			* \htmlinclude kernels_1d_conv_gauss_1_deriv.htm
			* 
			* \htmlinclude KernelsNorm.htm 	
			* 
			**********************************************************************************/
			template<typename T,unsigned Deriv=0,Vtl::Plf::Kernel::Type kernel_type_=Vtl::Plf::Kernel::KT_Conv>
			class Gauss
				:public Vtl::Plf::Kernel::Base<T,kernel_type_>
			{
				T sigma;
			public:
				Gauss() : Vtl::Plf::Kernel::Base<T,kernel_type_>("Gauss(t)=exp(-x^2/(2*sigma^2))") {}
				Gauss(T sigmaInPixels,T sigmas_number)
					: Vtl::Plf::Kernel::Base<T,kernel_type_>("Gauss(t)=exp(-x^2/(2*sigma^2))")
				{	
					Init(sigmaInPixels,sigmas_number);
				}
				Gauss(int left, int right,T sigmaInPixels)
					: Vtl::Plf::Kernel::Base<T,kernel_type_>("Gauss(t)=exp(-x^2/(2*sigma^2))")
				{	
					Init(left, right,sigmaInPixels);
				}
				virtual ~Gauss() {}
				/*!
				* @param sigmaInPixels - exp(-x^2/(2*sigma^2)) - value of sigma in pixels
				* @param sigmas_number -   ()  
				*
				*/
				void Init(T sigmaInPixels,T sigmas_number)
				{
					sigma=sigmaInPixels;
					int wnd_half_size=int(sigmaInPixels*sigmas_number+0.5);
					_Init(wnd_half_size,wnd_half_size);
					//sigma=wnd_half_size/sigmas_number;
					typedef Vtl::ImAlg::WndFun::Gauss<T,Deriv> Func;
					Func gauss(sigma,sigmas_number); 
					build(gauss);
					normalize<Func::conv_norm_alg>();
				}
				/*!
				* @param left -   ()      ,   .
				* @param right -   ()      ,   .
				* @param sigmaInPixels - exp(-x^2/(2*sigma^2)) - value of sigma in pixels
				* \see Vtl::Plf::Kernel::Base::_Init()
				*/
				void Init(int left, int right,T sigmaInPixels)
				{
					sigma=sigmaInPixels;
					_Init(left,right);
					//sigma=wnd_half_size/sigmas_number;
					typedef F::Gauss<T,Deriv> Func;
					Func gauss(sigma,T(1)); 
					build(gauss);
					normalize<Func::conv_norm_alg>();
				}
			private:
				template<typename Func>
				void build(Func func)
				{
					iterator d=begin_internal();
					if(kernel_type==Vtl::Plf::Kernel::KT_Conv)
					{
						for(int i=0; i<size(); i++)
							*d++=func(left()-i);
					}
					else
					{
						for(int i=0; i<size(); i++)
							*d++=func(i-left());
					}
				}
			};
		}

		template<class K>
		void save(const K* kernel)
		{
			std::string name=std::string("kernel_")+kernel->Name()+".txt";
			std::string s=kernel->SaveForMathLab();
			std::ofstream os;
			os.open(name.c_str(),std::ios_base::out | std::ios_base::trunc );
				os << s;
			os.flush();
			os.close();
		}


	int test_plf_kernels(int argc, char* argv[])
	{
		//ConvKernel<float> ker; //    -       !
		std::string str;
		const Kernel::Gauss<float,0,Vtl::Plf::Kernel::KT_Corr> ker(3,1.5f); 
		const Kernel::Gauss<float,1,Vtl::Plf::Kernel::KT_Corr> ker_deriv(3,1.5f); 
	//	ker.Init(0.7,1);
		str=ker.SaveForMathLab();
		printf("GaussFilter:\n%s===\n",str.c_str());
		str=ker_deriv.SaveForMathLab();
		printf("GaussDerivFilter:\n%s===\n",str.c_str());

		Kernel::Gauss<float,0,Vtl::Plf::Kernel::KT_Corr>::const_iterator f(ker.begin());
		const int sz=17;
		float v[sz]={1,1,1,4,4,4,4,1,1,1,4,4,4,4,1,1,1};
		double fff=Vtl::ImAlg::Conv::Conv1pt<double>(v,ker.begin(),ker.end());
		printf("fff=%lf\n",fff);
		float out[sz]={1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0};
		printf("--Gauss-----\n");
		Vtl::ImAlg::Conv::ConvolveNoExt(v,v+sz-ker.size()+1,out,ker);
		for(int i=0; i<sizeof(out)/sizeof(out[0]);i++)
			printf("%d : %f\n",i,out[i]);
		printf("--GaussDeriv-----\n");
		Vtl::ImAlg::Conv::ConvolveNoExt(v,v+sz-ker_deriv.size()+1,out,ker_deriv);
		for(int i=0; i<sizeof(out)/sizeof(out[0]);i++)
			printf("%d : %f\n",i,out[i]);
		printf("-------\n");

		{
			std::cout << "--Hamming Window-----" <<std::setprecision(4) << std::endl;
			const int sz=11;
			Vtl::ImAlg::WndFun::Hamming<float,0> h0(sz);
			Vtl::ImAlg::WndFun::Hamming<float,1> h1(sz);
			Vtl::ImAlg::WndFun::Hamming<float,2> h2(sz);
			Vtl::ImAlg::WndFun::Hamming<float,3> h3(sz);
			Vtl::ImAlg::WndFun::Hamming<float,4> h4(sz);
			for(int i=0; i < h0.size(); ++i)
			{
				std::cout << i << "  " << std::setw(12) << h0(i) << " "<< std::setw(12) << h1(i) << " "<< std::setw(12) << h2(i) << " "
					<< std::setw(12) << h3(i) << " "<< std::setw(12) << h4(i) <<std::endl;
			}
		}
		{
			std::cout << "--Hann Window-----" << std::setprecision(4) << std::endl;
			const int sz=11;
			Vtl::ImAlg::WndFun::Hann<float,0> h0(sz);
			Vtl::ImAlg::WndFun::Hann<float,1> h1(sz);
			Vtl::ImAlg::WndFun::Hann<float,2> h2(sz);
			Vtl::ImAlg::WndFun::Hann<float,3> h3(sz);
			Vtl::ImAlg::WndFun::Hann<float,4> h4(sz);
			for(int i=0; i< h0.size() ; ++i)
			{
				std::cout << i << "  " << std::setw(12) << h0(i) << " "<< std::setw(12) << h1(i) << " "<< std::setw(12) << h2(i) << " "
					<< std::setw(12) << h3(i) << " "<< std::setw(12) << h4(i) <<std::endl;
			}
		}
		{
			std::cout << "Gauss functor, Hermite polinom coefficients: "<< std::endl;

			std::cout << "c<0,0> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<0,0>() << std::endl;
			std::cout << "c<1,1> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<1,1>() << std::endl;
			std::cout << "c<2,2> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<2,2>() << std::endl;
			std::cout << "c<2,0> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<2,0>() << std::endl;
			std::cout << "c<3,3> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<3,3>() << std::endl;
			std::cout << "c<3,1> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<3,1>() << std::endl;
			std::cout << "c<4,4> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<4,4>() << std::endl;
			std::cout << "c<4,2> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<4,2>() << std::endl;
			std::cout << "c<4,0> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<4,0>() << std::endl;
			std::cout << "c<5,5> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<5,5>() << std::endl;
			std::cout << "c<5,3> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<5,3>() << std::endl;
			std::cout << "c<5,1> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<5,1>() << std::endl;
			std::cout << "c<6,6> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<6,6>() << std::endl;
			std::cout << "c<6,4> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<6,4>() << std::endl;
			std::cout << "c<6,2> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<6,2>() << std::endl;
			std::cout << "c<6,0> = "<< Vtl::ImAlg::WndFun::Impl::Gauss::c<6,0>() << std::endl;

			std::cout << "===== "<< std::endl;

			std::cout << "Q<0>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<0>(0.0f,0.0f) << std::endl;
			std::cout << "Q<1>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<1>(0.0f,0.0f) << std::endl;
			std::cout << "Q<2>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<2>(0.0f,0.0f) << std::endl;
			std::cout << "Q<3>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<3>(0.0f,0.0f) << std::endl;
			std::cout << "Q<4>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<4>(0.0f,0.0f) << std::endl;
			std::cout << "Q<5>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<5>(0.0f,0.0f) << std::endl;
			std::cout << "Q<6>(0)"<< Vtl::ImAlg::WndFun::Impl::Gauss::Q<6>(0.0f,0.0f) << std::endl;

			Vtl::ImAlg::WndFun::Gauss<float,5> g(4.2f,2.0f);
			Gauss_<float,5> g_(4.2f);
			for(int i=-20; i<20 ; ++i)
			{
				std::cout << "i= "<< i << " diff= " << g(i)-g_(i) << "     g= " << g(i) << "     g_= " << g_(i) << std::endl;
			}

		}
		{
			std::cout << "Sinc functor: "<< std::endl;
			for(int half_size=1; half_size <= 5 ; ++half_size)
			{
				std::cout << "half_size= " << half_size << std::endl;
				Vtl::ImAlg::WndFun::Sinc<float,0> sinc(half_size,2);
//				typedef Vtl::ImAlg::WndFun::Kaiser<float,0> WndKaiser;
//				typedef Vtl::ImAlg::WndFun::Nuttall<float,0> WndNuttall;
//				Vtl::ImAlg::WndFun::Sinc<float,0,WndKaiser> sinc_kaiser(half_size,2,WndKaiser(3,3));
//				Vtl::ImAlg::WndFun::Sinc<float,0,WndNuttall> sinc_nuttall(half_size,2,WndNuttall(3));
				int left=int(sinc.left());
				for(int i=left; i<sinc.right(); ++i)
				{
					std::cout << "i= "<< i << "\t sinc=\t" << sinc(i) 
//						<<  "\t sinc_kaiser=\t" << sinc_kaiser(i) 
//						<< "\t sinc_nuttall= \t" << sinc_nuttall(i)  
						<< std::endl;
				}
			}

		}
		{
			std::cout << "Compile-time check Functor/Traits cast: "<< std::endl;
			Vtl::ImAlg::WndFun::Hamming<float,0> h(2); //This is OK
			Vtl::ImAlg::WndFun::Traits<float,true,true,0,0> * ptr=&h; //Not generates error, this is UNDESIRED!
			ptr=new Vtl::ImAlg::WndFun::Hamming<float,0>(2); //Not generates error - this is BAD!!!
			//delete ptr; //Generates Error - this is OK!!!
		}
		{
			std::cout << "Compile-time check Direct_Ref, Direct, Laws5: "<< std::endl;
			const int dr_data[]={1,2,5,2,1};
			Vtl::ImAlg::WndFun::Direct_Ref<float,int> dr(5,2,dr_data);
			Vtl::ImAlg::WndFun::Direct<float,int> dr_cpy(5,2,dr_data);
			Vtl::ImAlg::WndFun::Laws5<float>::L5 L;
		}
		{
			std::cout << "Check Vtl::Plf: "<< std::endl;
			typedef Vtl::ImAlg::WndFun::Gauss<float,0> F0;
			F0 gauss(2.5,2.5);
			Vtl::Plf::Kernel::FKern<float> g0(gauss);
			std::cout << "Gauss" <<std::endl <<g0.SaveForMathLab() <<std::endl;

			typedef Vtl::ImAlg::WndFun::Gauss<float,1> F1;
			F1 gauss_d1(2.5,2);
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv> g1(gauss_d1);
			std::cout << "Gauss 1d conv" <<std::endl <<g1.SaveForMathLab() <<std::endl;
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Corr> g1a(gauss_d1);
			std::cout << "Gauss 1d corr" <<std::endl <<g1a.SaveForMathLab() <<std::endl;

			typedef Vtl::ImAlg::WndFun::Hann<float,0> F2;
			F2 hann8(8);
			Vtl::Plf::Kernel::FKern<float> h1(hann8);
			std::cout << "Hann 8" <<std::endl <<h1.SaveForMathLab() <<std::endl;
			F2 hann9(9);
			Vtl::Plf::Kernel::FKern<float> h2(hann9);
			std::cout << "Hann 9" <<std::endl <<h2.SaveForMathLab() <<std::endl;
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win> h3(hann9,8);
			std::cout << "Hann window" <<std::endl <<h3.SaveForMathLab() <<std::endl;

			typedef Vtl::ImAlg::WndFun::Hann<float,1> F3;
			F3 hann8_d1(8);
			Vtl::Plf::Kernel::FKern<float> h4(hann8_d1);
			std::cout << "Hann 8 d/dx" <<std::endl <<h4.SaveForMathLab() <<std::endl;
			F3 hann9_d1(9);
			Vtl::Plf::Kernel::FKern<float> h5(hann9_d1);
			std::cout << "Hann 9 d/dx" <<std::endl <<h5.SaveForMathLab() <<std::endl;
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win> h6(hann9_d1,8);
			std::cout << "Hann window d/dx" <<std::endl <<h6.SaveForMathLab() <<std::endl;
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win> h61(hann9_d1,8,2.0);
			std::cout << "Hann window d/dx center at 2.0" <<std::endl <<h61.SaveForMathLab() <<std::endl;
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win> h62(hann9_d1,8,3.5);
			std::cout << "Hann window d/dx center at 3.5" <<std::endl <<h62.SaveForMathLab() <<std::endl;

			typedef Vtl::ImAlg::WndFun::Hamming<float,0> F4;
			F4 ham8(8);
			Vtl::Plf::Kernel::FKern<float> ha1(ham8);
			std::cout << "Hamming 8" <<std::endl <<ha1.SaveForMathLab() <<std::endl;
			F4 ham9(9);
			Vtl::Plf::Kernel::FKern<float> ha2(ham9);
			std::cout << "Hamming 9" <<std::endl <<ha2.SaveForMathLab() <<std::endl;
			Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win> ha3(ham9,8);
			std::cout << "Hamming window" <<std::endl <<ha3.SaveForMathLab() <<std::endl;

		}
		{
			//float a[]={0.3635819f ,	0.4891775f , 0.1365995f , 0.0106411f};
			float a[]={5.0f ,	7.0f , 11.0f , 13.0f};
			Vtl::ImAlg::WndFun::Impl::common_Trig::cos_Member<float,0,4,1> cm0(1,a);
			std::cout << "result=" << cm0(0,0) <<std::endl;
			Vtl::ImAlg::WndFun::Impl::common_Trig::cos_Member<float,1,4,1> cm1(1,a);
			std::cout << "result=" << cm1(0,0) <<std::endl;
			Vtl::ImAlg::WndFun::Impl::common_Trig::cos_Member<float,2,4,1> cm2(1,a);
			std::cout << "result=" << cm2(0,0) <<std::endl;
			Vtl::ImAlg::WndFun::Impl::common_Trig::cos_Member<float,3,4,1> cm3(1,a);
			std::cout << "result=" << cm3(0,0) <<std::endl;

			const int N=10;
			Vtl::ImAlg::WndFun::Hann<float,0> h0(N);
			Hann_<float,0> h_0(N);
			Vtl::ImAlg::WndFun::Hann<float,1> h1(N);
			Hann_<float,1> h_1(N);
			Vtl::ImAlg::WndFun::Hann<float,2> h2(N);
			Hann_<float,2> h_2(N);
			Vtl::ImAlg::WndFun::Hann<float,3> h3(N);
			Hann_<float,3> h_3(N);
			for(int i=0; i<N; ++i)
			{
				std::cout 
					<< std::setw(7) << h0(i)  << " " 
					<< std::setw(7) << h_0(i) << " "
					<< std::setw(7) << h1(i)  << " "
					<< std::setw(7) << h_1(i) << " "
					<< std::setw(7) << h2(i)  << " "
					<< std::setw(7) << h_2(i) << " "
					<< std::setw(7) << h3(i)  << " "
					<< std::setw(7) << h_3(i) 
					<< std::endl;
			}

		}
		std::cout << "FKern testing" <<std::endl;
		{
			const int N=10;
			std::cout << "=== Infinite support" <<std::endl;
			{
				typedef Vtl::ImAlg::WndFun::Gauss<float,0> FF;
				FF::value_type sigma=FF::value_type(N/5);
				FF func(sigma,FF::value_type(N/(2*sigma)));
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			{
				typedef Vtl::ImAlg::WndFun::Sinc<float,0> FF;
				FF func(4,3);
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			std::cout << "=== Finite support, N=" << N <<std::endl;
			{
				typedef Vtl::ImAlg::WndFun::Hann<float,0> FF;
				FF func(N);
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			std::cout << "" <<std::endl;
			{
				typedef Vtl::ImAlg::WndFun::Hamming<float,0> FF;
				FF func(N);
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			{
				typedef Vtl::ImAlg::WndFun::Kaiser<float,0> FF;
				FF func(N,FF::value_type(3));
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			{
				typedef Vtl::ImAlg::WndFun::Blackman<float,0> FF;
				FF func(N);
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			{
				typedef Vtl::ImAlg::WndFun::Nuttall<float,0> FF;
				FF func(N);
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			{
				typedef Vtl::ImAlg::WndFun::Bartlett<float,0> FF;
				FF func(N);
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			std::cout << "=== Finite support table" <<std::endl;
			{
				typedef Vtl::ImAlg::WndFun::Laws5<float>::E5 FF;
				FF func;
				std::cout << func.Name()<<std::endl;
				Vtl::Plf::Kernel::FKern<float> ker(func);
				for( Vtl::Plf::Kernel::FKern<float>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
					std::cout << *it << "  ";
				std::cout << std::endl;
				save(&ker);
			}
			std::cout << "=== Compare conv,corr & win kernels" <<std::endl;
			{
				typedef Vtl::ImAlg::WndFun::Nuttall<float,1> FF;
				FF func(N);
				std::cout << func.Name()<<std::endl;
				std::cout << "Conv"<<std::endl;
				{
					Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv> ker(func);
					for( Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Conv>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
						std::cout << *it << "  ";
					std::cout << std::endl;
				}
				std::cout << "Corr"<<std::endl;
				{
					Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Corr> ker(func);
					for( Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Corr>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
						std::cout << *it << "  ";
					std::cout << std::endl;
				}
				std::cout << "Win"<<std::endl;
				{
					Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win> ker(func,N);
					for( Vtl::Plf::Kernel::FKern<float,Vtl::Plf::Kernel::KT_Win>::const_iterator it=ker.begin(); it!=ker.end(); ++it)
						std::cout << *it << "  ";
					std::cout << std::endl;
				}
			}

		}

		return 0;
	}
}//namespace Tests

