#pragma warning( disable : 4786 )

#include <xercesc/util/PlatformUtils.hpp>

#include <xercesc/dom/DOM.hpp>
#include <xercesc/dom/DOMImplementation.hpp>
#include <xercesc/dom/DOMImplementationLS.hpp>
#include <xercesc/dom/DOMWriter.hpp>

#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp> 

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLUni.hpp>

#include <xercesc/util/XMLChar.hpp>

#define YXML_DRIVER
#include "xml_interface_imp.h"



using namespace XERCES_CPP_NAMESPACE;

//			catch(const SAXException& toCatch) \
//			{	*exception_message = XMLString::transcode(toCatch.getMessage());	}\
//

#define TRY_XML *exception_message=0; try {
#define CATCH_XML }	\
	    catch(const XMLException& toCatch) \
			{	*exception_message = XMLString::transcode(toCatch.getMessage());	}\
			catch(const DOMException& toCatch) \
			{	*exception_message = XMLString::transcode(toCatch.msg);			}\
			catch(...) 			\
			{	StrXML s("Unexpected Exception");	*exception_message=XMLString::transcode(s.GetXMLCh());	}
#define THROW(cstring) {	StrXML s(cstring);	*exception_message=XMLString::transcode(s.GetXMLCh()); return;	}
////////////////////////////////////////////////////////////////////
	class StrXML
	{
		XMLCh* ch;
	public:
		StrXML(const char *data) { ch=XMLString::transcode(data);}
		~StrXML() {if(ch) XMLString::release(&ch); /*delete[] ch;*/}
		operator XMLCh*() {return ch;}
		const XMLCh* GetXMLCh() const {return ch;}
	};
////////////////////////////////////////////////////////////////////
static YXML_DRV_DomCString TranscodeStrX(const XMLCh* const toTranscode) 
{ 
	return (toTranscode)?XMLString::transcode(toTranscode):0; 
}
static void RemoveSpaceTextNodes(DOMNode  *node)
{
	DOMNode* n=0;
	n=node->getFirstChild();
	while(n)
	{
		if(DOMNode::TEXT_NODE==n->getNodeType())
		{
			const XMLCh* const str=n->getNodeValue();
			if(XMLChar1_0::isAllSpaces(str,XMLString::stringLen(str)))
			//if(XMLString::isAllWhiteSpace(str))
			{ //remove node
				DOMNode* n2=n;
				n=n->getNextSibling();
				n2=node->removeChild(n2);
				n2->release(); 
				//delete n2;
				continue;
			}
			else
			{
				//remove trailing and multiplie WhiteSpaces
				XMLCh *newstr=XMLString::replicate(str);  
				XMLString::collapseWS(newstr); 
				n->setNodeValue(newstr);
				XMLString::release(&newstr); 
				//XMLString::collapseWS(str); //hack, crack data into XERCES! Correctly need copy to new buffer and set!
			}
		}
		else
			RemoveSpaceTextNodes(n);
		n=n->getNextSibling();
	}
}

////////////////////////////////////////////////////////////////////
void YXMLDRVFUN YXML_DRV_ReleaseCString(YXML_DRV_DomCString * cstr)
{ 
	if(*cstr) XMLString::release(cstr);   
}
////////////////////////////////////////////////////////////////////
int YXML_DRV_reference_counter=0;
int YXMLDRVFUN YXML_DRV_InitializeXercesc(int true_false,const char *encoding_system,YXML_DRV_DomCString *exception_message)
{
	TRY_XML
	if(true_false)
	{
		if(YXML_DRV_reference_counter==0)
			XMLPlatformUtils::Initialize(encoding_system);
		++YXML_DRV_reference_counter;
	}
	else
	{
		if(--YXML_DRV_reference_counter==0)
			XMLPlatformUtils::Terminate();
	}
	CATCH_XML
	return YXML_DRV_reference_counter;
}
////////////////////////////////////////////////////////////////////
int YXMLDRVFUN YXML_DRV_OpenClose(YXML_DRV_DocHandle* handle,unsigned op_opt,const char *fname,const char *docname,const char *encoding,YXML_DRV_DomCString *exception_message)
{
	TRY_XML
	switch(op_opt&YXML_DOC_DO_MASK)
	{
	case YXML_DOC_DO_CREATE:
		{
			DOMImplementation* impl =  DOMImplementationRegistry::getDOMImplementation(StrXML("Core"));
			DOMDocument* doc = impl->createDocument(
									0,                    // root element namespace URI.
									StrXML(docname),         // root element name
									0);                   // document type object (DTD).
			//doc->setEncoding(StrXML("cp866")); // DOS
			doc->setEncoding(StrXML(encoding));// Windows
			doc->setStandalone(true);
			//doc->setDocumentURI(StrXML("http://www.optecs.ru"));
			*handle=doc;
		}
		break;
	case YXML_DOC_DO_LOAD:
		{
			const bool   gDoNamespaces          = false;
			const bool   gDoSchema              = true;//false;
			const bool   gSchemaFullChecking    = true;//false;
			const bool   gDoCreate              = true;
			// options for DOMWriter's features
			//const XMLCh*  gOutputEncoding        = 0;
			//const XMLCh*  gMyEOLSequence         = 0;

			//const bool    gSplitCdataSections    = true;
			//const bool    gDiscardDefaultContent = true;
			//const bool    gUseFilter             = false;
			//const bool    gFormatPrettyPrint     = false;

			const XercesDOMParser::ValSchemes    gValScheme  = XercesDOMParser::Val_Auto;

			XercesDOMParser * parser = new XercesDOMParser;
			parser->setValidationScheme(gValScheme);
			parser->setDoNamespaces(gDoNamespaces);
			parser->setDoSchema(gDoSchema);
			parser->setValidationSchemaFullChecking(gSchemaFullChecking);
			parser->setCreateEntityReferenceNodes(gDoCreate);
			parser->setIncludeIgnorableWhitespace(true); 
			parser->parse(fname); //connect to XML file

			DOMDocument* doc=parser->adoptDocument(); 
			delete parser;

			DOMNode* node=doc->getDocumentElement();
			if(node)
			{
	//			if(0!=XMLString::compareString(node->getNodeName(),docname))
	//				THROW("this is another document");
				if(op_opt&YXML_DOC_OPT_EAT_WS)
					RemoveSpaceTextNodes(node);
			}
			*handle=doc;
		}
		break;
	case YXML_DOC_DO_SAVE:
		{
			DOMNode  *node=((DOMDocument*)(*handle))->getDocumentElement();
			DOMNode  *node_main = node->getOwnerDocument();

			XMLCh tempStr[100];
			XMLString::transcode("LS", tempStr, 99);
			DOMImplementation *impl = DOMImplementationRegistry::getDOMImplementation(tempStr);
			DOMWriter* theSerializer = ((DOMImplementationLS*)impl)->createDOMWriter();

			// optionally you can set some features on this serializer
			if (theSerializer->canSetFeature(XMLUni::fgDOMWRTDiscardDefaultContent, true))
					theSerializer->setFeature(XMLUni::fgDOMWRTDiscardDefaultContent, true);

			if(op_opt&YXML_DOC_OPT_FORMATTED_OUTPUT)
			{
				if (theSerializer->canSetFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true))
						 theSerializer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, true);
			}
					// optionally you can implement your DOMWriterFilter (e.g. MyDOMWriterFilter)
					// and set it to the serializer
	//        DOMWriterFilter* myFilter = new myDOMWriterFilter();
	//        theSerializer->setFilter(myFilter);

					// optionally you can implement your DOMErrorHandler (e.g. MyDOMErrorHandler)
					// and set it to the serializer
	//        DOMErrorHandler* errHandler = new myDOMErrorHandler();
	//        theSerializer->setErrorHandler(myErrorHandler);

			// StdOutFormatTarget prints the resultant XML stream
			// to stdout once it receives any thing from the serializer.
			//XMLFormatTarget *myFormTarget = new StdOutFormatTarget();
			XMLFormatTarget *myFormTarget = new LocalFileFormatTarget(fname);

			 // do the serialization through DOMWriter::writeNode();
					theSerializer->writeNode(myFormTarget, *node_main);

			theSerializer->release();
			//delete myErrorHandler;
			//delete myFilter;
			delete myFormTarget;
		}
		break;
	case YXML_DOC_DO_FREE:
		{
				if(*handle)
				{
					((DOMDocument*)(*handle))->release();
					*handle=0;
				}
		}
		break;
	default:
		;
	};
	CATCH_XML
	return 0;
}
////////////////////////////////////////////////////////////////////
YXML_DRV_NodeHandle YXMLDRVFUN YXML_DRV_GetRootNode(YXML_DRV_DocHandle handle,YXML_DRV_DomCString *exception_message)
{
	DOMNode  *node=0;
	TRY_XML
	node=((DOMDocument*)handle)->getDocumentElement();
	CATCH_XML
	return node;
}
////////////////////////////////////////////////////////////////////
// Node operation
////////////////////////////////////////////////////////////////////
#define NODE(node) ((DOMNode*)(node)) 
//
YXML_DRV_NodeHandle YXMLDRVFUN YXML_DRV_TreeWalk(YXML_DRV_NodeHandle current,unsigned walk,YXML_DRV_DomCString *exception_message)
{
	YXML_DRV_NodeHandle node_new=0;
	TRY_XML
	switch(walk)
	{
	case YXML_DRV_WALK_PARENT:
		node_new=NODE(current)->getParentNode();
		break;
	case YXML_DRV_WALK_CHILD_FIRST:
		node_new=NODE(current)->getFirstChild();
		break;
	case YXML_DRV_WALK_CHILD_LAST:
		node_new=NODE(current)->getLastChild();
		break;
	case YXML_DRV_WALK_SIBLING_NEXT:
		node_new=NODE(current)->getNextSibling();
		break;
	case YXML_DRV_WALK_SIBLING_PREV:
		node_new=NODE(current)->getPreviousSibling();
		break;
	default:
		node_new=0;
	};
	CATCH_XML
	return node_new;
}

////////////////////////////////////////////////////////////////////
YXML_DRV_NodeHandle YXMLDRVFUN YXML_DRV_Insert(YXML_DRV_NodeHandle current,unsigned where,unsigned type,const char *data,YXML_DRV_DomCString *exception_message)
{
	TRY_XML
		StrXML val(data);
		DOMDocument* doc=NODE(current)->getOwnerDocument();
		DOMNode*    elem=0;
		switch(type)
		{
		case YXML_DRV_NODETYPE_ELEMENT:
			elem = doc->createElement(val);
			break;
		case YXML_DRV_NODETYPE_TEXT:
			elem = doc->createTextNode(val);
			break;
		case YXML_DRV_NODETYPE_COMMENT:
			elem = doc->createComment(val);
			break;
		default:
			break;
		};
		switch(where)
		{
		case YXML_DRV_INSERT_DELETE:
			{
				DOMNode*    p=NODE(current)->getParentNode();
				if(p)
				{
					p=p->removeChild(NODE(current));
					if(p)
						p->release();
				}
			}
			break;
		case YXML_DRV_INSERT_APPEND_CHILD:
			return NODE(current)->appendChild(elem);
		case YXML_DRV_INSERT_BEFORE:
			return NODE(current)->getParentNode()->insertBefore(elem,NODE(current));
		case YXML_DRV_INSERT_AFTER:
			{
				DOMNode *np=NODE(current)->getParentNode();
				if(np)
				{
					DOMNode *nn=NODE(current)->getNextSibling();
					if(nn)
						return np->insertBefore(elem,nn);
					else
						return np->appendChild(elem);
				}
			}
			break;
		};
	CATCH_XML
	return 0;
}


unsigned YXMLDRVFUN YXML_DRV_NodeType(YXML_DRV_NodeHandle current,YXML_DRV_DomCString *exception_message)
{
	TRY_XML
	switch(NODE(current)->getNodeType())
	{
	case DOMNode::ELEMENT_NODE:
		return YXML_DRV_NODETYPE_ELEMENT;
	case DOMNode::TEXT_NODE:
		return YXML_DRV_NODETYPE_TEXT;
	case DOMNode::ATTRIBUTE_NODE:
		return YXML_DRV_NODETYPE_ATTRIBUTE;
	case DOMNode::COMMENT_NODE:
		return YXML_DRV_NODETYPE_COMMENT;
	default:
		break;
	};
	CATCH_XML
	return YXML_DRV_NODETYPE_UNSUPPORTED;
}
YXML_DRV_DomCString YXMLDRVFUN YXML_DRV_NodeGetSet(YXML_DRV_NodeHandle current,unsigned what,const char *attr_name,const char *data,YXML_DRV_DomCString *exception_message)
{
	TRY_XML
	switch(what)
	{
//	case YXML_DRV_NODE_SET_NAME:
	case YXML_DRV_NODE_SET_VALUE:
		NODE(current)->setNodeValue(StrXML(data));
		return 0;
	case YXML_DRV_NODE_GET_NAME:
		return TranscodeStrX(NODE(current)->getNodeName());
	case YXML_DRV_NODE_GET_VALUE:
		return TranscodeStrX(NODE(current)->getNodeValue());
	case YXML_DRV_NODE_GET_ATTR:
		{
			if(!current)
				return 0;
			if(!NODE(current)->hasAttributes())
				return 0;
			DOMNamedNodeMap *attr=NODE(current)->getAttributes();
			DOMNode*  n=attr->getNamedItem(StrXML(attr_name)); 
			if(n)
				return TranscodeStrX(n->getNodeValue());
			else
				return 0;
		}
	case YXML_DRV_NODE_SET_ATTR:
		{
			DOMDocument* doc=NODE(current)->getOwnerDocument();
			DOMAttr*  attr=doc->createAttribute(StrXML(attr_name));
			attr->setValue(StrXML(data));
			DOMNamedNodeMap* nm=NODE(current)->getAttributes();
			nm->setNamedItem(attr);
		}
		return 0;
	default:
		break;
	};
	CATCH_XML
	return 0;
}
////////////////////////////////////////////////////////////////////
