#ifndef _ALEXANDRESKU_H
#define _ALEXANDRESKU_H
/////////////////////////////////////////////////////////////////////////
#ifdef __BORLANDC__
	#include <system.hpp>
	#include <stdint.h>
#elif _MSC_VER
	typedef unsigned char        Byte;
	#include "stdint.h"
#elif __GNUC__
	typedef unsigned char        Byte;
	#include <stdint.h>
#else
	#error "Unknown compiler"
#endif
#include <sstream>
#include <algorithm>
#define _USE_MATH_DEFINES
#include <cmath>
#include "loki_TypeManip.h"

/** @addtogroup TplDoxGroup  Templates library of algorithms. 
 *
 *  @brief      ,
 *	        
 *	     .
 *  @{
 *  @}
 ***********************************************************************/


/*!
 * \brief Computer <b>V</b>ision <b>T</b>emplate <b>L</b>ibrary - ++ 
 *          .
 *
 ***********************************************************************/
namespace Vtl
{
/** @addtogroup TplBasicsDoxGroup  Some basic templates
 *  @ingroup TplDoxGroup  
 *  @{
 ***********************************************************************/
	//--------General macros:
	template<typename T> T sqr(T v) {return v*v;}
	//#ifndef MAX3
	template<typename T>  T max(T a,T b,T c) {return std::max(std::max(a,b),c);}
	//#endif
	//#ifndef MAX4
	template<typename T>  T max(T a,T b,T c,T d) {return std::max(std::max(a,b),std::max(c,d));}
	//#endif
	//#ifndef MIN3
	template<typename T>  T min(T a,T b,T c) {return std::min(std::min(a,b),c); }
	//#endif
	//#ifndef MIN4
		template<typename T>  T min(T a,T b,T c,T d) {return std::min(std::min(a,b),std::min(c,d)); }
	//#endif
	template<typename T1,typename T2,typename T3> 
	T1 ToBounds(T1 var,T2 min,T3 max)  { return (var<min)?min:((var>max)?max:var); }
	//------FORTRAN definitions:
	//   "numalg.h"
	//template<typename T1,typename T2>	T1 sign(T1 x1,T2 x2) { return (x2 >= 0) ? std::abs(x1) : -std::abs(x1); }
/** 
 *  TplBasicsDoxGroup  End
 *  
 *  @}   
 ***********************************************************************/
   /*!
	* \namespace Vtl::Types
	* \brief   Vtl::Types      
	*   :
	*
	*   - struct EmptyType; -    " ".
	*   - class NullType;   -    " ".
	*   - template<int v> struct Int2Type; -    .
	*   - template<typename T> struct Type2Type; -     
	*        ,   - .   
	*              , 
	*          (  ,     
	*         ,    ). 
	*     . Lit_Alexandresku  53-54.
	*   - template<bool flag,typename Type1,typename Type2> struct Select; - 
	*            .
	*   - template<class T,class U> class Conversion; - ,   
	*        (cast)  T  U.
	*   -  ,   ,     .
	*   -       (namespace Vtl::Types::TL).
	*   -    (Traits).
	*
	*       \ref Lit_Alexandresku .
	* \cond 
	* .     ++.  
	* ++ In-Depth, .3.:.  . -.:  "",2002. -336 . 
	* . -. . . ISBN 5-8459-035103 (.).
	* \endcond
	*
	*******************************************************************/
	namespace Types
	{
		///,    ,    EmptyType 
		///          (  ).
		struct EmptyType {};
		///  
		class NullType {};
		///       
		///  .
		template<int v>
		struct Int2Type
		{
			enum {value=v};
		};
		///      
		///         
		///  String* s=Create(Type2Type<String>());
		template<typename T>
		struct Type2Type
		{
			typedef T OriginalType;
		};
		/*! 
		 *	@class Select
		 *	@brief     , if flag==true -  ,  - 
		 *
		 *  :
		 *  template<class Pixel> void test(Pixel *p)
		 *  {   //     MyType 
		 *      //  ,   Pixel  -  uint16_t
		 *      //    uint32_t
		 *  	typedef Vtl::Types::Select<sizeof(Pixel)==1,uint16_t,uint32_t>::Type MyType;
		 *  	MyType1 aa1=*p;
		 *  }
		***********************************************************/
		template<bool flag,typename Type1,typename Type2>
		struct Select
		{
			typedef Type1 Type;
		};
		//@cond 0
		template<typename Type1,typename Type2>
		struct Select<false,Type1,Type2>
		{
			typedef Type2 Type;
		};
		//@endcond
		/*!
		* @class Conversion
		* @brief ,      (cast)   T   U
		* 
		***********************************************************/
		template<class T,class U>
		class Conversion
		{
		//@cond 0
			typedef char Small;
			class Big {Small dummy[2];};
			static Small Test(U);
			static Big Test(...);
			static T MakeT();
		//@endcond 
		public:
			enum {exists = sizeof(Test(MakeT()))==sizeof(Small)};
			enum {sameType=false};
		};
		//@cond 0
		template<class T>
		class Conversion<T,T>
		{
		public:
			enum {exists = true};
			enum {sameType=true};
		};
		//@endcond
		/*!
		* @class isSameType
		* @brief ,   .    
		*   -   value=1,  value=0.
		* 
		***********************************************************/
		template<class T1,class T2> struct isSameType
		{
			enum {value=0};
		};
		//@cond 0
		template<class T> struct isSameType<T,T>
		{
			enum {value=1};
		};
		//@endcond

		/*!
		 *	@class Typelist
		 *	@brief  . .  
		 *	@link dxgrp_TypeLists     
		 *	  @endlink 
		 *	 namespace Vtl::Types::TL 
		 ***************************************************************/
		template<class T, class U>
		struct Typelist
		{
			typedef T Head;
			typedef U Tail;
		};
/*! 
 * @defgroup dxgrp_TypeLists        
 * @ingroup TplDoxGroup  
 *
 * @{
 ********************************************************************************/
		/*! 
		     Typelist   .\n
		 : \n
		 typedef TYPELIST_2(signed char,short) MyTList_1; \n
		 typedef TYPELIST_3(signed char,short,int) MyTList_2; 
		 */
		#define TYPELIST_1(T1) Vtl::Types::Typelist<T1,Vtl::Types::NullType>
		#define TYPELIST_2(T1,T2) Vtl::Types::Typelist<T1,TYPELIST_1(T2) >
		#define TYPELIST_3(T1,T2,T3) Vtl::Types::Typelist<T1,TYPELIST_2(T2,T3) >
		#define TYPELIST_4(T1,T2,T3,T4) Vtl::Types::Typelist<T1,TYPELIST_3(T2,T3,T4) >
		#define TYPELIST_5(T1,T2,T3,T4,T5) Vtl::Types::Typelist<T1,TYPELIST_4(T2,T3,T4,T5) >
		#define TYPELIST_6(T1,T2,T3,T4,T5,T6) Vtl::Types::Typelist<T1,TYPELIST_5(T2,T3,T4,T5,T6) >
		#define TYPELIST_7(T1,T2,T3,T4,T5,T6,T7) Vtl::Types::Typelist<T1,TYPELIST_6(T2,T3,T4,T5,T6,T7) >
		#define TYPELIST_8(T1,T2,T3,T4,T5,T6,T7,T8) Vtl::Types::Typelist<T1,TYPELIST_7(T2,T3,T4,T5,T6,T7,T8) >
		#define TYPELIST_9(T1,T2,T3,T4,T5,T6,T7,T8,T9) Vtl::Types::Typelist<T1,TYPELIST_8(T2,T3,T4,T5,T6,T7,T8,T9) >
		#define TYPELIST_10(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10) Vtl::Types::Typelist<T1,TYPELIST_9(T2,T3,T4,T5,T6,T7,T8,T9,T10) >
		#define TYPELIST_11(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11) Vtl::Types::Typelist<T1,TYPELIST_10(T2,T3,T4,T5,T6,T7,T8,T9,T10,T11) >
		#define TYPELIST_12(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12) Vtl::Types::Typelist<T1,TYPELIST_11(T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12) >
		#define TYPELIST_13(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13) Vtl::Types::Typelist<T1,TYPELIST_12(T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13) >
		#define TYPELIST_14(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14) Vtl::Types::Typelist<T1,TYPELIST_13(T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14) >
		#define TYPELIST_15(T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15) Vtl::Types::Typelist<T1,TYPELIST_14(T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15) >
//@}
		/*!******************************************************************
		 * \namespace Vtl::Types::TL 
		 * \brief   Vtl::Types::TL  
		 *     ,  Vtl::Types::Typelist
		 *
		 *   Vtl::Types::TL   
		 *    .  ,  
		 *    ,     
		 *       .
		 *
		 *      
         *
		 * .     ++.  
		 * ++ In-Depth, .3.:.  . -.:  "",2002. -336 . 
		 * . -. . . ISBN 5-8459-035103 (.).
		 *
		 * . \ref Lit_Alexandresku
		 *******************************************************************/
		namespace TL
		{
			/*!
			*	class template MakeTypelist
			*	Takes a number of arguments equal to its numeric suffix
			*	The arguments are type names.
			*	@code
			*	MakeTypelist<T1, T2, ...>::Result
			*	@endcode
			*	returns a typelist that is of T1, T2, ...
			**************************************************************/

			template
			<
				typename T1  = NullType, typename T2  = NullType, typename T3  = NullType,
				typename T4  = NullType, typename T5  = NullType, typename T6  = NullType,
				typename T7  = NullType, typename T8  = NullType, typename T9  = NullType,
				typename T10 = NullType, typename T11 = NullType, typename T12 = NullType,
				typename T13 = NullType, typename T14 = NullType, typename T15 = NullType,
				typename T16 = NullType, typename T17 = NullType, typename T18 = NullType
			> 
			struct MakeTypelist
			{
			private:
				typedef typename MakeTypelist
				<
					T2 , T3 , T4 , 
					T5 , T6 , T7 , 
					T8 , T9 , T10, 
					T11, T12, T13,
					T14, T15, T16, 
					T17, T18
				>
				::Result TailResult;

			public:
				typedef Typelist<T1, TailResult> Result;
			};

			template<>
			struct MakeTypelist<>
			{
				typedef NullType Result;
			};
			/*!
			*	@class Vtl::Types::TL::Length
			*	@brief    
			*
			*      .
			*   @code
			*	Vtl::Types::TL::Length<class TList>
			*	@endcode
			*         Vtl::Types::Typelist
			*    Vtl::Types::NullType .    - 
			*    enum ,   value ,   
			*    .      !
			*            
			*      - "  ..."
			*	:
			*	@code
			*	//- 
			*	template<int num> class Test 
			*	{ 
			*	public: 
			*		int get() const {return num;} 
			*	};
			*	//  
			*	typedef TYPELIST_3(int,float,char) MyTList;
			*	//      
			*	// MyTList, ..    
			*	//  
			*	Test<Vtl::Types::TL::Length<MyTList>::value > temp;
			*   //      :
			*   float a1[Vtl::Types::TL::Length<MyTList>::value];
			*   int len=sizeof(my_array)/sizeof(my_array[0]);
			*   std::cout << "Length of array a1 is" << len << std::endl;
			*   std::cout << "Class initialized by template integer param " << temp.Get() << std::endl;
			*	@endcode
			**************************************************************/
			template <class TList> struct Length;
			// @cond 0
			template <> struct Length<NullType>
			{
				enum {value=0};
			};
			template <class T, class U> struct Length<Typelist<T,U> >
			{
				enum {value=1+Length<U>::value};
			};
			// @endcond
			/*!
			*	@class Vtl::Types::TL::TypeAt
			*	@brief     
			*
			*      .
			*   @code
			*	Vtl::Types::TL::TypeAt<class TList, unsigned int index>
			*	@endcode
			*        Vtl::Types::Typelist
			*    Vtl::Types::NullType . 
			*    -  , ,   0   
			*       .
			*       - 
			*   typedef Result,       .
			*    (.. )     !
			*           
			*        
			*      - "  ..."
			*	:
			*	@code
			*	//- 
			*	template<class T> class Test {};
			*	//  
			*	typedef TYPELIST_5(int,float,char,Test<float>,double) MyTList;
			*	//      Test<T> 
			*	// T -     MyTList
			*	Test<Vtl::Types::TL::TypeAt<MyTList,2>::Result > temp;
			*	@endcode
			**************************************************************/
			template<class TList, unsigned int index> struct TypeAt;
			// @cond 0
			template<class Head,class Tail>	struct TypeAt<Typelist<Head,Tail>,0>
			{
				typedef Head Result;
			};
			template<class Head,class Tail, unsigned int i> struct TypeAt<Typelist<Head,Tail>,i>
			{
				typedef typename TypeAt<Tail,i-1>::Result Result;
			};
			/*!
			 * class template TypeAtNonStrict
			 * Finds the type at a given index in a typelist
			 * Invocations (TList is a typelist and index is a compile-time integral 
			 *     constant):
			 *
			 * a) 
 			 *	@code
 			 *	TypeAt<TList, index>::Result
			 *	@endcode
			 * returns the type in position 'index' in TList, or NullType if index is 
			 *     out-of-bounds
			 *
			 * b) 
 			 *	@code
			 *	TypeAt<TList, index, D>::Result
			 *	@endcode
			 * returns the type in position 'index' in TList, or D if index is out-of-bounds
			 ********************************************************************************/

			template <class TList, unsigned int index,
				typename DefaultType = NullType>
			struct TypeAtNonStrict
			{
				typedef DefaultType Result;
			};
	        
			template <class Head, class Tail, typename DefaultType>
			struct TypeAtNonStrict<Typelist<Head, Tail>, 0, DefaultType>
			{
				typedef Head Result;
			};
	        
			template <class Head, class Tail, unsigned int i, typename DefaultType>
			struct TypeAtNonStrict<Typelist<Head, Tail>, i, DefaultType>
			{
				typedef typename 
					TypeAtNonStrict<Tail, i - 1, DefaultType>::Result Result;
			};
 				// @endcond
			/*!
			*	@class Vtl::Types::TL::IndexOf
			*	@brief      ,    -  -1;
			*
			*      .
			*   @code
			*	Vtl::Types::TL::IndexOf<TList, class T>
			*	@endcode
			*	:
			*	@code
			*	//- 
			*	template<unsigned num> class Test {};
			*	//  
			*	typedef TYPELIST_5(int,float,char,Test<3>,double) MyTList;
			*	//      Test<T> 
			*	// T -     MyTList
			*	Test<Vtl::Types::TL::IndexOf<MyTList,float>::value > temp;
			*   //   Test<1> temp; .  MyTList .
			*	@endcode
			**************************************************************/
			template<class TList, class T> struct IndexOf;
			// @cond 0
			template<class T> struct IndexOf<NullType,T>
			{
				enum {value=-1};
			};
			template<class T,class Tail> struct IndexOf<Typelist<T,Tail>,T>
			{
				enum {value=0};
			};
			template<class Head,class Tail,class T> struct IndexOf<Typelist<Head,Tail>,T>
			{
			private:
				enum {temp=IndexOf<Tail,T>::value};
			public:
				enum {value=temp==-1?-1:1+temp};
			};
			// @endcond
		   /*!
			*	@class Vtl::Types::TL::Append
			*	@brief     
			*
			*   @code
			*	Vtl::Types::TL::Append<class TList,class T> 
			*	@endcode
			*	:
			*	@code
			*	@endcode
			********************************************************/
			template<class TList,class T> struct Append;
			// @cond 0
			template<> struct Append<NullType,NullType>
			{
				typedef NullType Result;
			};
			template<typename T> struct Append<NullType,T>
			{
				typedef TYPELIST_1(T) Result;
			};
			template<class Head,class Tail> struct Append<NullType,Typelist<Head,Tail> >
			{
				typedef Typelist<Head,Tail> Result;
			};
			template<class Head,class Tail,typename T> struct Append<Typelist<Head,Tail>,T>
			{
				typedef Typelist<Head,typename Append<Tail,T>::Result> Result;
			};
			// @endcond
			///   ,     - TRUE
			template<class TList,class U,bool append> struct AppendIf
			{
				typedef TList Result;
			};
			// @cond 0
			template<class TList,class U> struct AppendIf<TList,U,true>
			{
				typedef typename Append<TList,U>::Result Result;
			};
			// @endcond
			/*!
			*	@class Vtl::Types::TL::FindIf
			*	@brief        , 
			*       Comparator .
			*
			*     3 .
			*   @code
			*	Vtl::Types::TL::FindIf<<class TList, class T,template <class,class> Comparator >
			*	@endcode
			*            -
			*      value  , :
			*   @code
			*   template<class T1,class T2> struct Comp2
			*   {
			*		enum { value=sizeof(typename T1::sample_type)==sizeof(T2)?1:0};
			*   };
			*	@endcode
			*	 :
			*	@code
			*	typedef TYPELIST_4(char,short,unsigned short,int) MyTypesList;
			*	typedef Vtl::Types::TL::FindIf<MyTypesList,short,Comp2>::Result my_typelist;
			*	std::cout << Vtl::Types::TL::Print<my_typelist>::str()<< std::endl;
			*	@endcode
			**************************************************************/
			template<class TList, class T,template <class,class> class Comparator> struct FindIf;
			// @cond 0
			template<class T,template <class,class> class Comparator> struct FindIf<NullType,T,Comparator>
			{
				typedef NullType Result;
			};
			template<class Head,class Tail,class T,template <class,class> class Comparator> struct FindIf<Typelist<Head,Tail>,T,Comparator>
			{
			private:
				enum { temp=Comparator<Head,T>::value};
			public:
				typedef typename AppendIf<typename FindIf<Tail,T,Comparator>::Result,Head,temp>::Result Result;
			};
			// @endcond
			/*!
			*	@class Vtl::Types::TL::Erase
			*	@brief     .
			*
			*      .
			*   @code
			*	Vtl::Types::TL::Erase<class TList,typename T>
			*	@endcode
			*	:
			*	@code
			*	@endcode
			**************************************************************/
			template<class TList,typename T> struct Erase;
			// @cond 0
			template<typename T> struct Erase<NullType,T>
			{
				typedef NullType Result;
			};
			template<typename T,class Tail> struct Erase<Typelist<T,Tail>,T>
			{
				typedef Tail Result;
			};
			template<class Head,class Tail,typename T> struct Erase<Typelist<Head,Tail>,T>
			{
				typedef Typelist<Head,typename Erase<Tail,T>::Result> Result;
			};
			// @endcond
			/*!
			*	@class Vtl::Types::TL::EraseAll
			*	@brief       .
			*
			*      .
			*   @code
			*	Vtl::Types::TL::EraseAll<class TList,typename T>
			*	@endcode
			*	:
			*	@code
			*	@endcode
			**************************************************************/
			template<class TList,typename T> struct EraseAll;
			// @cond 0
			template<typename T> struct EraseAll<NullType,T>
			{
				typedef NullType Result;
			};
			template<typename T,class Tail> struct EraseAll<Typelist<T,Tail>,T>
			{
				typedef typename EraseAll<Tail,T>::Result Result;
			};
			template<class Head,class Tail,typename T> struct EraseAll<Typelist<Head,Tail>,T>
			{
				typedef Typelist<Head,typename EraseAll<Tail,T>::Result> Result;
			};
			// @endcond
			/*!
			*	@class Vtl::Types::TL::NoDuplicates
			*	@brief       .
			*
			*      .
			*   @code
			*	Vtl::Types::TL::NoDuplicates<class TList>
			*	@endcode
			*	:
			*	@code
			*	@endcode
			**************************************************************/
			template<class TList> struct NoDuplicates;
			//@cond 0
			template<> struct NoDuplicates<NullType>
			{
				typedef NullType Result;
			};
			template<class Head,class Tail> struct NoDuplicates<Typelist<Head,Tail> >
			{
			private:
				typedef typename NoDuplicates<Tail>::Result L1;
				typedef typename Erase<L1,Head>::Result L2;
			public:
				typedef Typelist<Head,L2> Result;
			};
			//@endcond
			///   ?
			template<class TList,typename T> struct isIn
			{
				enum{ Result=(TL::IndexOf<TList,T>::value) >= 0 ? true:false}; 
			};
			/*!
			*	@class Vtl::Types::TL::TypeLargerThen
			*	@brief       sizeof , 
			*               .
			*
			*      .
			*   @code
			*	Vtl::Types::TL::TypeLargerThen<class TList,unsigned size> 
			*	@endcode
			*	:
			*	@code
			*	@endcode
			**************************************************************/
			template<size_t s1,size_t s2> struct GT
			{
				enum {value=(s1>s2) };
			};
			template<size_t s1,size_t s2> struct LT
			{
				enum {value=(s1<s2) };
			};
			template<size_t s1,size_t s2> struct GE
			{
				enum {value=(s1>=s2) };
			};
			template<size_t s1,size_t s2> struct LE
			{
				enum {value=(s1<=s2) };
			};
			template<size_t s1,size_t s2> struct EQ
			{
				enum {value=(s1==s2) };
			};
			template<size_t s1,size_t s2> struct NE
			{
				enum {value=(s1!=s2) };
			};
			template<class TList,size_t s,template<size_t s1,size_t s2> class Comaparator > struct FilteredTypes;
			// @cond 0
			template<size_t s,template<size_t s1,size_t s2> class Comaparator > struct FilteredTypes<NullType,s,Comaparator>
			{
				typedef NullType Result;
			};
			template<class Head,class Tail,size_t s,template<size_t s1,size_t s2> class Comaparator > struct FilteredTypes<Typelist<Head,Tail>,s,Comaparator>
			{
				typedef typename Select<
					Comaparator<sizeof(Head),s>::value
					,typename Typelist<Head,
						typename FilteredTypes<Tail,s,Comaparator>::Result> 
					,typename FilteredTypes<Tail,s,Comaparator>::Result
					>::Type Result;
			};
			// @endcond
			/*!
			*	@class Vtl::Types::TL::SmallestType
			*	@brief        sizeof.
			*
			*      .
			*   @code
			*	Vtl::Types::TL::SmallestType<class TList>
			*	@endcode
			*	:
			*	@code
			*	@endcode
			**************************************************************/
			template<class TList,template<size_t s1,size_t s2> class Comaparator > struct SmallestType;
			// @cond 0
			template<template<size_t s1,size_t s2> class Comaparator > struct SmallestType<NullType,Comaparator >
			{
				enum {size=0};
				typedef NullType Result;
			};
			template<typename T,template<size_t s1,size_t s2> class Comaparator > struct SmallestType<Typelist<T,NullType>,Comaparator >
			{
				enum {size=sizeof(T)};
				typedef T Result;
			};
			template<typename Head,class Tail,template<size_t s1,size_t s2> class Comaparator> struct SmallestType<Typelist<Head,Tail>,Comaparator >
			{
				typedef SmallestType<Tail,Comaparator> NextType;
				typedef typename Select<
					Comaparator<sizeof(Head),NextType::size>::value 
					,Head 
					,typename NextType::Result
					>::Type Result;
				enum {size=sizeof(Result)};
			};
			// @endcond


			/*!
			*	@class Vtl::Types::TL::Print
			*	@brief  -  :     sizeof  .
			*
			*      .
			*   @code
			*	Vtl::Types::TL::Print<class TList>
			*	@endcode
			*	:
			*	@code
			*	typedef TYPELIST_2(char,int) MyList;
			*	std::cout << Vtl::Types::TL::Print<MyList>::str()<< std::endl;
			*	@endcode
			**************************************************************/
			template<class TList> struct Print
			{//  
				static std::string str() 
				{ 
					std::ostringstream os;
					os <<"sizeof(" << typeid(TList).name() <<")=" << sizeof(TList) << std::endl; 
					return os.str();
				}
			};
			// @cond 0
			template<> struct Print<NullType >
			{//  - NullType -   
				static std::string str() { return "NullType" ;}
			};
			template<class Head,class Tail> struct Print<Typelist<Head,Tail> >
			{//  -  , Typelist
				static std::string str() 
				{ 
					std::ostringstream os;
					os <<"sizeof(" << typeid(Head).name() <<")=" << sizeof(Head) << std::endl << Print<Tail>::str() ; 
					return os.str();
				}
			};
			// @endcond
		};

/*! 
 *  @defgroup StdTypeLists   ()  
 *  @ingroup TplDoxGroup  
 *   @brief    :
 ********************************************************************************************/
//@{
		///     
		typedef TL::NoDuplicates<TYPELIST_8(signed char,short,int,long,int8_t,int16_t,int32_t,int64_t)>::Result SignedIntegerTypes;
		///     
		typedef TL::NoDuplicates<TYPELIST_8(unsigned char,unsigned short,unsigned int,unsigned long,uint8_t,uint16_t,uint32_t,uint64_t)>::Result UnsignedIntegerTypes;
		///     
		typedef TL::NoDuplicates<TYPELIST_3(float,double,long double)>::Result RealTypes;
		///   "" ,   bool,char,wchar_t
		typedef TL::NoDuplicates<TYPELIST_3(bool,char,wchar_t)>::Result OtherIntegerTypes;
		///       .
		typedef TL::NoDuplicates<TL::Append<SignedIntegerTypes,UnsignedIntegerTypes>::Result>::Result IntegerTypes;
		///       ,  OtherIntegerTypes (  'char').
		typedef TL::NoDuplicates<TL::Append<IntegerTypes,OtherIntegerTypes>::Result>::Result AllIntegerTypes;
		///     -     .
		typedef TL::NoDuplicates<TL::Append<SignedIntegerTypes,RealTypes>::Result>::Result SignedTypes;
		///     -        . 
		typedef TL::NoDuplicates<TL::Append<IntegerTypes,RealTypes>::Result>::Result ArithTypes;
		///     -        ,  OtherIntegerTypes (  'char').
		typedef TL::NoDuplicates<TL::Append<ArithTypes,OtherIntegerTypes>::Result>::Result AllArithTypes;
//@}
		
		///
		/// \class Traits
		///   
		///
		///
		template<typename T>
		class Traits
		{
		private:
			// @cond 0  -hide from documentation realization details
			template<class U> struct PointerTraits
			{
				enum {Result=false};
				typedef NullType PointeeType;
			};
			template<class U> struct PointerTraits<U*>
			{
				enum {Result=true};
				typedef U PointeeType;
			};
			template <class U> struct ReferenceTraits
			{
				enum { Result = false };
				typedef U ReferredType;
			};
	        
			template <class U> struct ReferenceTraits<U&>
			{
				enum { Result = true };
				typedef U ReferredType;
			};
	        
			template <class U> struct PToMTraits
			{
				enum { Result = false };
			};
	        
			template <class U, class V>
			struct PToMTraits<U V::*>
			{
				enum { Result = true };
			};
			template <class U> struct UnConst
			{
				typedef U Result;
				enum { isConst = 0 };
			};
	        
			template <class U> struct UnConst<const U>
			{
				typedef U Result;
				enum { isConst = 1 };
			};
			template <class U> struct UnVolatile
			{
				typedef U Result;
				enum { isVolatile = 0 };
			};
		    
			template <class U> struct UnVolatile<volatile U>
			{
				typedef U Result;
				enum { isVolatile = 1 };
			};
			// @endcond
		public:
			/// True,   T - ,  False.
			enum {isPointer=PointerTraits<T>::Result};
			/// @brief   T - ,  PointeeType    
			///  .    PointeeType
			///   NullType. 
			typedef typename PointerTraits<T>::PointeeType PointeeType;
			/// True,   T -  ,  False.
			enum { isReference = ReferenceTraits<T>::Result };
			/// @brief  T -    const. 
			///   T - ,  ReferredType -    
			///  T.    ReferredType 
			///      T. 
			typedef typename ReferenceTraits<T>::ReferredType ReferredType;
			///  @brief True,   T -   - .
			enum { isMemberPointer = PToMTraits<T>::Result };
			/*! @brief True,   T -  . 
				: @code
				template<typename T> void foo(T var)
				{
					bool c=Vtl::Types::Traits<T>::isConst;
					std::cout << "Type of 'var'" is " << c?"true":"false" <<std::endl;
				}
				const char* name="asd";
				foo(asd);

				@endcode
			*/
			enum { isConst = UnConst<T>::isConst };
			/*!  @brief  T -    const. 

				: @code
				template<typename T> T fun(T a)
				{
					typedef Vtl::Types::Traits<T>::UnConst MyType;
					MyType b=a;
					b+=5; //  b   
					      //,     . 
					return b; // T -  , 
					          //     .
				}
				const float f=3.2;
				const float f2=fun(f);
				@endcode
			*/
			typedef typename UnConst<T>::Result NonConstType;
			/// . Traits<T>::isConst 
			enum { isVolatile = UnVolatile<T>::isVolatile };
			/// . Traits<T>::UnConst 
			typedef typename UnVolatile<T>::Result NonVolatileType;
			enum { isQualifiedType = isVolatile||isConst };
			///  T     const  volatile
			typedef typename UnVolatile<typename UnConst<T>::Result>::Result 
				UnqualifiedType;
			/// True,  T -     ,  int, signed char.
			enum {isStdSignedInt=TL::isIn<SignedIntegerTypes,T>::Result};
			/// True,  T -     ,  unsigned long, unsigned char.
			enum {isStdUnsignedInt=TL::isIn<UnsignedIntegerTypes,T>::Result};
			/// True,  T -    ,  int, unsigned long, unsigned char. 
			enum {isStdInteger=isStdUnsignedInt||isStdSignedInt};
			/// True,  T -      ,  float, long double.
			enum {isStdReal=TL::isIn<RealTypes,T>::Result};
			/// True,  T -         ,  float, short.
			enum {isStdArith=isStdInteger||isStdReal};
			/// @brief True,  T -     ,    , 
			///  .  char,wchar_t,bool.
			enum {isStdOtherInt=TL::isIn<OtherIntegerTypes,T>::Result};
			/// True,  T        isStdOtherInt.
			enum {isStdIntegerAll=isStdInteger||isStdOtherInt};
			/// True,  T     : isStdInteger isStdReal isStdOtherInt
			enum {isStdArithAll=isStdArith||isStdOtherInt};
			///   ,   isStdArithAll   void.
			enum {isStdFundamental=isStdArithAll||Conversion<T,void>::sameType};
		};
		///    
		struct IntOrReal 
		{
			enum Types
			{
				Int,
				Real,
				Unknown=-1
			};
		};
		///    
		struct SUIntOrReal
		{
			enum Types
			{
				SInt,
				UInt,
				Real,
				Unknown=-1
			};
		};
		template<typename T>
		class PixelTraits
		{
			enum { group=isSUIntOrReal<T>::Result };
/*
			typedef typename Select<Traits<T>::isStdSignedInt,
				SignedIntegerTypes,
				typename Select<Traits<T>::isStdUnsignedInt,
					UnsignedIntegerTypes,
					typename Select<Traits<T>::isStdReal,
						RealTypes,
						NullType>::Type >::Type >::Type TypesGroup;
*/
			template<typename P,unsigned types_group> struct LargerTypeSelector;
			//  
			template<typename P> struct LargerTypeSelector<P,SUIntOrReal::SInt>
			{
				typedef SignedIntegerTypes TypesGroup;
			};
			//  
			template<typename P> struct LargerTypeSelector<P,SUIntOrReal::UInt>
			{
				typedef UnsignedIntegerTypes TypesGroup;
			};
			//  
			template<typename P> struct LargerTypeSelector<P,SUIntOrReal::Real>
			{
				typedef RealTypes TypesGroup;
			};
		public:
			typedef typename LargerTypeSelector<T,group>::TypesGroup TypesGroup;
			typedef T PixelType;
			typedef typename Vtl::Types::TL::FilteredTypes<TypesGroup,sizeof(PixelType)*2,Vtl::Types::TL::GE>::Result listForSqr;
			typedef typename Vtl::Types::TL::SmallestType<listForSqr,Vtl::Types::TL::LT>::Result SqrType;
		};

/** \example HowTo_MakeDifferentRealizationForIntegersAnsFloats.hxx
 * ,       ,   
 *      .   
 *    ,  -     
 *     -  .
 */
		///         T  Vtl::Types::IntOrReal
		template<typename T> struct isIntOrReal
		{
			enum{ Result=Traits<T>::isStdInteger
					?IntOrReal::Int
					:(Traits<T>::isStdReal
						?IntOrReal::Real
					    :IntOrReal::Unknown)
				}; 
		};
		///         T  Vtl::Types::SUIntOrReal
		template<typename T> struct isSUIntOrReal
		{
			enum{ Result=Traits<T>::isStdSignedInt
					?SUIntOrReal::SInt
					:(Traits<T>::isStdUnsignedInt
						?SUIntOrReal::UInt
					    :(Traits<T>::isStdReal
							?SUIntOrReal::Real
							:SUIntOrReal::Unknown))
				}; 
		};
/*

		template<typename T,int X> struct _IntXbytesLarger
		{
			enum {size=sizeof(T)};
			enum {size_X=size+X};

		};
		template<typename T> struct Sqr;
		template<typename T> 
		struct Sqr
		{
			typedef NullType Type;
		};
*/
	};

	namespace Alg
	{

		/*!
		*	@struct Vtl::Alg::Factorial
		*	@brief     .
		*
		*      .
		*   @code
		*	int n=Vtl::Alg::Factorial<12>::value;
		*	std::cout << "12!="<< n << std::endl;
		*	@endcode
		*	:
		*	@code
		*	template<unsigned N> class X 
		*	{
		*		//class realization
		*	public:
		*		//class realization
		*		unsigned v() const {return N;}
		*	};
		*	//class using:
		*	X<Vtl::Alg::Factorial<5>::value> x; //   N=5!=120.
		*   int result=x.v(); // result  120.
		*	@endcode
		*
		*   \b !         
		*    , ,        unsigned (enum).
		**************************************************************/
		template<unsigned N> struct Factorial
		{
			enum {value = N*Factorial<N-1>::value};
		};
		template<> struct Factorial<0>
		{
			enum {value = 1};
		};
//		template<unsigned N,typename T> T n_pow(T x,Vtl::Types::Int2Type<N> );
//		template<typename T> T n_pow(T x,Vtl::Types::Int2Type<0> ) { return 1; }
		template<typename T> T n_pow(T x,Vtl::Types::Int2Type<1> ) { return x; }
		template<typename T,unsigned N> T n_pow(T x,Vtl::Types::Int2Type<N> ) { return x*n_pow(x,Vtl::Types::Int2Type<N-1>()); }
		/*!
		*	@struct Vtl::Alg::isOdd
		*	@brief        .  
		*   ,  Vtl::Alg::isOdd::value  1,  0.
		*
		**************************************************************/
		template<unsigned N> struct isOdd
		{
			enum {value = N%2 };
		};
		/*!
		*	@brief minus_if_odd<N>(x)  x,  N   -x,  N  . 
		*
		**************************************************************/
		template<unsigned N,typename T> T minus_if_odd(T x)
		{
			return Impl::minus_if_odd(x,Loki::Int2Type<isOdd<N>::value>());
		}
		/*!
		*	@brief mult_if_odd<N>(x,mult)  x,  N   x*mult,  N  . 
		*
		**************************************************************/
		template<unsigned N,typename T,typename TT> T mult_if_odd(T x,TT mult)
		{
			return mult_if<bool(isOdd<N>::value)>(x,mult);
		}
		/*!
		*	@brief mult_if<do_mult>(x,mult)  x,  do_mult=false  x*mult,  do_mult=true. 
		*
		**************************************************************/
		template<bool do_mult,typename T,typename TT> T mult_if(T x,TT mult)
		{
			return Impl::mult_if(x,mult,Loki::Int2Type<do_mult>());
		}
		namespace Impl
		{
			template<typename T> T minus_if_odd(T x,Loki::Int2Type<isOdd<0>::value> )
			{
				return x;
			}
			template<typename T> T minus_if_odd(T x,Loki::Int2Type<isOdd<1>::value> )
			{
				return -x;
			}
			template<typename T,typename TT> T mult_if(T x,TT mult,Loki::Int2Type<false> )
			{
				return x;
			}
			template<typename T,typename TT> T mult_if(T x,TT mult,Loki::Int2Type<true> )
			{
				return x*mult;
			}
			template<typename T> T sincos(T x,Loki::Int2Type<isOdd<0>::value> )
			{
				return std::sin(x);
			}
			template<typename T> T sincos(T x,Loki::Int2Type<isOdd<1>::value> )
			{
				return std::cos(x);
			}
		}
		/*!
		 *	\brief sincos<N>(x)  sin(x),  N-,  cos(x),  N-.
		 *        N-   sin(x).
		 * \see dsin()
		 *   
		 *
		**************************************************************/
		template<unsigned N,typename T> T sincos(T x)
		{
			return Impl::sincos(x,Loki::Int2Type<isOdd<(N&1)>::value>() );
		}
		/*!
		 *	\brief dsin<N>(x)   N-   sin(x)  . 
		 *     \f[ dsin<n>(x)= { d^n sin \over d x^n}(x) = (-1)^{ [ n/2 ] } \left\lbrace   \begin{array}{cc}
		 *		sin( x ) , \qquad if \quad n \quad is \quad even \\ 
		 *		cos( x ) , \qquad if \quad n \quad is \quad odd
		 *	\end{array} \right. \f] 
		 *  \f$ { [ ... ] } \f$         . 
		 *   
		 *
		**************************************************************/
		template<unsigned N,typename T> T dsin(T x)
		{
			return minus_if_odd<N/2>( sincos<N>(x) );
		}
		/*!
		*	@struct Vtl::Alg::x_to_N
		*	@brief  x^N, .. pow(x,N). N -     .
		*
		**************************************************************/
		template<unsigned N> struct x_to_N
		{
			template<typename T> static T value(T x)
			{
				return x_to_N__<N,isOdd<N>::value >::v(x);
			}
		private:
			template<unsigned K,unsigned M> struct x_to_N__;
			template<unsigned K> struct x_to_N__<K,0> //N is Even
			{
				template<typename T> static T sqr(T x) {return x*x;}
				template<typename T> static T v(T x)
				{
					return sqr(x_to_N__<K/2,isOdd<K/2>::value>::v(x));
				}
			};
			template<unsigned K> struct x_to_N__<K,1> //N is Odd
			{
				template<typename T> static T sqr(T x) {return x*x;}
				template<typename T> static T v(T x)
				{
					return x*x_to_N__<K-1,isOdd<K-1>::value>::v(x);
				}
			};
			template<> struct x_to_N__<1,1> //N is Odd
			{
				template<typename T> static T v(T x)
				{
					return x;
				}
			};
			template<> struct x_to_N__<0,0> //N is Odd
			{
				template<typename T> static T v(T x)
				{
					return 1;
				}
			};
		};
		/*!
		 *
		 ******************************************************************/
		template<typename T>  T Return1stArg(T arg1,T arg2,Loki::Int2Type<true>)  { return arg1;} 
		template<typename T>  T Return1stArg(T arg1,T arg2,Loki::Int2Type<false>) { return arg2;} 
		///
		/// \brief    first  true -   ,   - .
		///    .
		///
		template<bool first,typename T>  T Return1stArg(T arg1,T arg2) { return Return1stArg(arg1,arg2,Loki::Int2Type<first>());}

		///
		/// \brief     n.    
		/// .     0 ().
		///
		template<typename T>  T ReturnArg_(T arg1,T arg2,Loki::Int2Type<0>) { return arg1;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,Loki::Int2Type<1>) { return arg2;} 
		template<int n,typename T>  T ReturnArg(T arg1,T arg2) { return ReturnArg_(arg1,arg2,Loki::Int2Type<n>());}
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,Loki::Int2Type<0>) { return arg1;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,Loki::Int2Type<1>) { return arg2;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,Loki::Int2Type<2>) { return arg3;} 
		template<int n,typename T>  T ReturnArg(T arg1,T arg2,T arg3) { return ReturnArg_(arg1,arg2,arg3,Loki::Int2Type<n>());}
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,Loki::Int2Type<0>) { return arg1;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,Loki::Int2Type<1>) { return arg2;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,Loki::Int2Type<2>) { return arg3;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,Loki::Int2Type<3>) { return arg4;} 
		template<int n,typename T>  T ReturnArg(T arg1,T arg2,T arg3,T arg4) { return ReturnArg_(arg1,arg2,arg3,arg4,Loki::Int2Type<n>());}
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,T arg5,Loki::Int2Type<0>) { return arg1;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,T arg5,Loki::Int2Type<1>) { return arg2;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,T arg5,Loki::Int2Type<2>) { return arg3;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,T arg5,Loki::Int2Type<3>) { return arg4;} 
		template<typename T>  T ReturnArg_(T arg1,T arg2,T arg3,T arg4,T arg5,Loki::Int2Type<4>) { return arg4;} 
		template<int n,typename T>  T ReturnArg(T arg1,T arg2,T arg3,T arg4,T arg5) { return ReturnArg_(arg1,arg2,arg3,arg4,arg5,Loki::Int2Type<n>());}
	};

};

/////////////////////////////////////////////////////////////////////////
#endif //_ALEXANDRESKU_H

