#ifndef _DATASTRING_H
#define _DATASTRING_H
///////////////////////////////////////////////////////////////////////////////
//file:
// #include "datastring.h"
//
///////////////////////////////////////////////////////////////////////////////
#include <string>
#include <sstream>
#include <iomanip> 
#include "ytime.h"
/** @defgroup DataStringGroup       . 
 *  @ingroup GeneralGroup
 *  @{
 */
#include "loki_EmptyType.h"
/// \brief   .
namespace Gen
{
/// \brief       , -
namespace Data
{
	class SaveFormat
	{
		std::iostream& stream_;
		std::ios_base::fmtflags flags;
	public:
		SaveFormat(std::iostream& stream)
			: stream_(stream),flags(stream.flags())
		{
			
		}
		~SaveFormat()
		{
			stream_.flags(flags);
		}
	};
	namespace Impl
	{
		inline int Sscanf(std::istream& is
			,Loki::EmptyType v1,Loki::EmptyType v2,Loki::EmptyType v3,Loki::EmptyType v4,Loki::EmptyType v5
			,Loki::EmptyType v6,Loki::EmptyType v7,Loki::EmptyType v8,Loki::EmptyType v9,Loki::EmptyType v10
			,Loki::EmptyType v11,Loki::EmptyType v12,Loki::EmptyType v13,Loki::EmptyType v14,Loki::EmptyType v15
			,Loki::EmptyType v16
			)
		{
			return 0;
		}
		template<typename T1,typename T2,typename T3,typename T4,typename T5
			,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14,typename T15,typename T16
		>
		int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5
			,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10
			,T11 & v11,T12 & v12,T13 & v13,T14 & v14,T15 & v15,T16 & v16
			)
		{
			try
			{
				is >> v1;
			}
			catch(...)
			{
				return 0;
			}
			if(is.fail())
				return 0;
			else
				return 1+Sscanf(is,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,v11,v12,v13,v14,v15
							   ,v16
							   ,Loki::EmptyType());
		}
	}
	/// \brief   - STL (std::istream, std::isringstream, ...)   
	///  ::sscanf().   . type-safe.     .
	/// :
	/// \code
	///		float v1;
	///		int v2;
	///		int v3;
	///		double v4;
	///		int n=Gen::Data::Sscanf("12.34 13",v1);
	///		std::cout << n << "  " << v1 << "  " << v2 << "  " << v3 << "  " << v4 <<std::endl;
	///   ///  n=1.     v1=12.34
	///		n=Gen::Data::Sscanf("12.34 13",v1,v2);
	///		std::cout << n << "  " << v1 << "  " << v2 << "  " << v3 << "  " << v4 <<std::endl;
	///   ///  n=2.       v1=12.34 v2=13
	///		n=Gen::Data::Sscanf("12.34 13 33",v1,v2,v3,v4);
	///		std::cout << n << "  " << v1 << "  " << v2 << "  " << v3 << "  " << v4 <<std::endl;
	///   ///  n=3.   ,     v1=12.34 v2=13 v3=33.      .
	///		n=Gen::Data::Sscanf("12.34 13 33 4.0e-14",v1,v2,v3,v4);
	///		std::cout << n << "  " << v1 << "  " << v2 << "  " << v3 << "  " << v4 <<std::endl;
	///   ///  n=4.   ,   4 .
	/// \endcode
	///
	///     ,    operator>>(std::istream, Tx& ),   ,
	///     (throw) ,   ,         
	///  ,       .   ,   
	///  ,         is.fail().
	///
	template<typename T1>
	int Sscanf(std::istream& is,T1 & v1)
	{
		return Impl::Sscanf(is,v1,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
			                   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2)
	{
		return Impl::Sscanf(is,v1,v2,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
			                   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3)
	{
		return Impl::Sscanf(is,v1,v2,v3,Loki::EmptyType(),Loki::EmptyType()
			                   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,Loki::EmptyType()
			                   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,Loki::EmptyType()
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,v11,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,v11,v12,Loki::EmptyType(),Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,v11,v12,v13,Loki::EmptyType(),Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13,T14 & v14)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,v11,v12,v13,v14,Loki::EmptyType()
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14,typename T15>
	int Sscanf(std::istream& is,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13,T14 & v14,T15 & v15)
	{
		return Impl::Sscanf(is,v1,v2,v3,v4,v5
			                   ,v6,v7,v8,v9,v10
							   ,v11,v12,v13,v14,v15
							   ,Loki::EmptyType()
							   );
	}
	template<typename T1>
	int Sscanf(const char* str,T1 & v1)
	{
		std::istringstream is(str);
		return Sscanf(is,v1);
	}
	template<typename T1>
	int Sscanf(std::string str,T1 & v1)
	{
		std::istringstream is(str);
		return Sscanf(is,v1);
	}
	template<typename T1,typename T2>
	int Sscanf(const char* str,T1 & v1,T2 & v2)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2);
	}
	template<typename T1,typename T2>
	int Sscanf(std::string str,T1 & v1,T2 & v2)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2);
	}
	template<typename T1,typename T2,typename T3>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3);
	}
	template<typename T1,typename T2,typename T3>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3);
	}
	template<typename T1,typename T2,typename T3,typename T4>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4);
	}
	template<typename T1,typename T2,typename T3,typename T4>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13,T14 & v14)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13,T14 & v14)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14,typename T15>
	int Sscanf(const char* str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13,T14 & v14,T15 & v15)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14,v15);
	}
	template<typename T1,typename T2,typename T3,typename T4,typename T5,typename T6,typename T7,typename T8,typename T9,typename T10,typename T11,typename T12,typename T13,typename T14,typename T15>
	int Sscanf(std::string str,T1 & v1,T2 & v2,T3 & v3,T4 & v4,T5 & v5,T6 & v6,T7 & v7,T8 & v8,T9 & v9,T10 & v10,T11 & v11,T12 & v12,T13 & v13,T14 & v14,T15 & v15)
	{
		std::istringstream is(str);
		return Sscanf(is,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14,v15);
	}
}//namespace Data
}//namespace Gen
///////////////////////////////////////////////////////////////////////////////
/*!
 *    STL         ,
 *    ,   ,  DataString   
 * -,    .       
 *     std::string.
 *       std::string    
 *     .
 */
class DataString
:public std::string
{
	int sec_digits;
	template<typename T> DataString& Set(T val)
	{
		sec_digits=3;
		std::ostringstream os;
		os << val;
		*this=os.str();
		return *this;
	}
	DataString& Set(double val)
	{
		sec_digits=3;
		std::ostringstream os;
		os << std::setprecision(18) << val;
		*this=os.str();
		return *this;
	}
/*!
 * International standard ISO/IEC 14882, first edition 1998-09-01, Programming Languages - C++
 * P.255. 14.5.5.1 Function template overloading.
 * "The signature of a function template consists of its function signature, its return type and 
 *  its template parameter list. The names of template parameters are significant only for 
 *  establishing the relationship between the template parameters and the rest of the signature. 
 *  [Note: two distinct function templates may have identical functions return types and function 
 *  parameter lists, even if overload resolution alone cannot distinguish them.
 *
 *  template<class T> void f();
 *  template<int I> void f();  //OK: overloads the first template
 *                             //distinguishable with an explicit template argument list 
 * --end note]"
 */
	template<typename T> T Get() const
	{
		std::istringstream is(*this);
		T val;
		is >> val;
		if(is.fail())
			throw std::string("DataString::Get<T>() : value reading from string is FAIL");
		return val;
	}
	static std::string DataString::TimeToString(YDataTime val,YTimeInterval zone,int sec_digits=3);
	static void DataString::StringToTime(std::string str,YDataTime& val,YTimeInterval& zone);
public:
	DataString() {}
	DataString(const char* val) : std::string(val) {}
	DataString(const std::string val) : std::string(val) {}

	DataString(int val) { Set(val); }
	DataString& operator=(int val) { return Set(val);}
	operator int() const { return Get<int>();}

	DataString(unsigned val) { Set(val); }
	DataString& operator=(unsigned val) { return Set(val);}
	operator unsigned() const { return Get<unsigned>();}

	DataString(float val) { Set(val); }
	DataString& operator=(float val) { return Set(val);}
	operator float() const { return Get<float>();}

	DataString(double val) { Set(val); }
	DataString& operator=(double val) { return Set(val);}
	operator double() const { return Get<double>();}

	DataString(YDataTime val) { *this=val; }
//	DataString& operator=(YDataTime val); 
	DataString& operator=(YDataTime val)
	{
		YTimeInterval zone(0);
		*this=DataString::TimeToString(val,zone,sec_digits);
		return *this;
	}
//	operator YDataTime() const;
	operator YDataTime() const
	{
		YDataTime dt;
		YTimeInterval zone;
		DataString::StringToTime(*this,dt,zone);
		dt-=zone;//convert to grinvich
		return dt;
	}

	DataString(YDataTimeZone val) { *this=val; }
	DataString& operator=(YDataTimeZone val)
	{
		YTimeInterval zone(0);
		*this=DataString::TimeToString(val.DataTime(),val.Zone(),sec_digits);
		return *this;
	}
	operator YDataTimeZone() const
	{
		YDataTime dt;
		YTimeInterval zone;
		DataString::StringToTime(*this,dt,zone);
		return YDataTimeZone(dt,zone);
	}
	void setDataTime_Precision(int digits)
	{
		if(digits<1 || digits>6)
			digits=3;
		sec_digits=digits;
	}


/*
//	friend std::ostream& operator<<(std::ostream& s,const DataString& ds) { return s<<ds.c_str();}
	template<typename T> std::ostream operator<<(T val)
	{
		std::ostringstream os(*this);
		os << val;
		return os;
	}
*/
};
/** @} */ // end of DataStringGroup
///////////////////////////////////////////////////////////////////////////////
#endif //_DATASTRING_H
