#ifndef _YMEM_H
#define _YMEM_H
///////////////////////////////////////////////////////////////////
#include <assert.h>
#include "ytypes.h"
/** @defgroup MemoryGroup    . 
 *  @ingroup GeneralGroup
 *  @{
 */
//------------------------------------------------------------------------------
// This class used to guarantee dynamically created object will destroyed when
// program leaves current block {}.
//------------------------------------------------------------------------------
template<class T> 
class PtrDelete
{
	T* ptr;
	int len;
public:
	PtrDelete() { ptr=0;len=0;}
	PtrDelete(T* p,int n=1) { ptr=p;len=n;}
	virtual ~PtrDelete() { clear();}
	void clear() {if(ptr) if(len>1) delete[] ptr; else delete ptr; ptr=0;}
	void Set(T*p,int n=1) {clear();ptr=p;len=n;}
	T* Make(int n) { if(n!=len ) {clear(); if(n>0) { ptr=(n==1)?(new T):(new T[n]); len=n;} } return ptr; }
	T* Get() const {return ptr;}
	int Size() const {	return len;	}
};

template <class T> 
class PtrArrayDel
{
	T** ptr;
	int size;
public:
	PtrArrayDel() {	ptr = 0; size = 0; };
	virtual ~PtrArrayDel()	{clear();}
	void clear()
	{	if(ptr)
		{
			for ( int i = 0 ; i< size; i++) 
				if (ptr[i]) { delete ptr[i]; }
			delete[] ptr; ptr=0;
		}
	}
	T** Get()  {	return ptr;	};
	const T** Get() const { return ptr; };
//	T** Set(T** p, int n=1) { clear();ptr=p;size=n; return ptr;}// External pointers array will delete by destructor????
	int Size() const {	return size;	}
	void Make(int n)
	{
		assert(n >0 );
		clear();
		try
		{
			ptr = new T* [n];
			if (ptr == 0) throw 1;
		}
		catch(...)
		{
			throw "Bad Memory Allocation in PtrArrayDel::Init()";
		};
		size = n;
		for (int i = 0; i< size; i++) ptr[i]= 0;
	}

	T*& operator[](int i) 	{ assert( i == IN_BOUNDS(i, 0, size) );	return ( ptr[i] );}
	T* const operator[](int i) const { assert( i == IN_BOUNDS(i, 0, size) ); return ( ptr[i] );}
};

//------------------------------------------------------------------------------
// This class used when need allocate a lot number of small objects,
// maximal number of object is known in advance. It is fast and use a minimum 
// of spare memory if sizeof(object)>=sizeof(Byte*).
//------------------------------------------------------------------------------
class YSmallAllocLtd
{
	typedef uint8_t * BytePtr;
	uint8_t *data;
	uint8_t *f;
	unsigned objsz;
	unsigned num;
	unsigned count;
	unsigned memsz;
private:
	void nullify();
public:
	YSmallAllocLtd();
	YSmallAllocLtd(unsigned size_of_object,unsigned maximal_number_of_objects,unsigned max_tail=0);
	virtual ~YSmallAllocLtd();
	void Init(unsigned size_of_object,unsigned maximal_number_of_objects,unsigned max_tail=0);
	void DeInit();
	void * malloc();
	void free(void *ptr);
	void erase_all();
	unsigned NumberUsed() const { return count;}
	unsigned NumberFree() const { return num-count;}
	unsigned NumberMax() const { return num;}
	unsigned SizeOfObject() const { return objsz;}
	bool IsOk() const { return data!=0; }
};
template<class T>
class YTArray
{
	int block_size;
	int step;
	int size;
protected:
	T* data;
private:
	int calc_blk_size(int sz) const { return sz<=0?0:((MAX(sz,1)+step-1)/step)*step;}
	//minimal block size is T[grow_step]
public:
	YTArray(int grow_step=128) {SetGrowStep(grow_step); data=0;block_size=0; size=0;}
	// cycle in reverse order to prevent realloc;
	YTArray(const YTArray& ar) {SetGrowStep(ar.step); data=0;block_size=0; size=0; SetSize(ar.size); for(int i=ar.size-1;i>=0; i--) (*this)[i]=ar[i]; }
	YTArray& operator=(const YTArray& ar) {SetGrowStep(ar.step);  SetSize(ar.size); for(int i=0; i<ar.size;i++) (*this)[i]=ar[i]; return *this;}
	virtual ~YTArray() { if(data) delete[] data; }
	void SetGrowStep(int grow_step=128) {step=MAX(1,grow_step);}
	int  GetGrowStep() const {return step;}
	int  GetBlockSize() const {return block_size;}
	void SetSize(int sz) 
		{ int bs=calc_blk_size(sz); 
			if(bs!=block_size)
			{
				T* d=0;
				if(bs>0)
				{
					d=new T[bs]; 
					if(!d)
						throw "err";
					if(data) 
					{ 
						int k=MIN(sz,size); 
						while(k--) d[k]=data[k];
					}
				}
				if(data)
					delete[] data;
				data=d; 
				block_size=bs;
			}
			size=sz;
		}
	int Size() const { return size;}
	T* Get_Ptr() { return data;}
	const T* Get_Ptr() const { return data;}
	virtual T& operator[](int i) { return data[i]; }
	virtual const T& operator[](int i) const { return data[i]; }
	virtual T& Insert(const T& elem,int pos)
		{ for(int i=size-1; i>pos; i--) data[i]=data[i-1]; data[pos]=elem; return data[pos];}
	virtual void Remove(int pos) {for(int i=pos+1; i<size; i++) data[i-1]=data[i]; }
	virtual void SetAll(const T * values,int num)
	{
		SetSize(num);
		for(int i=0; i<num; i++)
			data[i]=values[i];
	}
};
template<class T>
class YTDynArray
: public YTArray<T>
{
public:
	YTDynArray(int grow_step=100) :YTArray<T>(grow_step) {}
	YTDynArray(const YTDynArray& ar) :YTArray<T>(ar) {}
	virtual ~YTDynArray() {}
	YTDynArray& operator=(const YTArray<T>& ar) {SetGrowStep(ar.step); SetSize(ar.size); for(int i=0; i<ar.size;i++) (*this)[i]=ar[i]; return *this; }
	virtual T& operator[](int i) {if(i<0) throw "err"; if(i>=YTArray<T>::Size()) YTArray<T>::SetSize(i+1); return YTArray<T>::data[i]; }
	virtual const T& operator[](int i) const {if(i<0||i>=YTArray<T>::Size()) throw "err";  return YTArray<T>::data[i]; }
	virtual T& Append(const T& elem)
		{ 
			(*this)[YTArray<T>::Size()]=elem; 
			return (*this)[YTArray<T>::Size()-1];
		}
	virtual T& Insert(const T& elem,int pos)
	{	YTArray<T>::SetSize(YTArray<T>::Size()+1); 	return YTArray<T>::Insert(elem,pos);	}
	virtual void Remove(int pos) {YTArray<T>::Remove(pos); YTArray<T>::SetSize(YTArray<T>::Size()-1);}
};

////////////////////////////////////////////////////////////////////////
// class YTHistList 
// This class used for 1-D histogram build and store.
// No space reserved for data values which not happened.
//
////////////////////////////////////////////////////////////////////////
template <class T>
class YTHistListTag
{
public:
	//ATTANTION, function YTHistList<T>::Sort() 
	//   (via static int __cdecl compare_HstTag(elem1,elem2) ) 
	// suppose, that 'count' is FIRST MEMBER VARIABLE in class YTHistList<T>
	unsigned count;
	T value;
public:
	YTHistListTag() { count=0; }
	virtual ~YTHistListTag() {count=0;}
	bool operator==(const YTHistListTag& t2) const{ return count==t2.count;} 
	bool operator<(const YTHistListTag& t2) const { return count>=t2.count;}//sort in descendant 
};

template <class T>
class YTHistList 
{
public:
	typedef YTHistListTag<T> Tag;
	YTDynArray<Tag> array;
public:
	void Add(const T& val);
	int Count() const {return array.Size();}
	void Sort() {}
	Tag& operator[](int i) { return array[i]; }
};

template <class T>
void YTHistList<T>::Add(const T& val)
{
	const int sz=array.Size();
	for(int i=0; i<sz; i++)
	{
		Tag t=array[i];
		if(t.value==val)
		{
			array[i].count++;
			t.count++;
			// find position;
			int j=i-1;
			if(j>=0 && array[j].count< t.count)
			{ //previous historgamm entry has smaller count, move to top
				//test if a lot of previous 
				int k=j;
				while(j)
				{
					j--;
					if(t.count<array[j].count)
					{
						k=j+1;
						break;
					}
				}
				array[i]=array[k];
				array[k]=t;
			}
			return;
		}
	}
	Tag x;
	x.count=1;
	x.value=val;
	array[sz]=x;
}

/*
static int __cdecl compare_HstTag(const void *elem1, const void *elem2 ) 
{
	//ATTANTION, this functions suppose, that count is FIRST MEMBER VARIABLE in class YTHistList<T>
	typedef YTHistList<unsigned>::Tag TT;
	return ((TT*)(elem2))->count-((TT*)(elem1))->count;
}
template <class T>
void YTHistList<T>::Sort() 
{
	::qsort(array.Get_Ptr(),array.Size(),sizeof(Tag),compare_HstTag);
}
*/
/** @} */ // end of MemoryGroup
///////////////////////////////////////////////////////////////////
#endif //_YMEM_H
