#include "ytime.h"

const static int MONTH_IN_YEAR=12;
const static int DAYS_IN_YEAR=365;
const static int DAYS_IN_YEAR_LEAP=366;
const static int SEC_IN_HOUR=60*60;
const static int SEC_IN_DAY=24*60*60;
const static int DAY_IN_MONTH[MONTH_IN_YEAR]     ={31,28,31,30, 31, 30, 31, 31, 30, 31, 30, 31};
const static int DAY_IN_MONTH_LEAP[MONTH_IN_YEAR]={31,28,31,30, 31, 30, 31, 31, 30, 31, 30, 31};
const static int DAY_TO_MONTH[MONTH_IN_YEAR]     ={ 0,31,59,90,120,151,181,212,243,273,304,334};
const static int DAY_TO_MONTH_LEAP[MONTH_IN_YEAR]={ 0,31,60,91,121,152,182,213,244,274,305,335};
const static int32_t SEC_TO_MKS=1000000;
const static int32_t SEC_IN_YEAR     =DAYS_IN_YEAR     *SEC_IN_DAY;
const static int32_t SEC_IN_YEAR_LEAP=DAYS_IN_YEAR_LEAP*SEC_IN_DAY;
const static int64_t MKS_IN_YEAR     =int64_t(SEC_IN_YEAR)     *int64_t(SEC_TO_MKS);
const static int64_t MKS_IN_YEAR_LEAP=int64_t(SEC_IN_YEAR_LEAP)*int64_t(SEC_TO_MKS);


int YDtTm::MonthInYear()
{
	return MONTH_IN_YEAR;
}
bool YDtTm::IsLeapYear(int _year)
{
	return _year%4==0 && (_year%100 != 0 || _year%400 == 0);
}
int YDtTm::DaysInYear(int _year)
{
	return YDtTm::IsLeapYear(_year)?DAYS_IN_YEAR_LEAP:DAYS_IN_YEAR;
}
int32_t YDtTm::SecondsInYear(int _year)
{
	return YDtTm::IsLeapYear(_year)?SEC_IN_YEAR_LEAP:SEC_IN_YEAR;
}
int64_t YDtTm::MicroSecondsInYear(int _year)
{
	return YDtTm::IsLeapYear(_year)?MKS_IN_YEAR_LEAP:MKS_IN_YEAR;
}

int YDtTm::DaysInMonth(int _year,int _month)
{
	const int *day_in_month=IsLeapYear(_year)?DAY_IN_MONTH_LEAP:DAY_IN_MONTH;
	return day_in_month[_month-1];
}

int YDtTm::YearDay(int _year,int _month,int _day)
{
	const int *full_month=IsLeapYear(_year)?DAY_TO_MONTH_LEAP:DAY_TO_MONTH;
  int yday=full_month[(_month-1)]+_day;
  return yday;
}
int YDtTm::YearDayMonth(int _year_day,int _year)
{
	const int *full_month=IsLeapYear(_year)?DAY_TO_MONTH_LEAP:DAY_TO_MONTH;
	for(int i=1; i<MONTH_IN_YEAR; i++)
	{
		if(_year_day<=full_month[i])
			return i; //because month numbered from one!
	}
  return MONTH_IN_YEAR;
}
int YDtTm::YearDayDay(int _year_day,int _year)
{
	int month=YDtTm::YearDayMonth(_year_day,_year);
	int d=YDtTm::YearDay(_year,month,1);
	return _year_day-d+1;
}

void YDtTm::BCDdate2date(int _bcd,int& _year,int& _month,int& _day)
{// bcd like 19991231
	_year=_bcd/10000;
	_bcd-=_year*10000;
	_month=_bcd/100;
	_bcd-=_month*100;
	_day=_bcd;
	if(_month<1 || _month>MONTH_IN_YEAR || _day<1 || _day > YDtTm::DaysInMonth(_year,_month) )
		throw "Bad BCD date";
}
int YDtTm::BCDdate2date(int _bcd,int& _year_day)
{// returns year. bcd like 19991231
	int year,month,day;
	YDtTm::BCDdate2date(_bcd,year,month,day);
	_year_day=YDtTm::YearDay(year,month,day);
	return year;
}
int YDtTm::date2BCDdate(int _year,int _month,int _day)
{// returns bcd date like 19991231
	return (_year*100+_month)*100+_day;
}
int YDtTm::date2BCDdate(int _year,int _year_day)
{// returns bcd date like 19991231
	int month=YDtTm::YearDayMonth(_year_day,_year);
	int day  =YDtTm::YearDayDay(_year_day,_year);
	return YDtTm::date2BCDdate(_year,month,day);
}


//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////
void YDataTime::Set(int _year,int _month,int _day,int _hour,int _minutes,double _seconds)
{
	year=_year;     //sign value, negative is before Crist, 0-year is forbidden,nessesary for leap year monitoring.
	// time in microseconds from NewYear
	microsec=YDtTm::YearDay(_year,_month,_day); //in days
	microsec=microsec*24+_hour;                 //in hours
	microsec=microsec*60+_minutes;              //in minutes
	microsec=microsec*60*SEC_TO_MKS+int64_t(_seconds*SEC_TO_MKS+0.5);// in microseconds
}
void YDataTime::Get(int& _year,int& _month,int& _day,int& _hour,int& _minutes,double& _seconds)
const
{
	_year=year;     //sign value, negative is before Crist, 0-year is forbidden,nessesary for leap year monitoring.
	// time in microseconds from NewYear
	int64_t mks=microsec;
	_seconds=double(int32_t(mks%SEC_TO_MKS))/SEC_TO_MKS;
	mks/=SEC_TO_MKS;
	_seconds+=int32_t(mks%60);
	mks/=60;
	_minutes=int(mks%60);
	mks/=60;
	_hour=int(mks%24);
	mks/=24;
	int yday=int(mks);
	if(yday>YDtTm::DaysInYear(_year))
		throw "YDataTime::Get() : bad days in year";
	_month=YDtTm::YearDayMonth(yday,_year);
	_day=YDtTm::YearDayDay(yday,_year);
}

YTimeInterval YDataTime::operator-(const YDataTime& dt) 
const
{
	int64_t interval=0;
	if(year==dt.year)
		interval=microsec-dt.microsec;
	else
	{
		const YDataTime* t1=this;
		const YDataTime* t2=&dt;
		bool sign=true;
		if(year<dt.year)
		{
			const YDataTime* tmp=t1;
			t1=t2;
			t2=tmp;
			sign=false;
		}
		interval=0;
		int y=t2->year;
		while(y < t1->year)
		{
			interval+=YDtTm::DaysInYear(y);
			y++;
		}
		interval*=SEC_IN_DAY;
		interval*=SEC_TO_MKS;
		interval-=t2->microsec;
		interval+=t1->microsec;
		if(!sign)
			interval=-interval;
	}
	return YTimeInterval(interval);
}

YDataTime& YDataTime::operator+=(const YTimeInterval& ti)
{
	microsec+=ti.microsec;
	if(microsec>0)
	{
		int64_t mks=YDtTm::MicroSecondsInYear(year);
		while(microsec > mks)
		{
			year++;
			microsec -= mks;
			mks=YDtTm::MicroSecondsInYear(year);
		}
	}
	else if(microsec<0)
	{
		do
		{
			year--;
			microsec+=YDtTm::MicroSecondsInYear(year);
		} while(microsec<0);
	}
	return *this;
}
bool YDataTime::operator<(const YDataTime& dt2)
const
{
	if(year<dt2.year)
		return true;
	else if(year>dt2.year)
		return false;
	else 
		return (microsec<dt2.microsec);
}

bool YDataTime::operator<=(const YDataTime& dt2)
const
{
	if(year<dt2.year)
		return true;
	else if(year>dt2.year)
		return false;
	else 
		return (microsec<=dt2.microsec);
}

bool YDataTime::operator==(const YDataTime& dt2)
const
{
	return ( (year==dt2.year) && (microsec==dt2.microsec) );
}


void YDataTime::Between(const YDataTime& dt1,const YDataTime& dt2,int cnt1,int cnt2,int cnt)
{
	(*this)=dt1;
	if(cnt1!=cnt2)
	{
		YTimeInterval ti=dt2-dt1;
		ti*=cnt-cnt1;
		ti/=cnt2-cnt1;
		(*this)+=ti;
	}
}

bool YDataTime::IsLeapYear() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::IsLeapYear(_year);
}
//static 
int YDataTime::MonthInYear()
{
	return YDtTm::MonthInYear();
}
int  YDataTime::DaysInMonth() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::DaysInMonth(_year,_month);
}
int  YDataTime::DaysInYear() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::DaysInYear(_year);
}
int32_t YDataTime::SecondsInYear() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::SecondsInYear(_year);
}
int64_t YDataTime::MicroSecondsInYear() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::MicroSecondsInYear(_year);
}
int  YDataTime::Year() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return _year;
}
int  YDataTime::Month() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return _month;
}
int  YDataTime::Day() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return _day;
}
int  YDataTime::Hour() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return _hour;
}
int  YDataTime::Minutes() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return _minutes;
}
double  YDataTime::Seconds() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return _seconds;
}
int  YDataTime::YearDay() 
const
{
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::YearDay(_year,_month,_day);
}
int  YDataTime::date2BCDdate() 
const
{// returns bcd date like 19991231
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	return YDtTm::date2BCDdate(_year,_month,_day);
}
double  YDataTime::YearDaySec() 
const
{ //returns yearday with fractional accuracy up to microseconds
	int _year,_month,_day,_hour,_minutes;
	double _seconds;
	YDataTime::Get(_year,_month,_day,_hour,_minutes,_seconds);
	double yd=YDtTm::YearDay(_year,_month,_day);
	yd+=(double(_hour)+(double(_minutes)+_seconds/60)/60)/24;
	return yd;
}

YTimeInterval& YTimeInterval::Set(int days,int hour,int minutes,double seconds)
{
	int ss=int(seconds);
	int mks=int((seconds-ss)*SEC_TO_MKS);
	microsec=(((days*24+hour)*60+minutes)*60+ss);
	microsec*=SEC_TO_MKS;
	microsec+=mks;
	return *this;
}
const YTimeInterval& YTimeInterval::Get(int& days,int& hour,int& minutes,double& seconds) 
const
{
	int64_t mks=microsec;
	seconds=double(int32_t(mks%SEC_TO_MKS))/SEC_TO_MKS;
	mks/=SEC_TO_MKS;
	seconds+=int32_t(mks%60);
	mks/=60;
	minutes=int(mks%60);
	mks/=60;
	hour=int(mks%24);
	mks/=24;
	days=int(mks);
	return *this;
}

void YTimeInterval::Get(double& seconds) 
const
{
	int64_t mks=microsec;
	seconds=double(int32_t(mks%SEC_TO_MKS))/SEC_TO_MKS;
	mks/=SEC_TO_MKS;
	seconds+=int32_t(mks);
}
