#include "bmp.h"
using namespace IFF;

static uint32_t _calcRowLen(uint16_t bits,uint16_t wid)
{
	uint32_t bytes_per_row=0;
  switch(bits)
	{
  case 1:
    bytes_per_row=(wid+7)/8;
    break;
  case 4:
    bytes_per_row=(wid+1)/2;
    break;
  case 8:
    bytes_per_row=wid;
    break;
  case 24:
    bytes_per_row=3*wid;
    break;
  default:
    throw "BMP supported only 1,4,8 and 24 bits per pixel";
  };
  bytes_per_row=((bytes_per_row+3)/4)*4;
	return bytes_per_row;
}

BmpRead::BmpRead(const char *filename)
: YFileRead(filename)
{
	_pal=0;
	pal_entries=0;
	try
	{
		YFileRead::read(0,&header,sizeof(header));
		if(header.bfType!=0x4d42)
			throw "not BMP file";

		YFileRead::read(&info,sizeof(info));

		// this function throw if unsupported BitsPerPixel
		bytes_per_row=_calcRowLen(info.biBitCount,(uint16_t)info.biWidth);

//		img = new BmpImage(io,(uint16_t)info.biBitCount,(int2)info.biWidth,(int2)info.biHeight,
//					bytes_per_row,header.bfOffBits,sizeof(header)+info.biSize);

		base_offset=header.bfOffBits;

		if(info.biBitCount==24)
		{
			img_class=BT_RGB;
			pal_entries=0;
			_pal=0;
		}
		else
		{
			if(info.biBitCount!=8)
				throw "temporary indexed images supported only 8 bits per pixel";
			if(info.biBitCount==1)
				img_class=BT_BILEVEL;
			else
				img_class=BT_INDEX;

			//Read palette
			//_pal=new BmpPalette(_io,bits,pal_offset);
			uint32_t pal_offset=sizeof(header)+info.biSize;
			pal_entries=0x1FF & (1<<info.biBitCount);
			_pal=new RgbQuad[pal_entries];
			if(_pal==0) throw "BmpRead: insufficient memory to allocate palette";
			YFileRead::read(pal_offset,_pal,sizeof(_pal[0])*pal_entries);
		}
	}
	catch(...)
	{
		if(_pal)
		{
			delete[] _pal;
			_pal=0;
			pal_entries=0;
		}
		_close();
		throw;
	};

}
//virtual
BmpRead::~BmpRead()
{
	if(_pal)
	{
		delete[] _pal;
		_pal=0;
		pal_entries=0;
	}
}

void BmpRead::read(int x, int y,Byte *data,int len)
{
	YFileRead::read(base_offset+(info.biHeight-1-y)*bytes_per_row+x,data,len);
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
IFF::BmpWrite::BmpWrite(const char *filename,int width,int height,IFF::BmpWrite::Type iclass,const RgbQuad *pal)
:YFileWrite(filename)
{
	pal_entries=0;
	pal_offset=0;
	try
	{
    //check param
		int bits=0;
		switch(iclass)
		{
		case BT_BILEVEL:
			bits=1;
			break;
		case BT_INDEX:
			bits=8;
			break;
		case BT_RGB:
			bits=24;
			break;
		default:
			throw "inadmissible image class";
		};
    bytes_per_row=_calcRowLen(bits,width);
    if(width<0 || height<0)
      throw "Invalid image size";
    //set headers
    header.bfType=0x4d42;// ( 'BM')
	  //uint32_t header.bfSize;  // see low    // file size in bytes (disk space)
	  header.bfReserved1=0;
	  header.bfReserved2=0;
	  header.bfOffBits=sizeof(header)+sizeof(info)+((bits==24)?0:(4*(1<<bits)));   // offset for raster data beginning
    info.biSize=sizeof(info);          // this structure size,=40. Use for found RgbQuad data.
	  info.biWidth=(uint32_t)width;
	  info.biHeight=(uint32_t)height;
	  info.biPlanes=1;           // always = 1
	  info.biBitCount=bits;      // 1,4,8 or 24.
    info.biCompression=BI_RGB; // BI_RGB (no compression), BI_RLE8 or BI_RLE4.
	  info.biSizeImage=bytes_per_row*height;    // size of image in bytes, may be 0 for BI_RGB.
	  info.biXPelsPerMeter=0;     // pixels per meter
	  info.biYPelsPerMeter=0;
	  info.biClrUsed=0;  // 0 - used all colors
	  info.biClrImportant=0; // n - first n colors (index) important, 0 -all.
	  header.bfSize=header.bfOffBits+info.biSizeImage;      // file size in bytes (disk space)
		//
		pal_entries=0x1FF & (1<<info.biBitCount);
		pal_offset=sizeof(header)+info.biSize;
		base_offset=header.bfOffBits;
		//
		YFileWrite::write(0,&header,sizeof(header));
    YFileWrite::write(&info,sizeof(info));
		if(pal && pal_entries>0 )
		{
			YFileWrite::write(pal_offset,pal,sizeof(pal[0])*pal_entries);
		}
    if(1!=YFileWrite::write(header.bfSize-1,&header.bfReserved1,1)) // write last byte of file for size will be correct.
		{
			throw "can't write file, probably insufficient disk space";
		}
	}
	catch(...)
	{
		_close();
		throw;
	};
}
//virtual
BmpWrite::~BmpWrite()
{
}

void BmpWrite::write(int x, int y,const Byte *data,int len)
{
	YFileWrite::write(base_offset+(info.biHeight-1-y)*bytes_per_row+x,data,len);
}
