#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
//#include <mem.h>
#include <memory.h>
#include <sys\stat.h>
//Yurin C++ fix
#include "Tiff.h"
//#include "ImgDrive.h"
using namespace IFF;

//////////////////////////////////////////////////////////////////
//################################################################
//################################################################
//////////////////////////////////////////////////////////////////
void TiffReadIFD::readIFD()
{
	width=0;
	length=0;
	bits=0;
	spi=0; //strips per image
	rowsPS=0;
	str_offset=0;
	str_length=0;
	pal_offset=0;
	pal_length=0;

	struct {
		uint16_t ByteOrder;
		uint16_t Version;
		uint32_t IFD0offset;
	} header;
	this->read(0,&header,sizeof(header));
	if( (header.ByteOrder != MSB && header.ByteOrder != LSB) ||
					header.Version != 0x2a )
	{
		throw "not TIFF format";
		//printf("%s not Tiff format\n",filename);
		//exit(-2);
	}
	this->read(header.IFD0offset,&entries,sizeof(entries));
	this->read(&ifd,entries*sizeof(TiffTag));
}

TiffReadIFD::TiffReadIFD(const char* filename) : YFileRead(filename)
{
	readIFD();
	int count=entries;
	while(count--)
	{
		switch(ifd[count].tag)
		{
		case IMAGE_WIDTH_TAG:
						width  = (uint16_t)ifd[count].value_or_offset;
						break;
		case IMAGE_LENGTH_TAG:
						length = (uint16_t)ifd[count].value_or_offset;
						break;
		case SAMPLES_PER_PIXEL_TAG:
						if(1!=(uint16_t)ifd[count].value_or_offset)
							throw "Currently supported only 8-bit GrayScale TIFF files";
						break;
		case BITS_PER_SAMPLE_TAG:
//@@						if(8!=(uint16_t)ifd[count].value_or_offset)
//@@							throw "Currently supported only 8-bit GrayScale TIFF files";
							bits=(uint16_t)ifd[count].value_or_offset;
							if(bits!=8 && bits!=16)
								throw "Currently supported only 8 and 16-bit GrayScale and Palette TIFF files";
						break;
		case TAG_COMPRESSION:
						if(1!=(uint16_t)ifd[count].value_or_offset)
							throw "Currently supported No-Compressed TIFF files";
						break;
		case ROWS_PER_STRIP_TAG:
						rowsPS = (uint16_t)ifd[count].value_or_offset;
						break;
		case STRIPS_OFFSET_TAG:
						spi    = (uint16_t)ifd[count].length;
						str_offset = ifd[count].value_or_offset;
						break;
		case STRIPS_BYTE_COUNTS_TAG:
						str_length = ifd[count].value_or_offset;
						break;
		case PALETTE_TAG:
						pal_offset = ifd[count].value_or_offset;
						pal_length = uint16_t(ifd[count].length);
						break;
		} // end switch
	} // end while
	if(rowsPS==0) rowsPS=length;
	if(spi==0) spi=uint16_t((length+rowsPS-1)/rowsPS);
}

int TiffReadIFD::GetPalette(uint32_t *paletta)
{
	if(paletta==NULL) return 0;
	if(pal_offset==0) return 0; //no palette exist
	uint32_t *cur_p = paletta;
	uint16_t rgb[768];
	int len=MIN(pal_length,256);
	this->read(pal_offset,rgb,sizeof(short)*768);
	for(int i = 0;i < len; i++)
	{
		*cur_p++= ((unsigned long)rgb[i]>>8)
				|(((unsigned long)rgb[i+256]>>8)<<8)
				|(((unsigned long)rgb[i+512]>>8)<<16);
	}
return pal_length;
}
Tiff8Read::~Tiff8Read(void)
{
	if(buf_alloc) delete[] buf_alloc;
}

Tiff8Read::Tiff8Read(const char* filename) : TiffReadIFD(filename)
{
	int spi=this->StripsPerImage();
	buf_alloc=new uint32_t[spi*2*sizeof(uint32_t)];
	if(buf_alloc==NULL)
	{
		throw "Tiff8Read: Insufficient memory";
		//printf("Tiff8Read: Insufficient memory.\n");
		//exit(-3);
	}
	StripLength=buf_alloc+sizeof(uint32_t)*spi;
	StripOffset=buf_alloc;
	if(spi==1)
	{
		StripOffset[0]=str_offset;
		StripLength[0]=Width()*Length()*((Bits()==8)?1:((Bits()==16)?2:0));
	}
	else
	{
		this->read(str_offset,StripOffset,spi*sizeof(StripOffset[0]));
		uint32_t imsize=Width()*Length()*((Bits()==8)?1:((Bits()==16)?2:0));
		uint32_t strsize=Width()*RowsPerStrip()*((Bits()==8)?1:((Bits()==16)?2:0));
		for(int i=0; i<spi; i++)
		{
			if(imsize>strsize)
			{
				StripLength[i]=strsize;
				imsize-=strsize;
			}
			else
			{
				StripLength[i]=imsize;
			}
		}
	}
}

int Tiff8Read::readStrip(int nStrip,Byte* buf)
{
	return read(StripOffset[nStrip],buf,StripLength[nStrip]);
}

int Tiff8Read::readImage(Byte* buf)
{
int imsize=0;
	for(int i=0; i<StripsPerImage() ; i++ )
	{
		int count=readStrip(i,buf);
		buf+=count;
		imsize+=count;
	}
return imsize;
}


//////////////////////////////////////////////////////////////////
//################################################################
//################################################################
//////////////////////////////////////////////////////////////////
uint16_t TiffWriteIFD::fillIFD(void)
{
	int count=0;
	ifd[count].tag=NEW_SUBFILE_TYPE_TAG;
	ifd[count].type=TTYP_LONG;
	ifd[count].length=1;
	ifd[count].value_or_offset=0;
	count++;
	ifd[count].tag=SUBFILE_TYPE_TAG;
	ifd[count].type=TTYP_SHORT;
	ifd[count].length=1;
	ifd[count].value_or_offset=1;
	count++;
	ifd[count].tag=IMAGE_WIDTH_TAG;
	ifd[count].type=TTYP_LONG;
	ifd[count].length=1;
	ifd[count].value_or_offset=(uint32_t)width;
	count++;
	ifd[count].tag=IMAGE_LENGTH_TAG;
	ifd[count].type=TTYP_LONG;
	ifd[count].length=1;
	ifd[count].value_or_offset=(uint32_t)length;
	count++;
	ifd[count].tag=BITS_PER_SAMPLE_TAG;
	ifd[count].type=TTYP_SHORT;
	ifd[count].length=1;
	ifd[count].value_or_offset=bits;
	count++;
	ifd[count].tag=TAG_COMPRESSION;
	ifd[count].type=TTYP_SHORT;
	ifd[count].length=1;
	ifd[count].value_or_offset=1;
	count++;
	ifd[count].tag=PHOTOMETRIC_INTERPRETATION_TAG;
	ifd[count].type=TTYP_SHORT;
	ifd[count].length=1;
	if(pal_length)
	{
		ifd[count].value_or_offset=3;//palette
		count++;
	}
	else
	{
		ifd[count].value_or_offset=1;//grayscale
		count++;
	}
	ifd[count].tag=STRIPS_OFFSET_TAG;
	ifd[count].type=TTYP_LONG;
	ifd[count].length=spi;
	ifd[count].value_or_offset=str_offset;
	count++;
	ifd[count].tag=ROWS_PER_STRIP_TAG;
	ifd[count].type=TTYP_LONG;
	ifd[count].length=1;
	ifd[count].value_or_offset=(uint32_t)rowsPS;
	count++;
	ifd[count].tag=STRIPS_BYTE_COUNTS_TAG;
	ifd[count].type=TTYP_LONG;
	ifd[count].length=1;
	ifd[count].value_or_offset=str_length;
	count++;
	ifd[count].tag=TAG_PLANAR_CONFIGURATION;
	ifd[count].type=TTYP_SHORT;
	ifd[count].length=1;
	ifd[count].value_or_offset=1;//R->R->B order
	count++;

	if(pal_length)
	{
		ifd[count].tag=317;//predictor
		ifd[count].type=TTYP_SHORT;
		ifd[count].length=1;
		ifd[count].value_or_offset=1;
		count++;
		ifd[count].tag=PALETTE_TAG;
		ifd[count].type=TTYP_SHORT;
		ifd[count].length=pal_length*3;
		ifd[count].value_or_offset=pal_offset;
		count++;
	}
	entries=uint16_t(count);
return entries;
}

void TiffWriteIFD::writeIFD(void)
{
////
	//this->write(header.IFD0offset,&entries,sizeof(entries));
	//this->SetPosition(8);
	this->write(&entries,sizeof(entries));
	this->write(&ifd,entries*sizeof(TiffTag));
	uint32_t tt=0;
	this->write(&tt,sizeof(tt));
}

TiffWriteIFD::TiffWriteIFD(const char* filename,
	int wid,int len,int bits_ps,uint32_t *palette) : YFileWrite(filename)
{
	struct {
		uint16_t ByteOrder;
		uint16_t Version;
		uint32_t IFD0offset;
	} header;
	entries=0;
	width=uint16_t(wid);
	length=uint16_t(len);
	bits=(unsigned short)((bits_ps<=8)?8:16);
	rowsPS=length;
	spi=1; //strips per image
//	spi=(length+rowsPS-1)/rowsPS;
	if(palette) pal_length=256;
	else pal_length=0;
	str_length=width*length*(bits/8);
	entries= fillIFD(); // to determine 'entries'
	str_offset=sizeof(header)+sizeof(TIFF_IFDentry)*entries+2+4;
	if(palette) pal_offset=str_offset+str_length;
	else pal_offset=0;
	if(pal_offset&0x00000001) pal_offset++;
// write header
	header.ByteOrder=LSB;
	header.Version=0x2a;
	header.IFD0offset=sizeof(header);
	entries= fillIFD();
	this->SetPosition(0);
	this->write(0,&header,sizeof(header));
	writeIFD();
	if(pal_offset) this->PutPalette(palette);
	this->write(&header,sizeof(header));
}

int TiffWriteIFD::PutPalette(uint32_t *paletta)
{
	if(paletta==NULL)
	{
		pal_offset=0; //no palette exist
		pal_length=0;
	}
	else
	{
		//pal_offset=GetPosition(); //no palette exist
		pal_length=256;
		uint32_t *cur_p = paletta;
		uint16_t rgb[768];
		for(int i = 0;i < pal_length; i++)
		{
			uint16_t t;
			t=(Byte)(*cur_p); t=t&uint16_t(0x00FF); t|=uint16_t(t<<8);
			rgb[i+512]=t;
			t=(Byte)((*cur_p)>>8); t=t&uint16_t(0x00FF); t|=uint16_t(t<<8);
			rgb[i+256]=t;
			t=(Byte)((*cur_p)>>16); t=t&uint16_t(0x00FF); t|=uint16_t(t<<8);
			rgb[i]=t;
			cur_p++;
		}
		this->write(pal_offset,rgb,sizeof(short)*768);
	}
return pal_length;
}

Tiff8Write::Tiff8Write(const char* filename,int wid,int len,int bits_ps,uint32_t *paletta)
	: TiffWriteIFD(filename,wid,len,bits_ps,paletta)
{
	;
}

int Tiff8Write::writeStrip(int /*nStrip*/,Byte* buf)
{
/*
	{ //test for 10-bit images
		uint16_t *p=(uint16_t *)buf;
		int i=str_length/2;
		while(i--)
		{
			uint16_t v=*p;
			*p++=v<<6;
		}
	}
*/
	return this->write(str_offset,buf,str_length);
}

int Tiff8Write::writeImage(Byte* buf)
{
int imsize=0;
	for(int i=0; i<StripsPerImage() ; i++ )
	{
		int count=writeStrip(i,buf);
		buf+=count;
		imsize+=count;
	}
return imsize;
}

//////////////////////////////////////////////////////////////////
//################################################################
//################################################################
//////////////////////////////////////////////////////////////////

extern "C" int __cdecl writetiff(const char *filename,uint16_t width, uint16_t length,
             Byte *buffer,uint32_t *paletta)
{
	int pict_number = 1; // return value,probably fool cycle.
	Tiff8Write tf(filename,(int)width,(int)length,8,paletta);
	tf.writeImage(buffer);
//	tf.PutPalette(paletta);
return pict_number;
}


extern "C" int __cdecl readtiff(const char *filename,uint16_t *width, uint16_t *length,
             Byte *buffer,uint32_t *paletta)
{
	int pict_number = 1; // return value,probably fool cycle.
	Tiff8Read tf(filename);
	*width=uint16_t(tf.Width());
	*length=uint16_t(tf.Length());
	tf.readImage(buffer);
	tf.GetPalette(paletta);
return pict_number;
}

/*
ImgInfo * TIFF_OpenImageFile(const char *fname)
{
	ImgInfo * imInf=new ImgInfo;
	if(imInf)
	{
		::memset(imInf,0,sizeof(ImgInfo));
		imInf->IsOk=1;
		imInf->struct_size=sizeof(ImgInfo);
		//::strcpy(imInf->format,"TIFF");
		imInf->ErrMsg[0]=0;

		Tiff8Read *tf =new Tiff8Read(fname);
		imInf->img_file_obj=tf;
		if(tf)
		{
			imInf->img_class=AImage::IMGC_GRAYSCALE; //numerically equal enum AImage::YImageClass
			imInf->num_images=1;//number of different images in file
			imInf->image_id=0;  //Get info on this image (0<=image_id<num_images)
			imInf->bits=8;
			imInf->width=tf->Width();
			imInf->height=tf->Length();
			imInf->next_line_shift=tf->Width();
		}
	}
return imInf;
}
int TIFF_GetImageInfo(ImgInfo * imInf,int image_id)
{
	if(imInf)
	{
		Tiff8Read *tf =(Tiff8Read*)(imInf->img_file_obj);
		if(tf)
		{
			imInf->img_class=AImage::IMGC_GRAYSCALE; //numerically equal enum AImage::YImageClass
			imInf->num_images=1;//number of different images in file
			imInf->image_id=0;  //Get info on this image (0<=image_id<num_images)
			imInf->bits=8;
			imInf->width=tf->Width();
			imInf->height=tf->Length();
			imInf->next_line_shift=tf->Width();
		}
	}
	return 0;
}
int TIFF_LoadImage(ImgInfo * imInf,void *topleft,int next_line_shift,ProgressCounter* indicator)
{
	if(imInf)
	{
		Tiff8Read *tf =(Tiff8Read*)(imInf->img_file_obj);
		if(tf)
		{
			tf->readImage((unsigned char *)topleft);
			//tf->GetPalette(paletta);
		}
	}
	return 0;
}
int TIFF_CloseImageFile(ImgInfo ** imInf)
{//** because nullify pointer to ImgInfo
	if(imInf && *imInf)
	{
		Tiff8Read *tf =(Tiff8Read*)((*imInf)->img_file_obj);
		if(tf)
		{
			delete tf;
			(*imInf)->img_file_obj=0;
		}
		delete (*imInf);
		(*imInf)=0;
	}
	return 0;
}
*/
