#include <math.h> //sqrt()
#include <io.h>
#include <fcntl.h>
#include <sys\stat.h>
#pragma hdrstop

#include "VirtDib.h"
#ifndef MIN
	#define MIN(a,b) ( ((a)<(b)) ? (a) : (b) )
#endif
typedef unsigned short Pix16;
//!!!! CODE of functions build_line?????? differ ONLY by SECOND ARGUMENT TYPE
static void build_line_zoom_in(long *dst,BYTE* src,int left,int right,int zoom_x); //
static void build_line_zoom_in(long *dst,Pix16* src,int left,int right,int zoom_x);//
static void build_line_zoom_out(long *dst,BYTE* src,int left,int right,int zoom_x,int img_width); //
static void build_line_zoom_out(long *dst,Pix16* src,int left,int right,int zoom_x,int img_width);//
static void build_line_zoom_no(long *dst,BYTE* src,int left,int right);
static void build_line_zoom_no(long *dst,Pix16* src,int left,int right);
static void build_line(long *dst,BYTE* src,int left,int right,int image_width,int zoom_x);
static void build_line(long *dst,Pix16* src,int left,int right,int image_width,int zoom_x);
//static void build_line(long *dst,unsigned short* src,int left,int right,int image_width,int zoom_x)
//					{ build_line(dst,(short* )src,left,right,image_width,zoom_x); }
static void build_line_zoom_in_add(long *dst,BYTE* src,int left,int right,int zoom_x); //
static void build_line_zoom_in_add(long *dst,Pix16* src,int left,int right,int zoom_x);//
static void build_line_zoom_out_add(long *dst,BYTE* src,int left,int right,int zoom_x,int img_width); //
static void build_line_zoom_out_add(long *dst,Pix16* src,int left,int right,int zoom_x,int img_width);//
static void build_line_zoom_no_add(long *dst,Pix16* src,int left,int right);
static void build_line_zoom_no_add(long *dst,BYTE* src,int left,int right);
static void build_line_add(long *dst,BYTE* src,int left,int right,int image_width,int zoom_x);
static void build_line_add(long *dst,Pix16* src,int left,int right,int image_width,int zoom_x);
//static void build_line_add(long *dst,unsigned short* src,int left,int right,int image_width,int zoom_x)
//					{ build_line_add(dst,(short* )src,left,right,image_width,zoom_x); }
static void normalize_line(BYTE *dst,long *src,int count,int divisor,short bits,const BYTE * map_table);
static void normalize_line_rgb(BYTE *dst,long *src,int count,int divisor,short bits,const BYTE * map_table);


long VirtDib::BytesPerRow(short width,short bits)
{
	long   bytes_per_row;
  switch(bits)
  {
  case 1:
    bytes_per_row=(width+7)/8;
    break;
  case 4:
    bytes_per_row=(width+1)/2;
    break;
  case 8:
    bytes_per_row=width;
    break;
  case 24:
    bytes_per_row=3*width;
    break;
  default:
    throw "Bmp::Error::errInadmissibleBitsPerPixel";
  };
	bytes_per_row=(bytes_per_row*4+3)/4; //SetDIBitsToDevice() required DWORD alignment
return bytes_per_row;
}

void VirtDib::SetGrayPalette()
{
  if(info)
  {
    for(int i=0; i<pal_entries; i++)
    {
      info->bmiColors[i].rgbBlue
         =info->bmiColors[i].rgbGreen
         =info->bmiColors[i].rgbRed=BYTE(i);
      info->bmiColors[i].rgbReserved=0;
    }
	}
}
void VirtDib::SetImagePalette(const AImageMem8* img8)
{
  if(info && img8)
  {
		int ent=MIN(GetPaletteEntries(),img8->GetPaletteEntries());
		RGBQUAD* q=GetPalette();
		AImage::Rgb *p=img8->Palette();
		if(ent >0 && p && q)
		{
			for(int i=0; i<ent; i++)
			{
				q[i].rgbBlue=p[i].b;
				q[i].rgbGreen=p[i].g;
				q[i].rgbRed=p[i].r;
				q[i].rgbReserved=0;
			}
		}
	}
}

void VirtDib::bmiHeaderInit(const Inf& param)
//void VirtDib::bmiHeaderInit(int width,int height,int bits,int bytes_per_row)
{
  info->bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  info->bmiHeader.biWidth=param.width;
  info->bmiHeader.biHeight=param.height;
  info->bmiHeader.biPlanes=1;
  info->bmiHeader.biBitCount=param.bits;
  info->bmiHeader.biCompression=BI_RGB;
  info->bmiHeader.biSizeImage=param.bytes_per_row*param.height;
  info->bmiHeader.biXPelsPerMeter=0;
  info->bmiHeader.biYPelsPerMeter=0;
  info->bmiHeader.biClrUsed=0;
  info->bmiHeader.biClrImportant=0;
}

VirtDib::VirtDib()
{
	//>====== Calculate parameters ==========
	unsigned width=512;
	unsigned height=64;
	short bits=8;
	bytes_per_row=BytesPerRow(width,bits);
	pal_entries = short(0x1FF & (1<<bits));
	data_sz=bytes_per_row*height;
	//-----------------
	inf8pal.width=width;
	inf8pal.height=height;
	inf8pal.bits=bits;
	inf8pal.pal_entries=pal_entries;
	inf8pal.bytes_per_row=BytesPerRow(inf8pal.width,inf8pal.bits);
	//-----------------
	bits=24;
	//width=short(::sqrt(data_sz/3));
	while( data_sz < unsigned(height*BytesPerRow(width,bits)) )	
		width--;
	inf8rgb.width=width;
	inf8rgb.height=height;
	inf8rgb.bits=24;
	inf8rgb.pal_entries=0;
	inf8rgb.bytes_per_row=BytesPerRow(inf8rgb.width,inf8rgb.bits);
	//-----------------
	//>====== Allocate space ================
	info=(BITMAPINFO*)new BYTE[sizeof(BITMAPINFOHEADER)+sizeof(RGBQUAD)*pal_entries];
  if(info)
  {
//		bmiHeaderInit(width,height,bits,bytes_per_row);
		line_buffer_sz=bytes_per_row;
		data = new BYTE[data_sz];
    if(data==NULL)
    {
    	delete[] info; info=NULL;
      throw "VirtDib::insufficientMemory";
    }
    line_buffer = new long[line_buffer_sz];
    if(line_buffer==NULL)
    {
    	delete[] data; data=NULL;
    	delete[] info; info=NULL;
      throw "VirtDib::insufficientMemory";
    }
  }
  else
      throw "VirtDib::insufficientMemory";
	//>====== Initialize data ===============
	last_type=BMT_GRAY8;
	bmiHeaderInit(inf8pal);		
	SetGrayPalette();
}

VirtDib::~VirtDib()
{
	if(data)
  {
  	delete[] data;
    data=NULL;
  }
  if(line_buffer)
  {
  	delete[] line_buffer;
    line_buffer=NULL;
  }
  if(info)
  {
  	delete[] info;
    info=NULL;
  }
}
/*
void VirtDib::Load(const char *fname)
{
}
*/
/*
void VirtDib::Save(const char *fname)
{
	BITMAPFILEHEADER header;
  header.bfType=0x4d42;// ( 'BM')
  header.bfReserved1=0;
  header.bfReserved2=0;
  header.bfOffBits=sizeof(header)+sizeof(info)+pal_entries*sizeof(RGBQUAD);   // offset for raster data beginning
  header.bfSize=header.bfOffBits+info->bmiHeader.biSizeImage;      // file size in bytes
  int handle=open(fname, O_WRONLY | O_CREAT | O_TRUNC,
                         S_IREAD | S_IWRITE);
  if(handle!=-1)
  {
	  write(handle,&header,sizeof(header));
	  write(handle,info,sizeof(BITMAPINFOHEADER)+pal_entries*sizeof(RGBQUAD));
	  write(handle,data,info->bmiHeader.biSizeImage);
  }
  close(handle);
}
*/
void VirtDib::Paint(HDC hdc,const RECT& rect)
{
	const int r_width=rect.right-rect.left;
	const int r_height=rect.bottom-rect.top;
  int u_XSrc=0;//rect2.left;
  int u_YSrc=Height()-r_height;
  unsigned int u_uStartScan=0;
  unsigned int u_cScanLines=Height();
	::SetDIBitsToDevice(hdc,
  				rect.left,rect.top,
          r_width,r_height,
          u_XSrc,u_YSrc,
          u_uStartScan,u_cScanLines,
          data,info,DIB_RGB_COLORS);
}


static void PaintRect(HDC hdc,HBRUSH hBrush,const RECT& rect)
{
		//CPoint pt;
		//::SetBrushOrgEx(pDC->m_hDC,0,0,&pt);	
		HGDIOBJ oldBrush=::SelectObject(hdc,hBrush);
		if(rect.left<rect.right)
		{
			::PatBlt(hdc,rect.left,rect.top,rect.right-rect.left,rect.bottom-rect.top,PATCOPY);
		}
		::SelectObject(hdc,oldBrush);
}
static void PaintNotImageArea(HDC hdc,HBRUSH hBrush,const RECT& rect,const SIZE& img_zoomed_size)
{ //paint(Erase) Background
	RECT r_right(rect);
	RECT r_bottom(r_right);
	r_right.left+=img_zoomed_size.cx;
	r_bottom.top+=img_zoomed_size.cy;
	r_bottom.right=r_right.left;

	HGDIOBJ oldBrush=::SelectObject(hdc,hBrush);
	if(r_right.left<r_right.right)
		::PatBlt(hdc,r_right.left,r_right.top,r_right.right-r_right.left,r_right.bottom-r_right.top,PATCOPY);
	if(r_bottom.top<r_bottom.bottom && r_bottom.left<r_bottom.right )
		::PatBlt(hdc,r_bottom.left,r_bottom.top,r_bottom.right-r_bottom.left,r_bottom.bottom-r_bottom.top,PATCOPY);
	::SelectObject(hdc,oldBrush);
}


void VirtDib::DefaultPaint(HWND hWnd,HDC hdc,HBRUSH hBrush,const AImage *img,const ZoomFactor2D& zoom,const BYTE *map_table)
{
	//Get visible client rect due to scroll
	POINT org;
	::GetViewportOrgEx(hdc, &org);
	RECT rect2;
	::GetClientRect(hWnd,&rect2);
	rect2.left   -= org.x;
	rect2.right  -= org.x;
	rect2.top    -= org.y;
	rect2.bottom -= org.y;
	const RECT rect(rect2);//Now rect-is client rectangle due to Scroll origin
	if(!img )
	{
		::PaintRect(hdc,hBrush,rect);
		return;
	}

	SIZE img_size={img->Width(),img->Height()};
	const SIZE img_zoomed_size=zoom.ZoomValue(img_size);
	const RECT img_rect={0,0,img_zoomed_size.cx,img_zoomed_size.cy};
	//pDC->RectVisible(&rect2);
	int err=::GetClipBox(hdc,&rect2);
	if( err == NULLREGION )
		return; //this situation take place VERY FREQUENTLY, probbably need repaint all?
	else if(err==ERROR)
		rect2=rect;//probably not take place?
	//rect2=rect;
	//rect2 &=img_rect;
	if(!::IntersectRect(&rect2,&rect2,&img_rect))
	{
		::PaintRect(hdc,hBrush,rect);
	}
	else
	{
		const int y0=rect2.top;
		const int ymax=rect2.bottom;
		const int x0=rect2.left;
		const int xmax=rect2.right;
		switch(img->GetImageStore())
		{
		case AImage::IMGS_AImageMem8:
			{
				::PaintNotImageArea(hdc,hBrush,rect,img_zoomed_size);
				const AImageMem8  *img8=dynamic_cast<const AImageMem8 *>(img);
				if(AImage::IMGC_PALETTE==img8->ImageClass() )
				{
					if(img8->Palette())
					{
						if(last_type!=BMT_PAL8)
						{
							last_type=BMT_PAL8;
							bmiHeaderInit(inf8pal);		
						}
						this->SetImagePalette(img8);
					}
					else
						this->SetGrayPalette();
				}
				else if(AImage::IMGC_GRAYSCALE==img8->ImageClass() )
				{
						if(last_type!=BMT_GRAY8)
						{
							last_type=BMT_GRAY8;
							bmiHeaderInit(inf8pal);		
							this->SetGrayPalette();
						}
				}
				const int x_step = this->Width();
				const int y_step = this->Height();
				for(int y=y0 ; y < ymax ; y+=y_step )
				for(int x=x0 ; x < xmax ; x+=x_step )
				{
					RECT portion={x,y,x+x_step,y+y_step};
					//portion &= rect2;
					if(::IntersectRect(&portion,&portion,&rect2))
					//if(!portion.IsRectEmpty())
					{
						this->Build(*img8,portion,zoom,map_table);
						this->Paint(hdc,portion);
					}
				}
			}
			break;
		case AImage::IMGS_AImageMem16:
			{
				::PaintNotImageArea(hdc,hBrush,rect,img_zoomed_size);
				const AImageMem16 *img16=dynamic_cast<const AImageMem16 *>(img);
					if(last_type!=BMT_GRAY8)
					{
						last_type=BMT_GRAY8;
						bmiHeaderInit(inf8pal);		
						this->SetGrayPalette();
					}
					const int x_step = this->Width();
					const int y_step = this->Height();
					for(int y=y0 ; y < ymax ; y+=y_step )
					for(int x=x0 ; x < xmax ; x+=x_step )
					{
						RECT portion={x,y,x+x_step,y+y_step};
						//portion &= rect2;
						if(::IntersectRect(&portion,&portion,&rect2))
						//if(!portion.IsRectEmpty())
						{
							this->Build(*img16,portion,zoom,map_table);
							this->Paint(hdc,portion);
						}
					}
			}
			break;
		case AImage::IMGS_AImageMemFloat:
			{
				::PaintRect(hdc,hBrush,rect);
			}
			break;
		case AImage::IMGS_AImageMemRgb8:
			{
				::PaintNotImageArea(hdc,hBrush,rect,img_zoomed_size);
				const AImageMemRgb8 *img_rgb=dynamic_cast<const AImageMemRgb8 *>(img);
					if(last_type!=BMT_RGB8)
					{
						last_type=BMT_RGB8;
						bmiHeaderInit(inf8rgb);		
					}
					const int x_step = this->Width();
					const int y_step = this->Height();
					for(int y=y0 ; y < ymax ; y+=y_step )
					for(int x=x0 ; x < xmax ; x+=x_step )
					{
						RECT portion={x,y,x+x_step,y+y_step};
						//portion &= rect2;
						if(::IntersectRect(&portion,&portion,&rect2))
						//if(!portion.IsRectEmpty())
						{
							this->Build(*img_rgb,portion,zoom,map_table);
							this->Paint(hdc,portion);
						}
					}
			}
			break;
		case AImage::IMGS_AImageMemRgb16:
			{
				::PaintRect(hdc,hBrush,rect);
			}
			break;
		default:
			{
				::PaintRect(hdc,hBrush,rect);
			}
		};
	}
}

/*
void VirtDib::Build(AImageMem16& img,TRect& portion,BYTE * map_table)
{ //Build when no zoom !!!
  int maxx=portion.Width();
  int maxy=portion.Height();
  if(map_table)
  {
    for(int y=0; y<maxy; y++)
    for(int x=0; x<maxx; x++)
    {
        int im_x=portion.left+x;
        int im_y=portion.top+y;
        (*this)[y][x]=map_table[(img[im_y][im_x])&0x3FF];
    }
  }
  else
  {
    for(int y=0; y<maxy; y++)
    for(int x=0; x<maxx; x++)
    {
        int im_x=portion.left+x;
        int im_y=portion.top+y;
        (*this)[y][x]=BYTE((img[im_y][im_x])>>2);
    }
  }
}
*/
void VirtDib::BuildViaDecimation(const AImageMem8& img,const RECT& portion,const ZoomFactor2D zoom,const BYTE *map_table)
{
	if(zoom.Y<-1)
	{ //
	}
	else if(zoom.Y>1)
	{
	}
	else
	{//no zoom in Y
	}
}
void VirtDib::Build(const AImageMem8& img,const RECT& portion,const ZoomFactor2D zoom,const BYTE *map_table)
{
	if(zoom.Y<-1)
  {
  	int zoom_y=-zoom.Y;
    int count_center= portion.bottom-portion.top;
    int count_bottom=portion.bottom*zoom_y;
    if(img.Height() < count_bottom)
    {
      count_bottom=img.Height()%zoom_y;
      count_center--;
    }
    else
    {
      count_bottom=0;
    }
    int y_src = portion.top*zoom_y;
    int y_dst=0;
    if(count_center>0)
    {
      while(count_center--)
      {
        build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        int count=zoom_y-1;
        while(count--)
          build_line_add(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        normalize_line((*this)[y_dst++],line_buffer,portion.right-portion.left,zoom_y,img.Bits(),map_table);
      }
    }
    if(count_bottom>0)
    {
        build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        int count=count_bottom-1;
        while(count--)
          build_line_add(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        normalize_line((*this)[y_dst++],line_buffer,portion.right-portion.left,count_bottom,img.Bits(),map_table);
    }
  }
  else if(zoom.Y >1)
  {
    int top_org      = portion.top    / zoom.Y;
    int count_top    = portion.top    % zoom.Y;
    int count_center = portion.bottom / zoom.Y;
    int count_bottom = portion.bottom % zoom.Y;
    count_center -= top_org;
    int y_src = top_org;
    int y_dst=0;
    if( 0==count_center )
    {
      count_top=count_bottom-count_top;
      //count_center=0;
      count_bottom=0;
    }
    else
    {
      count_top=zoom.Y-count_top;
      count_center--;
    }
    if(count_top)
    {
      build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
      BYTE * good_src=(*this)[y_dst++];
      normalize_line(good_src,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
      while(--count_top)
      	::memcpy((*this)[y_dst++],good_src,sizeof(*good_src)*(portion.right-portion.left));
    }
    while(count_center--)
    {
      build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
      BYTE * good_src=(*this)[y_dst++];
      normalize_line(good_src,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
      int count = zoom.Y;// remember zoom.Y>1 !!!
      while(--count)
      	::memcpy((*this)[y_dst++],good_src,sizeof(*good_src)*(portion.right-portion.left));
    }
    if(count_bottom)
    {
      build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
      BYTE * good_src=(*this)[y_dst++];
      normalize_line(good_src,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
      while(--count_bottom)
      	::memcpy((*this)[y_dst++],good_src,sizeof(*good_src)*(portion.right-portion.left));
    }
	}
  else
  {// no zoom in y (zoom.Y==1)
    int maxy=portion.bottom-portion.top;
    for(int y=0; y<maxy; y++)
    {
      int im_y=portion.top+y;
      build_line(line_buffer,img[im_y],portion.left,portion.right,img.Width(),zoom.X);
      normalize_line((*this)[y],line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
    }
  }
}
void VirtDib::Build(const AImageMem16& img,const RECT& portion,const ZoomFactor2D zoom,const BYTE *map_table)
{
	
	if(zoom.Y<-1)
  {
  	int zoom_y=-zoom.Y;
    int count_center= portion.bottom-portion.top;
    int count_bottom=portion.bottom*zoom_y;
    if(img.Height() < count_bottom)
    {
      count_bottom=img.Height()%zoom_y;
      count_center--;
    }
    else
    {
      count_bottom=0;
    }
    int y_src = portion.top*zoom_y;
    int y_dst=0;
    if(count_center>0)
    {
      while(count_center--)
      {
        build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        int count=zoom_y-1;
        while(count--)
          build_line_add(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        normalize_line((*this)[y_dst++],line_buffer,portion.right-portion.left,zoom_y,img.Bits(),map_table);
      }
    }
    if(count_bottom>0)
    {
        build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        int count=count_bottom-1;
        while(count--)
          build_line_add(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
        normalize_line((*this)[y_dst++],line_buffer,portion.right-portion.left,count_bottom,img.Bits(),map_table);
    }
  }
  else if(zoom.Y >1)
  {
    int top_org      = portion.top    / zoom.Y;
    int count_top    = portion.top    % zoom.Y;
    int count_center = portion.bottom / zoom.Y;
    int count_bottom = portion.bottom % zoom.Y;
    count_center -= top_org;
    int y_src = top_org;
    int y_dst=0;
    if( 0==count_center )
    {
      count_top=count_bottom-count_top;
      //count_center=0;
      count_bottom=0;
    }
    else
    {
      count_top=zoom.Y-count_top;
      count_center--;
    }
    if(count_top)
    {
      build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
      BYTE * good_src=(*this)[y_dst++];
      normalize_line(good_src,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
      while(--count_top)
      	::memcpy((*this)[y_dst++],good_src,sizeof(*good_src)*(portion.right-portion.left));
    }
    while(count_center--)
    {
      build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
      BYTE * good_src=(*this)[y_dst++];
      normalize_line(good_src,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
      int count = zoom.Y;// remember zoom.Y>1 !!!
      while(--count)
      	::memcpy((*this)[y_dst++],good_src,sizeof(*good_src)*(portion.right-portion.left));
    }
    if(count_bottom)
    {
      build_line(line_buffer,img[y_src++],portion.left,portion.right,img.Width(),zoom.X);
      BYTE * good_src=(*this)[y_dst++];
      normalize_line(good_src,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
      while(--count_bottom)
      	::memcpy((*this)[y_dst++],good_src,sizeof(*good_src)*(portion.right-portion.left));
    }
  }
  else
  {// no zoom in y (zoom.Y==1)
    int maxy=portion.bottom-portion.top;
    for(int y=0; y<maxy; y++)
    {
      int im_y=portion.top+y;
      build_line(line_buffer,img[im_y],portion.left,portion.right,img.Width(),zoom.X);
      normalize_line((*this)[y],line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
    }
  }
}
void VirtDib::Build(const AImageMemRgb8& img,const RECT& portion,const ZoomFactor2D zoom,const BYTE *map_table)
{
	if(zoom.Y<-1)
  {
  	int zoom_y=-zoom.Y;
    int count_center= portion.bottom-portion.top;
    int count_bottom=portion.bottom*zoom_y;
    if(img.Height() < count_bottom)
    {
      count_bottom=img.Height()%zoom_y;
      count_center--;
    }
    else
    {
      count_bottom=0;
    }
    int y_src = portion.top*zoom_y;
    int y_dst=0;
    if(count_center>0)
    {
      while(count_center--)
      {
				int y;
				BYTE * good_src=(*this)[y_dst++];
				for(int rgb=0; rgb<3; rgb++)
				{
					y=y_src;
					AImageMem8* im=img.GetPlane8(2-rgb);
					build_line(line_buffer,(*im)[y++],portion.left,portion.right,img.Width(),zoom.X);
		      int count=zoom_y-1;
			    while(count--)
						build_line_add(line_buffer,(*im)[y++],portion.left,portion.right,img.Width(),zoom.X);
					normalize_line_rgb(good_src+rgb,line_buffer,portion.right-portion.left,zoom_y,img.Bits(),map_table);
				}
				y_src=y;
      }
    }
    if(count_bottom>0)
    {
	    BYTE * good_src=(*this)[y_dst];
			for(int rgb=0; rgb<3; rgb++)
			{
				int y=y_src;
				AImageMem8* im=img.GetPlane8(2-rgb);
				build_line(line_buffer,(*im)[y++],portion.left,portion.right,img.Width(),zoom.X);
        int count=count_bottom-1;
        while(count--)
          build_line_add(line_buffer,(*im)[y++],portion.left,portion.right,img.Width(),zoom.X);
				normalize_line_rgb(good_src+rgb,line_buffer,portion.right-portion.left,count_bottom,img.Bits(),map_table);
			}
    }
	}
  else if(zoom.Y >1)
	{
    int top_org      = portion.top    / zoom.Y;
    int count_top    = portion.top    % zoom.Y;
    int count_center = portion.bottom / zoom.Y;
    int count_bottom = portion.bottom % zoom.Y;
    count_center -= top_org;
    int y_src = top_org;
    int y_dst=0;
    if( 0==count_center )
    {
      count_top=count_bottom-count_top;
      //count_center=0;
      count_bottom=0;
    }
    else
    {
      count_top=zoom.Y-count_top;
      count_center--;
    }
    if(count_top)
    {
	    BYTE * good_src=(*this)[y_dst];
			for(int rgb=0; rgb<3; rgb++)
			{
				AImageMem8* im=img.GetPlane8(2-rgb);
				build_line(line_buffer,(*im)[y_src],portion.left,portion.right,img.Width(),zoom.X);
				normalize_line_rgb(good_src+rgb,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
			}
			y_src++;
			y_dst++;
			while(--count_top)
      	::memcpy((*this)[y_dst++],good_src,3*sizeof(*good_src)*(portion.right-portion.left));
    }
    while(count_center--)
    {
	    BYTE * good_src=(*this)[y_dst];
			for(int rgb=0; rgb<3; rgb++)
			{
				AImageMem8* im=img.GetPlane8(2-rgb);
				build_line(line_buffer,(*im)[y_src],portion.left,portion.right,img.Width(),zoom.X);
				normalize_line_rgb(good_src+rgb,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
			}
			y_src++;
			y_dst++;
      int count = zoom.Y;// remember zoom.Y>1 !!!
      while(--count)
      	::memcpy((*this)[y_dst++],good_src,3*sizeof(*good_src)*(portion.right-portion.left));
    }
    if(count_bottom)
    {
	    BYTE * good_src=(*this)[y_dst];
			for(int rgb=0; rgb<3; rgb++)
			{
				AImageMem8* im=img.GetPlane8(2-rgb);
				build_line(line_buffer,(*im)[y_src],portion.left,portion.right,img.Width(),zoom.X);
				normalize_line_rgb(good_src+rgb,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
			}
			y_src++;
			y_dst++;
      while(--count_bottom)
      	::memcpy((*this)[y_dst++],good_src,3*sizeof(*good_src)*(portion.right-portion.left));
    }
	}
  else
	{// no zoom in y (zoom.Y==1)
		const int maxy=portion.bottom-portion.top;
		for(int rgb=0; rgb<3; rgb++)
		{
			AImageMem8* im=img.GetPlane8(2-rgb);
			for(int y=0; y<maxy; y++)
			{
				int im_y=portion.top+y;
				build_line(line_buffer,(*im)[im_y],portion.left,portion.right,img.Width(),zoom.X);
				normalize_line_rgb((*this)[y]+rgb,line_buffer,portion.right-portion.left,0,img.Bits(),map_table);
			}
		}
  }
}
/*
void build_line_zoom_in(long *dst,short* src,int left,int right,int zoom_x);//
void build_line_zoom_out(long *dst,short* src,int left,int right,int zoom_x,int img_width);//
void build_line_zoom_no(long *dst,short* src,int left,int right);
void build_line(long *dst,short* src,int left,int right,int image_width,int zoom_x);
void build_line_zoom_in_add(long *dst,short* src,int left,int right,int zoom_x);//
void build_line_zoom_out_add(long *dst,short* src,int left,int right,int zoom_x,int img_width);//
void build_line_zoom_no_add(long *dst,short* src,int left,int right);
void build_line_add(long *dst,short* src,int left,int right,int image_width,int zoom_x);
void normalize_line(BYTE *dst,long *src,int count,int divisor=1);
void normalize_line(BYTE *dst,long *src,int count,int divisor,short bits,const BYTE * map_table);
void normalize_line_rgb(BYTE *dst,long *src,int count,int divisor,short bits,const BYTE * map_table);
*/
static void normalize_line_rgb(BYTE *dst,long *src,int count,int divisor,short bits,const BYTE * map_table)
{
/*
	for(int i=0; i<count; i++)
	{
		*dst++=255;
		*dst++=0;
		*dst++=255;
	};
	return;
*/
	if(map_table)
  {
		unsigned long mask=(1UL<<bits)-1;
    if(divisor>1)
    {
      while(count--)
      {
        unsigned long val=*src++;
        val/=divisor;
        *dst=map_table[val&mask];
				dst+=3;
      }
    }
    else
    {
      while(count--)
      {
        unsigned long val=*src++;
        *dst=map_table[val&mask];
				dst+=3;
      }
    }
  }
  else
  {
		int shift=bits-8;
		shift=(shift<0)?0:shift;
    if(divisor>1)
    {
      while(count--)
      {
        unsigned long val=*src++;
        val/=divisor;
        *dst=BYTE(val>>shift);
				dst+=3;
      }
    }
    else
    {
      while(count--)
      {
        unsigned long val=*src++;
        *dst=BYTE(val>>shift);
				dst+=3;
      }
    }
  }
}

static void normalize_line(BYTE *dst,long *src,int count,int divisor,short bits,const BYTE * map_table)
{
	if(map_table)
  {
		unsigned long mask=(1UL<<bits)-1;
    if(divisor>1)
    {
      while(count--)
      {
        unsigned long val=*src++;
        val/=divisor;
        *dst++=map_table[val&mask];
      }
    }
    else
    {
      while(count--)
      {
        unsigned long val=*src++;
        *dst++=map_table[val&mask];
      }
    }
  }
  else
  {
		int shift=bits-8;
		shift=(shift<0)?0:shift;
    if(divisor>1)
    {
      while(count--)
      {
        unsigned long val=*src++;
        val/=divisor;
        *dst++=BYTE(val>>shift);
      }
    }
    else
    {
      while(count--)
      {
        unsigned long val=*src++;
        *dst++=BYTE(val>>shift);
      }
    }
  }
}

static void build_line_zoom_in(long *dst,BYTE* src,int left,int right,int zoom_x)
{
  int  left_org   =  left/zoom_x;
  int count_left  =  left % zoom_x;
  int count_center= right/zoom_x;
  int count_right = right % zoom_x;
  count_center -= left_org;
  src += left_org;
//  dst += left;
  if( 0==count_center )
  {
    count_left=count_right-count_left;
    //count_center=0;
    count_right=0;
  }
  else
  {
    count_left=zoom_x-count_left;
    count_center--;
  }
  {
    long val=long(*src++);
    while(count_left--)
      *dst++ = val;
  }
  while(count_center--)
  {
    long val=long(*src++);
    int count=zoom_x;
    while(count--)
      *dst++ = val;
  }
  if(count_right)
  {
    long val=long(*src);
    while(count_right--)
      *dst++ = val;
  }
}

static void build_line_zoom_in(long *dst,Pix16* src,int left,int right,int zoom_x)
{
  int  left_org   =  left/zoom_x;
  int count_left  =  left % zoom_x;
  int count_center= right/zoom_x;
  int count_right = right % zoom_x;
  count_center -= left_org;
  src += left_org;
//  dst += left;
  if( 0==count_center )
  {
    count_left=count_right-count_left;
    //count_center=0;
    count_right=0;
  }
  else
  {
    count_left=zoom_x-count_left;
    count_center--;
  }
  {
    long val=long(*src++);
    while(count_left--)
      *dst++ = val;
  }
  while(count_center--)
  {
    long val=long(*src++);
    int count=zoom_x;
    while(count--)
      *dst++ = val;
  }
  if(count_right)
  {
    long val=long(*src);
    while(count_right--)
      *dst++ = val;
  }
}

static void build_line_zoom_in_add(long *dst,BYTE* src,int left,int right,int zoom_x)
{
  int  left_org   =  left/zoom_x;
  int count_left  =  left % zoom_x;
  int count_center= right/zoom_x;
  int count_right = right % zoom_x;
  count_center -= left_org;
  src += left_org;
//  dst += left;
  if( 0==count_center )
  {
    count_left=count_right-count_left;
    //count_center=0;
    count_right=0;
  }
  else
  {
    count_left=zoom_x-count_left;
    count_center--;
  }
  {
    long val=long(*src++);
    while(count_left--)
      (*dst++) += val;
  }
  while(count_center--)
  {
    long val=long(*src++);
    int count=zoom_x;
    while(count--)
      (*dst++) += val;
  }
  if(count_right)
  {
    long val=long(*src);
    while(count_right--)
      (*dst++) += val;
  }
}

static void build_line_zoom_in_add(long *dst,Pix16* src,int left,int right,int zoom_x)
{
  int  left_org   =  left/zoom_x;
  int count_left  =  left % zoom_x;
  int count_center= right/zoom_x;
  int count_right = right % zoom_x;
  count_center -= left_org;
  src += left_org;
//  dst += left;
  if( 0==count_center )
  {
    count_left=count_right-count_left;
    //count_center=0;
    count_right=0;
  }
  else
  {
    count_left=zoom_x-count_left;
    count_center--;
  }
  {
    long val=long(*src++);
    while(count_left--)
      (*dst++) += val;
  }
  while(count_center--)
  {
    long val=long(*src++);
    int count=zoom_x;
    while(count--)
      (*dst++) += val;
  }
  if(count_right)
  {
    long val=long(*src);
    while(count_right--)
      (*dst++) += val;
  }
}


static void build_line_zoom_out(long *dst,BYTE* src,int left,int right,int zoom_x,int img_width)
{
  int count_center= right-left;
  int count_right=right*zoom_x;
  if(img_width < count_right)
  {
		count_right=img_width%zoom_x;
   	count_center--;
  }
  else
  {
		count_right=0;
  }
  src += left*zoom_x;
  if(count_center>0)
  {
    while(count_center--)
    {
      long val=0;
      int count=zoom_x;
      while(count--)
        val += long(*src++);
      val= (val + (zoom_x>>1) ) / zoom_x;
      (*dst++) = val;
    }
  }
  if(count_right>0)
  {
      long val=0;
      int count=count_right;
      while(count--)
        val += long(*src++);
      val= (val + (count_right>>1) ) / count_right;
      (*dst) = val;
  }
}
static void build_line_zoom_out(long *dst,Pix16* src,int left,int right,int zoom_x,int img_width)
{
  int count_center= right-left;
  int count_right=right*zoom_x;
  if(img_width < count_right)
  {
		count_right=img_width%zoom_x;
   	count_center--;
  }
  else
  {
		count_right=0;
  }
  src += left*zoom_x;
  if(count_center>0)
  {
    while(count_center--)
    {
      long val=0;
      int count=zoom_x;
      while(count--)
        val += long(*src++);
      val= (val + (zoom_x>>1) ) / zoom_x;
      (*dst++) = val;
    }
  }
  if(count_right>0)
  {
      long val=0;
      int count=count_right;
      while(count--)
        val += long(*src++);
      val= (val + (count_right>>1) ) / count_right;
      (*dst) = val;
  }
}

static void build_line_zoom_out_add(long *dst,BYTE* src,int left,int right,int zoom_x,int img_width)
{
  int count_center= right-left;
  int count_right=right*zoom_x;
  if(img_width < count_right)
  {
		count_right=img_width%zoom_x;
   	count_center--;
  }
  else
  {
		count_right=0;
  }
  src += left*zoom_x;
  if(count_center>0)
  {
    while(count_center--)
    {
      long val=0;
      int count=zoom_x;
      while(count--)
        val += long(*src++);
      val= (val + (zoom_x>>1) ) / zoom_x;
      (*dst++) += val;
    }
  }
  if(count_right>0)
  {
      long val=0;
      int count=count_right;
      while(count--)
        val += long(*src++);
      val= (val + (count_right>>1) ) / count_right;
      (*dst) += val;
  }
}

static void build_line_zoom_out_add(long *dst,Pix16* src,int left,int right,int zoom_x,int img_width)
{
  int count_center= right-left;
  int count_right=right*zoom_x;
  if(img_width < count_right)
  {
		count_right=img_width%zoom_x;
   	count_center--;
  }
  else
  {
		count_right=0;
  }
  src += left*zoom_x;
  if(count_center>0)
  {
    while(count_center--)
    {
      long val=0;
      int count=zoom_x;
      while(count--)
        val += long(*src++);
      val= (val + (zoom_x>>1) ) / zoom_x;
      (*dst++) += val;
    }
  }
  if(count_right>0)
  {
      long val=0;
      int count=count_right;
      while(count--)
        val += long(*src++);
      val= (val + (count_right>>1) ) / count_right;
      (*dst) += val;
  }
}
static void build_line_zoom_no(long *dst,BYTE* src,int left,int right)
{
  src+=left;
  int count=right-left;
  while(count-- > 0)
  	*dst++= long(*src++);
}
static void build_line_zoom_no(long *dst,Pix16* src,int left,int right)
{
  src+=left;
  int count=right-left;
  while(count-- > 0)
  	*dst++= long(*src++);
}

static void build_line_zoom_no_add(long *dst,BYTE* src,int left,int right)
{
  src+=left;
  int count=right-left;
  while(count-- > 0)
  	(*dst++) += long(*src++);
}

static void build_line_zoom_no_add(long *dst,Pix16* src,int left,int right)
{
  src+=left;
  int count=right-left;
  while(count-- > 0)
  	(*dst++) += long(*src++);
}

static void build_line(long *dst,BYTE* src,int left,int right,int image_width,int zoom_x)
{
	if(zoom_x>1)
  {
		build_line_zoom_in(dst,src,left,right,zoom_x);
  }
  else if(zoom_x<-1)
  {
  	zoom_x=-zoom_x;
		build_line_zoom_out(dst,src,left,right,zoom_x,image_width);
  }
  else
  {
		build_line_zoom_no(dst,src,left,right);
  }
}
static void build_line(long *dst,Pix16* src,int left,int right,int image_width,int zoom_x)
{
	if(zoom_x>1)
  {
		build_line_zoom_in(dst,src,left,right,zoom_x);
  }
  else if(zoom_x<-1)
  {
  	zoom_x=-zoom_x;
		build_line_zoom_out(dst,src,left,right,zoom_x,image_width);
  }
  else
  {
		build_line_zoom_no(dst,src,left,right);
  }
}

static void build_line_add(long *dst,BYTE* src,int left,int right,int image_width,int zoom_x)
{
	if(zoom_x>1)
  {
		build_line_zoom_in_add(dst,src,left,right,zoom_x);
  }
  else if(zoom_x<-1)
  {
  	zoom_x=-zoom_x;
		build_line_zoom_out_add(dst,src,left,right,zoom_x,image_width);
  }
  else
  {
		build_line_zoom_no_add(dst,src,left,right);
  }
}

static void build_line_add(long *dst,Pix16* src,int left,int right,int image_width,int zoom_x)
{
	if(zoom_x>1)
  {
		build_line_zoom_in_add(dst,src,left,right,zoom_x);
  }
  else if(zoom_x<-1)
  {
  	zoom_x=-zoom_x;
		build_line_zoom_out_add(dst,src,left,right,zoom_x,image_width);
  }
  else
  {
		build_line_zoom_no_add(dst,src,left,right);
  }
}

