#ifndef _VTL_IMAGE_H
#define _VTL_IMAGE_H
///////////////////////////////////////////////////////////////////////////////////////////
#include "loki_typeManip.h"
#include "loki_Typelist.h"
#include "loki_static_check.h" 
///////////////////////////////////////////////////////////////////////////////////////////
#include "alexandresku.h"
#include "vtl_meta_base.h"
#include "vtl_image_types.h"
///////////////////////////////////////////////////////////////////////////////////////////

namespace Vtl
{
namespace Img
{
	typedef int offset_type;

	template<class T,int T_stride>
	class Iterator_1D
	{
		T *data;
	public:
		typedef T value_type;
		enum {stride=T_stride};

		Iterator_1D() :data(0){}
		Iterator_1D(T *data_) :data(data_) {}
		Iterator_1D& operator++() {data+=stride; return *this;}
		Iterator_1D& operator--() {data-=stride; return *this;}
		Iterator_1D& operator+=(offset_type i) {data+=stride*i; return *this;}
		Iterator_1D& operator-=(offset_type i) {data-=stride*i; return *this;}
		T& operator[](offset_type i) {return data[stride*i]; }
		T operator[](offset_type i) const {return data[stride*i]; }
		T& operator*() {return *data;}
		T operator*() const {return *data;}
	};
	template<class T>
	class Iterator_1D<T,0>
	{
		T *data;
	public:
		typedef T value_type;
		const offset_type stride;

		Iterator_1D() :data(0),stride(0) {}
		Iterator_1D(T *data_,offset_type stride_) :data(data_),stride(stride_) {}
		Iterator_1D& operator++() {data+=stride; return *this;}
		Iterator_1D& operator--() {data-=stride; return *this;}
		Iterator_1D& operator+=(offset_type i) {data+=stride*i; return *this;}
		Iterator_1D& operator-=(offset_type i) {data-=stride*i; return *this;}
		T& operator[](offset_type i) {return data[stride*i]; }
		T operator[](offset_type i) const {return data[stride*i]; }
		T& operator*() {return *data;}
		T operator*() const {return *data;}
	};
	//   NDimension+1  1D ,  0,1, .. .   -   (, bands). 
	template<class T_Layout>
	class BuildIter1DList
	{
		typedef typename T_Layout::value_type value_type;
		typedef typename T_Layout::SList::IList IList;
		enum {ND=Vtl::Meta::IL::Length<IList>::value-2 };
		//-----------------
		template<typename T,int i>
		struct Iterator_1D_selector
		{
			typedef Iterator_1D<T,i> Result;
		};
		template<typename T>
		struct Iterator_1D_selector<T,1>
		{
			typedef T* Result;
		};
	public:
		typedef typename Iterator_1D_selector<value_type,Vtl::Meta::IL::Elem<ND+1,IList>::value>::Result BandsIterator;
	private:
		//-----------------
		template<int i>
		struct build
		{
			typedef Loki::Typelist<typename Iterator_1D_selector<value_type,Vtl::Meta::IL::Elem<i,IList>::value >::Result,typename build<i+1>::Result > Result;
		};
		template<>
		struct build<ND>
		{
			typedef Loki::Typelist< BandsIterator ,Loki::NullType > Result;
		};
		//-----------------
	public:
		typedef typename build<0>::Result Result;
	};

	///
	///   -  ,  
	///
	template<class T_Layout,bool isGray_>
	class Iterator_ND_base
		: private T_Layout
	{
	public:
		typedef typename T_Layout::value_type value_type;
		typedef typename T_Layout::SList::IList IList;
		typedef typename T_Layout::SList SList;
		enum {Ndim=T_Layout::Ndim};
		enum {isGray=isGray_};
		//enum {isGray=Vtl::Meta::IL::Elem<Ndim,IList>::value==1};
	public:
		Iterator_ND_base(T_Layout strides_)
			:T_Layout(strides_)
		{
			Meta::set<T_Layout::Ndim>(offset,0);
		}
		void print()
		{
			Meta::print_array<Ndim>(offset,"offset");
		}
		template<int I,int stride>
		class MoveX
		{
			Iterator_ND_base& base;
		private:
			offset_type& ofs() {return base.offset[I-1]; }
		public:
			MoveX(Iterator_ND_base& base_,offset_type /*stride_*/) : base(base_) {}
			Iterator_ND_base& operator++() {ofs()+=stride; return base;}
			Iterator_ND_base& operator--() {ofs()_-=stride; }
			Iterator_ND_base& operator+=(offset_type move_by) {ofs()+=stride*move_by; }
			Iterator_ND_base& operator-=(offset_type i)       {ofs()-=stride*move_by; }
	//		T& operator[](offset_type i) {return data[stride*i]; }
	//		T operator[](offset_type i) const {return data[stride*i]; }
			bool operator==(const MoveX& rhs) const { return ofs() == rhs.ofs();}
			bool operator!=(const MoveX& rhs) const { return ofs() != rhs.ofs();}
		};
		template<int I>
		class MoveX<I,0>
		{
			Iterator_ND_base& base;
			const int stride;
		private:
			offset_type& ofs() {return base.offset[I-1]; }
		public:
			MoveX(Iterator_ND_base& base_,offset_type stride_) : base(base_),stride(stride_) {}
			Iterator_ND_base& operator++() {ofs()+=stride; return base;}
			Iterator_ND_base& operator--() {ofs()_-=stride; }
			Iterator_ND_base& operator+=(offset_type move_by) {ofs()+=stride*move_by; }
			Iterator_ND_base& operator-=(offset_type i)       {ofs()-=stride*move_by; }
	//		T& operator[](offset_type i) {return data[stride*i]; }
	//		T operator[](offset_type i) const {return data[stride*i]; }
			bool operator==(const MoveX& rhs) const { return ofs() == rhs.ofs();}
			bool operator!=(const MoveX& rhs) const { return ofs() != rhs.ofs();}
		};
	private:
		class BuildMoveXList
		{
			typedef typename T_Layout::value_type value_type;
			typedef typename T_Layout::SList::IList IList;
			enum {ND=Vtl::Meta::IL::Length<IList>::value-2 };
			//-----------------
			template<int i>
			struct MoveX_selector
			{
				typedef MoveX<i,Vtl::Meta::IL::Elem<i,IList>::value> Result;
			};
			//template<>
			//struct MoveX_selector<0,1>
			//{
			//	typedef T* Result;
			//};
			//-----------------
			template<int i>
			struct build
			{
				typedef Loki::Typelist<typename MoveX_selector<i>::Result,typename build<i+1>::Result > Result;
			};
			template<>
			struct build<ND>
			{
				typedef Loki::Typelist<typename MoveX_selector<ND>::Result,Loki::NullType > Result;
			};
			//-----------------
		public:
			typedef typename build<0>::Result Result;
		};
	protected:
		//// SList strides;
		offset_type offset[Ndim];
		value_type *data;
		typedef typename BuildMoveXList::Result MoveXList;
	public:
		template<int i>
		typename Loki::TL::TypeAt<MoveXList,i>::Result x()
		{
			return Loki::TL::TypeAt<MoveXList,i>::Result(*this,T_Layout::stride<i>());
		}
		//template<>
		//class It
		//{
		//	int i;

		//};
		//template<int i>
		//x()
		//{
		//}
	};
	//template<class T_Layout,bool isGray,bool isSubSpace> 
	//class Iterator_ND_spec;
	template<class T_Layout,bool isGray,bool isSubSpace> 
	class Iterator_ND_spec
		:public Iterator_ND_base<T_Layout,isGray>
	{
	//public:
	//	Iterator_ND_spec(SList strides_)
	//		:Iterator_ND_base(strides_)
	//	{
	//	}
	};
	///
	///  ,  ,  
	///
	template<class T_Layout>
	class Iterator_ND_spec<T_Layout,false,false>
		:public Iterator_ND_base<T_Layout,false>
	{
	public:
		typedef typename BuildIter1DList<T_Layout>::BandsIterator BandsIterator;
		typedef typename BuildIter1DList<T_Layout>::Result Iter1DList;
	public:
		//template<int i>
		//struct Iter1D
		//{
		//	typedef typename Loki::TL::TypeAt<Iter1DList,i>::Result Type;
		//};
		struct IterBands
		{
			typedef typename Loki::TL::TypeAt<Iter1DList,Ndim>::Result Type;
		};
	private:
		offset_type ofs() const { return Vtl::Meta::sum<Ndim-1,offset_type>(offset); }
	public:
		Iterator_ND_spec(T_Layout strides_)
			:Iterator_ND_base(strides_)
		{
		}
		BandsIterator operator*() { return BandsIterator(data+ofs(),strides.x<Ndim>()); }
	};
	///
	///  ,  ,   
	///
	template<class T_Layout>
	class Iterator_ND_spec<T_Layout,false,true>
		:public Iterator_ND_base<T_Layout,false>
	{
	public:
		typedef typename BuildIter1DList<T_Layout>::BandsIterator BandsIterator;
		typedef typename BuildIter1DList<T_Layout>::Result Iter1DList;
	public:
		//template<int i>
		//struct Iter1D
		//{
		//	typedef typename Loki::TL::TypeAt<Iter1DList,i>::Result Type;
		//};
		struct IterBands
		{
			typedef typename Loki::TL::TypeAt<Iter1DList,Ndim>::Result Type;
		};
	private:
		offset_type ofs() const { return Vtl::Meta::sum<Ndim-1,offset_type>(offset); }
	public:
		Iterator_ND_spec(T_Layout strides_)
			:Iterator_ND_base(strides_)
		{
		}
		BandsIterator operator*() { return BandsIterator(data+ofs(),strides.x<Ndim>()); }
	};
	///
	///  ,    ,  
	///
	template<class T_Layout>
	class Iterator_ND_spec<T_Layout,true,false>
		:public Iterator_ND_base<T_Layout,true>
	{
	private:
		offset_type ofs() const { return Vtl::Meta::sum<Ndim-1,offset_type>(offset); }
	public:
		Iterator_ND_spec(T_Layout strides_)
			:Iterator_ND_base(strides_)
		{
		}
		value_type& operator*() { return data[ofs()]; }
	};
	///
	///  ,    ,   
	///
	template<class T_Layout>
	class Iterator_ND_spec<T_Layout,true,true>
		:public Iterator_ND_base<T_Layout,true>
	{
	private:
		offset_type ofs() const { return Vtl::Meta::sum<Ndim-1,offset_type>(offset); }
	public:
		Iterator_ND_spec(T_Layout strides_)
			:Iterator_ND_base(strides_)
		{
		}
		value_type& operator*() { return data[ofs()]; }
	};
	///
	///
	///
	template<class T_Layout>
	class Iterator_ND
		:public Iterator_ND_spec<T_Layout,bool(T_Layout::isGray!=0),false>
	{
	public:
		Iterator_ND(T_Layout strides_)
			:Iterator_ND_spec(strides_)
		{
		}
	};
	///
	///
	///
	template<class T_SList,class T_Layout> 
	class ImageNDref
	{
	public:
		enum {Ndim=Vtl::Meta::IL::Length<T_SList::IList>::value };
		typedef Iterator_ND<T_Layout> IterND;
	private:
		T_SList sizes;
		T_Layout layout;

		typedef typename T_Layout::value_type value_type;
		value_type* data;
		//typedef Loki::Typelist<Iterator_1D<value_type,1>,Loki::Typelist<Iterator_1D<value_type,2>,Loki::Typelist<Iterator_1D<value_type,0>,Loki::NullType> > > I1D;
		typedef typename BuildIter1DList<T_Layout>::Result Iter1DList;
	public:
		template<int i>
		struct Iter1D
		{
			typedef typename Loki::TL::TypeAt<Iter1DList,i>::Result Type;
		};
		struct IterBands
		{
			typedef typename Loki::TL::TypeAt<Iter1DList,Ndim>::Result Type;
		};
		virtual ~ImageNDref() {}
		ImageNDref(value_type* data_,const T_SList& size_list,const T_Layout& layout_)
			:sizes(size_list),layout(layout_),data(data_)
		{

		}
		IterND NIter() const 
		{
			return IterND(layout);
		}
	};
}//namespace Img
}//namespace Vtl
///////////////////////////////////////////////////////////////////////////////////////////
#endif //_VTL_IMAGE_H
