#include <stdlib.h>
#include "fourier.H"
#include "imagemem.h"
#include <math.h>
#define  SWAP(a,b) tvalue=(a);(a)=(b);(b)=tvalue;


AImageMemFloat* FFT_2D::Image2Float(AImage* img,bool make2power)
{
	AImageMem8* imt8;
	AImageMem16* imt16;
	bool flag8=false;
	bool flag16=false;
	if(imt8=dynamic_cast<AImageMem8 *>(img))flag8=true;
	if(!flag8)
		if(imt16=dynamic_cast<AImageMem16 *>(img)) flag16=true;
	if((flag8)||(flag16))
	{
		int width=img->Width();
		int height=img->Height();
		int Fourier_width=width;
		int Fourier_height=height;
		int dx=0;
		int dy=0;
		if(make2power)
		{
			CalcSize(img,&Fourier_width,&Fourier_height);
			GetLeftTopPosAtFFT(img,&dx,&dy);
		};
		AImage *im1;
		AImageMemFloat* img1;
		im1=::AImageCreate(0,AImage::IMGC_FLOAT,Fourier_width,Fourier_height,sizeof(float)*8);
		img1=dynamic_cast<AImageMemFloat *>(im1);
		int y;
		for(y=0;y<Fourier_height;y++)
		for(int x=0;x<Fourier_width;x++)
			(*img1)[y][x]=0;
		for(y=0;y<height;y++)
		for(int x=0;x<width;x++)
		{
			if(flag8)
			(*img1)[y+dy][x+dx]=(*imt8)[y][x];
			else
			(*img1)[y+dy][x+dx]=(*imt16)[y][x];
		}
		return img1;
	}
	else
		throw "wrong type of data in FFT_2D::CreateImage ";
}


void FFT_2D::Float2Image(AImageMemFloat *fi,AImageMem8* img,bool logarithmic,bool autoscale,float A,float B)
{
	if(ChekImages(img,fi,false))
	{
		int height=img->Height();
		int width=img->Width();
		const Pixel8 max_col=Pixel8((1<<img->Bits())-1);
		if(autoscale)
		{
			float max=0;
			int x;
			for(x=0;x<width;x++)
				for(int y=0;y<height;y++)
				{
					float tmp=(*fi)[y][x];
					tmp=(tmp>=0)?tmp:-tmp;
					if(logarithmic)
						tmp = (tmp<1)? 0 : float(1+::log(tmp));					//
						max=(tmp>max)? tmp : max;
				}
			for(x=0;x<width;x++)
				for(int y=0;y<height;y++)
				{
					float tmp=(*fi)[y][x];
					tmp=(tmp>=0)?tmp:-tmp;
					if(logarithmic)
						tmp = (tmp<1)? 0 : float(1+::log(tmp));					//
					(*img)[y][x]=Pixel8(tmp*max_col/max);
				}
		}
		else
		{
			for(int x=0;x<width;x++)
				for(int y=0;y<height;y++)
				{
					float tmp=(*fi)[y][x];
					tmp=(tmp>=0)?tmp:-tmp;
					if(logarithmic)
					{
						tmp = (tmp<1)? A : float(A+1+::log(tmp)*B);
						(*img)[y][x]=(tmp>max_col) ? max_col : Pixel8(tmp);
					}
					else
					{
						tmp=A+tmp*B;
						(*img)[y][x]=(tmp>max_col) ? max_col : Pixel8(tmp);
					}
				}
		}
	}
	else throw "dimensions must be equal for transform Float2Image";

}

bool FFT_2D::ChekImages(AImage *ReAmp,AImage *ImPhase,bool chekpower)
{
	bool flag=true;
	if(ReAmp->Height()!=ImPhase->Height()||
		(ReAmp->Width()!=ImPhase->Width()))
		flag=false;
	if(chekpower)
	{
		int width=ReAmp->Width();
		int height=ReAmp->Height();
		CalcSize(ReAmp,&width,&height);
		if((width!=ReAmp->Width())||
			(height!=ReAmp->Height()))flag=false;
	}
	return flag;
}


void FFT_2D::ReIm2AmpPhase(AImageMemFloat *ReAmp,AImageMemFloat *ImPhase)//
{
	if(ChekImages(ReAmp,ImPhase,false))
	{
		PixelFloat *DataR = (*ReAmp)[0];
		PixelFloat *DataI = (*ImPhase)[0];
		int cnt=ReAmp->Width()*ReAmp->Height();
		while(cnt--)
		{
			double re=*DataR;
			double im=*DataI;
			*(DataR++)=float(::sqrt(re*re+im*im));
			*(DataI++)=float(::atan2(im,re));
		}
	}
	else
		throw"dimensions must be equal for transform ReIm2AmpPhase";
}

void FFT_2D::AmpPhase2ReIm(AImageMemFloat *ReAmp,AImageMemFloat *ImPhase)//
{
	if(ChekImages(ReAmp,ImPhase,false))
	{
		int width=ReAmp->Width();
		int height=ImPhase->Height();
		PixelFloat *DataR = (*ReAmp)[0];
		PixelFloat *DataI = (*ImPhase)[0];
		for(int y=0;y<height;y++)
			for(int x=0;x<width;x++)
			{
				float amp=*DataR;
				float phase=*DataI;
				*(DataR++)=float(amp*cos(phase));
				*(DataI++)=float(amp*sin(phase));
			}
	}
	else
		throw"dimensions must be equal for transform AmpPhase2ReIm";
}



void FFT_2D::GetLeftTopPosAtFFT(AImage* img,int *Fourier_left,int *Fourier_top)
{
	*Fourier_left=(Transform2Power(img->Width())-(img->Width()))/2;
	*Fourier_top =(Transform2Power(img->Height())-(img->Height()))/2;
}



int FFT_2D::Transform2Power(int size)
{
	if(size<=0) throw "wrong parametr in Transform2Power (negativ), must=be positiv";
	unsigned int v=size;
	unsigned int t=1;
	const int n=sizeof(int)*8;
	for(int x=0;x<n;x++)
	{
		t<<=1;
		if(t>=v) break;
	}
	return t;
}

void FFT_2D::CalcSize(AImage *img,int *Fourier_width,int *Fourier_height)
{//
	*Fourier_width=Transform2Power(img->Width());
	*Fourier_height=Transform2Power(img->Height());
}

void FFT_2D::TransposeQuarters( AImageMemFloat *img)
{
    const int xsize=img->Width()/2;
	const int ysize=img->Height()/2;
	const int low_half=2*xsize*ysize;
	PixelFloat * const p0=(*img)[0];
	//  1 | 2
	//  -----
	//  4 | 3
	{//xchange 1 & 3 quarters
		PixelFloat *p1=p0;//(*img)[0];
		PixelFloat *p2=p0+low_half+xsize;//(*img)[ysize]+xsize;
		int ycnt=ysize;
		while(ycnt--)
		{
			int cnt=xsize;
			while(cnt--)
			{
				float t=*p2;
				*p2++=*p1;
				*p1++=t;
			}
			p1+=xsize;
			p2+=xsize;
		}
	}
	{//xchange 2 & 4 quarters
		PixelFloat *p1=p0+xsize;//(*img)[0]+xsize;
		PixelFloat *p2=p0+low_half;//(*img)[ysize];
		int ycnt=ysize;
		while(ycnt--)
		{
			int cnt=xsize;
			while(cnt--)
			{
				float t=*p2;
				*p2++=*p1;
				*p1++=t;
			}
			p1+=xsize;
			p2+=xsize;
		}
	}
}

void FFT_2D::fourn(float *Redata, float *Imdata,int nn[],bool reverse)
{
	int idim;
	unsigned long i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
	unsigned long ibit,k1,k2,n,nprev,nrem,ntot;
	float tempi,tempr,tvalue;
	double theta,wi,wpi,wpr,wr,wtemp;
	const double M_2PI=4*(::atan2(1.,0.));
	for (ntot=1,idim=1;idim<=2;idim++)
		ntot *= nn[idim];
	nprev=1;
	for (idim=2;idim>=1;idim--)
	{
		n=nn[idim];
		nrem=ntot/(n*nprev);
		ip1=nprev;
		ip2=ip1*n;
		ip3=ip2*nrem;
		i2rev=1;
		for (i2=1;i2<=ip2;i2+=ip1)
		{
			if (i2 < i2rev)
			{
				for (i1=i2;i1<=i2+ip1-1;i1++)
				{
					for (i3=i1;i3<=ip3;i3+=ip2)
					{
						i3rev=i2rev+i3-i2;
						SWAP(Redata[i3],Redata[i3rev]);
						SWAP(Imdata[i3],Imdata[i3rev]);
					}
				}
			}
			ibit=ip2 >> 1;
			while ((ibit >=ip1)&&(i2rev > ibit))
			{
				i2rev -= ibit;
				ibit >>= 1;
			}
			i2rev += ibit;
		}
		ifp1=ip1;
		while (ifp1 < ip2)
		{
			ifp2=ifp1 << 1;
			theta=(reverse)?M_2PI/(ifp2/ip1):-M_2PI/(ifp2/ip1);
			wtemp=sin(0.5*theta);
			wpr=-2.0*wtemp*wtemp;
			wpi=sin(theta);
			wr=1.0;
			wi=0.0;
			for (i3=1;i3<=ifp1;i3+=ip1)
			{
				for (i1=i3;i1<=i3+ip1-1;i1++)
				{
					for (k1=i1;k1<=ip3;k1+=ifp2)
					{
						//k1=i2;
						k2=k1+ifp1;
						tempr=float(wr*Redata[k2]-wi*Imdata[k2]);
						tempi=float(wr*Imdata[k2]+wi*Redata[k2]);
						Redata[k2]=Redata[k1]-tempr;
						Imdata[k2]=Imdata[k1]-tempi;
						Redata[k1]+=tempr;
						Imdata[k1]+=tempi;
					}
				}
				wr=(wtemp=wr)*wpr-wi*wpi+wr;
				wi=wi*wpr+wtemp*wpi+wi;
			}
			ifp1=ifp2;
		}
		nprev *= n;
	}
}

/*
void FFT_2D::fourn(float *Redata, float *Imdata,int nn[],bool reverse)
{
	int idim;
	unsigned long ip1,ip2,ip3,ifp1,ifp2;
	unsigned long ibit,n,nprev,nrem,ntot;
	//float tvalue;
	double theta,wi,wpi,wpr,wr,wtemp;
	const double M_2PI=4*(::atan2(1,0));
	for (ntot=1,idim=1;idim<=2;idim++)
		ntot *= nn[idim];
	nprev=1;
	for (idim=2;idim>=1;idim--)
	{
		n=nn[idim];
		nrem=ntot/(n*nprev);
		ip1=nprev;
		ip2=ip1*n;
		ip3=ip2*nrem;
		unsigned i2rev=1;
		for (unsigned i2=1;i2<=ip2;i2+=ip1)
		{
			if (i2 < i2rev)
			{
				{
					for (unsigned i1=i2;i1<=i2+ip1-1;i1++)
					{
						float *Re1=Redata+i1;
						//float *Re2=Redata+i2rev+i1-i2;
						int kkk=i2rev-i2;
						for (unsigned i3=i1;i3<=ip3;i3+=ip2)
						{
							float t=*Re1;
							*Re1=Re1[kkk];
							Re1[kkk]=t;
							Re1+=ip2;
							//Re2+=ip2;
							//unsigned i3rev=i2rev+i3-i2;
							//SWAP(Redata[i3],Redata[i3rev]);
							//SWAP(Imdata[i3],Imdata[i3rev]);
						}
					}
				}
				{
					for (unsigned i1=i2;i1<=i2+ip1-1;i1++)
					{
						float *Im1=Imdata+i1;
						//float *Im2=Imdata+i2rev+i1-i2;
						int kkk=i2rev-i2;
						for (unsigned i3=i1;i3<=ip3;i3+=ip2)
						{
							float t=*Im1;
							*Im1=Im1[kkk];
							Im1[kkk]=t;
							Im1+=ip2;
							//Im2+=ip2;
							//unsigned i3rev=i2rev+i3-i2;
							//SWAP(Redata[i3],Redata[i3rev]);
							//SWAP(Imdata[i3],Imdata[i3rev]);
						}
					}
				}
			}
			ibit=ip2 >> 1;
			while ((ibit >=ip1)&&(i2rev > ibit))
			{
				i2rev -= ibit;
				ibit >>= 1;
			}
			i2rev += ibit;
		}
		ifp1=ip1;
		while (ifp1 < ip2)
		{
			ifp2=ifp1 << 1;
			theta=(reverse)?M_2PI/(ifp2/ip1):-M_2PI/(ifp2/ip1);
			wtemp=sin(0.5*theta);
			wpr=-2.0*wtemp*wtemp;
			wpi=sin(theta);
			wr=1.0;
			wi=0.0;
			for (unsigned i3=1;i3<=ifp1;i3+=ip1)
			{
				for (unsigned i1=i3;i1<=i3+ip1-1;i1++)
				{
					unsigned k2=i1+ifp1;
					float *Re2=Redata+k2;
					float *Im2=Imdata+k2;
					//////unsigned cnt=(ip3-i1)/ifp2+1;
					//////while(cnt--)
					for (unsigned k1=i1;k1<=ip3;k1+=ifp2)
					{
						float tempr=float(wr*(*Re2)-wi*(*Im2));
						float tempi=float(wr*(*Im2)+wi*(*Re2));
						(*Re2)=(*(Re2-ifp1))-tempr;
						(*Im2)=(*(Im2-ifp1))-tempi;
						(*(Re2-ifp1))+=tempr;
						(*(Im2-ifp1))+=tempi;
						Re2+=ifp2;
						Im2+=ifp2;
						//unsigned k2=k1+ifp1;
						//float tempr=float(wr*Redata[k2]-wi*Imdata[k2]);
						//float tempi=float(wr*Imdata[k2]+wi*Redata[k2]);
						//Redata[k2]=Redata[k1]-tempr;
						//Imdata[k2]=Imdata[k1]-tempi;
						//Redata[k1]+=tempr;
						//Imdata[k1]+=tempi;
					}
				}
				wr=(wtemp=wr)*wpr-wi*wpi+wr;
				wi=wi*wpr+wtemp*wpi+wi;
			}
			ifp1=ifp2;
		}
		nprev *= n;
	}
}
*/
void FFT_2D::FFT(AImageMemFloat *Re,AImageMemFloat *Im,bool reverse)
{
	if(ChekImages(Re,Im,true))
	{
		float *Redata=(*Re)[0]-1;
		float *Imdata=(*Im)[0]-1;
		int nn[3];
		int xsize=Im->Width();
		int ysize=Im->Height();
		if((ysize!=Transform2Power(ysize))||(xsize!=Transform2Power(xsize)))
			throw " wrong size of data's massiv for FFT_2D(must be power of =)";
		nn[1]=ysize;
		nn[2]=xsize;
		fourn(Redata,Imdata,nn,reverse);
		if (reverse)
		{
			for(int y=0;y<ysize;y++)
				for(int x=0;x<xsize;x++)
				{
					(*Re)[y][x]/=xsize*ysize;
					(*Im)[y][x]/=xsize*ysize;
				}
		}
	} else
		throw "wrong sizes of data's massives for FFT_2D - must be equal, a=d be power of 2";
}
