#include <math.h>
#include "aimage.h"
#include "imfilter.h"

//////////////////////////////////////////////////////////////////////////
//////////Roberts Sqrt////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
template<class T,class P> void tpl_AImageFilter_LineDetect_RobertsSqrt(const T* in,T* out,P max_color,ProgressCounter *ind,const char *name)
{
	const int width=in->Width();
	const int height=in->Height();
	if(ind) ind->start(name,0,height);
	//----------------------------------
	{
		P* o=(*out)[0];
		int w=width;
		while(w--)
			*o++=0;
	}
	double cof=1.;
	for(int y=1; y<(height-1); y++)
	{
		(*out)[y][0]=0;
		(*out)[y][width-1]=0;
		for(int x=1; x<(width-1); x++)
		{
			//Roberts
			double X=(*in)[y][x]-(*in)[y+1][x+1];
			double Y=(*in)[y+1][x]-(*in)[y][x+1];
			double v=cof*::sqrt(X*X+Y*Y)+0.5; //Roberts sqrt

			(*out)[y][x]=v>=max_color ? max_color : P(v);
		}
		if(ind)
		{
			if(ind->current(y))
			{
				ind->finish();
				throw "User abort";
			}
		}
	}
	//----------------------------------
	{
		P* o=(*out)[height-1];
		int w=width;
		while(w--)
			*o++=0;
	}
	if(ind) ind->finish();
}

void AImageFilter_LineDetect_RobertsSqrt(const AImage *in,AImage *out,ProgressCounter *ind)
{
	const char *name=(ind && ind->russian() )? " ,  (.)":"Line detect, Roberts (sqrt)";
	const AImageMem8 *in8=dynamic_cast<const AImageMem8  *>(in);
	AImageMem8 *out8=dynamic_cast<AImageMem8 *>(out);
	Pixel8 MAX_COLOR=Pixel8((1<<in->Bits())-1);
	if(in8 && out8)
	{
		tpl_AImageFilter_LineDetect_RobertsSqrt(in8,out8,MAX_COLOR,ind,name);
	}
	else
	{
		const AImageMem16 *in16=dynamic_cast<const AImageMem16  *>(in);
		AImageMem16 *out16=dynamic_cast<AImageMem16 *>(out);
		Pixel16 MAX_COLOR=Pixel16((1<<in->Bits())-1);
		if(in16 && out16)
		{
			tpl_AImageFilter_LineDetect_RobertsSqrt(in16,out16,MAX_COLOR,ind,name);
		}
		else
		{
			throw "Unsupported image type for Roberts";
		}
	}
}
//////////////////////////////////////////////////////////////////////////
////////////Roberts Abs///////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
template<class T,class P> void tpl_AImageFilter_LineDetect_RobertsAbs(const T* in,T* out,P max_color,ProgressCounter *ind,const char *name)
{
	const int width=in->Width();
	const int height=in->Height();
	if(ind) ind->start(name,0,height);
	//----------------------------------
	{
		P* o=(*out)[0];
		int w=width;
		while(w--)
			*o++=0;
	}
//	double cof=1.;
	for(int y=1; y<(height-1); y++)
	{
		(*out)[y][0]=0;
		(*out)[y][width-1]=0;
		for(int x=1; x<(width-1); x++)
		{
			//Roberts
			int X=(*in)[y][x]-(*in)[y+1][x+1];
			int Y=(*in)[y+1][x]-(*in)[y][x+1];
			int v=ABS(X)+ABS(Y);                //Roberts abs 
			(*out)[y][x]=v>=max_color ? max_color : P(v);
		}
		if(ind)
		{
			if(ind->current(y))
			{
				ind->finish();
				throw "User abort";
			}
		}
	}
	//----------------------------------
	{
		P* o=(*out)[height-1];
		int w=width;
		while(w--)
			*o++=0;
	}
	if(ind) ind->finish();
}

void AImageFilter_LineDetect_RobertsAbs(const AImage *in,AImage *out,ProgressCounter *ind)
{
	const char *name=(ind && ind->russian() )? " ,  ()":"Line detect, Roberts (abs)";
	const AImageMem8 *in8=dynamic_cast<const AImageMem8  *>(in);
	AImageMem8 *out8=dynamic_cast<AImageMem8 *>(out);
	Pixel8 MAX_COLOR=Pixel8((1<<in->Bits())-1);
	if(in8 && out8)
	{
		tpl_AImageFilter_LineDetect_RobertsAbs(in8,out8,MAX_COLOR,ind,name);
	}
	else
	{
		const AImageMem16 *in16=dynamic_cast<const AImageMem16  *>(in);
		AImageMem16 *out16=dynamic_cast<AImageMem16 *>(out);
		Pixel16 MAX_COLOR=Pixel16((1<<in->Bits())-1);
		if(in16 && out16)
		{
			tpl_AImageFilter_LineDetect_RobertsAbs(in16,out16,MAX_COLOR,ind,name);
		}
		else
		{
			throw "Unsupported image type for Roberts";
		}
	}
}
//////////////////////////////////////////////////////////////////////////
//////////////Sobel///////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
template<class T,class P> void tpl_AImageFilter_LineDetect_Sobel(const T* in,T* out,P max_color,ProgressCounter *ind,const char *name)
{
	const int width=in->Width();
	const int height=in->Height();
	if(ind) ind->start(name,0,height);
	//----------------------------------
	{
		P* o=(*out)[0];
		int w=width;
		while(w--)
			*o++=0;
	}
	double cof=1.;
	for(int y=1; y<(height-1); y++)
	{
		(*out)[y][0]=0;
		(*out)[y][width-1]=0;
		for(int x=1; x<(width-1); x++)
		{
			//Sobel
			double X=((*in)[y-1][x+1]+2*(*in)[y  ][x+1]+(*in)[y+1][x+1])
				   -((*in)[y-1][x-1]+2*(*in)[y  ][x-1]+(*in)[y+1][x-1]);
			double Y=((*in)[y-1][x-1]+2*(*in)[y-1][x  ]+(*in)[y-1][x+1])
				   -((*in)[y+1][x-1]+2*(*in)[y+1][x  ]+(*in)[y+1][x+1]);
			double v=cof*::sqrt(X*X+Y*Y)+0.5; 

			(*out)[y][x]=v>=max_color ? max_color : P(v);
		}
		if(ind)
		{
			if(ind->current(y))
			{
				ind->finish();
				throw "User abort";
			}
		}
	}
	//----------------------------------
	{
		P* o=(*out)[height-1];
		int w=width;
		while(w--)
			*o++=0;
	}
	if(ind) ind->finish();
}

void AImageFilter_LineDetect_Sobel(const AImage *in,AImage *out,ProgressCounter *ind)
{
	const char *name=(ind && ind->russian() )? " , ":"Line detect, Sobel";
	const AImageMem8 *in8=dynamic_cast<const AImageMem8  *>(in);
	AImageMem8 *out8=dynamic_cast<AImageMem8 *>(out);
	Pixel8 MAX_COLOR=Pixel8((1<<in->Bits())-1);
	if(in8 && out8)
	{
		tpl_AImageFilter_LineDetect_Sobel(in8,out8,MAX_COLOR,ind,name);
	}
	else
	{
		const AImageMem16 *in16=dynamic_cast<const AImageMem16  *>(in);
		AImageMem16 *out16=dynamic_cast<AImageMem16 *>(out);
		Pixel16 MAX_COLOR=Pixel16((1<<in->Bits())-1);
		if(in16 && out16)
		{
			tpl_AImageFilter_LineDetect_Sobel(in16,out16,MAX_COLOR,ind,name);
		}
		else
		{
			throw "Unsupported image type for Sobel";
		}
	}
}
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
template<class T,class P> void tpl_AImageFilter_LineThining(const T* _in,T* out,P max_color,ProgressCounter *ind,const char *name)
{
	T* in=(T*)_in;
	const int width=out->Width();
	const int height=out->Height();
	if(ind) ind->start(name,0,(4+1+1)*MAX(width,height));
	int count=0;
	//----------------------------------
	{
		P* o1=(*out)[0];
		P* o2=(*out)[height-1];
		int w=width;
		while(w--)
		{
			*o1++=0;
			*o2++=0;
		}
		for(int y=1; y<(height-1); y++)
		{
			(*out)[y][0]=0;
			(*out)[y][width-1]=0;
		}
	}
	//----------------------------------
	int colors_changed=0;
	do
	{
		colors_changed=0;
		int y;
		//Filling holes
		AImageFilter_Duplicate(out,in,0);//copy back
		for(y=1; y<(height-1); y++)
		{
			for(int x=1; x<(width-1); x++)
			{
				if(((*in)[y][x])==0)
				{
					int cnt=0;
					for(int j=-1; j<=1; j++)
					for(int i=-1; i<=1; i++)
						if(((*in)[y+j][x+i])!=0)
							cnt++;
					if(cnt>=5)
						(*out)[y][x]=0;
				}
			}
			if(ind)
			{
				if(ind->current(count++))
				{
					ind->finish();
					throw "User abort";
				}
			}
		}
		//Tail remove
		AImageFilter_Duplicate(out,in,0);//copy back
		for(y=1; y<(height-1); y++)
		{
			for(int x=1; x<(width-1); x++)
			{
				if(((*in)[y][x])!=0)
				{
					if( (0==((*in)[y  ][x-1])) && (0==((*in)[y-1][x-1])) && (0==((*in)[y-1][x  ])) && (0==((*in)[y-1][x+1])) && (0==((*in)[y  ][x+1]))  )
					{ //Upper end
							(*out)[y][x]=0; 
					}
					else if( (0==((*in)[y+1][x  ])) && (0==((*in)[y+1][x-1])) && (0==((*in)[y  ][x-1])) && (0==((*in)[y-1][x-1])) && (0==((*in)[y-1][x  ]))  )
					{ //Left end
							(*out)[y][x]=0; 
					}
					else if( (0==((*in)[y  ][x-1])) && (0==((*in)[y+1][x-1])) && (0==((*in)[y+1][x  ])) && (0==((*in)[y+1][x+1])) && (0==((*in)[y  ][x+1]))  )
					{ //Down end
							(*out)[y][x]=0; 
					}
					else if( (0==((*in)[y-1][x  ])) && (0==((*in)[y-1][x+1])) && (0==((*in)[y  ][x+1])) && (0==((*in)[y+1][x+1])) && (0==((*in)[y+1][x  ]))  )
					{ //Right end
							(*out)[y][x]=0; 
					}
				}
			}
			if(ind)
			{
				if(ind->current(count++))
				{
					ind->finish();
					throw "User abort";
				}
			}
		}
		//Up Thining
		AImageFilter_Duplicate(out,in,0);//copy back
		for(y=1; y<(height-1); y++)
		{
			for(int x=1; x<(width-1); x++)
			{
				if(((*in)[y][x])!=0)
				{
					if( (0==((*in)[y-1][x])) && (0!=((*in)[y+1][x])) )
					{ //Upper bound
						if( ( (0==((*in)[y-1][x+1]))  &&  (0!=((*in)[y  ][x-1])) )
							||( (0==((*in)[y-1][x-1]))  &&  (0!=((*in)[y  ][x+1])) )
							||( (0!=((*in)[y  ][x+1]))  &&  (0!=((*in)[y  ][x-1])) ) )
						{
							(*out)[y][x]=0;
							colors_changed++;
						}
					}
				}
			}
			if(ind)
			{
				if(ind->current(count++))
				{
					ind->finish();
					throw "User abort";
				}
			}
		}
		//Left Thining
		AImageFilter_Duplicate(out,in,0);//copy back
		for(y=1; y<(height-1); y++)
		{
			for(int x=1; x<(width-1); x++)
			{
				if( (0!=((*in)[y][x])) )
				{
					if( (0==((*in)[y][x-1])) && (0!=((*in)[y][x+1])) )
					{ //left bound
						if( ( (0==((*in)[y-1][x-1]))  &&  (0!=((*in)[y+1][x  ])) )
							||( (0==((*in)[y+1][x-1]))  &&  (0!=((*in)[y-1][x  ])) )
							||( (0!=((*in)[y-1][x  ]))  &&  (0!=((*in)[y+1][x  ])) ) )
						{
							(*out)[y][x]=0;
							colors_changed++;
						}
					}
				}
			}
			if(ind)
			{
				if(ind->current(count++))
				{
					ind->finish();
					throw "User abort";
				}
			}
		}
		//Down Thining
		AImageFilter_Duplicate(out,in,0);//copy back
		for(y=1; y<(height-1); y++)
		{
			for(int x=1; x<(width-1); x++)
			{
				if((*in)[y][x]!=0)
				{
					if( (0==((*in)[y+1][x])) && ((*in)[y-1][x]!=0))
					{ //Down bound
						if( ( (0==((*in)[y+1][x-1]))  &&  (0!=((*in)[y  ][x+1])) )
							||( (0==((*in)[y+1][x+1]))  &&  (0!=((*in)[y  ][x-1])) )
							||( (0!=((*in)[y  ][x-1]))  &&  (0!=((*in)[y  ][x+1])) ) )
						{
							(*out)[y][x]=0;
							colors_changed++;
						}
					}
				}
			}
			if(ind)
			{
				if(ind->current(count++))
				{
					ind->finish();
					throw "User abort";
				}
			}
		}
		//Right Thining
		AImageFilter_Duplicate(out,in,0);//copy back
		for(y=1; y<(height-1); y++)
		{
			for(int x=1; x<(width-1); x++)
			{
				if( 0!=((*in)[y][x]) )
				{
					if( (0==((*in)[y][x+1])) && (0!=((*in)[y][x-1])) )
					{ //Right bound
						if( ( (0==((*in)[y+1][x+1]))  &&  (0!=((*in)[y-1][x  ])) )
							||( (0==((*in)[y-1][x+1]))  &&  (0!=((*in)[y+1][x  ])) )
							||( (0!=((*in)[y+1][x  ]))  &&  (0!=((*in)[y-1][x  ])) ) )
						{
							(*out)[y][x]=0;
							colors_changed++;
						}
					}
				}
			}
			if(ind)
			{
				if(ind->current(count++))
				{
					ind->finish();
					throw "User abort";
				}
			}
		}
		//
	} while(colors_changed);
	//----------------------------------
	if(ind) ind->finish();
}

void AImageFilter_LineThining(const AImage *in,AImage *out,ProgressCounter *ind)
{
	AImageFilter_Duplicate(in,out,ind);
	///
	const char *name=(ind && ind->russian() )? " ":"Line thining";
	const AImageMem8 *in8=dynamic_cast<const AImageMem8 *>(in);
	AImageMem8 *out8=dynamic_cast<AImageMem8 *>(out);
	Pixel8 MAX_COLOR=Pixel8((1<<in->Bits())-1);
	if(out8)
	{
		tpl_AImageFilter_LineThining(in8,out8,MAX_COLOR,ind,name);
	}
	else
	{
		const AImageMem16 *in16=dynamic_cast<const AImageMem16 *>(in);
		AImageMem16 *out16=dynamic_cast<AImageMem16 *>(out);
		Pixel16 MAX_COLOR=Pixel16((1<<in->Bits())-1);
		if(out16)
		{
			tpl_AImageFilter_LineThining(in16,out16,MAX_COLOR,ind,name);
		}
		else
		{
			throw "Unsupported image type for LineThining";
		}
	}
}
