
	/**
	* class Gauss
	* 
	* \brief       \f$ \exp(-x^2/(2*\sigma^2))\f$ 
	*    6  .    
	*   (float, double),  -   (0- ), 
	*   (. Plf::Kernel::Type).  wnd_half_size -    
	*    sigma       
	* wnd_half_size=sigma*sigmas_number. .   void Init()  .
	* 
	* \htmlinclude kernels_1d_conv_gauss.htm
	*	
	* \htmlinclude kernels_1d_conv_gauss_1_deriv.htm
    * 
	* \htmlinclude KernelsNorm.htm 	
    * 
	**********************************************************************************/
	template<typename T,unsigned Deriv=0,Kernel::Type kernel_type_=KT_Conv>
	class Gauss
		:public Base<T,kernel_type_>
	{
		T sigma;
	public:
		Gauss() : Base<T,kernel_type_>("Gauss(t)=exp(-x^2/(2*sigma^2))") {}
		Gauss(T sigmaInPixels,T sigmas_number)
			: Base<T,kernel_type_>("Gauss(t)=exp(-x^2/(2*sigma^2))")
		{	
			Init(sigmaInPixels,sigmas_number);
		}
		Gauss(int left, int right,T sigmaInPixels)
			: Base<T,kernel_type_>("Gauss(t)=exp(-x^2/(2*sigma^2))")
		{	
			Init(left, right,sigmaInPixels);
		}
		virtual ~Gauss() {}
		/*!
		* @param sigmaInPixels - exp(-x^2/(2*sigma^2)) - value of sigma in pixels
		* @param sigmas_number -   ()  
		*
		*/
		void Init(T sigmaInPixels,T sigmas_number)
		{
			sigma=sigmaInPixels;
			int wnd_half_size=int(sigmaInPixels*sigmas_number+0.5);
			_Init(wnd_half_size,wnd_half_size);
			//sigma=wnd_half_size/sigmas_number;
			typedef F::Gauss<T,Deriv> Func;
			Func gauss(sigma,sigmas_number); 
			build(gauss);
			normalize<Func::conv_norm_alg>();
		}
		/*!
		* @param left -   ()      ,   .
		* @param right -   ()      ,   .
		* @param sigmaInPixels - exp(-x^2/(2*sigma^2)) - value of sigma in pixels
		* \see Plf::Kernel::Base::_Init()
		*/
		void Init(int left, int right,T sigmaInPixels)
		{
			sigma=sigmaInPixels;
			_Init(left,right);
			//sigma=wnd_half_size/sigmas_number;
			typedef F::Gauss<T,Deriv> Func;
			Func gauss(sigma,T(1)); 
			build(gauss);
			normalize<Func::conv_norm_alg>();
		}
	};
	/**
	* class Hamming
	* 
	* \brief \f$ h(x)=0.54-0.46\cdot\cos(\frac{2\pi x}{n-1})\f$ ; for x=0..n-1
	*  :
	*  \f$ h(x)=0.54+0.46\cdot\cos(\frac{2\pi x}{n-1})\f$ ;  
	*  for  \f$ x=-n/2,\dots,n/2-1 \f$; for n=size() -  
    * 
	**********************************************************************************/
	template<typename T,Kernel::Type kernel_type_=KT_Conv>
	class Hamming
		:public Base<T,kernel_type_>
	{
	public:
		Hamming() : Base<T,kernel_type_>() {}
		Hamming(int wnd_half_size)
			: Base<T,kernel_type_>("Hamming window")
		{	
			Init(wnd_half_size);
		}
		virtual ~Hamming() {}
		/**
		* @param wnd_half_size - 
		*/
		void Init(int wnd_half_size)
		{
			_Init(wnd_half_size,wnd_half_size);

			typedef F::Hamming<T,0>  Func;
			Func h(wnd_half_size+1+wnd_half_size); ; 
			build(h);
			normalize<Func::conv_norm_alg>();
		}
	};
	/**
	* class Hann
	* 
	* \brief \f$ h(x)= { 1 \over 2 } [1-cos(\frac{2\pi x}{n-1})]\f$ ; for x=0..n-1
	*  :
	*  \f$  h(x)= { 1 \over 2 } [1-cos(\pi+\frac{2\pi x}{n-1})] = { 1 \over 2 } [1+cos(\frac{2\pi x}{n-1})] \f$ ;  
	*  for  \f$ x=-n/2,\dots,n/2-1 \f$; for n=size() -  
	*  
    * 
	**********************************************************************************/
	template<typename T,Kernel::Type kernel_type_=KT_Conv>
	class Hann
		:public Base<T,kernel_type_>
	{
	public:
		Hann() : Base<T,kernel_type_>() {}
		Hann(int wnd_half_size)
			: Base<T,kernel_type_>("Hamming window")
		{	
			Init(wnd_half_size);
		}
		virtual ~Hann() {}
		/**
		* @param wnd_half_size - 
		*/
		void Init(int wnd_half_size)
		{
			_Init(wnd_half_size,wnd_half_size);

			typedef F::Hann<T,0>   Func;
			Func h(wnd_half_size+1+wnd_half_size); ; 
			build(h);
			normalize<Func::conv_norm_alg>();
		}
	};