#ifndef _IO_FILTER_H
#define _IO_FILTER_H
/////////////////////////////////////////////////////////////////////////////
#include <list>
#include "aimage.h"
#include "plf_filters.h"
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup IoFiltersDoxGroup  Source and target filters
 *  @ingroup PipeLineFiltersDoxGroup  
 *
 *  @brief      ()  
 *     (Pipe Line Filtering) 
 *
 *  @{
 */
	///////////////////////////////////////////////////////////////////////////
	// class SrcArray
	///        
	/// ,     .
	/// \param T -    ()  
	/// \param I -   ,    P   P  
	///                .
	/// \param P -   ,   -  ,
	///                   v 
	///                v=P(v).
	/// 
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class I,class P=Vtl::Plf::F11::DoNothing<T> >
	class SrcArray
		: public FilterElementary
	{
	public:
		OutPort<T,P> out;
	private:
		I * data;
		int y;
		int w;
		int h;
		int nls;
	public:
		SrcArray()
			: FilterElementary(0,1),data(0)
		{
			y=0;
			w=0;
			h=0;
			nls=0;
		}
		SrcArray(std::string name,I * ptr,int width,int height,int next_line_shift,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(0,1),data(0)
		{
			SrcArray::Init(name,ptr,width,height,next_line_shift,post_proc);
		}
		virtual ~SrcArray()
		{
		}
		void Init(std::string name,I * ptr,int width,int height,int next_line_shift,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			setName(name);
			data=ptr;
			w=width;
			h=height;
			nls=next_line_shift;
			out.setPostProc(post_proc);
			AddPort(out,0);
			Restart();
		}
		void Restart()
		{
			y=0;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			out.set_bufferImgSize(ImgSize(w,h));
		}
		virtual void processLine() 
		{
			if(y<h)
			{
				out.beginUpdate();
				OutPortNoPostProc<T>::BufType::Iter_Line it=out.begin();
				I *p=data+y*nls;
				for(int x=0; x<w; x++)
					*it++=*p++;
				y++;
				out.endUpdate();
			}
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class SrcAImage
	/// \brief        
	/// ,      
	///  AImageMem8, AImageMem16, AImageMemFloat.
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class I,class P=Vtl::Plf::F11::DoNothing<T> >
	class SrcAImage
		: public SrcArray<T,typename I::Pixel,P>
	{
	public:
		SrcAImage()
			: SrcArray<T,I::Pixel,P>(),img(0)
		{
		}
		SrcAImage(std::string name,I * ptr,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: SrcArray<T,I::Pixel,P>(name,(*ptr)[0],ptr->Width(),ptr->Height(),ptr->NextLineShift(),post_proc)
		{
		}
		virtual ~SrcAImage()
		{
		}
		void Init(std::string name,I * ptr,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			SrcArray::Init(name,(*ptr)[0],ptr->Width(),ptr->Height(),ptr->NextLineShift(),post_proc);
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class SrcAImageRgb
	///        
	/// ,      
	///  AImageMemRgb8, AImageMemRgb16.
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class I,class P=Vtl::Plf::F11::DoNothing<T> >
	class SrcAImageRgb
		: public FilterElementary
	{
	public:
		OutPort<T,P> out_1;
		OutPort<T,P> out_2;
		OutPort<T,P> out_3;
	private:
		typename I::PlaneImageType * r;
		typename I::PlaneImageType * g;
		typename I::PlaneImageType * b;
		int y;
		int w;
		int h;
	public:
		SrcAImageRgb()
			: FilterElementary(0,1),img(0)
		{
		}
		SrcAImageRgb(std::string name,I * ptr,P post_proc=Vtl::Plf::F11::DoNothing<T>())
			: FilterElementary(name,0,1),img(ptr)
		{
			///
			w=img->Width();
			h=img->Height();
			AddPort(out_1,0);
			AddPort(out_2,1);
			AddPort(out_3,2);
			Restart();
		}
		virtual ~SrcAImageRgb()
		{
		}
		void Init(std::string name,I * ptr,P post_proc=Vtl::Plf::F11::DoNothing<T>())
		{
			img=ptr;
			setName(name);
			out_1.setPostProc(post_proc);
			out_2.setPostProc(post_proc);
			out_3.setPostProc(post_proc);
			///
			w=img->Width();
			h=img->Height();
			AddPort(out_1,0);
			AddPort(out_2,1);
			AddPort(out_3,2);
			Restart();
		}
		void Restart()
		{
			y=0;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{
			out_1.set_bufferImgSize(ImgSize(w,h));
			out_2.set_bufferImgSize(ImgSize(w,h));
			out_3.set_bufferImgSize(ImgSize(w,h));
		}
		virtual void processLine() 
		{
			if(y<h)
			{
				out_1.beginUpdate();
				out_2.beginUpdate();
				out_3.beginUpdate();
				copy_line(out_1.begin(),(*(img->GetPlane8(0)))[y]);
				copy_line(out_2.begin(),(*(img->GetPlane8(1)))[y]);
				copy_line(out_3.begin(),(*(img->GetPlane8(2)))[y]);
				y++;
				out_3.endUpdate();
				out_2.endUpdate();
				out_1.endUpdate();
			}
		}
	private:
		void copy_line(typename OutPortNoPostProc<T>::BufType::Iter_Line it,typename I::PlaneImageType::Pixel *p)
		{
			for(int x=0; x<w; x++)
				*it++=*p++;
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class DstArray
	/// \brief  -   . 
	///           
	///  ,   .
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class Pixel>
	class DstArray
		: public FilterElementary
	{
	public:
		InPort<T> in;
	private:
		Pixel * img;
		int y;
		int w;
		int h;
		int nls;
	public:
		DstArray(std::string name)
			: FilterElementary(name,1,0),img(0)
		{
			AddPort(in,0,Extent(0,0,0,0));
			w=0;
			h=0;
			Restart();
		}
		virtual ~DstArray()
		{
		}
		void setDstImage(Pixel* ptr,int width,int height, int next_line_shift)
		{
			img=ptr;
			w=width;
			h=height;
			nls=next_line_shift;
			Restart();
		}
		void Restart()
		{
			y=0;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{//Must be empty for destination filter, NOT using base definition!
		}
		virtual void processLine() 
		{
			if(y<h)
			{
				ImageRef<T> im=in.getImg(); 
				const T* input=im[0];
				Pixel *p=img + nls*y;
				for(int x=0; x<w; x++)
					*p++=Pixel(*input++);
				y++;
			}
			in.ToNextLine(y);
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class DstAImage
	/// \brief  -   . 
	///     -  ,   aimage.h  :
	/// AImageMem8, AImageMem16, AImageMemFloat.
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class I>
	class DstAImage
		: public DstArray<T,typename I::Pixel>
	{
	public:
		DstAImage(std::string name)
			: DstArray<T,I::Pixel>(name)
		{
		}
		virtual ~DstAImage()
		{
		}
		void setDstImage(I * ptr)
		{
			DstArray<T,I::Pixel>::setDstImage((*ptr)[0],ptr->Width(),ptr->Height(),ptr->NextLineShift());
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class DstContainer
	/// \brief  -   .     
	///  ,      threshold.   
	///         -     .
	///   -   .
	///////////////////////////////////////////////////////////////////////////
	template<typename T,class C,class P=Vtl::Plf::F11::DoNothing<T> >
	class DstContainer
		: public FilterElementary
	{
	public:
		InPort<T> in;
	private:
		int y;
		int w;
		int h;
		int tmp;
		C& container;
		T thresh;
	public:
		DstContainer(std::string name,C& container_,T threshold=T(0))
			: FilterElementary(name,1,0),container(container_),thresh(threshold)
		{
			AddPort(in,0,Extent(0,0,0,0));
			w=0;
			h=0;
			Restart();
		}
		virtual ~DstContainer()
		{
		}
		void setDstImage(Vtl::Plf::ImgSize is)
		{
			w=is.width;
			h=is.height;
			Restart();
		}
		void Restart()
		{
			y=0;
			tmp=0;
		}
		virtual void set_OutPortsImagesSizeAndInPortsExtents()
		{//Must be empty for destination filter, NOT using base definition!
		}
		virtual void processLine() 
		{
			if(y<h)
			{
				ImageRef<T> im=in.getImg(); 
				const T* input=im[0];
				for(int x=0; x<w; x++)
				{
					T value = *input++;
					if(value>thresh)
					{
						container.Add(x,y,value);
						//std::cout << tmp++ << "  " << x-(w/2) << " , "<< y-(h/2) << " , " << value <<std::endl;
					}
				}
				y++;
			}
			in.ToNextLine(y);
		}
	};

/** @} */ // end of IoFiltersDoxGroup
} //namespace Plf
} //namespace Vtl
#endif //_IO_FILTER_H
