#ifndef _PLF_SYS_H
#define _PLF_SYS_H
////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <fstream>
#include "ytypes.h"
#include "plf_filters.h"
#include "datastring.h"
#include "yfilesys.h"
////////////////////////////////////////////////////////////////////////////
namespace Vtl
{
// @brief     plf_filters.h,   
namespace Plf 
{
/** @addtogroup PipeLineFiltersDoxGroup  Pipe Line Filters. 
 *  @{
 */
/**
 * \brief  .     
 *  .        
 *  .      .  
 ******************************************************************************/
namespace Impl
{
	//////////////////////////////////////////////////////////////////
	// struct Edge
	///
	/// @brief  ,      
	///   Node.      
	///     (id_outp, id_of)  
	///         (id_inp, id_if)
	//////////////////////////////////////////////////////////////////
	struct Edge
	{
		int id_of;  //!<  ,      id_outp,  (id_of,id_outp) 
		int id_outp;//!<      
		int id_if;  //!<  ,       ,  (id,id_inp) 
		int id_inp; //!<      
		int id_edge;//!<  (   'edges'  
	};
	///////////////////////////////////////////////////////////////////////////
	// struct Node
	///
	/// @brief  ,    FiltersSystem  GraphBuilder
	///////////////////////////////////////////////////////////////////////////
	struct Node
	{
		enum Role
		{
			Source,
			Target,
			Transformer
		};
		Vtl::Plf::FilterElementary* fil;
		int id;
		int n_in;
		int n_out;
		int n_oto; //    ()  n_out .
		Role role;
		int rank;
		std::list<int> Adj; //!<   (  edges)   (OutPortNoPostProc->InPort)    
		std::list<int> AdjPrev;//!<   (  edges)  ,     .
		std::list<int> Adj_Next;//!< ,    ,   .
		std::list<int> Adj_Prev;//!< ,    ,   .
		std::vector<int> extra_delay;//!<  -  .,  - . .
		//----------------------
		Node() {fil=0;n_in=0;n_out=0;n_oto=0;}
		Node(FilterElementary* filter,int id) 
		{
			this->id=id;
			fil=filter;
			fil->id=id;
			n_in=fil->NumInputs();
			n_out=fil->NumOutputs();
			if(n_in==0 && n_out==0)
			{
				std::ostringstream os;
				os << "Filter '" << fil->Name() << "' has NO inputs NO outputs"; 
				throw os.str();
			}
			n_oto=0;
			for(int i=0; i<n_out; ++i)
			{
				n_oto+=fil->getOutPort(i)->NumConnectedTo();
			}
			if(n_in==0)
				role=Source;
			else if(n_out==0)
				role=Target;
			else
				role=Transformer;
			rank=-1;
		}
		void buildAdj(const std::vector<Node> & graph,std::vector<Edge>& edges)
		{
			Adj.clear();
			AdjPrev.clear();
			Adj_Next.clear();
			Adj_Prev.clear();
			Fifo<Vtl::Plf::FilterElementary*> Q;
			for(int i=0; i<fil->NumOutputs(); ++i)
			{
				//Adj.merge(fil->getOutPort(i)->getConnectedToIds());
				fil->getOutPort(i)->putConnectionsInQueue(Q);
				while(!Q.empty())
				{
					FilterElementary *f=Q.get();
					Edge e;
					e.id_of=fil->id;
					e.id_if=f->id;
					e.id_outp=i;
					for(int j=0; j<f->NumInputs(); ++j)
						if(fil==f->getInPort(j)->BaseConnectedTo()->getFilter())
						{
							e.id_inp=j;
							break;
						}
					int id_edge=edges.size();
					e.id_edge=id_edge;
					edges.push_back(e);
					Adj.push_back(id_edge);
				}
			}
			for(std::list<int>::const_iterator it=Adj.begin(); it!=Adj.end(); ++it)
			{
				int id_edge=*it;
				int id_if=edges[id_edge].id_if;
				if(Adj_Next.end()==std::find(Adj_Next.begin(),Adj_Next.end(),id_if))
					Adj_Next.push_back(id_if);
			}
			for(int j=0; j<fil->NumInputs(); ++j)
			{
				OutPortBase* outp=fil->getInPort(j)->BaseConnectedTo();
				int id_prev=outp->getFilter()->id;
				//    ,     
				for(std::list<int>::const_iterator it=graph[id_prev].Adj.begin(); it!=graph[id_prev].Adj.end(); ++it)
				{
					int i=*it;
					Edge e=edges[i];
					if(e.id_if == id)
					AdjPrev.push_back(i);
				}
			}
			for(std::list<int>::const_iterator it=AdjPrev.begin(); it!=AdjPrev.end(); ++it)
			{
				int id_of=edges[(*it)].id_of;
				if(Adj_Prev.end()==std::find(Adj_Prev.begin(),Adj_Prev.end(),id_of))
					Adj_Prev.push_back(id_of);
			}
		}
	};
	///////////////////////////////////////////////////////////////////////////
	// class GraphBuilder
	///
	/// \brief     ,   
	///  ,    ,      
	///  ,  "" .
	///    std::vector<Node> graph;  FiltersSystem
	///////////////////////////////////////////////////////////////////////////
	class GraphBuilder
	{
		std::vector<Node> & graph;
		std::vector<Edge> & edges;
		std::list<Node> src;
		std::list<Node> sorted;
		std::list<Node> dst;
		int id;
		std::ostringstream os_back;
	public:
		std::string BackEdges()
		{
			return os_back.str();
		}
		GraphBuilder(std::vector<Node> & _graph,std::vector<Impl::Edge>& _edges) :graph(_graph),edges(_edges) {}
		template<typename IterSrc> void build(IterSrc begin,IterSrc end)
		{
			id=0;
			graph.resize(0);
			//DFS + topologicallySort
			for(;begin!=end ;++begin)
				DFS_Visit(*begin);

			fillTopologicallySortedVector();
		}
		void build(FilterElementary * begin,FilterElementary * end)
		{
			id=0;
			graph.resize(0);
			//DFS + topologicallySort
			for(;begin!=end ;++begin)
				DFS_Visit(begin);
			fillTopologicallySortedVector();
		}
		template<typename Iter> FilterElementary * checkInGraph(Iter begin,Iter end)
		{
			for(;begin!=end ;++begin)
			{
				if(isWhite(*begin))
					return *begin;
			}
			return 0;
		}
		template<typename Iter> FilterElementary * checkInGraph(Iter* begin,Iter* end)
		{
			for(;begin!=end ;++begin)
			{
				if(isWhite(begin))
					return begin;
			}
			return 0;
		}
		void findParentNotInGraph()
		{
			for(size_t i=0; i<graph.size() ; ++i)
			{
				FilterElementary * f=graph[i].fil;
				for(int i=0; i< f->NumInputs(); ++i)
				{
					FilterElementary * parent=f->getInPort(i)->BaseConnectedTo()->getFilter();
					if(isWhite(parent))
					{
						std::ostringstream os;
						os	<< "Filter \"" << parent->Name() 
							<< "\" is parent to filter \"" << f->Name() 
							<< "\" but not distinguishable from directed sources. Check your code!"; 
						throw os.str();
					}
				}
			}
		}
	private:
		void fillTopologicallySortedVector()
		{
			int n=0;
			for(std::list<Node>::iterator it=src.begin(); it!=src.end(); it++ )
			{
				Node node=*it;
				node.id=n;
				node.fil->id=n;
				graph[n++]=node;
			}
			for(std::list<Node>::iterator it=sorted.begin(); it!=sorted.end(); it++ )
			{
				Node node=*it;
				node.id=n;
				node.fil->id=n;
				graph[n++]=node;
			}
			for(std::list<Node>::iterator it=dst.begin(); it!=dst.end(); it++ )
			{
				Node node=*it;
				node.id=n;
				node.fil->id=n;
				graph[n++]=node;
			}
			edges.resize(0);
			for(size_t i=0; i<graph.size() ; ++i)
				graph[i].buildAdj(graph,edges);
		}
		///  DFS_Visit(), checkInGraph(), findParentNotInGraph()
		bool isWhite(FilterElementary * filter)
		{
			if(0<= filter->id && filter->id < int(graph.size()) )
				if(graph[filter->id].fil == filter)
					return false;//          .
			return true;
		}
		///  DFS_Visit(), <b>   !isWhite()</b>, 
		///     !
		bool isGray(FilterElementary * filter)
		{
			return graph[filter->id].rank==-1;
		}
		///  23.10         , 
		///          .
		///   -            .
		///     -          .
		void DFS_Visit(FilterElementary * filter)
		{
			std::list<Vtl::Plf::FilterElementary *> Q;
			//  (     graph[]), 
			// id    time   .447
			// Node.id  filter.id -  d[u]
			graph.push_back(Node(filter,id++));
			//      .
			//    
			for(int i=0; i<filter->NumOutputs(); i++)
			{
				OutPortBase* o=filter->getOutPort(i);
				o->putConnectionsInQueue(Q);
			}
			//    ,        
			//    ,    , 
			//        .
			//  :
			for(std::list<Vtl::Plf::FilterElementary *>::iterator it=Q.begin(); it!=Q.end(); ++it)
			{
				std::list<Vtl::Plf::FilterElementary *>::iterator it2=it;
				++it2; //   it
				while( it2!=Q.end())
				{
					//     - 
					std::list<Vtl::Plf::FilterElementary *>::iterator it3=it2++;
					if(*it==*it3)
						Q.erase(it3); // , it3       (++ Standard ISO/IEC 14882 p.477)
				}
			}
			//  Q     
			for(std::list<Vtl::Plf::FilterElementary *>::iterator it=Q.begin(); it!=Q.end(); ++it)
			{
				if(isWhite(*it))
					DFS_Visit(*it);
				else if(isGray(*it))
				{
					//std::ostringstream os_back;
					os_back	<< "Back edge found: from filter \"" << filter->Name() 
						<< "\" to filter \"" << (*it)->Name() << "\". Check your code!"
						<< std::endl; 
					//throw os_back.str();
				}
			}
			// ,   
			graph[filter->id].rank=0;
			if(graph[filter->id].role==Node::Transformer)
				sorted.push_front(graph[filter->id]);
			else if(graph[filter->id].role==Node::Target)
				dst.push_front(graph[filter->id]);
			else //	Node::Source
				src.push_front(graph[filter->id]);
		}
	};
	class DelayAdjuster
	{
		std::vector<Node> & graph;
		const std::vector<Edge> & edges;
	public:
		std::vector<int> d;
		std::vector<int> b;
	public:
		///   - ,    
		///  .
		DelayAdjuster(std::vector<Node>& graph_,const std::vector<Edge> & edges_) 
			: graph(graph_),edges(edges_)
		{
		}
		void Run()
		{
			//process in topologically sorted order
			//!!! !!!    !!!
			d.assign(graph.size(),0);
			b.assign(edges.size(),0);
			for(size_t i=0; i<graph.size(); ++i)
			{
				if(graph[i].role == Node::Source)
				{
					d[i]=0;
				}
				else
				{
					for(std::list<int>::const_iterator it=graph[i].AdjPrev.begin(); it!=graph[i].AdjPrev.end(); ++it)
					{
						const int id_edge= *it;
						const Edge& e=edges[id_edge];
						assert(e.id_if==i);
						int dd=d[e.id_of]+w(e.id_if,e.id_inp);
						dd=std::max(d[i],dd);
						d[i]=dd;
					}
					for(std::list<int>::const_iterator it=graph[i].AdjPrev.begin(); it!=graph[i].AdjPrev.end(); ++it)
					{
						const int id_edge= *it;
						const Edge& e=edges[id_edge];
						int delta=d[i];
						delta=delta-(d[e.id_of]+w(e.id_if,e.id_inp));
						b[id_edge]=delta;
					}
				}
			}
/*
			for(int i=0; i<int(graph.size()); ++i)
			{
				printf("d[%d]=%d\n",i,int(d[i]));
			}
*/
			//--Eat delays already in buffer
//			printf("Eat delays already in buffer\n");
			std::vector<Extent> ext_fin;
			std::vector<int32_t> min_delay;
			for(size_t i=0; i<graph.size(); ++i)
			{//for each node
				Impl::Node & n=graph[i];

				//n.extra_delay.assign(n.n_out,0);
				min_delay.assign(n.n_out,INT32_MAX);
				ext_fin.assign(n.n_out,Extent(0));

//				printf("Filter id=%d \n",i);
				for(std::list<int>::const_iterator it=n.Adj.begin(); it!=n.Adj.end(); ++it)
				{
					const int id_edge= *it;
					const Edge& e=edges[id_edge];
					assert(e.id_of==i);
					assert(e.id_outp<n.n_out);
					int32_t del=b[e.id_edge];
					Extent ex=graph[e.id_if].fil->getInPort(e.id_inp)->ext();
					ex.forward+=b[id_edge];
					ext_fin[e.id_outp]=ex.max(ext_fin[e.id_outp]);
					min_delay[e.id_outp]=std::min(min_delay[e.id_outp],del);

				}
				for(int j=0; j<n.n_out; ++j)
				{
					Extent ext=ext_fin[j];
/*
					printf("OutPort=%d, Ext(%3d,%3d,%3d,%3d)\n",j
					,ext.left
					,ext.right
					,ext.before
					,ext.forward
					);
*/
					n.fil->getOutPort(j)->init_buffer(ext);
				}
//				Vtl::Plf::OutPortBase* o=n.fil->getOutPort(e.id_outp);
//				Extent ext=o->calcRequiredExtent();
				
			}


			//--   2 : 
			//  1) ,    
			//  2)  ,   .
		}
	private:
		int w(int id_fil,int id_port)
		{
			return graph[id_fil].fil->getInPort(id_port)->ext().forward;
		}
	};
}//namespace Impl
	///////////////////////////////////////////////////////////////////////////
	// class FiltersSystem
	///
	/// @brief " ",     
	/// 
	///
	///////////////////////////////////////////////////////////////////////////
	class FiltersSystem
	{
		std::vector<Impl::Node> graph;
		std::vector<Impl::Edge> edges;
		std::vector<int> sources;
		std::vector<int> multi_inputs;
		std::vector<int> multi_connected_to;
		std::vector<int> targets;
		std::string fname_dot;
	private:
		template<typename IterSrc> void _fillVec(std::vector<std::pair<FilterElementary *,bool> >& vec,IterSrc begin,IterSrc end)
		{
			for(;begin!=end ;++begin)
				vec.push_back(std::pair<FilterElementary *,bool>(*begin,false));
		}
		template<typename IterSrc> void _fillVec(std::vector<std::pair<FilterElementary *,bool> >& vec,IterSrc* begin,IterSrc* end)
		{
			for(;begin!=end ;++begin)
				vec.push_back(std::pair<FilterElementary *,bool>(begin,false));
		}
	public:
		virtual ~FiltersSystem() {}
		FiltersSystem(const char* dot_file_name=0) 
			: fname_dot(dot_file_name?dot_file_name:"")
		{
		}
		FiltersSystem(std::string dot_file_name) 
			: fname_dot(dot_file_name)
		{
		}
		void set_DotFName(std::string dot_file_name){ fname_dot=dot_file_name;}
		void set_DotFName(const char* dot_file_name=0) {fname_dot=dot_file_name?dot_file_name:"";}
		std::string get_DotFName() {return fname_dot;}
		//---------------------------------------------------------------
		template<typename IterSrc> 
			void Assign(IterSrc src_begin,IterSrc src_end,FilterElementary * dst)
		{
			Assign(src_begin,src_end,dst,dst+1);
		}
		template<typename IterDst> 
			void Assign(FilterElementary * src,IterDst dst_begin,IterDst dst_end)
		{
			Assign(src,src+1,dst_begin,dst_end);
		}
		void Assign(FilterElementary * src,FilterElementary * dst)
		{
			Assign(src,src+1,dst,dst+1);
		}
		template<typename IterSrc,typename IterDst> 
			void Assign(IterSrc src_begin,IterSrc src_end,IterDst dst_begin,IterDst dst_end)
		{
			Impl::GraphBuilder gb(graph,edges);
			gb.build(src_begin,src_end );
			if(fname_dot.length()!=0)
				SaveGraphAsDot(fname_dot.c_str());
			//   
			if(gb.BackEdges().length()>0)
			{
				throw gb.BackEdges();
			}
			//   Targets   
			FilterElementary * f=gb.checkInGraph(dst_begin,dst_end);
			if(f)
			{
				std::ostringstream os;
				os	<< "Target filter \"" << f->Name() 
					<< "\" is not distinguishable from directed sources. Check your code!"; 
				throw os.str();
			}
			//      , ...  
			gb.findParentNotInGraph();
			//  std::vector<int> sources,multi_inputs,targets;
			findSourcesTargetsAndJunctions();
		//=====   ,     ===============
			propagateImageSize();
			//Adjust StripBuffer's Extents
			init_buffersExtents();
		//Adjust dalays
			Impl::DelayAdjuster da(graph,edges);
			da.Run();
			{
				EdgeLabel_DelayB el_b(da.b);
				std::string fn=YFileSys::getNameWithoutExt(fname_dot)+"_b.dot";
				_SaveGraphAsDot(fn.c_str(),el_b);
			}

			//=====  ,       ===


			InitStripBuffers(); //-    DelayAdjuster 
		}
		//---------------------------------------------------------------
		void show_lists()
		{
#if 0
			for(size_t i=0; i<graph.size(); ++i)
			{
				Impl::Node n=graph[i];
				printf("Filter id=%d  name=%s\n",n.id,n.fil->Name().c_str());
				printf("Adj_Next: ");
				for(std::list<int>::iterator it=n.Adj_Next.begin(); it!=n.Adj_Next.end(); ++it)
				{
					int id=*it;
					printf("%d, ",id);
				}
				printf("\n");
				printf("Adj_Prev: ");
				for(std::list<int>::iterator it=n.Adj_Prev.begin(); it!=n.Adj_Prev.end(); ++it)
				{
					int id=*it;
					printf("%d, ",id);
				}
				printf("\n");
				printf("Adj: (Edge.id,Edge.id_inp,Edge.id_outp) : ");
				for(std::list<int>::const_iterator it=n.Adj.begin(); it!=n.Adj.end(); ++it)
				{
					Impl::Edge e=edges[*it];
					printf("(%d,%d,%d,%d) ",e.id_if,e.id_inp,e.id_of,e.id_outp);
				}
				printf("\n");
			}
#endif
		}
		//---------------------------------------------------------------
		void Run(ProgressCounter* indicator=0,const char* name="")
		{
			//     :
			int dst_max_haight=0;
			int id_dst_max=-1;
			int id_dst_max_inp=-1;
			for(size_t i=0; i<targets.size(); ++i)
			{
				int n=graph[targets[i]].fil->NumInputs();
				for(int j=0; j<n; j++)
				{
					int d=graph[targets[i]].fil->getInPort(j)->getImageSize().height;
					if(d>dst_max_haight)
					{
						dst_max_haight=d;
						id_dst_max=graph[targets[i]].id;
						id_dst_max_inp=j;
					}
				}
			}
			if(indicator)
			{
				indicator->start(name,0,dst_max_haight);
			}
			int y=0;
			Fifo<int> Q;
			//  ,      .
			for(size_t i=0; i<sources.size(); ++i)
				graph[sources[i]].fil->put_in_queue(Q);
			//,    
			while(!Q.empty())
			{
				int id=Q.get();
				graph[id].fil->do_1_step();
				if(id==id_dst_max)
				{
					if(indicator)
					{
						indicator->current(graph[id].fil->getInPort(id_dst_max_inp)->Y());
					}
				}

//				printf("y=%d,  id=%d\n",y++,id);
				//         ,
				//   -     .
				for(std::list<int>::iterator it=graph[id].Adj_Next.begin(); it!=graph[id].Adj_Next.end(); ++it)
				{
					int id_next=*it;
					graph[id_next].fil->TryPutInQueue(Q);
				}
				//         ,
				//   -     .
				//         
				// (  ) -      .
				//    -           
				//  -,    . 
				if(!graph[id].fil->isFinished())
				{
					for(std::list<int>::iterator it=graph[id].Adj_Prev.begin(); it!=graph[id].Adj_Prev.end(); ++it)
					{
						int id_prev=*it;
						graph[id_prev].fil->TryPutInQueue(Q);
					}
				}
				//      ? 
				//  ,       
				//     ..      
				//      .
				graph[id].fil->TryPutInQueue(Q);
			}
			if(indicator)
			{
				indicator->finish();
			}
			show_lists();
		}
		//---------------------------------------------------------------
		void SaveGraphAsDot(const char* fname) const
		{
			SaveGraphAsDot(std::string(fname));
		}
		void SaveGraphAsDot(std::string fname) const
		{
			EdgeLabel_Unknown el;
			_SaveGraphAsDot(fname,el);
		}
		void SaveGraphAsDot(std::ostream& os) const
		{
			EdgeLabel_Unknown el;
			_SaveGraphAsDot(os,el);
		}
		template<class E> 
		void _SaveGraphAsDot(std::string fname,E& efunc) const
		{
			std::ofstream os;
			os.open(fname.c_str(),std::ios_base::out | std::ios_base::trunc );
			_SaveGraphAsDot(os,efunc);
			os.flush();
			os.close();
		}
		template<class E>
		void _SaveGraphAsDot(std::ostream& os,E& efunc) const
		{
			const char* fillcolor="\"#FCDDB4\"";
			const char* fillcolor_src="\"#CBFCB4\"";
			const char* fillcolor_dst="\"#FCB4B4\"";
			os << "digraph " << "test123" << " {" << std::endl;
			os << "   node [fontname=Helvetica, fontsize=12];"<< std::endl;
			os << "   edge [style=\"solid\", color=\"black\" ,arrowhead=\"open\"];"<< std::endl;
			{
				for(size_t i=0; i<graph.size(); i++)
				{
					os	<< "   node" << i 
						<< " [label=\""<< graph[i].fil->Name() << " \\n(" << graph[i].id <<")" <<"\""
						<< ", shape=" << ((graph[i].role==Impl::Node::Transformer)?"record":"oval")
						<< ", color=" << "black"
//						<< ", color=" << ((graph[i].role==Node::Source)? "green":(graph[i].role==Node::Target)?"red":"black")
						<< ", fillcolor=" << ((graph[i].role==Impl::Node::Source)? fillcolor_src:(graph[i].role==Impl::Node::Target)?fillcolor_dst:fillcolor)
						<< ", fontcolor=black"
						<< ", style=filled"
						<< "];" << std::endl;
				}
				for(size_t i=0; i<graph.size(); i++)
				{
					for(std::list<int>::const_iterator it=graph[i].Adj.begin(); it!=graph[i].Adj.end(); ++it)
					{
						const Impl::Edge& e=edges[*it]; 
						//int v=(*it).id;
						os	<< "   node" << i << "-> " << "node" << e.id_if
							<< " [label=\""<< efunc(e) <<"\"" 
							<< " taillabel=\"" << e.id_outp <<"\"" 
							<< " headlabel=\"" << e.id_inp  <<"\"" 
							<< "];"
							<< std::endl;
					}
				}
			}
			os << "}" << std::endl;
		}
		struct EdgeLabel_Unknown
		{
			EdgeLabel_Unknown() {}
			std::string operator()(const Impl::Edge& e) {return "a";}
		};
		class EdgeLabel_DelayB
		{
			const std::vector<int>& b;
		public:
			EdgeLabel_DelayB(const std::vector<int>& b_) : b(b_) {}
			std::string operator()(const Impl::Edge& e) {return std::string("b=")+DataString(b[e.id_edge]);}
		};

	private:
		void printList(std::string name,const std::vector<int>& lst) const
		{
			
			std::cout << "===="<< name << "===="<< std::endl;
			for(size_t i=0; i<lst.size(); ++i)
			{
				std::cout << graph[lst[i]].fil->Name() << " " << graph[lst[i]].id << std::endl;;
			}
			std::cout << "========"<< std::endl;;
		}
		//---------------------------------------------------------------
		void findSourcesTargetsAndJunctions()
		{
			sources.resize(0);
			multi_inputs.resize(0);
			multi_connected_to.resize(0);
			targets.resize(0);
			for(size_t i=0; i<graph.size(); i++)
			{//    graph[i].id == i
				if(graph[i].role==Impl::Node::Source)
					sources.push_back(graph[i].id);
				else if(graph[i].role==Impl::Node::Target)
					targets.push_back(graph[i].id);
				if(graph[i].fil->NumInputs()>1) // It may be target too !!!
					multi_inputs.push_back(graph[i].id);
				if(graph[i].n_oto>1) // It may be target too !!!
					multi_connected_to.push_back(graph[i].id);
			}
			//printList("sources",sources);
			//printList("targets",targets);
			//printList("multi_inputs",multi_inputs);
			//printList("multi_connected_to",multi_connected_to);
		}

		void propagateImageSize()
		{
			for(size_t i=0; i<graph.size() ; ++i)
			{
				//graph[i].fil->set_buffersImgSize();
				graph[i].fil->set_OutPortsImagesSizeAndInPortsExtents();
			}
		}
		void init_buffersExtents()
		{
			for(size_t i=0; i<graph.size() ; ++i)
			{
				graph[i].fil->init_OutPortsExtents();
			}
		}
		void InitStripBuffers()
		{
			for(size_t i=0; i<graph.size(); ++i)
			{
				Vtl::Plf::FilterElementary * filter=graph[i].fil;
				filter->init_buffers();
			}
		}

	};
/** @} */ // end of PipeLineFiltersDoxGroup
} //namespace Plf 
} //namespace Vtl
////////////////////////////////////////////////////////////////////////////
#endif //_PLF_SYS_H
